<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    cns_multi_moderations
 */

/*EXTRA FUNCTIONS: zip_\w+*/

require_code('crud_module');

/**
 * Module page class.
 */
class Module_admin_cns_multi_moderations extends Standard_crud_module
{
    protected $lang_type = 'MULTI_MODERATION';
    protected $select_name = 'NAME';
    protected $archive_entry_point = '_SEARCH:forumview';
    protected $archive_label = 'SECTION_FORUMS';
    protected $menu_label = 'MULTI_MODERATIONS';
    protected $table = 'f_multi_moderations';
    protected $orderer = 'mm_name';
    protected $donext_entry_content_type = 'multi_moderation';
    protected $donext_category_content_type = null;

    /**
     * Find entry-points available within this module.
     *
     * @param  boolean $check_perms Whether to check permissions
     * @param  ?MEMBER $member_id The member to check permissions as (null: current user)
     * @param  boolean $support_crosslinks Whether to allow cross links to other modules (identifiable via a full-page-link rather than a screen-name)
     * @param  boolean $be_deferential Whether to avoid any entry-point (or even return null to disable the page in the Sitemap) if we know another module, or page_group, is going to link to that entry-point. Note that "!" and "browse" entry points are automatically merged with container page nodes (likely called by page-groupings) as appropriate.
     * @return ?array A map of entry points (screen-name=>language-code/string or screen-name=>[language-code/string, icon-theme-image]) (null: disabled)
     */
    public function get_entry_points(bool $check_perms = true, ?int $member_id = null, bool $support_crosslinks = true, bool $be_deferential = false) : ?array
    {
        if (!addon_installed('cns_multi_moderations')) {
            return null;
        }

        if (get_forum_type() != 'cns') {
            return null;
        }

        if ($be_deferential || $support_crosslinks) {
            return null;
        }

        return [
            'browse' => [do_lang_tempcode('menus:ITEMS_HERE', do_lang_tempcode('MULTI_MODERATIONS'), make_string_tempcode(escape_html(integer_format(intval($GLOBALS['FORUM_DB']->query_select_value('f_multi_moderations', 'COUNT(*)')), 0)))), 'menu/adminzone/structure/forum/multi_moderations'],
        ] + parent::get_entry_points();
    }

    public $title;

    /**
     * Module pre-run function. Allows us to know metadata for <head> before we start streaming output.
     *
     * @param  boolean $top_level Whether this is running at the top level, prior to having sub-objects called
     * @param  ?ID_TEXT $type The screen type to consider for metadata purposes (null: read from environment)
     * @return ?Tempcode Tempcode indicating some kind of exceptional output (null: none)
     */
    public function pre_run(bool $top_level = true, ?string $type = null) : ?object
    {
        if (get_forum_type() != 'cns') {
            warn_exit(do_lang_tempcode('NO_CNS'));
        }

        $error_msg = new Tempcode();
        if (!addon_installed__messaged('cns_multi_moderations', $error_msg)) {
            return $error_msg;
        }

        if (!addon_installed('cns_forum')) {
            warn_exit(do_lang_tempcode('MISSING_ADDON', escape_html('cns_forum')));
        }

        $type = get_param_string('type', 'browse');

        require_lang('cns');
        require_lang('cns_multi_moderations');
        require_css('cns_admin');

        set_helper_panel_tutorial('tut_moderation');

        if ($type == 'browse') {
            breadcrumb_set_parents([['_SEARCH:admin_cns_forums:browse', do_lang_tempcode('MANAGE_FORUMS')]]);
        } elseif ($type == 'import') {
            breadcrumb_set_parents([['_SEARCH:admin_cns_forums:browse', do_lang_tempcode('MANAGE_FORUMS')], ['_SELF:_SELF:browse', do_lang_tempcode('MULTI_MODERATIONS')]]);
        } elseif ($type == '_import') {
            breadcrumb_set_parents([['_SEARCH:admin_cns_forums:browse', do_lang_tempcode('MANAGE_FORUMS')], ['_SELF:_SELF:browse', do_lang_tempcode('MULTI_MODERATIONS')], ['_SELF:_SELF:import', do_lang_tempcode('IMPORT_STOCK_RESPONSES')]]);
        } else {
            breadcrumb_set_parents([['_SEARCH:admin_cns_forums:browse', do_lang_tempcode('MANAGE_FORUMS')], ['_SELF:_SELF:browse', do_lang_tempcode('MULTI_MODERATIONS')]]);
        }

        if ($type == 'import' || $type == '_import') {
            $this->title = get_screen_title('IMPORT_STOCK_RESPONSES');
        }

        return parent::pre_run($top_level);
    }

    /**
     * Standard crud_module run_start.
     *
     * @param  ID_TEXT $type The type of module execution
     * @return Tempcode The output of the run
     */
    public function run_start(string $type) : object
    {
        cns_require_all_forum_stuff();

        require_code('cns_moderation_action');
        require_code('cns_moderation_action2');
        require_code('cns_general_action2');

        $this->add_one_label = do_lang_tempcode('ADD_MULTI_MODERATION');
        $this->edit_this_label = do_lang_tempcode('EDIT_THIS_MULTI_MODERATION');
        $this->edit_one_label = do_lang_tempcode('EDIT_MULTI_MODERATION');

        if ($type == 'browse') {
            return $this->browse();
        }
        if ($type == 'import') {
            return $this->import();
        }
        if ($type == '_import') {
            return $this->_import();
        }
        return new Tempcode();
    }

    /**
     * The do-next manager for before content management.
     *
     * @return Tempcode The UI
     */
    public function browse() : object
    {
        require_code('templates_donext');
        return do_next_manager(
            get_screen_title('MULTI_MODERATIONS'),
            comcode_lang_string('DOC_MULTI_MODERATIONS'),
            [
                ['admin/add', ['_SELF', ['type' => 'add'], '_SELF'], do_lang('ADD_MULTI_MODERATION')],
                ['admin/edit', ['_SELF', ['type' => 'edit'], '_SELF'], do_lang('EDIT_MULTI_MODERATION')],
                ['admin/import', ['_SELF', ['type' => 'import'], '_SELF'], do_lang('IMPORT_STOCK_RESPONSES')],
            ],
            do_lang('MULTI_MODERATIONS')
        );
    }

    /**
     * The UI to import in bulk from an archive file.
     *
     * @return Tempcode The UI
     */
    public function import() : object
    {
        $post_url = build_url(['page' => '_SELF', 'type' => '_import', 'uploading' => 1], '_SELF');

        $fields = new Tempcode();

        $supported = 'tar';
        if (class_exists('ZipArchive', false)) {
            $supported .= ', zip';
        }
        $fields->attach(form_input_upload_multi(do_lang_tempcode('UPLOAD'), do_lang_tempcode('DESCRIPTION_ARCHIVE_TEXT_FILES', escape_html($supported), escape_html('txt')), 'file', true, null, null, true, 'txt,' . $supported));

        if (addon_installed('tickets')) {
            require_code('tickets');
            $ticket_forum_id = get_ticket_forum_id();
        } else {
            $ticket_forum_id = null;
        }
        require_code('cns_general_action2');
        $fields->attach(cns_get_forum_multi_code_field(($ticket_forum_id === null) ? '' : ('+' . strval($ticket_forum_id))));

        $text = paragraph(do_lang_tempcode('DESCRIPTION_IMPORT_STOCK_RESPONSES'));

        return do_template('FORM_SCREEN', [
            '_GUID' => 'bd30d8b0077567e3caf9239ed64204e5',
            'TITLE' => $this->title,
            'FIELDS' => $fields,
            'SUBMIT_ICON' => 'admin/import',
            'SUBMIT_NAME' => do_lang_tempcode('IMPORT_STOCK_RESPONSES'),
            'URL' => $post_url,
            'TEXT' => $text,
            'HIDDEN' => '',
        ]);
    }

    /**
     * The actualiser to import in bulk from an archive file.
     *
     * @return Tempcode The UI
     */
    public function _import() : object
    {
        require_code('files');
        require_lang('dearchive');

        require_code('files2');

        require_code('uploads');
        is_plupload(true);

        if ((!is_plupload(true)) && ((!array_key_exists('file_1', $_FILES)) || (!is_uploaded_file($_FILES['file_1']['tmp_name'])))) {
            warn_exit(do_lang_tempcode('IMPROPERLY_FILLED_IN_UPLOAD'));
        }

        set_mass_import_mode();

        $target_forum = read_multi_code('forum_multi_code');

        $multi_mods = $GLOBALS['FORUM_DB']->query_select('f_multi_moderations', ['id'], ['mm_move_to_forum_id' => null, 'mm_pin_state' => null, 'mm_open_state' => null, 'mm_title_suffix' => '', 'mm_forum_multi_code' => $target_forum]);
        require_code('cns_moderation_action2');
        foreach ($multi_mods as $multi_mod) {
            cns_delete_multi_moderation($multi_mod['id']);
        }

        foreach ($_FILES as $attach_name => $__file) {
            $file = $__file['name'];
            if ($file == '') {
                continue; // Not filled in this one
            }

            $tmp_name = $__file['tmp_name'];

            if ((!is_plupload()) && (!is_uploaded_file($tmp_name))) {
                $upload_error_message = get_upload_error_message($__file);
                attach_message($upload_error_message, 'warn');
                continue;
            }

            switch (get_file_extension($file)) {
                case 'zip':
                    if (!class_exists('ZipArchive', false)) {
                        warn_exit(do_lang_tempcode('ZIP_NOT_ENABLED'));
                    }

                    $zip_archive = new ZipArchive();
                    $myfile = $zip_archive->open($tmp_name);
                    if ($myfile !== true) {
                        require_code('failure');
                        warn_exit(zip_error($tmp_name, $myfile));
                    }

                    for ($i = 0; $i < $zip_archive->numFiles; $i++) {
                        $filename = $zip_archive->getNameIndex($i);

                        if (!check_extension($filename, false, null, true)) {
                            continue;
                        }

                        if ((cms_strtolower_ascii(substr($filename, -4)) == '.txt') && (!should_ignore_file($filename))) {
                                $data = $zip_archive->getFromIndex($i);
                                if ($data === false) { // Error reading
                                    warn_exit(zip_error($tmp_name, ZipArchive::ER_READ), false, true);
                                    break;
                                }
                                if ($data === '') { // No more to read
                                    continue;
                                }
                                $this->_import_stock_response($filename, $data, $target_forum);
                        }
                    }

                    $zip_archive->close();
                    unset($zip_archive);
                    break;
                case 'tar':
                    require_code('tar');
                    $myfile = tar_open($tmp_name, 'rb');
                    if ($myfile !== false) {
                        $directory = tar_get_directory($myfile);
                        foreach ($directory as $entry) {
                            $filename = $entry['path'];

                            if (!check_extension($filename, false, null, true)) {
                                continue;
                            }

                            if ((cms_strtolower_ascii(substr($filename, -4)) == '.txt') && (!should_ignore_file($filename))) {
                                // Load in file
                                $_in = tar_get_file($myfile, $entry['path'], false);

                                $this->_import_stock_response($filename, $_in['data'], $target_forum);
                            }
                        }

                        tar_close($myfile);
                    }
                    break;
                default:
                    if (!check_extension($file, false, null, true)) {
                        break;
                    }

                    if (cms_strtolower_ascii(substr($file, -4)) == '.txt') {
                        $this->_import_stock_response($file, cms_file_get_contents_safe($tmp_name, FILE_READ_LOCK | FILE_READ_UNIXIFIED_TEXT | FILE_READ_BOM), $target_forum);
                    } else {
                        attach_message(do_lang_tempcode('BAD_ARCHIVE_FORMAT'), 'warn');
                    }
            }
        }

        log_it('IMPORT_STOCK_RESPONSES');

        return $this->do_next_manager($this->title, do_lang_tempcode('SUCCESS'));
    }

    /**
     * Import a stock response.
     *
     * @param  PATH $path Path of the file (not on disk, just for reference as a title)
     * @param  string $data Data
     * @param  SHORT_TEXT $target_forum The forum multicode identifying where the multi-moderation is applicable
     */
    public function _import_stock_response(string $path, string $data, string $target_forum)
    {
        require_code('cns_moderation_action');

        $name = do_lang('STOCK_RESPONSE', cms_ucwords_ascii(str_replace(['/', '\\'], [': ', ': '], preg_replace('#\.txt$#', '', $path))));

        $data = fix_bad_unicode($data);

        cns_make_multi_moderation($name, $data, null, null, null, $target_forum, '');
    }

    /**
     * Get Tempcode for an adding form.
     *
     * @return mixed Either Tempcode; or a tuple of: (fields, hidden-fields[, delete-fields][, edit-text][, whether all delete fields are specified][, posting form text, more fields][, parsed WYSIWYG editable text])
     */
    public function get_form_fields_for_add()
    {
        return $this->get_form_fields();
    }

    /**
     * Get Tempcode for adding/editing form.
     *
     * @param  SHORT_TEXT $name The name of the Multi Moderation
     * @param  LONG_TEXT $post_text The text to place as a post in the topic when the Multi Moderation is performed
     * @param  ?AUTO_LINK $move_to Move the topic to this forum (null: don't move)
     * @param  ?BINARY $pin_state What to change the pin state to (null: don't change)
     * @param  ?BINARY $open_state What to change the open state to (null: don't change)
     * @param  SHORT_TEXT $forum_multi_code The forum multicode identifying where the multi-moderation is applicable
     * @param  SHORT_TEXT $title_suffix The title suffix
     * @return array A pair: The input fields, Hidden fields
     */
    public function get_form_fields(string $name = '', string $post_text = '', ?int $move_to = null, ?int $pin_state = null, ?int $open_state = null, string $forum_multi_code = '*', string $title_suffix = '') : array
    {
        require_code('cns_forums2');

        $fields = new Tempcode();
        $fields->attach(form_input_line(do_lang_tempcode('NAME'), do_lang_tempcode('DESCRIPTION_NAME'), 'multi_moderation_name', $name, true));
        $fields->attach(form_input_text_comcode(do_lang_tempcode('FORUM_POST'), do_lang_tempcode('DESCRIPTION_MULTI_MODERATION_POST'), 'post_text', $post_text, false));
        $fields->attach(form_input_tree_list(do_lang_tempcode('DESTINATION'), do_lang_tempcode('DESCRIPTION_DESTINATION_FORUM'), 'move_to', null, 'choose_forum', [], false, ($move_to === null) ? null : strval($move_to)));
        $pin_state_list = new Tempcode();
        $pin_state_list->attach(form_input_radio_entry('pin_state', '', ($pin_state === null), do_lang_tempcode('NA_EM')));
        $pin_state_list->attach(form_input_radio_entry('pin_state', '0', $pin_state === 0, do_lang_tempcode('UNPIN_TOPIC')));
        $pin_state_list->attach(form_input_radio_entry('pin_state', '1', $pin_state === 1, do_lang_tempcode('PIN_TOPIC')));
        $fields->attach(form_input_radio(do_lang_tempcode('PIN_STATE'), do_lang_tempcode('DESCRIPTION_PIN_STATE'), 'pin_state', $pin_state_list));
        $open_state_list = new Tempcode();
        $open_state_list->attach(form_input_radio_entry('open_state', '', ($open_state === null), do_lang_tempcode('NA_EM')));
        $open_state_list->attach(form_input_radio_entry('open_state', '0', $open_state === 0, do_lang_tempcode('CLOSE_TOPIC')));
        $open_state_list->attach(form_input_radio_entry('open_state', '1', $open_state === 1, do_lang_tempcode('OPEN_TOPIC')));
        $fields->attach(form_input_radio(do_lang_tempcode('OPEN_STATE'), do_lang_tempcode('DESCRIPTION_OPEN_STATE'), 'open_state', $open_state_list));
        $fields->attach(cns_get_forum_multi_code_field($forum_multi_code));
        $fields->attach(form_input_line(do_lang_tempcode('TITLE_SUFFIX'), do_lang_tempcode('DESCRIPTION_TITLE_SUFFIX'), 'title_suffix', $title_suffix, false));

        return [$fields, new Tempcode()];
    }

    /**
     * Standard crud_module table function.
     *
     * @param  array $url_map Details to go to build_url for link to the next screen
     * @return array A pair: The choose table, Whether re-ordering is supported from this screen
     */
    public function create_selection_list_choose_table(array $url_map) : array
    {
        require_code('templates_results_table');

        $current_ordering = get_param_string('sort', 'mm_name ASC', INPUT_FILTER_GET_COMPLEX);
        $sortables = [
            'mm_name' => do_lang_tempcode('NAME'),
            'mm_pin_state' => do_lang_tempcode('PIN_STATE'),
            'mm_open_state' => do_lang_tempcode('OPEN_STATE'),
        ];
        list($sql_sort, $sort_order, $sortable) = process_sorting_params('multi_moderation', $current_ordering);

        $header_row = results_header_row([
            do_lang_tempcode('NAME'),
            do_lang_tempcode('DESTINATION'),
            do_lang_tempcode('PIN_STATE'),
            do_lang_tempcode('OPEN_STATE'),
            do_lang_tempcode('ACTIONS'),
        ], $sortables, 'sort', $sortable . ' ' . $sort_order);

        $result_entries = new Tempcode();

        list($rows, $max_rows) = $this->get_entry_rows(false, $sql_sort);
        foreach ($rows as $row) {
            $pin_state = do_lang_tempcode('NA_EM');
            if ($row['mm_pin_state'] !== null) {
                switch ($row['mm_pin_state']) {
                    case 0:
                        $pin_state = do_lang_tempcode('UNPIN_TOPIC');
                        break;
                    case 1:
                        $pin_state = do_lang_tempcode('PIN_TOPIC');
                        break;
                }
            }
            $open_state = do_lang_tempcode('NA_EM');
            if ($row['mm_open_state'] !== null) {
                switch ($row['mm_open_state']) {
                    case 0:
                        $open_state = do_lang_tempcode('CLOSE_TOPIC');
                        break;
                    case 1:
                        $open_state = do_lang_tempcode('OPEN_TOPIC');
                        break;
                }
            }

            $destination = ($row['mm_move_to_forum_id'] === null) ? null : $GLOBALS['FORUM_DB']->query_select_value_if_there('f_forums', 'f_name', ['id' => $row['mm_move_to_forum_id']]);
            if ($destination === null) {
                $destination = do_lang_tempcode('NA_EM');
            }

            $edit_url = build_url($url_map + ['id' => $row['id']], '_SELF');

            $result_entries->attach(results_entry([get_translated_text($row['mm_name'], $GLOBALS['FORUM_DB']), $destination, $pin_state, $open_state, protect_from_escaping(hyperlink($edit_url, do_lang_tempcode('EDIT'), false, false, do_lang('EDIT') . ' #' . strval($row['id'])))], true));
        }

        return [results_table(do_lang($this->menu_label), either_param_integer('start', 0), 'start', either_param_integer('max', 20), 'max', $max_rows, $header_row, $result_entries, $sortables, $sortable, $sort_order), false];
    }

    /**
     * Standard crud_module list function.
     *
     * @return Tempcode The selection list
     */
    public function create_selection_list_entries() : object
    {
        $_m = $GLOBALS['FORUM_DB']->query_select('f_multi_moderations', ['id', 'mm_name']);
        $entries = new Tempcode();
        foreach ($_m as $m) {
            $entries->attach(form_input_list_entry(strval($m['id']), false, get_translated_text($m['mm_name'], $GLOBALS['FORUM_DB'])));
        }

        return $entries;
    }

    /**
     * Standard crud_module edit form filler.
     *
     * @param  ID_TEXT $id The entry being edited
     * @return mixed Either Tempcode; or a tuple of: (fields, hidden-fields[, delete-fields][, edit-text][, whether all delete fields are specified][, posting form text, more fields][, parsed WYSIWYG editable text])
     */
    public function fill_in_edit_form(string $id)
    {
        $m = $GLOBALS['FORUM_DB']->query_select('f_multi_moderations', ['*'], ['id' => intval($id)], '', 1);
        if (!array_key_exists(0, $m)) {
            warn_exit(do_lang_tempcode('MISSING_RESOURCE', 'multi_moderation'));
        }
        $r = $m[0];

        return $this->get_form_fields(get_translated_text($r['mm_name'], $GLOBALS['FORUM_DB']), $r['mm_post_text'], $r['mm_move_to_forum_id'], $r['mm_pin_state'], $r['mm_open_state'], $r['mm_forum_multi_code'], $r['mm_title_suffix']);
    }

    /**
     * Standard crud_module add actualiser.
     *
     * @return array A pair: The entry added, description about usage
     */
    public function add_actualisation() : array
    {
        $pin_state = post_param_integer('pin_state', null);
        $open_state = post_param_integer('open_state', null);

        $id = cns_make_multi_moderation(post_param_string('multi_moderation_name'), post_param_string('post_text'), post_param_integer('move_to', null), $pin_state, $open_state, read_multi_code('forum_multi_code'), post_param_string('title_suffix'));

        return [strval($id), null];
    }

    /**
     * Standard crud_module edit actualiser.
     *
     * @param  ID_TEXT $id The entry being edited
     * @return ?Tempcode Description about usage (null: none)
     */
    public function edit_actualisation(string $id) : ?object
    {
        $pin_state = post_param_integer('pin_state', null);
        $open_state = post_param_integer('open_state', null);

        cns_edit_multi_moderation(intval($id), post_param_string('multi_moderation_name'), post_param_string('post_text'), post_param_integer('move_to', null), $pin_state, $open_state, read_multi_code('forum_multi_code'), post_param_string('title_suffix'));

        return null;
    }

    /**
     * Standard crud_module delete actualiser.
     *
     * @param  ID_TEXT $id The entry being deleted
     */
    public function delete_actualisation(string $id)
    {
        cns_delete_multi_moderation(intval($id));
    }
}
