<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    core_language_editing
 */

/**
 * Module page class.
 */
class Module_admin_lang
{
    /**
     * Find details of the module.
     *
     * @return ?array Map of module info (null: module is disabled)
     */
    public function info() : ?array
    {
        $info = [];
        $info['author'] = 'Chris Graham';
        $info['organisation'] = 'Composr';
        $info['hacked_by'] = null;
        $info['hack_version'] = null;
        $info['version'] = 2;
        $info['locked'] = false;
        $info['min_cms_version'] = 11.0;
        $info['addon'] = 'core_language_editing';
        return $info;
    }

    /**
     * Find entry-points available within this module.
     *
     * @param  boolean $check_perms Whether to check permissions
     * @param  ?MEMBER $member_id The member to check permissions as (null: current user)
     * @param  boolean $support_crosslinks Whether to allow cross links to other modules (identifiable via a full-page-link rather than a screen-name)
     * @param  boolean $be_deferential Whether to avoid any entry-point (or even return null to disable the page in the Sitemap) if we know another module, or page_group, is going to link to that entry-point. Note that "!" and "browse" entry points are automatically merged with container page nodes (likely called by page-groupings) as appropriate.
     * @return ?array A map of entry points (screen-name=>language-code/string or screen-name=>[language-code/string, icon-theme-image]) (null: disabled)
     */
    public function get_entry_points(bool $check_perms = true, ?int $member_id = null, bool $support_crosslinks = true, bool $be_deferential = false) : ?array
    {
        $ret = [
            'browse' => ['TRANSLATE_CODE', 'menu/adminzone/style/language/language'],
        ];
        if (!$be_deferential) {
            $ret += [
                'criticise' => ['CRITICISE_LANGUAGE_PACK', 'menu/adminzone/style/language/criticise_language'],
            ];

            if ((multi_lang()) && (multi_lang_content())) {
                $ret += [
                    'content' => ['TRANSLATE_CONTENT', 'menu/adminzone/style/language/language_content'],
                ];
            }
        }
        return $ret;
    }

    /**
     * Uninstall the module.
     */
    public function uninstall()
    {
        require_code('files');

        $langs = find_all_langs(true);
        foreach (array_keys($langs) as $lang) {
            deldir_contents(get_custom_file_base() . '/caches/lang/' . $lang, true);
            // lang_custom purposely left
        }
    }

    public $title;

    /**
     * Module pre-run function. Allows us to know metadata for <head> before we start streaming output.
     *
     * @return ?Tempcode Tempcode indicating some kind of exceptional output (null: none)
     */
    public function pre_run() : ?object
    {
        require_code('form_templates'); // Needs to run high so that the anti-click-hacking header is sent

        require_code('input_filter_2');
        if (get_value('disable_modsecurity_workaround') !== '1') {
            modsecurity_workaround_enable();
        }

        $type = get_param_string('type', 'browse');

        require_lang('lang');
        require_code('lang2');
        require_code('lang3');
        require_code('lang_compile');

        set_helper_panel_tutorial('tut_intl');

        if ($type == 'criticise') {
            $this->title = get_screen_title('CRITICISE_LANGUAGE_PACK');
        }

        if ($type == 'browse') {
            if (get_param_string('lang', '') == '') {
                set_helper_panel_text(comcode_lang_string('DOC_FIND_LANG_STRING_TIP'));
            }

            $lang = filter_naughty_harsh(get_param_string('lang', ''));
            $lang_new = get_param_string('lang_new', $lang);
            if ($lang_new != '') {
                $lang = $lang_new;
            }
            $lang_file = get_param_string('lang_file', '');

            if (($lang == '') || ($lang_file == '')) {
                breadcrumb_set_self(do_lang_tempcode('TRANSLATE_CODE'));

                $this->title = get_screen_title('TRANSLATE_CODE');
            } else {
                $search = get_param_string('search', '', INPUT_FILTER_GET_COMPLEX);
                breadcrumb_set_parents([['_SELF:_SELF:browse', do_lang_tempcode('TRANSLATE_CODE')]]);
                if ($search != '') {
                    breadcrumb_set_self(do_lang_tempcode('RESULTS'));

                    $this->title = get_screen_title('TRANSLATE_CODE');
                } else {
                    breadcrumb_set_self(do_lang_tempcode('FILE'));

                    $this->title = get_screen_title('_TRANSLATE_CODE', true, [escape_html($lang_file), escape_html(lookup_language_full_name($lang))]);
                }
            }
        }

        if (($type == 'content') || ($type == '_content')) {
            $this->title = get_screen_title('TRANSLATE_CONTENT');
        }

        if (($type == '_code') || ($type == '_code2')) {
            $this->title = get_screen_title('TRANSLATE_CODE');
        }

        return null;
    }

    /**
     * Execute the module.
     *
     * @return Tempcode The result of execution
     */
    public function run() : object
    {
        require_code('input_filter_2');
        rescue_shortened_post_request();

        require_css('translations_editor');

        $type = get_param_string('type', 'browse');

        if ($type == 'content') {
            return $this->interface_content();
        }
        if ($type == '_content') {
            return $this->set_lang_content();
        }
        if ($type == 'criticise') {
            return $this->criticise();
        }
        if ($type == 'browse') {
            return $this->interface_code();
        }
        if ($type == '_code') {
            return $this->set_lang_code();
        }
        if ($type == '_code2') {
            return $this->set_lang_code_2(); // This is a language string setter called from an external source. Strings may be from many different files
        }

        return new Tempcode();
    }

    /**
     * The UI to choose a language.
     *
     * @param  Tempcode $title The title to show when choosing a language
     * @param  boolean $choose_lang_file Whether to also choose a language file
     * @param  boolean $add_lang Whether the user may add a language
     * @param  mixed $text Text message to show, provided in HTML format (string or Tempcode)
     * @param  ID_TEXT $param_name The name of the parameter for specifying language
     * @return Tempcode The UI
     */
    public function choose_lang(object $title, bool $choose_lang_file = false, bool $add_lang = false, $text = '', string $param_name = 'lang') : object
    {
        $langs = new Tempcode();
        $langs->attach(create_selection_list_langs(get_param_string('lang', null), $add_lang));

        $fields = new Tempcode();

        if ($add_lang) {
            $set_name = 'language';
            $set_title = do_lang_tempcode('LANGUAGE');
            $field_set = alternate_fields_set__start($set_name);

            $field_set->attach(form_input_list(do_lang_tempcode('EXISTING'), do_lang_tempcode('DESCRIPTION_LANGUAGE'), $param_name, $langs, null, false, false));

            $field_set->attach(form_input_codename(do_lang_tempcode('NEW'), do_lang_tempcode('DESCRIPTION_NEW_LANG'), 'lang_new', '', false));

            $fields->attach(alternate_fields_set__end($set_name, $set_title, '', $field_set, true));
        } else {
            $fields->attach(form_input_list(do_lang_tempcode('LANGUAGE'), do_lang_tempcode('DESCRIPTION_LANGUAGE'), $param_name, $langs, null, false, false));
        }

        if ($choose_lang_file) {
            $set_name = 'language_file';
            $set_title = do_lang_tempcode('LANGUAGE_FILE');
            $field_set = alternate_fields_set__start($set_name);

            $lang_files = new Tempcode();
            $lang_files->attach(create_selection_list_lang_files(null, get_param_string('lang_file', null)));
            $field_set->attach(form_input_list(do_lang_tempcode('CODENAME'), do_lang_tempcode('DESCRIPTION_LANGUAGE_FILE'), 'lang_file', $lang_files, null, true));

            $field_set->attach(form_input_line(do_lang('SEARCH'), '', 'search', '', false));

            $fields->attach(alternate_fields_set__end($set_name, $set_title, '', $field_set, true));
        }

        $post_url = get_self_url(false, false, [], false, true);

        return do_template('FORM_SCREEN', [
            '_GUID' => 'ee6bdea3661cb4736173cac818a769e5',
            'GET' => true,
            'SKIP_WEBSTANDARDS' => true,
            'HIDDEN' => '',
            'SUBMIT_ICON' => 'buttons/proceed',
            'SUBMIT_NAME' => do_lang_tempcode('PROCEED'),
            'TITLE' => $title,
            'FIELDS' => $fields,
            'URL' => $post_url,
            'TEXT' => $text,
        ]);
    }

    /**
     * Finds equivalents for a given string, in a different language, by automatic searching of codes and content.
     *
     * @param  string $old The language string we are searching for the equivalent of
     * @param  LANGUAGE_NAME $lang The language we want an equivalent in
     * @return string The match (or blank if no match can be found)
     */
    public function find_lang_matches(string $old, string $lang) : string
    {
        // Search for pretranslated content
        $potentials = $GLOBALS['SITE_DB']->query_select('translate', ['id'], ['text_original' => $old, 'language' => get_site_default_lang()]);
        foreach ($potentials as $potential) {
            $test = $GLOBALS['SITE_DB']->query_select_value_if_there('translate', 'text_original', ['id' => $potential['id'], 'language' => $lang]);
            if ($test !== null) {
                return $test;
            }
        }

        // Search code strings
        global $LANGUAGE_STRINGS_CACHE;

        if (!array_key_exists(user_lang(), $LANGUAGE_STRINGS_CACHE)) {
            return '';
        }
        $finds = array_keys($LANGUAGE_STRINGS_CACHE[user_lang()], $old);
        foreach ($finds as $find) {
            if ((array_key_exists($lang, $LANGUAGE_STRINGS_CACHE)) && (array_key_exists($find, $LANGUAGE_STRINGS_CACHE[$lang]))) {
                return $LANGUAGE_STRINGS_CACHE[$lang][$find];
            }
        }

        return '';
    }

    /**
     * The UI to criticise a language pack.
     *
     * @return Tempcode The UI
     */
    public function criticise() : object
    {
        $lang = get_param_string('crit_lang', '');
        if ($lang == '') {
            return $this->choose_lang($this->title, false, false, do_lang_tempcode('CHOOSE_CRITICISE_LIST_LANG_FILE'), 'crit_lang');
        }

        $files = '';

        $missing = [];

        if (fallback_lang() == $lang) {
            warn_exit(do_lang_tempcode('CANNOT_CRITICISE_BASE_LANG'));
        }

        $lang_files_base = get_lang_files(fallback_lang());
        $lang_files_criticise = get_lang_files($lang);

        $file = new Tempcode();
        foreach (array_keys($lang_files_base) as $file_base) {
            if (array_key_exists($file_base, $lang_files_criticise)) {
                // Process this file
                $base_map = get_lang_file_map(fallback_lang(), $file_base, true);
                $criticise_map = get_lang_file_map($lang, $file_base);

                foreach ($base_map as $key => $val) {
                    if (array_key_exists($key, $criticise_map)) {
                        $is = $criticise_map[$key];

                        // Perhaps we have a parameter mismatch?
                        if (strpos($val, '{3}') !== false) {
                            $num = 3;
                        } elseif (strpos($val, '{2}') !== false) {
                            $num = 2;
                        } elseif (strpos($val, '{1}') !== false) {
                            $num = 1;
                        } else {
                            $num = 0;
                        }

                        if (strpos($is, '{3}') !== false) {
                            $num_is = 3;
                        } elseif (strpos($is, '{2}') !== false) {
                            $num_is = 2;
                        } elseif (strpos($is, '{1}') !== false) {
                            $num_is = 1;
                        } else {
                            $num_is = 0;
                        }

                        if ($num_is != $num) {
                            $crit = do_template('TRANSLATE_LANGUAGE_CRITICISM', [
                                '_GUID' => '424388712f07bde0a04d89b0f349a0de',
                                'CRITICISM' => do_lang_tempcode('CRITICISM_PARAMETER_COUNT_MISMATCH', escape_html($key), escape_html($val)),
                            ]);
                            $file->attach($crit);
                        }

                        unset($criticise_map[$key]);
                    } elseif (trim($val) != '') {
                        $crit = do_template('TRANSLATE_LANGUAGE_CRITICISM', [
                            '_GUID' => '1c06d1d7c26ed73787eef6bfd912f57a',
                            'CRITICISM' => do_lang_tempcode('CRITICISM_MISSING_STRING', escape_html($key), escape_html($val)),
                        ]);
                        $file->attach($crit);
                    }
                }

                foreach ($criticise_map as $key => $val) {
                    $crit = do_template('TRANSLATE_LANGUAGE_CRITICISM', [
                        '_GUID' => '550018f24c0f677c50cd1bba96f24cc8',
                        'CRITICISM' => do_lang_tempcode('CRITICISM_EXTRA_STRING', escape_html($key)),
                    ]);
                    $file->attach($crit);
                }
            } else {
                $missing[] = $file_base;
            }

            if (!$file->is_empty()) {
                $file_result = do_template('TRANSLATE_LANGUAGE_CRITICISE_FILE', ['_GUID' => '925ae4a8dc34fed864c3072734a9abe5', 'COMPLAINTS' => $file, 'FILENAME' => $file_base]);
                $files .= $file_result->evaluate();/*Conserve memory*/
            }
        }

        if (!empty($missing)) {
            foreach ($missing as $missed) {
                $crit = do_template('TRANSLATE_LANGUAGE_CRITICISM', ['_GUID' => 'c19b1e83b5119495b52baf942e829336', 'CRITICISM' => do_lang_tempcode('CRITICISM_MISSING_FILE', escape_html($missed))]);
                $file->attach($crit);
            }
            $file_result = do_template('TRANSLATE_LANGUAGE_CRITICISE_FILE', ['_GUID' => '4ffab9265ea8c5a5e99a7b9fb23d15e1', 'COMPLAINTS' => $file, 'FILENAME' => do_lang_tempcode('NA_EM')]);
            $files .= $file_result->evaluate();/*Conserve memory*/
        }

        return do_template('TRANSLATE_LANGUAGE_CRITICISE_SCREEN', ['_GUID' => '62d6f40ca69609a8fd33704a8a38fb6f', 'TITLE' => $this->title, 'FILES' => $files]);
    }

    /**
     * The UI to translate content language strings.
     *
     * @return Tempcode The UI
     */
    public function interface_content() : object
    {
        if (!multi_lang()) {
            warn_exit(do_lang_tempcode('MULTILANG_OFF_PRACTICAL'));
        }

        if (!multi_lang_content()) {
            warn_exit(do_lang_tempcode('MULTILANG_OFF_CONTENT'));
        }

        push_query_limiting(false);

        $max = get_param_integer('max', 100);
        $start = get_param_integer('start', 0);

        $lang = choose_language($this->title);
        if (is_object($lang)) {
            return $lang;
        }

        // Find what we haven't translated
        switch (get_param_string('keep_translate_sort', 'priority')) {
            case 'recent':
                $order_by = 'a.id DESC';
                break;

            case 'priority':
            default:
                $order_by = 'a.importance_level,a.id DESC';
                break;
        }
        $subquery = 'SELECT * FROM ' . get_table_prefix() . 'translate b WHERE a.id=b.id AND b.broken=0 AND ' . db_string_equal_to('b.language', $lang);
        $query = 'SELECT id,importance_level,text_original,language FROM ' . get_table_prefix() . 'translate a WHERE NOT EXISTS(' . $subquery . ') AND ' . db_string_not_equal_to('language', $lang) . ' AND ' . db_string_not_equal_to('text_original', '') . ' ORDER BY ' . $order_by;
        $to_translate = $GLOBALS['SITE_DB']->query($query, $max/*reasonable limit*/, $start);
        $total = $GLOBALS['SITE_DB']->query_value_if_there('SELECT COUNT(*) AS count FROM (' . $query . ') AS subquery');
        if (empty($to_translate)) {
            inform_exit(do_lang_tempcode('NOTHING_TO_TRANSLATE'));
        }

        require_all_lang($lang, true);
        require_all_open_lang_files($lang);

        // Make our translation page
        require_code('translation');
        $has_translation = has_translation(get_site_default_lang(), $lang);
        if ($has_translation) {
            $translation_credit = get_translation_credit(get_site_default_lang(), $lang);
        } else {
            $translation_credit = '';
        }
        $lines = '';
        $actions = make_string_tempcode('&nbsp;');
        $last_level = null;
        $too_many = (count($to_translate) == $max);
        $ids_to_lookup = [];
        foreach ($to_translate as $it) {
            $ids_to_lookup[] = $it['id'];
        }
        $names = find_lang_content_names($ids_to_lookup);
        $_to_translate = [];
        foreach ($to_translate as $i => $it) {
            if ($it['importance_level'] == 0) {
                continue; // Corrupt data
            }

            $id = $it['id'];
            $name = $names[$id];
            if ($name === null) {
                continue; // Orphaned string
            }

            $_to_translate[] = $it;
        }
        $to_translate = $_to_translate;
        foreach ($to_translate as $i => $it) {
            $old = $it['text_original'];
            $current = $this->find_lang_matches($old, $lang);
            $priority = ($last_level === $it['importance_level']) ? null : do_lang('PRIORITY_' . strval($it['importance_level']));

            $id = $it['id'];
            $name = $names[$id];

            if ($has_translation) {
                $actions = do_template('TRANSLATE_ACTION', [
                    '_GUID' => 'f625cf15c9db5e5af30fc772a7f0d5ff',
                    'LANG_FROM' => $it['language'],
                    'LANG_TO' => $lang,
                    'NAME' => 'trans-' . strval($id),
                    'OLD' => $old,
                ]);
            }

            check_suhosin_request_quantity(2, strlen('trans_' . $name));

            $line = do_template('TRANSLATE_LINE_CONTENT', [
                '_GUID' => '87a0f5298ce9532839f3206cd0e06051',
                'NAME' => $name,
                'ID' => strval($id),
                'OLD' => $old,
                'CURRENT' => $current,
                'ACTIONS' => $actions,
                'TRANSLATION_CREDIT' => $translation_credit,
                'PRIORITY' => $priority,
                'LAST' => !isset($to_translate[$i + 1]),
            ]);
            $lines .= $line->evaluate();

            $last_level = $it['importance_level'];
        }

        pop_query_limiting();

        $url = build_url(['page' => '_SELF', 'type' => '_content', 'lang' => $lang, 'start' => $start], '_SELF');

        $_GET['lang'] = $lang;
        require_code('templates_pagination');
        $pagination = pagination(do_lang_tempcode('TRANSLATE_CONTENT'), $start, 'start', $max, 'max', $total, true);

        return do_template('TRANSLATE_SCREEN_CONTENT_SCREEN', [
            '_GUID' => 'af732c5e595816db1c6f025c4b8fa6a2',
            'MAX' => integer_format($max),
            'TOTAL' => integer_format($total - $max),
            'LANG_ORIGINAL_NAME' => get_site_default_lang(),
            'LANG_NICE_ORIGINAL_NAME' => lookup_language_full_name(get_site_default_lang()),
            'LANG_NICE_NAME' => lookup_language_full_name($lang),
            'TOO_MANY' => $too_many,
            'TRANSLATION_CREDIT' => $translation_credit,
            'LANG' => $lang,
            'LINES' => $lines,
            'TITLE' => $this->title,
            'URL' => $url,
            'PAGINATION' => $pagination,
        ]);
    }

    /**
     * The actualiser to translate content language strings.
     *
     * @return Tempcode The UI
     */
    public function set_lang_content() : object
    {
        $lang = choose_language($this->title);
        if (is_object($lang)) {
            return $lang;
        }

        foreach ($_POST as $key => $val) {
            if (!is_string($val)) {
                continue;
            }
            if ((is_string($key)) && (substr($key, 0, 6) != 'trans_')) {
                continue;
            }

            $lang_id = intval(substr($key, strlen('trans_')));

            if ($val != '') {
                $GLOBALS['SITE_DB']->query_delete('translate', ['language' => $lang, 'id' => $lang_id], '', 1);
                $importance_level = $GLOBALS['SITE_DB']->query_select_value_if_there('translate', 'importance_level', ['id' => $lang_id]);
                if ($importance_level !== null) {
                    $GLOBALS['SITE_DB']->query_insert('translate', ['id' => $lang_id, 'source_user' => get_member(), 'language' => $lang, 'importance_level' => $importance_level, 'text_original' => $val, 'text_parsed' => '', 'broken' => 0]);
                }
            }
        }

        log_it('TRANSLATE_CONTENT');

        require_code('caches3');
        erase_block_cache();
        erase_persistent_cache();

        if (get_param_integer('contextual', 0) == 1) {
            return inform_screen($this->title, do_lang_tempcode('SUCCESS_SAVE'));
        }

        // Show it worked / Refresh
        $url = post_param_string('redirect', '', INPUT_FILTER_URL_INTERNAL);
        if ($url == '') {
            $_url = build_url(['page' => '_SELF', 'type' => 'content', 'lang' => $lang, 'start' => get_param_integer('start', null)], '_SELF');
            $url = $_url->evaluate();
        }
        return redirect_screen($this->title, $url, do_lang_tempcode('SUCCESS_SAVE'));
    }

    /**
     * The UI to translate code.
     *
     * @return Tempcode The UI
     */
    public function interface_code() : object
    {
        $lang = filter_naughty_harsh(get_param_string('lang', ''));
        $lang_new = get_param_string('lang_new', $lang);
        if ($lang_new != '') {
            require_code('type_sanitisation');
            if (!is_alphanumeric($lang_new, true)) {
                warn_exit(do_lang_tempcode('BAD_CODENAME'));
            }

            if (strlen($lang_new) > 5) {
                warn_exit(do_lang_tempcode('INVALID_LANG_CODE'));
            }
            $lang = $lang_new;
        }

        $lang_file = get_param_string('lang_file', '');

        // UI to pick a language and lang file
        if (($lang == '') || ($lang_file == '')) {
            $choose_message = do_lang_tempcode(
                'CHOOSE_EDIT_LIST_LANG_FILE',
                escape_html(get_site_default_lang()),
                escape_html(lookup_language_full_name(get_site_default_lang())),
                [
                    get_base_url() . '/code_editor.php',
                    escape_html(user_lang()),
                    escape_html(lookup_language_full_name(user_lang())),
                ]
            );
            return $this->choose_lang($this->title, true, true, $choose_message);
        }

        // Search feature
        $search = get_param_string('search', '', INPUT_FILTER_GET_COMPLEX);
        if ($search != '') {
            $search = trim($search, '" ');

            require_lang($lang_file);
            require_all_open_lang_files();

            $fields = new Tempcode();
            global $LANGUAGE_STRINGS_CACHE;
            foreach ($LANGUAGE_STRINGS_CACHE[user_lang()] as $key => $value) {
                if ((stripos($value, $search) !== false) || (stripos($key, $search) !== false)) {
                    $fields->attach(form_input_text($key, '', 'trans_' . $key, str_replace('\n', "\n", $value), false, false));
                }
            }
            if ($fields->is_empty()) {
                inform_exit(do_lang_tempcode('NO_ENTRIES'));
            }
            $post_url = build_url(['page' => '_SELF', 'type' => '_code2'], '_SELF');
            $hidden = new Tempcode();
            $hidden->attach(form_input_hidden('redirect', static_evaluate_tempcode(protect_url_parameter(SELF_REDIRECT))));
            $hidden->attach(form_input_hidden('lang', $lang));
            return do_template('FORM_SCREEN', [
                '_GUID' => '2d7356fd2c4497ceb19450e65331c9c5',
                'TITLE' => $this->title,
                'HIDDEN' => $hidden,
                'FIELDS' => $fields,
                'URL' => $post_url,
                'TEXT' => '',
                'SUBMIT_ICON' => 'buttons/save',
                'SUBMIT_NAME' => do_lang('TRANSLATE_CODE'),
            ]);
        }

        // Upgrade to custom if not there yet (or maybe we are creating a new lang - same difference)
        $custom_dir = get_custom_file_base() . '/lang_custom/' . $lang;
        if (!file_exists($custom_dir)) {
            appengine_live_guard();

            // Work out what paths need to be writable...

            $writable_paths = ['lang_custom', 'lang_custom/' . $lang];

            $cached_dir = get_custom_file_base() . '/caches/lang/' . $lang;
            if (!file_exists($cached_dir)) {
                $writable_paths[] = 'caches/lang';
            }

            $zones = find_all_zones();

            require_code('themes2');
            $themes = find_all_themes();

            foreach ($zones as $zone) {
                $_special_dir = get_custom_file_base() . (($zone == '') ? '' : '/') . $zone . '/pages/comcode_custom/' . $lang;
                if (!file_exists($_special_dir)) {
                    $writable_paths[] = $zone . (($zone == '') ? '' : '/') . 'pages/comcode_custom';
                }
                $_special_dir = get_custom_file_base() . (($zone == '') ? '' : '/') . $zone . '/pages/html_custom/' . $lang;
                if (!file_exists($_special_dir)) {
                    $writable_paths[] = $zone . (($zone == '') ? '' : '/') . 'pages/html_custom';
                }
            }

            foreach (array_keys($themes) as $theme) {
                $_special_dir = get_custom_file_base() . '/themes/' . $theme . '/templates_cached/' . $lang;
                if (!file_exists($_special_dir)) {
                    $writable_paths[] = 'themes/' . $theme . '/templates_cached';
                }
            }

            // Connect the AFM...

            require_code('abstract_file_manager');
            force_have_afm_details($writable_paths);

            // Do the work...

            afm_make_directory('lang_custom/' . $lang, true);
            afm_make_file('lang_custom/' . $lang . '/index.html', '', false);

            if (!file_exists($cached_dir)) {
                afm_make_directory('caches/lang/' . $lang, true);
                afm_make_file('caches/lang/' . $lang . '/index.html', '', false);
            }

            // Make Comcode page dirs
            foreach ($zones as $zone) {
                $_special_dir = get_custom_file_base() . (($zone == '') ? '' : '/') . $zone . '/pages/comcode_custom/' . $lang;
                if (!file_exists($_special_dir)) {
                    afm_make_directory($zone . (($zone == '') ? '' : '/') . 'pages/comcode_custom/' . $lang, true);
                    afm_make_file($zone . (($zone == '') ? '' : '/') . 'pages/comcode_custom/' . $lang . '/index.html', '', false);
                }
                $_special_dir = get_custom_file_base() . (($zone == '') ? '' : '/') . $zone . '/pages/html_custom/' . $lang;
                if (!file_exists($_special_dir)) {
                    afm_make_directory($zone . (($zone == '') ? '' : '/') . 'pages/html_custom/' . $lang, true);
                    afm_make_file($zone . (($zone == '') ? '' : '/') . 'pages/html_custom/' . $lang . '/index.html', '', false);
                }
            }

            // Make templates_cached dirs
            foreach (array_keys($themes) as $theme) {
                $_special_dir = get_custom_file_base() . '/themes/' . $theme . '/templates_cached/' . $lang;
                if (!file_exists($_special_dir)) {
                    afm_make_directory('themes/' . $theme . '/templates_cached/' . $lang, true);
                    afm_make_file('themes/' . $theme . '/templates_cached/' . $lang . '/index.html', '', false);
                }
            }
        }

        // Get some stuff
        $for_lang = get_lang_file_map($lang, $lang_file);
        $for_base_lang = get_lang_file_map(fallback_lang(), $lang_file, true);
        $descriptions = get_lang_file_section(fallback_lang(), $lang_file);

        $max = get_param_integer('max', 30);
        $start = get_param_integer('start', 0);

        // Make our translation page
        $lines = '';
        require_code('translation');
        $has_translation = has_translation(fallback_lang(), $lang);
        if ($has_translation) {
            $translation_credit = get_translation_credit(fallback_lang(), $lang);
        } else {
            $translation_credit = '';
        }
        $i = 0;
        $actions = new Tempcode();
        foreach ($for_base_lang + $for_lang as $name => $old) {
            if (($i < $start) || ($i >= $start + $max)) {
                $i++;
                continue;
            }

            if (array_key_exists($name, $for_lang)) {
                $current = $for_lang[$name];
            } else {
                $current = '';//$this->find_lang_matches($old, $lang); Too slow / useless for code translation
            }
            $description = array_key_exists($name, $descriptions) ? $descriptions[$name] : '';
            $_current = str_replace('\n', "\n", $current);
            if ($_current == '') {
                $_current = str_replace('\n', "\n", $old);
            }

            if ($has_translation) {
                $actions = do_template('TRANSLATE_ACTION', [
                    '_GUID' => '9e9a68cb2c1a1e23a901b84c9af2280b',
                    'LANG_FROM' => get_site_default_lang(),
                    'LANG_TO' => $lang,
                    'NAME' => 'trans-' . $name,
                    'OLD' => $_current,
                ]);
            }

            $temp = do_template('TRANSLATE_LINE', [
                '_GUID' => '9cb331f5852ee043e6ad30b45aedc43b',
                'DESCRIPTION' => $description,
                'NAME' => $name,
                'OLD' => str_replace('\n', "\n", $old),
                'CURRENT' => $_current,
                'ACTIONS' => $actions,
            ]);
            $lines .= $temp->evaluate();

            $i++;
        }

        $url = build_url(['page' => '_SELF', 'type' => '_code', 'lang_file' => $lang_file, 'lang' => $lang, 'start' => $start, 'max' => $max, 'max_rows' => $i], '_SELF');

        require_code('templates_pagination');
        $pagination = pagination(do_lang_tempcode('LANGUAGE_STRINGS'), $start, 'start', $max, 'max', $i);

        return do_template('TRANSLATE_SCREEN', [
            '_GUID' => 'b3429f8bd0b4eb79c33709ca43e3207c',
            'PAGE' => $lang_file,
            'TRANSLATION_CREDIT' => $translation_credit,
            'LANG' => $lang,
            'LINES' => $lines,
            'TITLE' => $this->title,
            'URL' => $url,
            'PAGINATION' => $pagination,
        ]);
    }

    /**
     * The actualiser to translate code (called from this module).
     *
     * @return Tempcode The UI
     */
    public function set_lang_code() : object
    {
        $lang = get_param_string('lang');
        $lang_file = get_param_string('lang_file');

        $max = get_param_integer('max');
        $start = get_param_integer('start');
        $max_rows = get_param_integer('max_rows');

        $this->_set_lang_code($lang_file, $lang);

        log_it('TRANSLATE_CODE');

        require_code('caches3');
        erase_cached_language();
        erase_cached_templates(false, null, TEMPLATE_DECACHE_WITH_LANG);
        persistent_cache_delete('LANGS_LIST');
        erase_block_cache();
        delete_cache_entry('side_language');

        // Show it worked / Refresh
        if ($start + $max < $max_rows) {
            $url = build_url(['page' => '_SELF', 'type' => 'browse', 'lang' => $lang, 'lang_file' => $lang_file, 'start' => $start + $max, 'max' => ($max == 30) ? null : $max], '_SELF');
        } else {
            $url = build_url(['page' => '_SELF', 'type' => 'browse', 'lang' => $lang], '_SELF');
        }
        return redirect_screen($this->title, $url, do_lang_tempcode('SUCCESS_SAVE'));
    }

    /**
     * The actualiser to translate code (called externally, and may operate on many lang files).
     *
     * @return Tempcode The UI
     */
    public function set_lang_code_2() : object
    {
        $lang = post_param_string('lang');

        $lang_files = get_lang_files(fallback_lang());
        foreach (array_keys($lang_files) as $lang_file) {
            $this->_set_lang_code($lang_file, $lang);
        }

        log_it('TRANSLATE_CODE');

        require_code('caches3');
        erase_cached_language();
        erase_cached_templates(false, null, TEMPLATE_DECACHE_WITH_LANG);
        persistent_cache_delete('LANGS_LIST');
        erase_block_cache();

        // Show it worked / Refresh
        $url = post_param_string('redirect', '', INPUT_FILTER_URL_INTERNAL);
        if ($url == '') {
            return inform_screen($this->title, do_lang_tempcode('SUCCESS_SAVE'));
        }
        return redirect_screen($this->title, $url, do_lang_tempcode('SUCCESS_SAVE'));
    }

    /**
     * Save POSTed changes to a particular language file.
     *
     * @param  string $lang_file The language file
     * @param  LANGUAGE_NAME $lang The language
     */
    protected function _set_lang_code(string $lang_file, string $lang)
    {
        $for_base_lang_orig = get_lang_file_map(fallback_lang(), $lang_file, true);
        $for_base_lang_override = get_lang_file_map($lang, $lang_file, false);

        $descriptions = get_lang_file_section(fallback_lang(), $lang_file);
        $runtime_processing = get_lang_file_section(fallback_lang(), $lang_file, 'runtime_processing');

        $no_default_file = (!file_exists(get_file_base() . '/lang/' . fallback_lang() . '/' . $lang_file . '.ini')); // whole file is not in core software

        $out = '';
        $one_changed_from_saved = false;
        foreach ($for_base_lang_override + $for_base_lang_orig as $key => $disk_val) {
            $_disk_val = str_replace('\n', "\n", $disk_val);

            $val = post_param_string('trans_' . $key, $_disk_val);

            $changed_from_saved = ($val != $_disk_val); // was already changed in language file
            $not_a_default = (!array_key_exists($key, $for_base_lang_orig)); // not in core software
            $changed_from_default = !$not_a_default && ($for_base_lang_orig[$key] != $val); // changed from default

            if ($changed_from_saved || $not_a_default || $changed_from_default || $no_default_file) {
                $out .= $key . '=' . str_replace("\n", '\n', $val) . "\n";
            }

            if ($changed_from_saved) {
                $one_changed_from_saved = true;
            }
        }

        if (($out != '') && $one_changed_from_saved) {
            $path = get_custom_file_base() . '/lang_custom/' . filter_naughty($lang) . '/' . filter_naughty($lang_file) . '.ini';

            // Backup
            $path_backup = $path . '.' . strval(time());
            if (file_exists($path)) {
                @copy($path, $path_backup) or intelligent_write_error($path_backup);
                fix_permissions($path_backup);
                sync_file($path_backup);
            }

            // Generate overall output (includes metadata)
            $contents = '';
            if (!empty($descriptions)) {
                $contents .= "[descriptions]\n";
                foreach ($descriptions as $key => $description) {
                    $contents .= $key . '=' . $description . "\n";
                }
                $contents .= "\n";
            }
            if (!empty($runtime_processing)) {
                $contents .= "[runtime_processing]\n";
                foreach ($runtime_processing as $key => $flag) {
                    $contents .= $key . '=' . $flag . "\n";
                }
                $contents .= "\n";
            }
            $contents .= "[strings]\n";
            $contents .= $out;

            // Save to disk
            require_code('files');
            cms_file_put_contents_safe($path, $contents, FILE_WRITE_FIX_PERMISSIONS | FILE_WRITE_SYNC_FILE | FILE_WRITE_BOM);
            $path_backup2 = $path . '.latest_in_cms_edit';
            @copy($path, $path_backup2) or intelligent_write_error($path_backup2);
            fix_permissions($path_backup2);
            sync_file($path_backup2);
        }
    }
}
