<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    news
 */

require_code('crud_module');

/**
 * Module page class.
 */
class Module_cms_news extends Standard_crud_module
{
    protected $lang_type = 'NEWS';
    protected $select_name = 'TITLE';
    protected $code_require = 'news';
    protected $permissions_require = 'high';
    protected $permissions_module_require = 'news';
    protected $permissions_cat_name = 'main_news_category';
    protected $user_facing = true;
    protected $seo_type = 'news';
    protected $content_type = 'news';
    protected $possibly_some_kind_of_upload = true;
    protected $upload = 'image';
    protected $menu_label = 'NEWS';
    protected $table = 'news';
    protected $orderer = 'title';
    protected $supports_mass_delete = true;

    protected $donext_type = null;

    /**
     * Find entry-points available within this module.
     *
     * @param  boolean $check_perms Whether to check permissions
     * @param  ?MEMBER $member_id The member to check permissions as (null: current user)
     * @param  boolean $support_crosslinks Whether to allow cross links to other modules (identifiable via a full-page-link rather than a screen-name)
     * @param  boolean $be_deferential Whether to avoid any entry-point (or even return null to disable the page in the Sitemap) if we know another module, or page_group, is going to link to that entry-point. Note that "!" and "browse" entry points are automatically merged with container page nodes (likely called by page-groupings) as appropriate.
     * @return ?array A map of entry points (screen-name=>language-code/string or screen-name=>[language-code/string, icon-theme-image]) (null: disabled)
     */
    public function get_entry_points(bool $check_perms = true, ?int $member_id = null, bool $support_crosslinks = true, bool $be_deferential = false) : ?array
    {
        if (!addon_installed('news')) {
            return null;
        }

        $this->cat_crud_module = class_exists('Mx_cms_news_cat') ? new Mx_cms_news_cat() : new Module_cms_news_cat();

        if ($member_id === null) {
            $member_id = get_member();
        }

        $ret = [
            'browse' => ['MANAGE_NEWS', 'menu/rich_content/news'],
        ] + parent::get_entry_points();

        if ($support_crosslinks) {
            require_code('fields');
            $ret += manage_custom_fields_entry_points('news_category') + manage_custom_fields_entry_points('news');
        }

        if (has_privilege($member_id, 'mass_import')) {
            $ret += [
                'predefined_content' => ['PREDEFINED_CONTENT', 'admin/import'],
            ];
        }

        return $ret;
    }

    /**
     * Find privileges defined as overridable by this module.
     *
     * @return array A map of privileges that are overridable; privilege to 0 or 1. 0 means "not category overridable". 1 means "category overridable".
     */
    public function get_privilege_overrides() : array
    {
        require_lang('news');
        return [
            'view_private_content' => 0,

            'mass_import' => 0,

            'have_personal_category' => [0, 'HAVE_BLOG'],

            'submit_cat_highrange_content' => [0, 'ADD_NEWS_CATEGORY'],
            'edit_own_cat_highrange_content' => [0, 'EDIT_OWN_NEWS_CATEGORY'],
            'edit_cat_highrange_content' => [0, 'EDIT_NEWS_CATEGORY'],
            'delete_own_cat_highrange_content' => [0, 'DELETE_OWN_NEWS_CATEGORY'],
            'delete_cat_highrange_content' => [0, 'DELETE_NEWS_CATEGORY'],

            'submit_highrange_content' => [1, 'ADD_NEWS'],
            'bypass_validation_highrange_content' => [1, 'BYPASS_NEWS_VALIDATION'],
            'edit_own_highrange_content' => [1, 'EDIT_OWN_NEWS'],
            'edit_highrange_content' => [1, 'EDIT_NEWS'],
            'delete_own_highrange_content' => [1, 'DELETE_OWN_NEWS'],
            'delete_highrange_content' => [1, 'DELETE_NEWS'],

            // These are set as non-overridable. Technically they are, but that makes for confusion!
            //  Best to just keep the idea that only a blog owner or some vague 'staff' can edit a member's blog
            //   i.e. no need to make this editable per-blog! Just use a regular news category if you need that control.
            'submit_midrange_content' => [0, 'ADD_NEWS_BLOG'],
            'bypass_validation_midrange_content' => [0, 'BYPASS_NEWS_BLOG_VALIDATION'],
            'edit_own_midrange_content' => [0, 'EDIT_OWN_NEWS_BLOG'],
            'edit_midrange_content' => [0, 'EDIT_NEWS_BLOG'],
            'delete_own_midrange_content' => [0, 'DELETE_OWN_NEWS_BLOG'],
            'delete_midrange_content' => [0, 'DELETE_NEWS_BLOG'],
        ];
    }

    public $title;

    /**
     * Module pre-run function. Allows us to know metadata for <head> before we start streaming output.
     *
     * @param  boolean $top_level Whether this is running at the top level, prior to having sub-objects called
     * @param  ?ID_TEXT $type The screen type to consider for metadata purposes (null: read from environment)
     * @return ?Tempcode Tempcode indicating some kind of exceptional output (null: none)
     */
    public function pre_run(bool $top_level = true, ?string $type = null) : ?object
    {
        $error_msg = new Tempcode();
        if (!addon_installed__messaged('news', $error_msg)) {
            return $error_msg;
        }

        if (!addon_installed('news_shared')) {
            warn_exit(do_lang_tempcode('MISSING_ADDON', escape_html('news_shared')));
        }

        $this->cat_crud_module = class_exists('Mx_cms_news_cat') ? new Mx_cms_news_cat() : new Module_cms_news_cat();

        $type = get_param_string('type', 'browse');

        require_lang('news');

        inform_non_canonical_parameter('cat');

        set_helper_panel_tutorial('tut_news');

        if ($type == 'add' || $type == '_edit') {
            set_helper_panel_text(comcode_lang_string('DOC_NEWS'));
        }

        if ($type == 'import' || $type == '_import_news') {
            $this->title = get_screen_title('IMPORT_NEWS');
        }

        if ($type == '_import_news') {
            breadcrumb_set_parents([['_SELF:_SELF:import', do_lang_tempcode('IMPORT_NEWS')]]);
            breadcrumb_set_self(do_lang_tempcode('DONE'));
        }

        if ($type == 'predefined_content') {
        }

        if ($type == '_predefined_content') {
            breadcrumb_set_parents([['_SELF:_SELF:browse', do_lang_tempcode('MANAGE_NEWS')], ['_SELF:_SELF:predefined_content', do_lang_tempcode('PREDEFINED_CONTENT')]]);
            breadcrumb_set_self(do_lang_tempcode('DONE'));
        }

        if ($type == 'predefined_content' || $type == '_predefined_content') {
            $this->title = get_screen_title('PREDEFINED_CONTENT');
        }

        return parent::pre_run($top_level);
    }

    /**
     * Standard crud_module run_start.
     *
     * @param  ID_TEXT $type The type of module execution
     * @return Tempcode The output of the run
     */
    public function run_start(string $type) : object
    {
        $this->posting_form_title = do_lang_tempcode('NEWS_ARTICLE');

        require_code('input_filter_2');
        modsecurity_workaround_enable();

        require_code('news2');
        require_css('news');

        // Decide what to do
        if ($type == 'browse') {
            return $this->browse();
        }
        if ($type == 'import') {
            return $this->import_news();
        }
        if ($type == '_import_news') {
            return $this->_import_news();
        }
        if ($type == 'predefined_content') {
            return $this->predefined_content();
        }
        if ($type == '_predefined_content') {
            return $this->_predefined_content();
        }

        return new Tempcode();
    }

    /**
     * The do-next manager for before content management.
     *
     * @return Tempcode The UI
     */
    public function browse() : object
    {
        require_code('templates_donext');
        require_code('fields');
        return do_next_manager(
            get_screen_title('MANAGE_NEWS'),
            comcode_lang_string('DOC_NEWS'),
            array_merge([
                has_privilege(get_member(), 'submit_cat_highrange_content', 'cms_news') ? ['admin/add_one_category', ['_SELF', ['type' => 'add_category'], '_SELF'], do_lang('ADD_NEWS_CATEGORY')] : null,
                has_privilege(get_member(), 'edit_own_cat_highrange_content', 'cms_news') ? ['admin/edit_one_category', ['_SELF', ['type' => 'edit_category'], '_SELF'], do_lang('EDIT_NEWS_CATEGORY')] : null,
                has_privilege(get_member(), 'submit_highrange_content', 'cms_news') ? ['admin/add', ['_SELF', ['type' => 'add'], '_SELF'], do_lang('ADD_NEWS')] : null,
                has_privilege(get_member(), 'edit_own_highrange_content', 'cms_news') ? ['admin/edit', ['_SELF', ['type' => 'edit'], '_SELF'], do_lang('EDIT_NEWS')] : null,
                has_privilege(get_member(), 'mass_import', 'cms_news') ? ['admin/import', ['_SELF', ['type' => 'import'], '_SELF'], do_lang('IMPORT_NEWS')] : null,
                has_privilege(get_member(), 'mass_import') ? ['admin/install', ['_SELF', ['type' => 'predefined_content'], '_SELF'], do_lang('PREDEFINED_CONTENT')] : null,
            ], manage_custom_fields_donext_link('news')),
            do_lang('MANAGE_NEWS')
        );
    }

    /**
     * Standard crud_module table function.
     *
     * @param  array $url_map Details to go to build_url for link to the next screen
     * @return array A quartet: The choose table, Whether re-ordering is supported from this screen, Search URL, Archive URL
     */
    public function create_selection_list_choose_table(array $url_map) : array
    {
        require_code('templates_results_table');

        $current_ordering = get_param_string('sort', 'date_and_time DESC', INPUT_FILTER_GET_COMPLEX);
        $sortables = [
            'title' => do_lang_tempcode('TITLE'),
            'news_category' => do_lang_tempcode('MAIN_CATEGORY'),
            'date_and_time' => do_lang_tempcode('ADDED'),
            'news_views' => do_lang_tempcode('COUNT_VIEWS'),
            'submitter' => do_lang_tempcode('metadata:OWNER'),
        ];
        list($sql_sort, $sort_order, $sortable) = process_sorting_params('news', $current_ordering);

        $fh = [do_lang_tempcode('TITLE'), do_lang_tempcode('MAIN_CATEGORY')];
        $fh[] = do_lang_tempcode('ADDED');
        $fh[] = do_lang_tempcode('COUNT_VIEWS');
        if (addon_installed('validation')) {
            $fh[] = protect_from_escaping(do_template('COMCODE_ABBR', ['_GUID' => '5e5269c25a2eb44fd92b554a50f6007f', 'TITLE' => do_lang_tempcode('VALIDATED'), 'CONTENT' => do_lang_tempcode('VALIDATED_SHORT')]));
        }
        $fh[] = do_lang_tempcode('metadata:OWNER');
        $fh[] = do_lang_tempcode('ACTIONS');
        $header_row = results_header_row($fh, $sortables, 'sort', $sortable . ' ' . $sort_order);

        $result_entries = new Tempcode();

        $only_owned = has_privilege(get_member(), 'edit_highrange_content', 'cms_news') ? null : get_member();
        list($rows, $max_rows) = $this->get_entry_rows(false, $sql_sort, ($only_owned === null) ? [] : ['submitter' => $only_owned]);
        $news_cat_titles = [];
        foreach ($rows as $row) {
            $edit_url = build_url($url_map + ['id' => $row['id']], '_SELF');

            $fr = [];
            $fr[] = protect_from_escaping(hyperlink(build_url(['page' => 'news', 'type' => 'view', 'id' => $row['id']], get_module_zone('news')), get_translated_text($row['title']), false, true));
            if (array_key_exists($row['news_category'], $news_cat_titles)) {
                $nc_title = $news_cat_titles[$row['news_category']];
            } else {
                $nc_title = $GLOBALS['SITE_DB']->query_select_value_if_there('news_categories', 'nc_title', ['id' => $row['news_category']]);
                $news_cat_titles[$row['news_category']] = $nc_title;
            }
            if ($nc_title !== null) {
                $fr[] = protect_from_escaping(hyperlink(build_url(['page' => 'news', 'type' => 'browse', 'select' => $row['news_category']], get_module_zone('news')), get_translated_text($nc_title), false, true));
            } else {
                $fr[] = do_lang('UNKNOWN');
            }
            $fr[] = get_timezoned_date_time($row['date_and_time']);
            $fr[] = integer_format($row['news_views']);
            if (addon_installed('validation')) {
                $fr[] = ($row['validated'] == 1) ? do_lang_tempcode('YES') : do_lang_tempcode('NO');
            }
            $username = protect_from_escaping($GLOBALS['FORUM_DRIVER']->member_profile_hyperlink($row['submitter']));
            $fr[] = $username;
            $fr[] = protect_from_escaping(hyperlink($edit_url, do_lang_tempcode('EDIT'), false, true, do_lang('EDIT') . ' #' . strval($row['id'])));

            $result_entries->attach(results_entry($fr, true));
        }

        $search_url = build_url(['page' => 'search', 'id' => 'news'], get_module_zone('search'));
        $archive_url = build_url(['page' => 'news'], get_module_zone('news'));

        return [results_table(do_lang($this->menu_label), get_param_integer('start', 0), 'start', either_param_integer('max', 20), 'max', $max_rows, $header_row, $result_entries, $sortables, $sortable, $sort_order), false, $search_url, $archive_url];
    }

    /**
     * Standard crud_module list function.
     *
     * @return Tempcode The selection list
     */
    public function create_selection_list_entries() : object
    {
        $only_owned = has_privilege(get_member(), 'edit_highrange_content', 'cms_news') ? null : get_member();
        return create_selection_list_news(null, $only_owned, false);
    }

    /**
     * Get Tempcode for an adding form.
     *
     * @return mixed Either Tempcode; or a tuple of: (fields, hidden-fields[, delete-fields][, edit-text][, whether all delete fields are specified][, posting form text, more fields][, parsed WYSIWYG editable text])
     */
    public function get_form_fields_for_add()
    {
        return $this->get_form_fields();
    }

    /**
     * Get Tempcode for a news adding/editing form.
     *
     * @param  ?AUTO_LINK $id The news ID (null: new)
     * @param  ?AUTO_LINK $main_news_category The primary category for the news (null: personal)
     * @param  ?array $news_category A list of categories the news is in (null: not known)
     * @param  SHORT_TEXT $title The news title
     * @param  LONG_TEXT $news The news summary
     * @param  SHORT_TEXT $author The name of the author
     * @param  BINARY $validated Whether the news is validated
     * @param  ?BINARY $allow_rating Whether rating is allowed (null: decide statistically, based on existing choices)
     * @param  ?SHORT_INTEGER $allow_comments Whether comments are allowed (0=no, 1=yes, 2=review style) (null: decide statistically, based on existing choices)
     * @param  ?BINARY $allow_trackbacks Whether trackbacks are allowed (null: decide statistically, based on existing choices)
     * @param  BINARY $send_trackbacks Whether to show the "send trackback" field
     * @param  LONG_TEXT $notes Notes for the video
     * @param  URLPATH $image URL to the image for the news entry (blank: use cat image)
     * @param  ?array $scheduled Scheduled go-live time (null: N/A)
     * @param  array $regions The regions (empty: not region-limited)
     * @return array A tuple: The input fields, Hidden fields, ...
     */
    public function get_form_fields(?int $id = null, ?int $main_news_category = null, ?array $news_category = null, string $title = '', string $news = '', string $author = '', int $validated = 1, ?int $allow_rating = null, ?int $allow_comments = null, ?int $allow_trackbacks = null, int $send_trackbacks = 1, string $notes = '', string $image = '', ?array $scheduled = null, array $regions = []) : array
    {
        if ($id === null) {
            // Cloning support
            $id = get_param_integer('id', null);
            if ($id !== null) {
                if (method_exists($this, 'get_submitter')) {
                    list($submitter) = $this->get_submitter(strval($id));
                } else {
                    $submitter = null;
                }

                if ($this->permissions_require !== null) {
                    check_edit_permission($this->permissions_require, $submitter, [$this->permissions_module_require, ($this->permissions_cat_name === null) ? null : $this->get_cat(strval($id))], $this->privilege_page_name);
                }

                $ret = $this->fill_in_edit_form(strval($id));

                $ret[2] = null;
                $ret[3] = null;
                $ret[4] = null; // These strictly only relate to edits

                $this->posting_form_text = $ret[5];
                $ret[5] = null;

                $this->posting_form_text_parsed = $ret[7];
                $ret[7] = null;

                return $ret;
            }
        }

        list($allow_rating, $allow_comments, $allow_trackbacks) = $this->choose_feedback_fields_statistically($allow_rating, $allow_comments, $allow_trackbacks);

        if ($main_news_category === null) {
            $param_cat = get_param_string('cat', '');
            if ($param_cat == '') {
                $news_category = [];
                $main_news_category = null;
            } elseif (strpos($param_cat, ',') === false) {
                $news_category = [];
                $main_news_category = intval($param_cat);
            } else {
                require_code('selectcode');
                $_param_cat = explode(',', $param_cat);
                $_main_news_category = array_shift($_param_cat);
                $param_cat = implode(',', $_param_cat);
                $main_news_category = ($_main_news_category == '') ? null : intval($_main_news_category);
                $news_category = selectcode_to_idlist_using_db($param_cat, 'id', 'news_categories', 'news_categories', null, 'id', 'id');
            }

            $author = $GLOBALS['FORUM_DRIVER']->get_username(get_member());
        }

        $cats1 = create_selection_list_news_categories($main_news_category, false, true, is_integer($main_news_category), null, true);
        if ($cats1->is_empty()) {
            warn_exit(do_lang_tempcode('NO_CATEGORIES', 'news_category'));
        }
        $cats2 = create_selection_list_news_categories(($news_category === null) ? [] : $news_category, false, true, is_integer($main_news_category), null, true);

        $fields = new Tempcode();
        $fields2 = new Tempcode();
        $fields->attach(form_input_line_comcode(do_lang_tempcode('TITLE'), do_lang_tempcode('DESCRIPTION_TITLE'), 'title', $title, true));
        $fields->attach(form_input_list(do_lang_tempcode('MAIN_CATEGORY'), do_lang_tempcode('DESCRIPTION_MAIN_CATEGORY'), 'main_news_category', $cats1));
        if (addon_installed('authors')) {
            $fields->attach(form_input_author(do_lang_tempcode('SOURCE'), do_lang_tempcode('DESCRIPTION_SOURCE'), 'author', $author, true));
        }

        require_code('feedback2');

        $posting_form_tabindex = get_form_field_tabindex(null);

        $_validated = get_param_integer('validated', 0);
        if ($validated == 0) {
            if (($_validated == 1) && (addon_installed('validation'))) {
                $validated = 1;
                attach_message(do_lang_tempcode('WILL_BE_VALIDATED_WHEN_SAVING'), 'notice');
            }
        } elseif (($validated == 1) && ($_validated == 1) && ($id !== null)) {
            $action_log = build_url(['page' => 'admin_actionlog', 'type' => 'list', 'to_type' => 'VALIDATE_NEWS', 'param_a' => strval($id)]);
            attach_message(do_lang_tempcode('ALREADY_VALIDATED', escape_html($action_log->evaluate())), 'warn');
        }
        if (has_some_cat_privilege(get_member(), 'bypass_validation_' . $this->permissions_require . 'range_content', null, $this->permissions_module_require)) {
            if (addon_installed('validation')) {
                $fields2->attach(form_input_tick(do_lang_tempcode('VALIDATED'), do_lang_tempcode($GLOBALS['FORUM_DRIVER']->is_super_admin(get_member()) ? 'DESCRIPTION_VALIDATED_SIMPLE' : 'DESCRIPTION_VALIDATED', 'news'), 'validated', $validated == 1));
            }
        }

        $fields2->attach(do_template('FORM_SCREEN_FIELD_SPACER', ['_GUID' => '90e0f1f4557eb78d58b9a13c3e1e65dc', 'SECTION_HIDDEN' => $news == '' && $image == '' && ($scheduled === null) && ($title == ''/*=new entry and selected news cats was from URL*/ || (empty($news_category))), 'TITLE' => do_lang_tempcode('ADVANCED')]));

        $news_summary_required = (get_option('news_summary_required') == '1');
        $_summary_field = form_input_text_comcode(do_lang_tempcode('NEWS_SUMMARY'), $news_summary_required ? new Tempcode() : do_lang_tempcode('DESCRIPTION_NEWS_SUMMARY'), 'news', $news, $news_summary_required);
        if ($news_summary_required) {
            $fields->attach($_summary_field);
        } else {
            $fields2->attach($_summary_field);
        }

        $fields2->attach(form_input_multi_list(do_lang_tempcode('SECONDARY_CATEGORIES'), do_lang_tempcode('DESCRIPTION_SECONDARY_CATEGORIES', 'news'), 'news_category', $cats2));

        if (get_option('filter_regions') == '1') {
            require_code('locations');
            $fields2->attach(form_input_regions($regions));
        }

        $hidden = new Tempcode();

        if (get_value('disable_news_repimages') !== '1') {
            require_code('images');
            $fields2->attach(form_input_upload_multi_source(do_lang_tempcode('REPRESENTATIVE_IMAGE'), do_lang_tempcode('DESCRIPTION_NEWS_IMAGE_OVERRIDE'), $hidden, 'image', 'icons/news', true, $image, false, null, IMAGE_CRITERIA_WEBSAFE));
        }

        if ((addon_installed('calendar')) && (addon_installed('commandr')) && (has_privilege(get_member(), 'scheduled_publication_times'))) {
            $fields2->attach(form_input_date__cron(do_lang_tempcode('PUBLICATION_TIME'), do_lang_tempcode('DESCRIPTION_PUBLICATION_TIME'), 'schedule', false, ($scheduled === null), true, $scheduled, intval(date('Y')) - 1970 + 2, 1970));
        }

        require_code('syndication');
        $fields2->attach(get_syndication_option_fields('news', $id !== null));

        // Metadata
        require_code('content2');
        $seo_fields = seo_get_fields($this->seo_type, ($id === null) ? null : strval($id), false);
        require_code('feedback2');
        $feedback_fields = feedback_fields($this->content_type, $allow_rating == 1, $allow_comments == 1, $allow_trackbacks == 1, $send_trackbacks == 1, $notes, $allow_comments == 2, false, true, false);
        $fields2->attach(metadata_get_fields('news', ($id === null) ? null : strval($id), false, [], ($seo_fields->is_empty() && $feedback_fields->is_empty()) ? METADATA_HEADER_YES : METADATA_HEADER_FORCE));
        $fields2->attach($seo_fields);
        $fields2->attach($feedback_fields);

        if (addon_installed('content_privacy')) {
            require_code('content_privacy2');
            if ($id === null) {
                $fields2->attach(get_privacy_form_fields('news'));
            } else {
                $fields2->attach(get_privacy_form_fields('news', strval($id)));
            }
        }

        if (addon_installed('content_reviews')) {
            $fields2->attach(content_review_get_fields('news', ($id === null) ? null : strval($id)));
        }

        return [$fields, $hidden, null, null, null, null, make_string_tempcode($fields2->evaluate()), $posting_form_tabindex];
    }

    /**
     * Standard crud_module submitter getter.
     *
     * @param  ID_TEXT $id The entry for which the submitter is sought
     * @return array The submitter, and the time of submission (null submission time implies no known submission time)
     */
    public function get_submitter(string $id) : array
    {
        $rows = $GLOBALS['SITE_DB']->query_select('news', ['submitter', 'date_and_time'], ['id' => intval($id)], '', 1);
        if (!array_key_exists(0, $rows)) {
            return [null, null];
        }
        return [$rows[0]['submitter'], $rows[0]['date_and_time']];
    }

    /**
     * Standard crud_module category getter.
     *
     * @param  ID_TEXT $id The entry for which the category is sought
     * @return string The category
     */
    public function get_cat(string $id) : string
    {
        $temp = $GLOBALS['SITE_DB']->query_select_value_if_there('news', 'news_category', ['id' => intval($id)]);
        if ($temp === null) {
            warn_exit(do_lang_tempcode('MISSING_RESOURCE', 'news'));
        }
        return strval($temp);
    }

    /**
     * Standard crud_module edit form filler.
     *
     * @param  ID_TEXT $_id The entry being edited
     * @return mixed Either Tempcode; or a tuple of: (fields, hidden-fields[, delete-fields][, edit-text][, whether all delete fields are specified][, posting form text, more fields][, parsed WYSIWYG editable text])
     */
    public function fill_in_edit_form(string $_id)
    {
        $id = intval($_id);

        require_lang('menus');
        require_lang('zones');

        $rows = $GLOBALS['SITE_DB']->query_select('news', ['*'], ['id' => intval($id)], '', 1);
        if (!array_key_exists(0, $rows)) {
            warn_exit(do_lang_tempcode('MISSING_RESOURCE', 'news'));
        }
        $myrow = $rows[0];

        $cat = $myrow['news_category'];

        $categories = [];
        $category_query = $GLOBALS['SITE_DB']->query_select('news_category_entries', ['news_entry_category'], ['news_entry' => $id]);

        foreach ($category_query as $value) {
            $categories[] = $value['news_entry_category'];
        }

        $scheduled = null;

        if (addon_installed('calendar')) {
            require_code('calendar2');
            $scheduled = get_schedule_code_event_time('publish_news', strval($id));
        } else {
            $scheduled = null;
        }

        $regions = collapse_1d_complexity('region', $GLOBALS['SITE_DB']->query_select('content_regions', ['region'], ['content_type' => 'news', 'content_id' => strval($id)]));

        $ret = $this->get_form_fields($id, $cat, $categories, get_translated_text($myrow['title']), get_translated_text($myrow['news']), $myrow['author'], $myrow['validated'], $myrow['allow_rating'], $myrow['allow_comments'], $myrow['allow_trackbacks'], 0, $myrow['notes'], $myrow['news_image_url'], $scheduled, $regions);

        $ret[2] = new Tempcode();
        $ret[3] = '';
        $ret[4] = false;
        $ret[5] = get_translated_text($myrow['news_article']);
        $ret[7] = get_translated_tempcode('news', $myrow, 'news_article');
        return $ret;
    }

    /**
     * Standard crud_module add actualiser.
     *
     * @return array A pair: The entry added, description about usage
     */
    public function add_actualisation() : array
    {
        $author = post_param_string('author', $GLOBALS['FORUM_DRIVER']->get_username(get_member()));
        $news = post_param_string('news');
        $title = post_param_string('title');
        $validated = post_param_integer('validated', 0);
        $news_article = post_param_string('post');
        if (post_param_string('main_news_category') != 'personal') {
            $main_news_category = post_param_integer('main_news_category');
        } else {
            $main_news_category = null;
        }

        $news_category = [];
        if (array_key_exists('news_category', $_POST)) {
            foreach ($_POST['news_category'] as $val) {
                $news_category[] = ($val == 'personal') ? null : intval($val);
            }
        }

        $allow_rating = post_param_integer('allow_rating', 0);
        $allow_comments = post_param_integer('allow_comments', 0);
        $allow_trackbacks = post_param_integer('allow_trackbacks', 0);
        $notes = post_param_string('notes', '');

        require_code('images2');
        $image = post_param_image('image', 'uploads/repimages', 'icons/news', false);

        $schedule = post_param_date('schedule');
        if ((addon_installed('calendar')) && (addon_installed('commandr')) && (has_privilege(get_member(), 'scheduled_publication_times')) && ($schedule !== null) && ($schedule > time())) {
            $validated = 0;
        } else {
            $schedule = null;
        }

        if ($main_news_category !== null) {
            $owner = $GLOBALS['SITE_DB']->query_select_value('news_categories', 'nc_owner', ['id' => intval($main_news_category)]);
            if (($owner !== null) && ($owner != get_member())) {
                check_privilege('can_submit_to_others_categories', ['news', $main_news_category]);
            }
        }

        $metadata = actual_metadata_get_fields('news', null);

        $regions = isset($_POST['regions']) ? $_POST['regions'] : [];

        $id = add_news($title, $news, $author, $validated, $allow_rating, $allow_comments, $allow_trackbacks, $notes, $news_article, $main_news_category, $news_category, $metadata['add_time'], $metadata['submitter'], $metadata['views'], null, null, $image, '', '', $regions);

        require_code('feedback2');
        send_trackbacks(post_param_string('send_trackbacks', ''), $title, $news);

        set_url_moniker('news', strval($id));

        $main_news_category = $GLOBALS['SITE_DB']->query_select_value('news', 'news_category', ['id' => $id]);
        $this->donext_type = $main_news_category;

        if (addon_installed('content_privacy')) {
            require_code('content_privacy2');
            list($privacy_level, $additional_access) = read_privacy_fields();
            save_privacy_form_fields('news', strval($id), $privacy_level, $additional_access);
        }

        if (($validated == 1) || (!addon_installed('validation'))) {
            $is_blog = ($GLOBALS['SITE_DB']->query_select_value('news_categories', 'nc_owner', ['id' => $main_news_category]) !== null);

            require_code('users2');
            if (has_actual_page_access(get_modal_user(), 'news')) { // NB: no category permission check, as syndication choice was explicit, and news categorisation is a bit more complex
                $privacy_ok = true;
                if (addon_installed('content_privacy')) {
                    require_code('content_privacy');
                    $privacy_ok = has_privacy_access('news', strval($id), $GLOBALS['FORUM_DRIVER']->get_guest_id(), $title);
                }
                if ($privacy_ok) {
                    require_code('syndication');
                    syndicate_content('news', strval($id), [[$is_blog ? 'news:ACTIVITY_ADD_NEWS_BLOG' : 'news:ACTIVITY_ADD_NEWS', get_member()]]);
                }
            }
        }

        if ($schedule !== null) {
            if (!addon_installed('calendar')) {
                warn_exit(do_lang_tempcode('INTERNAL_ERROR', escape_html('d71b815fb8c550dcb21b2b79b29cbec2')));
            }

            require_code('calendar');

            $parameters = [];
            $start_year = intval(date('Y', $schedule));
            $start_month = intval(date('m', $schedule));
            $start_day = intval(date('d', $schedule));
            $start_hour = intval(date('H', $schedule));
            $start_minute = intval(date('i', $schedule));
            require_code('calendar2');
            schedule_code('publish_news', strval($id), $parameters, do_lang('PUBLISH_NEWS', $title), $start_year, $start_month, $start_day, $start_hour, $start_minute);
        }

        if (addon_installed('content_reviews')) {
            content_review_set('news', strval($id));
        }

        return [strval($id), null];
    }

    /**
     * Standard crud_module edit actualiser.
     *
     * @param  ID_TEXT $_id The entry being edited
     * @return ?Tempcode Description about usage (null: none)
     */
    public function edit_actualisation(string $_id) : ?object
    {
        $id = intval($_id);

        $validated = post_param_integer('validated', fractional_edit() ? INTEGER_MAGIC_NULL : 0);

        $news_article = post_param_string('post', STRING_MAGIC_NULL);
        if (post_param_string('main_news_category', null) !== 'personal') {
            $main_news_category = post_param_integer('main_news_category', fractional_edit() ? INTEGER_MAGIC_NULL : false);
        } else {
            warn_exit(do_lang_tempcode('INTERNAL_ERROR', escape_html('c43c41ad87e15bc58145aa3871927b0b')));
        }

        $news_category = [];
        if (array_key_exists('news_category', $_POST)) {
            foreach ($_POST['news_category'] as $val) {
                $news_category[] = intval($val);
            }
        }
        if (fractional_edit()) {
            $news_category = null;
        }

        $allow_rating = post_param_integer('allow_rating', fractional_edit() ? INTEGER_MAGIC_NULL : 0);
        $allow_comments = post_param_integer('allow_comments', fractional_edit() ? INTEGER_MAGIC_NULL : 0);
        $allow_trackbacks = post_param_integer('allow_trackbacks', fractional_edit() ? INTEGER_MAGIC_NULL : 0);
        $notes = post_param_string('notes', STRING_MAGIC_NULL);

        $this->donext_type = $main_news_category;

        if (!fractional_edit()) {
            require_code('images2');
            $image = post_param_image('image', 'uploads/repimages', 'icons/news', false);
        } else {
            $image = STRING_MAGIC_NULL;
        }

        $owner = $GLOBALS['SITE_DB']->query_select_value_if_there('news_categories', 'nc_owner', ['id' => $main_news_category]); // if_there in case somehow category setting corrupted
        if (($owner !== null) && ($owner != get_member())) {
            check_privilege('can_submit_to_others_categories', ['news', $main_news_category]);
        }

        $schedule = post_param_date('schedule');

        if ((addon_installed('calendar')) && (addon_installed('commandr')) && (!fractional_edit()) && (has_privilege(get_member(), 'scheduled_publication_times'))) {
            require_code('calendar2');
            unschedule_code('publish_news', strval($id));

            if (($schedule !== null) && ($schedule > time())) {
                $validated = 0;

                $parameters = [];
                $start_year = intval(date('Y', $schedule));
                $start_month = intval(date('m', $schedule));
                $start_day = intval(date('d', $schedule));
                $start_hour = intval(date('H', $schedule));
                $start_minute = intval(date('i', $schedule));
                schedule_code('publish_news', strval($id), $parameters, do_lang('PUBLISH_NEWS', post_param_string('title')), $start_year, $start_month, $start_day, $start_hour, $start_minute);
            }
        }

        $title = post_param_string('title', STRING_MAGIC_NULL);

        if (addon_installed('content_privacy')) {
            require_code('content_privacy2');
            list($privacy_level, $additional_access) = read_privacy_fields();
            save_privacy_form_fields('news', strval($id), $privacy_level, $additional_access);
        }

        if (($validated == 1) || (!addon_installed('validation'))) {
            require_code('users2');
            if (has_actual_page_access(get_modal_user(), 'news')) {
                $privacy_ok = true;
                if (addon_installed('content_privacy')) {
                    require_code('content_privacy');
                    $privacy_ok = has_privacy_access('news', strval($id), $GLOBALS['FORUM_DRIVER']->get_guest_id());
                }
                if ($privacy_ok) {
                    $just_validated = (addon_installed('validation')) && ($GLOBALS['SITE_DB']->query_select_value('news', 'validated', ['id' => $id]) == 0);
                    $is_blog = ($GLOBALS['SITE_DB']->query_select_value('news_categories', 'nc_owner', ['id' => $main_news_category]) !== null);
                    $submitter = $GLOBALS['SITE_DB']->query_select_value('news', 'submitter', ['id' => $id]);

                    $activities = [];
                    if ($just_validated) {
                        if ($submitter != get_member()) {
                            $activities[] = [($is_blog ? 'news:ACTIVITY_VALIDATE_NEWS_BLOG' : 'news:ACTIVITY_VALIDATE_NEWS'), get_member(), $title];
                        }
                        $activities[] = [($is_blog ? 'news:ACTIVITY_ADD_NEWS_BLOG' : 'news:ACTIVITY_ADD_NEWS'), $submitter, $title];
                    }

                    require_code('syndication');
                    syndicate_content('news', strval($id), $activities);
                }
            }
        }

        $metadata = actual_metadata_get_fields('news', strval($id));

        $regions = isset($_POST['regions']) ? $_POST['regions'] : [];

        edit_news($id, $title, post_param_string('news', STRING_MAGIC_NULL), post_param_string('author', STRING_MAGIC_NULL), $validated, $allow_rating, $allow_comments, $allow_trackbacks, $notes, $news_article, $main_news_category, $news_category, post_param_string('meta_keywords', STRING_MAGIC_NULL), post_param_string('meta_description', STRING_MAGIC_NULL), $image, $metadata['add_time'], $metadata['edit_time'], $metadata['views'], $metadata['submitter'], $regions, true);

        if (addon_installed('content_reviews')) {
            content_review_set('news', strval($id));
        }

        return null;
    }

    /**
     * Standard crud_module delete actualiser.
     *
     * @param  ID_TEXT $_id The entry being deleted
     */
    public function delete_actualisation(string $_id)
    {
        $id = intval($_id);

        delete_news($id);

        if (addon_installed('content_privacy')) {
            require_code('content_privacy2');
            delete_privacy_form_fields('news', strval($id));
        }

        require_code('syndication');
        unsyndicate_content('news', strval($id));
    }

    /**
     * The do-next manager for after download content management (events only).
     *
     * @param  ?Tempcode $title The title (output of get_screen_title) (null: don't use full page)
     * @param  Tempcode $description Some description to show, saying what happened
     * @param  ?ID_TEXT $id The ID of whatever we are working with (null: deleted)
     * @return Tempcode The UI
     */
    public function do_next_manager(?object $title, object $description, ?string $id = null) : object
    {
    return $this->cat_crud_module->_do_next_manager($title, $description, ($id === null) ? null : intval($id), $this->donext_type/*category*/);
    }

    /**
     * The UI to import news.
     *
     * @return Tempcode The UI
     */
    public function import_news() : object
    {
        check_privilege('mass_import');

        $lang = user_lang__with__translation_override(true);

        $post_url = build_url(['page' => '_SELF', 'type' => '_import_news', 'old_type' => get_param_string('type', '')], '_SELF');
        $submit_name = do_lang_tempcode('IMPORT_NEWS');

        require_code('news2');
        $fields = import_rss_fields(false);

        $hidden = new Tempcode();
        $hidden->attach(form_input_hidden('lang', $lang));
        handle_max_file_size($hidden);

        return do_template('FORM_SCREEN', [
            '_GUID' => '4ac8c667fa38c1e6338eedcb138e7fd4',
            'TITLE' => $this->title,
            'TEXT' => do_lang_tempcode('IMPORT_NEWS_TEXT'),
            'HIDDEN' => $hidden,
            'FIELDS' => $fields,
            'SUBMIT_ICON' => 'admin/import',
            'SUBMIT_NAME' => $submit_name,
            'URL' => $post_url,
        ]);
    }

    /**
     * The actualiser to import news.
     *
     * @return Tempcode The UI
     */
    public function _import_news() : object
    {
        check_privilege('mass_import');

        $is_validated = post_param_integer('auto_validate', 0);
        if (!addon_installed('validation')) {
            $is_validated = 1;
        }
        $download_images = post_param_integer('download_images', 0);
        if (!has_privilege(get_member(), 'draw_to_server')) {
            $download_images = 0;
        }
        $to_own_account = post_param_integer('add_to_own', 0);
        if (!$GLOBALS['FORUM_DRIVER']->is_super_admin(get_member())) {
            $to_own_account = 1;
        }
        $import_blog_comments = post_param_integer('import_blog_comments', 0);
        $import_to_blog = post_param_integer('import_to_blog', 0);

        require_code('uploads');
        is_plupload(true);
        if (((array_key_exists('file_anytype', $_FILES)) && ((is_plupload()) || (is_uploaded_file($_FILES['file_anytype']['tmp_name']))))) {
            $rss_feed = get_temporary_upload_path('file_anytype');
            $is_filesystem_path = true;
        } else {
            $rss_feed = post_param_string('rss_feed_url', null, INPUT_FILTER_URL_GENERAL);
            $is_filesystem_path = false;
        }
        if ($rss_feed === null) {
            warn_exit(do_lang_tempcode('IMPROPERLY_FILLED_IN'));
        }

        require_code('rss');
        $rss = new CMS_RSS($rss_feed, $is_filesystem_path);

        // Cleanup
        if ($is_filesystem_path) { // Means it is a temp file
            @unlink($rss_feed);
        }

        require_code('tasks');
        $ret = call_user_func_array__long_task(do_lang('IMPORT_NEWS'), $this->title, 'import_rss', [$is_validated, $download_images, $to_own_account, $import_blog_comments, $import_to_blog, $rss]);

        return $ret;
    }

    /**
     * UI for install/uninstall of predefined content.
     *
     * @return Tempcode The UI
     */
    public function predefined_content() : object
    {
        require_code('content2');
        return predefined_content_changes_ui('news', $this->title, build_url(['page' => '_SELF', 'type' => '_predefined_content'], '_SELF'));
    }

    /**
     * Actualise install/uninstall of predefined content.
     *
     * @return Tempcode The UI
     */
    public function _predefined_content() : object
    {
        require_code('content2');
        return predefined_content_changes_actualiser('news', $this->title);
    }
}

/**
 * Module page class.
 */
class Module_cms_news_cat extends Standard_crud_module
{
    protected $lang_type = 'NEWS_CATEGORY';
    protected $select_name = 'TITLE';
    protected $permissions_require = 'cat_high';
    protected $permission_module = 'news';
    protected $menu_label = 'NEWS';
    protected $table = 'news_categories';
    protected $orderer = 'nc_title';
    protected $is_chained_with_parent_browse = true;
    protected $do_preview = null;

    /**
     * Standard crud_module table function.
     *
     * @param  array $url_map Details to go to build_url for link to the next screen
     * @return array A pair: The choose table, Whether re-ordering is supported from this screen
     */
    public function create_selection_list_choose_table(array $url_map) : array
    {
        require_code('templates_results_table');

        $current_ordering = get_param_string('sort', 'nc_title ASC', INPUT_FILTER_GET_COMPLEX);
        $sortables = [
            'nc_title' => do_lang_tempcode('TITLE'),
            'entries_count' => do_lang_tempcode('COUNT_TOTAL'),
        ];
        list($sql_sort, $sort_order, $sortable) = process_sorting_params('news_category', $current_ordering);

        $columns = [];
        $columns[] = do_lang_tempcode('TITLE');
        if (get_value('disable_news_repimages') !== '1') {
            $columns[] = do_lang_tempcode('IMAGE');
        }
        $columns[] = do_lang_tempcode('COUNT_TOTAL_PRIMARY');
        $columns[] = do_lang_tempcode('COUNT_TOTAL');
        $columns[] = do_lang_tempcode('ACTIONS');
        $header_row = results_header_row($columns, $sortables, 'sort', $sortable . ' ' . $sort_order);

        $fields = new Tempcode();

        list($rows, $max_rows) = $this->get_entry_rows(false, $sql_sort);
        foreach ($rows as $row) {
            $edit_url = build_url($url_map + ['id' => $row['id']], '_SELF');

            $total_primary = $GLOBALS['SITE_DB']->query_select_value('news', 'COUNT(*)', ['news_category' => $row['id']]);
            $total = $total_primary + $GLOBALS['SITE_DB']->query_select_value('news_category_entries', 'COUNT(*)', ['news_entry_category' => $row['id']]);

            $values = [];
            $values[] = protect_from_escaping(hyperlink(build_url(['page' => 'news', 'type' => 'browse', 'select' => $row['id']], get_module_zone('news')), get_translated_text($row['nc_title']), false, true));
            if (get_value('disable_news_repimages') !== '1') {
                $values[] = protect_from_escaping(($row['nc_img'] == '') ? '' : ('<img width="15" alt="' . do_lang('IMAGE') . '" src="' . escape_html(get_news_category_image_url($row['nc_img'])) . '" />')); // XHTMLXHTML
            }
            $values[] = integer_format($total_primary);
            $values[] = integer_format($total);
            $values[] = protect_from_escaping(hyperlink($edit_url, do_lang_tempcode('EDIT'), false, true, do_lang('EDIT') . ' #' . strval($row['id'])));

            $fields->attach(results_entry($values, true));
        }

        return [results_table(do_lang($this->menu_label), get_param_integer('start', 0), 'start', either_param_integer('max', 20), 'max', $max_rows, $header_row, $fields, $sortables, $sortable, $sort_order), false];
    }

    /**
     * Standard crud_module list function.
     *
     * @return Tempcode The selection list
     */
    public function create_selection_list_entries() : object
    {
        return create_selection_list_news_categories(null, false, false, true);
    }

    /**
     * Get Tempcode for an adding form.
     *
     * @return mixed Either Tempcode; or a tuple of: (fields, hidden-fields[, delete-fields][, edit-text][, whether all delete fields are specified][, posting form text, more fields][, parsed WYSIWYG editable text])
     */
    public function get_form_fields_for_add()
    {
        return $this->get_form_fields();
    }

    /**
     * Get Tempcode for a news category adding/editing form.
     *
     * @param  ?AUTO_LINK $id The news category ID (null: new)
     * @param  SHORT_TEXT $title The title of the news category
     * @param  SHORT_TEXT $img The news category image
     * @param  LONG_TEXT $notes Notes relating to the news category
     * @param  ?MEMBER $owner The owner of the news category (null: public)
     * @param  ?AUTO_LINK $category_id The ID of this news category (null: we haven't added it yet)
     * @return array A pair: The input fields, Hidden fields
     */
    public function get_form_fields(?int $id = null, string $title = '', string $img = '', string $notes = '', ?int $owner = null, ?int $category_id = null) : array
    {
        $fields = new Tempcode();
        $hidden = new Tempcode();

        $fields->attach(form_input_line_comcode(do_lang_tempcode('TITLE'), do_lang_tempcode('DESCRIPTION_TITLE'), 'title', $title, true));

        if (get_value('disable_news_repimages') !== '1') {
            require_code('images');
            $fields->attach(form_input_upload_multi_source(do_lang_tempcode('IMAGE'), '', $hidden, 'image', 'icons/news', false, $img, false, null, IMAGE_CRITERIA_WEBSAFE));
        }

        if (get_option('enable_staff_notes') == '1') {
            $fields->attach(do_template('FORM_SCREEN_FIELD_SPACER', ['_GUID' => 'b88f1b286b05e18991ad51538812f7b2', 'SECTION_HIDDEN' => $notes == '', 'TITLE' => do_lang_tempcode('ADVANCED')]));
            $fields->attach(form_input_text(do_lang_tempcode('NOTES'), do_lang_tempcode('DESCRIPTION_NOTES'), 'notes', $notes, false, false));
        }

        $fields->attach(metadata_get_fields('news_category', ($id === null) ? null : strval($id)), false);

        if (addon_installed('content_reviews')) {
            $fields->attach(content_review_get_fields('news_category', ($id === null) ? null : strval($id)));
        }

        // Permissions
        $fields->attach($this->get_permission_fields(($category_id === null) ? '' : strval($category_id), null, ($title == '')));
        if (addon_installed('ecommerce')) {
            require_code('ecommerce_permission_products');
            $fields->attach(permission_product_form('news_category', ($category_id === null) ? null : strval($category_id)));
        }

        return [$fields, $hidden];
    }

    /**
     * Standard crud_module edit form filler.
     *
     * @param  ID_TEXT $_id The category being edited
     * @return mixed Either Tempcode; or a tuple of: (fields, hidden-fields[, delete-fields][, edit-text][, whether all delete fields are specified][, posting form text, more fields][, parsed WYSIWYG editable text])
     */
    public function fill_in_edit_form(string $_id)
    {
        $id = intval($_id);

        $rows = $GLOBALS['SITE_DB']->query_select('news_categories', ['*'], ['id' => $id], '', 1);
        if (!array_key_exists(0, $rows)) {
            warn_exit(do_lang_tempcode('MISSING_RESOURCE', 'news_category'));
        }
        $myrow = $rows[0];

        return $this->get_form_fields($id, get_translated_text($myrow['nc_title']), $myrow['nc_img'], $myrow['notes'], $myrow['nc_owner'], $myrow['id']);
    }

    /**
     * Standard crud_module add actualiser.
     *
     * @return array A pair: The category added, description about usage
     */
    public function add_actualisation() : array
    {
        $title = post_param_string('title');

        require_code('images2');
        $img = post_param_image('image', 'uploads/repimages', 'icons/news', false);

        $notes = post_param_string('notes', '');

        $metadata = actual_metadata_get_fields('news_category', null);

        $id = add_news_category($title, $img, $notes);

        set_url_moniker('news_category', strval($id));

        $this->set_permissions(strval($id));
        if (addon_installed('ecommerce')) {
            require_code('ecommerce_permission_products');
            permission_product_save('news_category', strval($id));
        }

        if (addon_installed('content_reviews')) {
            content_review_set('news_category', strval($id));
        }

        return [strval($id), null];
    }

    /**
     * Standard crud_module edit actualiser.
     *
     * @param  ID_TEXT $id The category being edited
     * @return ?Tempcode Description about usage (null: none)
     */
    public function edit_actualisation(string $id) : ?object
    {
        $title = post_param_string('title');

        if (!fractional_edit()) {
            require_code('images2');
            $img = post_param_image('image', 'uploads/repimages', 'icons/news', false, true);
        } else {
            $img = STRING_MAGIC_NULL;
        }

        $notes = post_param_string('notes', STRING_MAGIC_NULL);

        $metadata = actual_metadata_get_fields('news_category', $id);

        if ($metadata['submitter'] === null) { // We need to interpret this - if we didn't have specification permission, we need to copy through existing setting, as a null would imply a de-set
            $metadata['submitter'] = $GLOBALS['SITE_DB']->query_select_value_if_there('news_categories', 'nc_owner', ['id' => intval($id)]);
        }

        edit_news_category(intval($id), $title, $img, $notes, $metadata['submitter']);

        $this->set_permissions($id);
        if (addon_installed('ecommerce')) {
            require_code('ecommerce_permission_products');
            permission_product_save('news_category', $id);
        }

        if (addon_installed('content_reviews')) {
            content_review_set('news_category', $id);
        }

        return null;
    }

    /**
     * Standard crud_module submitter getter.
     *
     * @param  ID_TEXT $id The entry for which the submitter is sought
     * @return array The submitter, and the time of submission (null submission time implies no known submission time)
     */
    public function get_submitter(string $id) : array
    {
        return [$GLOBALS['SITE_DB']->query_select_value_if_there('news_categories', 'nc_owner', ['id' => intval($id)]), null];
    }

    /**
     * Standard crud_module delete actualiser.
     *
     * @param  ID_TEXT $id The category being deleted
     */
    public function delete_actualisation(string $id)
    {
        delete_news_category(intval($id));
    }

    /**
     * The do-next manager for after download content management (event types only).
     *
     * @param  ?Tempcode $title The title (output of get_screen_title) (null: don't use full page)
     * @param  Tempcode $description Some description to show, saying what happened
     * @param  ?ID_TEXT $id The ID of whatever we are working with (null: deleted)
     * @return Tempcode The UI
     */
    public function do_next_manager(?object $title, object $description, ?string $id = null) : object
    {
        return $this->_do_next_manager($title, $description, ($id === null) ? null : intval($id));
    }

    /**
     * The do-next manager for after news content management.
     *
     * @param  ?Tempcode $title The title (output of get_screen_title) (null: don't use full page)
     * @param  Tempcode $description Some description to show, saying what happened
     * @param  ?AUTO_LINK $id The ID of whatever was just handled (null: N/A)
     * @param  ?AUTO_LINK $cat The category ID we were working in (null: deleted)
     * @return Tempcode The UI
     */
    public function _do_next_manager(?object $title, object $description, ?int $id = null, ?int $cat = null) : object
    {
        require_code('templates_donext');

        if (($id === null) && ($cat === null)) {
            return do_next_manager(
                $title,
                $description,
                [],
                null,
                /* TYPED-ORDERED LIST OF 'LINKS' */
                ['_SELF', ['type' => 'add'], '_SELF'], // Add one
                null, // Edit this
                has_privilege(get_member(), 'edit_own_highrange_content', 'cms_news') ? ['_SELF', ['type' => 'edit'], '_SELF'] : null, // Edit one
                null, // View this
                ['news', ['type' => 'browse'], get_module_zone('news')], // View archive
                has_privilege(get_member(), 'submit_cat_highrange_content', 'cms_news') ? ['_SELF', ['type' => 'add_category'], '_SELF'] : null, // Add one category
                has_privilege(get_member(), 'edit_own_cat_highrange_content', 'cms_news') ? ['_SELF', ['type' => 'edit_category'], '_SELF'] : null, // Edit one category
                null, // Edit this category
                null, // View this category
                [],
                [],
                [],
                null,
                null,
                null,
                null,
                'news',
                'news_category'
            );
        }

        return do_next_manager(
            $title,
            $description,
            [],
            null,
            /* TYPED-ORDERED LIST OF 'LINKS' */
            ['_SELF', ['type' => 'add', 'cat' => $cat], '_SELF'], // Add one
            (($id === null) || (!has_privilege(get_member(), 'edit_own_highrange_content', 'cms_news', ['news', $cat]))) ? null : ['_SELF', ['type' => '_edit', 'id' => $id], '_SELF', do_lang_tempcode('EDIT_THIS_NEWS')], // Edit this
            has_privilege(get_member(), 'edit_own_highrange_content', 'cms_news') ? ['_SELF', ['type' => 'edit'], '_SELF'] : null, // Edit one
            ($id === null) ? null : ['news', ['type' => 'view', 'id' => $id], get_module_zone('news')], // View this
            ['news', ['type' => 'browse'], get_module_zone('news')], // View archive
            has_privilege(get_member(), 'submit_cat_highrange_content', 'cms_news') ? ['_SELF', ['type' => 'add_category'], '_SELF'] : null, // Add one category
            has_privilege(get_member(), 'edit_own_cat_highrange_content', 'cms_news') ? ['_SELF', ['type' => 'edit_category'], '_SELF'] : null, // Edit one category
            ($cat === null) ? null : (has_privilege(get_member(), 'edit_own_cat_highrange_content', 'cms_news') ? ['_SELF', ['type' => '_edit_category', 'id' => $cat], '_SELF', do_lang_tempcode('EDIT_THIS_NEWS_CATEGORY')] : null), // Edit this category
            null, // View this category
            [],
            [],
            [],
            null,
            null,
            null,
            null,
            'news',
            'news_category'
        );
    }
}
