<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    shopping
 */

/**
 * Module page class.
 */
class Module_shopping
{
    /**
     * Find details of the module.
     *
     * @return ?array Map of module info (null: module is disabled)
     */
    public function info() : ?array
    {
        $info = [];
        $info['author'] = 'Manuprathap';
        $info['organisation'] = 'Composr';
        $info['hacked_by'] = null;
        $info['hack_version'] = null;
        $info['version'] = 9;
        $info['update_require_upgrade'] = true;
        $info['locked'] = false;
        $info['min_cms_version'] = 11.0;
        $info['addon'] = 'shopping';
        return $info;
    }

    /**
     * Uninstall the module.
     */
    public function uninstall()
    {
        $tables = [
            'shopping_cart',
            'shopping_order_details',
            'shopping_orders',
            'shopping_logging',
        ];
        $GLOBALS['SITE_DB']->drop_table_if_exists($tables);

        require_code('menus2');
        delete_menu_item_simple('_SEARCH:catalogues:index:products');
    }

    /**
     * Install the module.
     *
     * @param  ?integer $upgrade_from What version we're upgrading from (null: new install)
     * @param  ?integer $upgrade_from_hack What hack version we're upgrading from (null: new-install/not-upgrading-from-a-hacked-version)
     */
    public function install(?int $upgrade_from = null, ?int $upgrade_from_hack = null)
    {
        if ($upgrade_from === null) {
            $GLOBALS['SITE_DB']->create_table('shopping_cart', [
                'id' => '*AUTO',
                'session_id' => 'ID_TEXT',
                'ordering_member' => 'MEMBER',
                'type_code' => 'ID_TEXT',
                'purchase_id' => 'ID_TEXT',
                'quantity' => 'INTEGER',
                'add_time' => 'TIME',
            ]);
            $GLOBALS['SITE_DB']->create_index('shopping_cart', 'session_id', ['session_id']);
            $GLOBALS['SITE_DB']->create_index('shopping_cart', 'type_code', ['type_code']);

            // Cart contents turns into order + details...

            $GLOBALS['SITE_DB']->create_table('shopping_orders', [
                'id' => '*AUTO',
                'session_id' => 'ID_TEXT',
                'member_id' => 'MEMBER',
                'add_date' => 'TIME',
                'total_price' => 'REAL',
                'total_tax_derivation' => 'LONG_TEXT', // Needs to be stored, as the product is dynamic and it's locked in time
                'total_tax' => 'REAL', // Needs to be stored, as the product is dynamic and it's locked in time
                'total_tax_tracking' => 'LONG_TEXT', // Needs to be stored, as the product is dynamic and it's locked in time
                'total_shipping_cost' => 'REAL',
                'total_shipping_tax' => 'REAL', // Needs to be stored, as the product is dynamic and it's locked in time
                'total_product_weight' => 'REAL',
                'total_product_length' => 'REAL',
                'total_product_width' => 'REAL',
                'total_product_height' => 'REAL',
                'order_currency' => 'ID_TEXT',
                'order_status' => 'ID_TEXT', // ORDER_STATUS_[awaiting_payment|payment_received|onhold|dispatched|cancelled|returned]
                'notes' => 'LONG_TEXT',
                'txn_id' => 'SHORT_TEXT',
                'purchase_through' => 'SHORT_TEXT', // cart|purchase_module
            ]);
            $GLOBALS['SITE_DB']->create_index('shopping_orders', 'finddispatchable', ['order_status']);
            $GLOBALS['SITE_DB']->create_index('shopping_orders', 'somember_id', ['member_id']);
            $GLOBALS['SITE_DB']->create_index('shopping_orders', 'sosession_id', ['session_id']);
            $GLOBALS['SITE_DB']->create_index('shopping_orders', 'soadd_date', ['add_date']);

            $GLOBALS['SITE_DB']->create_table('shopping_order_details', [ // individual products in an order
                'id' => '*AUTO',
                'p_order_id' => '?AUTO_LINK',
                'p_type_code' => 'ID_TEXT',
                'p_purchase_id' => 'ID_TEXT',
                'p_name' => 'SHORT_TEXT',
                'p_sku' => 'SHORT_TEXT',
                'p_quantity' => 'INTEGER',
                'p_price' => 'REAL', // Unit Price
                'p_tax_code' => 'ID_TEXT',
                'p_tax' => 'REAL', // Total tax of item (accounting for quantity);  We need this for accurate logging (we can't have it changing while looking at a past order); we use it for receipts
                'p_dispatch_status' => 'SHORT_TEXT',
            ]);
            $GLOBALS['SITE_DB']->create_index('shopping_order_details', 'order_id', ['p_order_id']);

            $GLOBALS['SITE_DB']->create_table('shopping_logging', [
                'id' => '*AUTO',
                'l_member_id' => '*MEMBER',
                'l_session_id' => 'ID_TEXT',
                'l_ip_address' => 'IP',
                'l_last_action' => 'SHORT_TEXT',
                'l_date_and_time' => 'TIME',
            ]);
        }

        if (($upgrade_from !== null) && ($upgrade_from < 7)) { // LEGACY
            $GLOBALS['SITE_DB']->alter_table_field('shopping_orders', 'session_id', 'ID_TEXT');
            $GLOBALS['SITE_DB']->alter_table_field('shopping_cart', 'session_id', 'ID_TEXT');
            $GLOBALS['SITE_DB']->alter_table_field('shopping_logging', 'l_session_id', 'ID_TEXT');

            $GLOBALS['SITE_DB']->change_primary_key('shopping_cart', ['id']);

            $GLOBALS['SITE_DB']->delete_index_if_exists('shopping_orders', 'recent_shopped');
        }

        if (($upgrade_from !== null) && ($upgrade_from < 8)) { // LEGACY
            $GLOBALS['SITE_DB']->delete_table_field('shopping_cart', 'price');
            $GLOBALS['SITE_DB']->delete_table_field('shopping_cart', 'price_pre_tax');
            $GLOBALS['SITE_DB']->delete_table_field('shopping_cart', 'product_name');
            $GLOBALS['SITE_DB']->delete_table_field('shopping_cart', 'product_code');
            $GLOBALS['SITE_DB']->delete_table_field('shopping_cart', 'product_description');
            $GLOBALS['SITE_DB']->delete_table_field('shopping_cart', 'product_type');
            $GLOBALS['SITE_DB']->delete_table_field('shopping_cart', 'product_weight');
            $GLOBALS['SITE_DB']->delete_table_field('shopping_cart', 'is_deleted');
            $GLOBALS['SITE_DB']->alter_table_field('shopping_cart', 'product_id', 'ID_TEXT', 'type_code');
            $GLOBALS['SITE_DB']->add_table_field('shopping_cart', 'purchase_id', 'ID_TEXT', '');
            $GLOBALS['SITE_DB']->add_table_field('shopping_cart', 'add_time', 'TIME');

            $GLOBALS['SITE_DB']->rename_table('shopping_order', 'shopping_orders');
            $GLOBALS['SITE_DB']->alter_table_field('shopping_orders', 'tot_price', 'REAL', 'total_price');
            $GLOBALS['SITE_DB']->add_table_field('shopping_orders', 'total_tax_derivation', 'LONG_TEXT', '');
            $GLOBALS['SITE_DB']->add_table_field('shopping_orders', 'total_tax', 'REAL', 0.00);
            $GLOBALS['SITE_DB']->add_table_field('shopping_orders', 'total_tax_tracking', 'LONG_TEXT', '');
            $GLOBALS['SITE_DB']->add_table_field('shopping_orders', 'total_shipping_cost', 'REAL', 0.00);
            $GLOBALS['SITE_DB']->add_table_field('shopping_orders', 'total_shipping_tax', 'REAL', 0.00);
            $GLOBALS['SITE_DB']->add_table_field('shopping_orders', 'total_product_weight', 'REAL', 0.00);
            $GLOBALS['SITE_DB']->add_table_field('shopping_orders', 'total_product_length', 'REAL', 0.00);
            $GLOBALS['SITE_DB']->add_table_field('shopping_orders', 'total_product_width', 'REAL', 0.00);
            $GLOBALS['SITE_DB']->add_table_field('shopping_orders', 'total_product_height', 'REAL', 0.00);
            $GLOBALS['SITE_DB']->alter_table_field('shopping_orders', 'c_member', 'MEMBER', 'member_id');
            $GLOBALS['SITE_DB']->delete_table_field('shopping_orders', 'tax_opted_out');
            $GLOBALS['SITE_DB']->add_table_field('shopping_orders', 'order_currency', 'ID_TEXT', get_option('currency'));
            $GLOBALS['SITE_DB']->alter_table_field('shopping_orders', 'transaction_id', 'SHORT_TEXT', 'txn_id');

            $GLOBALS['SITE_DB']->alter_table_field('shopping_order_details', 'p_price', 'REAL', 'p_price');
            $GLOBALS['SITE_DB']->add_table_field('shopping_order_details', 'p_tax_code', 'ID_TEXT', '0%');
            $GLOBALS['SITE_DB']->add_table_field('shopping_order_details', 'p_tax', 'REAL', 0.00);
            $GLOBALS['SITE_DB']->alter_table_field('shopping_order_details', 'order_id', '?AUTO_LINK', 'p_order_id');
            $GLOBALS['SITE_DB']->delete_table_field('shopping_order_details', 'included_tax');
            $GLOBALS['SITE_DB']->alter_table_field('shopping_order_details', 'dispatch_status', 'SHORT_TEXT', 'p_dispatch_status');
            $GLOBALS['SITE_DB']->alter_table_field('shopping_order_details', 'p_id', 'ID_TEXT', 'p_type_code');
            $GLOBALS['SITE_DB']->delete_table_field('shopping_order_details', 'p_type');
            $GLOBALS['SITE_DB']->add_table_field('shopping_order_details', 'p_purchase_id', 'ID_TEXT', '');
            $GLOBALS['SITE_DB']->alter_table_field('shopping_order_details', 'p_code', 'SHORT_TEXT', 'p_sku');

            $GLOBALS['SITE_DB']->alter_table_field('shopping_logging', 'e_member_id', '*MEMBER', 'l_member_id');
            $GLOBALS['SITE_DB']->alter_table_field('shopping_logging', 'session_id', 'ID_TEXT', 'l_session_id');
            $GLOBALS['SITE_DB']->alter_table_field('shopping_logging', 'ip', 'IP', 'l_ip');
            $GLOBALS['SITE_DB']->alter_table_field('shopping_logging', 'last_action', 'SHORT_TEXT', 'l_last_action');
            $GLOBALS['SITE_DB']->alter_table_field('shopping_logging', 'date_and_time', 'TIME', 'l_date_and_time');

            $GLOBALS['SITE_DB']->drop_table_if_exists('shopping_order_addresses');

            $GLOBALS['SITE_DB']->delete_index_if_exists('shopping_order', 'soc_member');
            $GLOBALS['SITE_DB']->create_index('shopping_orders', 'somember_id', ['member_id']);

            $option = get_option('shipping_cost_factor');
            $base = 0.00;
            $matches = [];
            if (preg_match('#\(([\d\.]+)\)#', $option, $matches) != 0) {
                $base = float_unformat($matches[1]);
                $option = str_replace($matches[0], '', $option);
            }
            if ($option == '') {
                $option = float_format(0.0);
            }
            $factor = float_unformat($option) / 100.0;
            set_option('shipping_cost_base', float_format($base));
            set_option('shipping_cost_factor', float_format($factor));

            $GLOBALS['SITE_DB']->delete_index_if_exists('shopping_cart', 'product_id');
            $GLOBALS['SITE_DB']->create_index('shopping_cart', 'type_code', ['type_code']);
        }

        if (($upgrade_from === null) || ($upgrade_from < 8)) {
            $GLOBALS['SITE_DB']->create_index('shopping_order_details', 'type_code', ['p_type_code']);
            $GLOBALS['SITE_DB']->create_index('shopping_logging', 'cart_log', ['l_date_and_time']);
        }

        if (($upgrade_from !== null) && ($upgrade_from < 9)) { // LEGACY: 11.beta1
            // Database consistency fixes
            $GLOBALS['SITE_DB']->alter_table_field('shopping_cart', 'ordered_by', 'MEMBER', 'ordering_member');
            $GLOBALS['SITE_DB']->alter_table_field('shopping_logging', 'l_ip', 'IP', 'l_ip_address');

            $GLOBALS['SITE_DB']->delete_index_if_exists('shopping_cart', 'ordered_by');
        }

        if (($upgrade_from === null) || ($upgrade_from < 9)) {
            $GLOBALS['SITE_DB']->create_index('shopping_cart', 'ordering_member', ['ordering_member']);
        }
    }

    /**
     * Find entry-points available within this module.
     *
     * @param  boolean $check_perms Whether to check permissions
     * @param  ?MEMBER $member_id The member to check permissions as (null: current user)
     * @param  boolean $support_crosslinks Whether to allow cross links to other modules (identifiable via a full-page-link rather than a screen-name)
     * @param  boolean $be_deferential Whether to avoid any entry-point (or even return null to disable the page in the Sitemap) if we know another module, or page_group, is going to link to that entry-point. Note that "!" and "browse" entry points are automatically merged with container page nodes (likely called by page-groupings) as appropriate.
     * @return ?array A map of entry points (screen-name=>language-code/string or screen-name=>[language-code/string, icon-theme-image]) (null: disabled)
     */
    public function get_entry_points(bool $check_perms = true, ?int $member_id = null, bool $support_crosslinks = true, bool $be_deferential = false) : ?array
    {
        if (!addon_installed('shopping')) {
            return null;
        }

        if (get_forum_type() != 'cns') {
            return null;
        }

        if ($member_id === null) {
            $member_id = get_member();
        }

        $ret = [
            'browse' => ['SHOPPING', 'menu/rich_content/ecommerce/shopping_cart'],
        ];
        if ((!$check_perms || !is_guest($member_id)) && (get_forum_type() != 'cns') && ($GLOBALS['SITE_DB']->query_select_value('shopping_orders', 'COUNT(*)', ['i_member_id' => $member_id]) > 0)) {
            $ret += [
                'my_orders' => ['MY_ORDERS', 'menu/rich_content/ecommerce/orders'],
            ];
        }
        return $ret;
    }

    public $title;

    /**
     * Module pre-run function. Allows us to know metadata for <head> before we start streaming output.
     *
     * @return ?Tempcode Tempcode indicating some kind of exceptional output (null: none)
     */
    public function pre_run() : ?object
    {
        $error_msg = new Tempcode();
        if (!addon_installed__messaged('shopping', $error_msg)) {
            return $error_msg;
        }

        $type = get_param_string('type', 'browse');

        require_javascript('core_form_interfaces');

        require_lang('shopping');
        require_lang('catalogues');

        $ecom_catalogue_count = $GLOBALS['SITE_DB']->query_select_value('catalogues', 'COUNT(*)', ['c_ecommerce' => 1]);
        $ecom_catalogue = $GLOBALS['SITE_DB']->query_select_value_if_there('catalogues', 'c_name', ['c_ecommerce' => 1]);
        $ecom_catalogue_id = $GLOBALS['SITE_DB']->query_select_value_if_there('catalogue_categories', 'MIN(id)', ['c_name' => $ecom_catalogue]);

        if ($type == 'browse') {
            if ($ecom_catalogue_count == 1) {
                breadcrumb_set_parents([['_SELF:catalogues:category:=' . strval($ecom_catalogue_id), do_lang_tempcode('DEFAULT_CATALOGUE_PRODUCTS_TITLE')]]);
            } else {
                breadcrumb_set_parents([['_SELF:catalogues:browse:ecommerce=1', do_lang_tempcode('CATALOGUES')]]);
            }

            $this->title = get_screen_title('SHOPPING');
        }

        if ($type == 'add_item') {
            $this->title = get_screen_title('SHOPPING');
        }

        if ($type == 'update_cart') {
            $this->title = get_screen_title('SHOPPING');
        }

        if ($type == 'empty_cart') {
            $this->title = get_screen_title('SHOPPING');
        }

        if ($type == 'my_orders') {
            $this->title = get_screen_title('MY_ORDERS');
        }
        if ($type == 'receipt') {
            $this->title = get_screen_title('ecommerce:RECEIPT');
        }

        if ($type == 'order_details') {
            breadcrumb_set_parents([['_SELF:_SELF:my_orders', do_lang_tempcode('MY_ORDERS')]]);

            $id = get_param_integer('id');
            $this->title = get_screen_title('_ORDER_DETAILS', true, [escape_html(strval($id))]);
        }

        return null;
    }

    /**
     * Execute the module.
     *
     * @return Tempcode The result of execution
     */
    public function run() : object
    {
        @ignore_user_abort(true); // Must keep going till completion

        // Static cache issue? (We can disable static caching for particular pages to avoid this)
        if (get_session_id() == '') {
            access_denied('NOT_AS_GUEST');
        }

        require_code('shopping');
        require_code('feedback');
        require_code('ecommerce');

        if (get_forum_type() != 'cns') {
            warn_exit(do_lang_tempcode('NO_CNS'));
        }

        // Kill switch
        if ((ecommerce_test_mode()) && (!$GLOBALS['IS_ACTUALLY_ADMIN']) && (!has_privilege(get_member(), 'access_ecommerce_in_test_mode'))) {
            warn_exit(do_lang_tempcode('PURCHASE_DISABLED'));
        }

        push_query_limiting(false);

        $type = get_param_string('type', 'browse');

        delete_incomplete_orders();

        if ($type == 'browse') {
            return $this->view_shopping_cart();
        }
        if ($type == 'add_item') {
            return $this->add_item();
        }
        if ($type == 'update_cart') {
            return $this->update_cart();
        }
        if ($type == 'empty_cart') {
            return $this->empty_cart();
        }
        if ($type == 'my_orders') {
            return $this->my_orders();
        }
        if ($type == 'order_details') {
            return $this->order_details();
        }
        if ($type == 'receipt') {
            return $this->receipt();
        }

        return new Tempcode();
    }

    /**
     * The UI to show shopping cart.
     *
     * @return Tempcode The UI
     */
    public function view_shopping_cart() : object
    {
        require_code('templates_results_table');
        require_code('form_templates');
        require_code('currency');
        require_css('shopping');
        require_css('ecommerce');
        require_javascript('shopping');

        $shopping_cart_rows = find_products_in_cart();
        $max_rows = count($shopping_cart_rows);

        $type_codes = [];
        $has_needed_fields = false;

        if ($max_rows > 0) {
            $shopping_cart = new Tempcode();

            $header_row = results_header_row([
                '',
                do_lang_tempcode('PRODUCT'),
                do_lang_tempcode('UNIT_PRICE'),
                do_lang_tempcode('QUANTITY'),
                do_lang_tempcode('PRICE'),
                do_lang_tempcode(get_option('tax_system')),
                do_lang_tempcode('AMOUNT'),
                do_lang_tempcode('REMOVE'),
            ]);

            list($total_price, $total_tax_derivation, $total_tax, $total_tax_tracking, $shopping_cart_rows_taxes, $total_shipping_cost, $total_shipping_tax, , , , ) = derive_cart_amounts($shopping_cart_rows);

            foreach ($shopping_cart_rows as $i => $item) {
                list($details, $product_object) = find_product_details($item['type_code']);

                if ($details === null) {
                    $GLOBALS['SITE_DB']->query_delete('shopping_cart', ['id' => $item['id']], '', 1);
                    continue;
                }

                $type_codes[] = $item['type_code'];

                // Check if we have fields that must be filled out
                if (!$has_needed_fields) {
                    $has_needed_fields = has_needed_fields($item['type_code']);
                }

                $this->show_cart_entry($shopping_cart, $details, $item, isset($shopping_cart_rows_taxes[$i]) ? $shopping_cart_rows_taxes[$i] : null);
            }

            $results_table = results_table(do_lang_tempcode('SHOPPING'), 0, 'cart_start', $max_rows, 'cart_max', $max_rows, $header_row, $shopping_cart, [], null, null, 'sort', null, [], 'cart');

            $update_cart_url = build_url(['page' => '_SELF', 'type' => 'update_cart'], '_SELF');
            $empty_cart_url = build_url(['page' => '_SELF', 'type' => 'empty_cart'], '_SELF');

            $fields = null;
            ecommerce_attach_memo_field_if_needed($fields);

            // Request additional information when needed, otherwise go directly to payment
            if ($has_needed_fields) {
                $next_url = build_url(['page' => 'purchase', 'type' => 'details', 'type_code' => 'CART_ORDER'], get_module_zone('purchase'));
            } else {
                $next_url = build_url(['page' => 'purchase', 'type' => 'pay', 'type_code' => 'CART_ORDER'], get_module_zone('purchase'));
            }
        } else {
            $total_price = 0.00;
            $total_tax = 0.00;
            $total_shipping_cost = 0.00;
            $total_shipping_tax = 0.00;

            $results_table = do_lang_tempcode('CART_EMPTY');

            $update_cart_url = new Tempcode();
            $empty_cart_url = new Tempcode();

            $fields = new Tempcode();
            $next_url = new Tempcode();
        }

        $grand_total = $total_price + $total_tax/*already included $total_shipping_tax*/ + $total_shipping_cost;

        $ecom_catalogue_count = $GLOBALS['SITE_DB']->query_select_value('catalogues', 'COUNT(*)', ['c_ecommerce' => 1]);
        $ecom_catalogue = $GLOBALS['SITE_DB']->query_select_value_if_there('catalogues', 'c_name', ['c_ecommerce' => 1]);
        $ecom_catalogue_id = $GLOBALS['SITE_DB']->query_select_value_if_there('catalogue_categories', 'MIN(id)', ['c_name' => $ecom_catalogue]);
        if ($ecom_catalogue_count == 1) {
            $continue_shopping_url = build_url(['page' => 'catalogues', 'type' => 'category', 'id' => $ecom_catalogue_id], get_module_zone('catalogues'));
        } else {
            $continue_shopping_url = build_url(['page' => 'catalogues', 'type' => 'browse', 'ecommerce' => 1], get_module_zone('catalogues'));
        }

        log_cart_actions(do_lang('VIEW_CART'));

        $tpl = do_template('ECOM_SHOPPING_CART_SCREEN', [
            '_GUID' => 'badff09daf52ee1c84b472c44be1bfae',
            'TITLE' => $this->title,
            'RESULTS_TABLE' => $results_table,
            'UPDATE_CART_URL' => $update_cart_url,
            'CONTINUE_SHOPPING_URL' => $continue_shopping_url,
            'MESSAGE' => '',
            'TYPE_CODES' => implode(',', array_unique($type_codes)),
            'EMPTY_CART_URL' => $empty_cart_url,
            'TOTAL_PRICE' => float_to_raw_string($total_price),
            'TOTAL_TAX' => float_to_raw_string($total_tax),
            'TOTAL_SHIPPING_COST' => float_to_raw_string($total_shipping_cost),
            'TOTAL_SHIPPING_TAX' => float_to_raw_string($total_shipping_tax),
            'GRAND_TOTAL' => float_to_raw_string($grand_total),
            'CURRENCY' => get_option('currency'),
            'FIELDS' => $fields,
            'NEXT_URL' => $next_url,
        ]);

        require_code('templates_internalise_screen');
        return internalise_own_screen($tpl);
    }

    /**
     * Produce a results table row for a particular shopping cart entry.
     *
     * @param  Tempcode $shopping_cart Tempcode object of shopping cart result table
     * @param  array $details Product details
     * @param  array $item Cart row
     * @param  ?array $tax_details Tax details (null: none known, look them up)
     */
    protected function show_cart_entry(object &$shopping_cart, array $details, array $item, ?array $tax_details = null)
    {
        $tpl_set = 'cart';

        $edit_quantity_link = do_template('ECOM_SHOPPING_ITEM_QUANTITY_FIELD', [
            '_GUID' => '37201554b0a4d7d61cb52fee27d5f2be',
            'TYPE_CODE' => $item['type_code'],
            'QUANTITY' => strval($item['quantity']),
        ]);

        $delete_item_link = do_template('ECOM_SHOPPING_ITEM_REMOVE_FIELD', ['_GUID' => '82d08bf4cd3bf668f4109907e19d4fea', 'TYPE_CODE' => $item['type_code']]);

        require_code('images');
        $product_image = do_image_thumb($details['item_image_url'], $details['item_name'], false, false, 50, 50);

        $price_singular = $details['price'];
        $price_multiple = $details['price'] * $item['quantity'];

        if ($tax_details === null) {
            list($tax_derivation, $tax, $tax_tracking) = calculate_tax_due($details, $details['tax_code'], $details['price'], 0.0, null, $item['quantity']);
        } else {
            list($tax_derivation, $tax, $tax_tracking) = $tax_details;
        }

        $amount = $price_multiple + $tax;

        $product_details_url = get_product_details_url($item['type_code'], false, get_member());
        $product_link = hyperlink($product_details_url, $details['item_name'], false, true, do_lang('INDEX'));

        require_code('templates_results_table');
        $shopping_cart->attach(results_entry([
            $product_image,
            $product_link,
            currency_convert_wrap($price_singular),
            $edit_quantity_link,
            currency_convert_wrap($price_multiple),
            currency_convert_wrap($tax),
            currency_convert_wrap($amount),
            $delete_item_link,
        ], false, $tpl_set));
    }

    /**
     * Add an item to the cart.
     *
     * @return Tempcode The UI
     */
    public function add_item() : object
    {
        if (is_guest()) {
            require_code('users_inactive_occasionals');
            set_session_id(get_session_id(), true); // Persist guest sessions longer
        }

        $prefix = either_param_string('prefix', '');
        $_type_code = $prefix . either_param_string('type_code');

        $purchase_id = either_param_string('purchase_id', '');

        $quantity = either_param_integer('quantity', 1);

        $purchase_through = either_param_string('purchase_through', 'cart');

        foreach (explode('|', $_type_code) as $type_code) {
            add_to_cart($type_code, $purchase_id, $quantity);
        }

        log_cart_actions(do_lang('ADD_TO_CART'));

        if ($purchase_through == 'cart') {
            $next_url = build_url(['page' => '_SELF', 'type' => 'browse'], '_SELF');
        } else {
            $next_url = build_url(['page' => 'purchase', 'type' => 'pay', 'type_code' => 'CART_ORDER'], get_module_zone('purchase'));
        }

        return redirect_screen($this->title, $next_url, do_lang_tempcode('ADDED_TO_CART'));
    }

    /**
     * Update the cart, editing quantities and deleting items.
     *
     * @return Tempcode The UI
     */
    public function update_cart() : object
    {
        $product_to_remove = [];

        $products_in_cart = [];

        $type_codes = explode(',', post_param_string('type_codes'));
        foreach ($type_codes as $type_code) {
            $quantity = post_param_integer('quantity_' . $type_code);

            list($details, $product_object) = find_product_details($type_code);

            $remove = (post_param_integer('remove_' . $type_code, 0) == 1);
            if ($remove) {
                $product_to_remove[] = $type_code;
            } else {
                if (method_exists($product_object, 'get_available_quantity')) {
                    $available_quantity = $product_object->get_available_quantity($type_code, false);
                    if (($available_quantity !== null) && ($available_quantity <= $quantity)) {
                        $quantity = $available_quantity;

                        attach_message(do_lang_tempcode('PRODUCT_QUANTITY_CHANGED', escape_html($details['item_name'])), 'warn');
                    }
                }

                $products_in_cart[] = [$type_code, $quantity];
            }
        }

        update_cart($products_in_cart);

        if (!empty($product_to_remove)) {
            remove_from_cart($product_to_remove);
        }

        log_cart_actions(do_lang('UPDATE_CART'));

        $cart_view_url = build_url(['page' => '_SELF', 'type' => 'browse'], '_SELF');
        return redirect_screen($this->title, $cart_view_url, do_lang_tempcode('CART_UPDATED'));
    }

    /**
     * Empty the shopping cart.
     *
     * @return Tempcode The UI
     */
    public function empty_cart() : object
    {
        empty_cart();

        log_cart_actions(do_lang('EMPTY_CART'));

        $cart_view_url = build_url(['page' => '_SELF', 'type' => 'browse'], '_SELF');
        return redirect_screen($this->title, $cart_view_url, do_lang_tempcode('CART_EMPTIED'));
    }

    /**
     * Show all my orders.
     *
     * @return Tempcode The interface
     */
    public function my_orders() : object
    {
        if (is_guest()) {
            access_denied('NOT_AS_GUEST');
        }

        $member_id = get_member();
        if (has_privilege(get_member(), 'assume_any_member')) {
            $member_id = get_param_integer('id', $member_id);
        }

        $orders = [];

        $rows = $GLOBALS['SITE_DB']->query_select('shopping_orders', ['*'], ['member_id' => $member_id], 'ORDER BY add_date');

        foreach ($rows as $row) {
            $order_details_url = build_url(['page' => '_SELF', 'type' => 'order_details', 'id' => $row['id']], '_SELF');

            if ($row['purchase_through'] == 'cart') {
                $order_title = do_lang('CART_ORDER', strval($row['id']));
            } else {
                $order_title = do_lang('PURCHASE_ORDER', strval($row['id']));
            }

            $price_linker = new Tempcode();
            if ($row['txn_id'] != '') {
                $receipt_url = build_url(['page' => '_SELF', 'type' => 'receipt', 'id' => $row['txn_id'], 'wide_high' => 1], '_SELF');
                $price_linker = hyperlink($receipt_url, ecommerce_get_currency_symbol($row['order_currency']) . escape_html(float_format($row['total_price'])), true, false, do_lang('RECEIPT'));
            } else {
                $price_linker->attach(ecommerce_get_currency_symbol($row['order_currency']) . escape_html(float_format($row['total_price'])));
            }

            $transaction_linker = build_transaction_linker($row['txn_id'], $row['order_status'] == 'ORDER_STATUS_awaiting_payment');

            $orders[] = [
                'ORDER_TITLE' => $order_title,
                'ID' => strval($row['id']),
                'TXN_ID' => $row['txn_id'],
                'TRANSACTION_LINKER' => $transaction_linker,
                'TOTAL_PRICE' => $price_linker,
                'TOTAL_TAX' => float_format($row['total_tax']),
                'TOTAL_SHIPPING_COST' => float_format($row['total_shipping_cost']),
                'CURRENCY' => $row['order_currency'],
                'DATE' => get_timezoned_date_time($row['add_date'], false),
                'STATUS' => do_lang_tempcode($row['order_status']),
                'NOTE' => '',
                'ORDER_DET_URL' => $order_details_url,
            ];
        }

        if (empty($orders)) {
            inform_exit(do_lang_tempcode('NO_ENTRIES'));
        }

        return do_template('ECOM_ORDERS_SCREEN', ['_GUID' => '79eb5f17cf4bc2dc4f0cccf438261c73', 'TITLE' => $this->title, 'ORDERS' => $orders]);
    }

    /**
     * Show an order details.
     *
     * @return Tempcode The interface
     */
    public function order_details() : object
    {
        if (is_guest()) {
            access_denied('NOT_AS_GUEST');
        }

        $id = get_param_integer('id');

        if (!has_privilege(get_member(), 'assume_any_member')) {
            $member_id = $GLOBALS['SITE_DB']->query_select_value_if_there('shopping_orders', 'member_id', ['id' => $id]);
            if ($member_id === null) {
                warn_exit(do_lang_tempcode('MISSING_RESOURCE'));
            }

            if ($member_id != get_member()) {
                access_denied('I_ERROR');
            }
        }

        require_code('ecommerce_reports');
        return build_order_details($this->title, $id, new Tempcode());
    }

    /**
     * Show a receipt for a transaction.
     *
     * @return Tempcode The result of execution
     */
    public function receipt() : object
    {
        $GLOBALS['SCREEN_TEMPLATE_CALLED'] = '';

        $txn_id = get_param_string('id');
        $receipt = display_receipt($txn_id, get_member());
        return $receipt;
    }
}
