<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    news
 */

/**
 * Block class.
 */
class Block_main_news
{
    /**
     * Find details of the block.
     *
     * @return ?array Map of block info (null: block is disabled)
     */
    public function info() : ?array
    {
        $info = [];
        $info['author'] = 'Chris Graham';
        $info['organisation'] = 'Composr';
        $info['hacked_by'] = null;
        $info['hack_version'] = null;
        $info['version'] = 2;
        $info['locked'] = false;
        $info['min_cms_version'] = 11.0;
        $info['addon'] = 'news';
        $info['parameters'] = [
            'param',
            'member_based',
            'select',
            'select_and',
            'display_slides',
            'display_summaries',
            'display_briefs',
            'blogs',
            'historic',
            'zone',
            'title',
            'show_in_full',
            'no_links',
            'attach_to_url_filter',
            'render_if_empty',
            'filter',
            'start',
            'pagination',
            'as_guest',
            'optimise',
            'check',
            'render_mode',
            'slider_interval',
            'news_articles_per_slide',
        ];
        return $info;
    }

    /**
     * Find caching details for the block.
     *
     * @return ?array Map of cache details (cache_on and ttl) (null: block is disabled)
     */
    public function caching_environment() : ?array
    {
        $info = [];
        $info['cache_on'] = <<<'PHP'
        ((preg_match('#<\w+>#', array_key_exists('filter', $map) ? $map['filter'] : '')) != 0)
        ?
        null
        :
        [
            array_key_exists('optimise', $map) ? $map['optimise'] : '0',
            (array_key_exists('pagination', $map) ? $map['pagination'] : '0' == '1'),
            array_key_exists('title', $map) ? escape_html($map['title']) : '(default title)',
            array_key_exists('as_guest', $map) ? ($map['as_guest'] == '1') : false,
            get_param_integer($block_id . '_start',
            array_key_exists('start', $map) ? intval($map['start']) : 0),
            array_key_exists('filter', $map) ? $map['filter'] : '',
            array_key_exists('show_in_full', $map) ? $map['show_in_full'] : '0',
            array_key_exists('render_if_empty', $map) ? $map['render_if_empty'] : '0',
            ((array_key_exists('attach_to_url_filter', $map) ? $map['attach_to_url_filter'] : '0') == '1'),
            array_key_exists('no_links', $map) ? $map['no_links'] : 0,
            array_key_exists('title', $map) ? $map['title'] : '',
            array_key_exists('member_based', $map) ? $map['member_based'] : '0',
            array_key_exists('blogs', $map) ? $map['blogs'] : '-1',
            array_key_exists('historic', $map) ? $map['historic'] : '',
            array_key_exists('param', $map) ? intval($map['param']) : 9,
            array_key_exists('display_slides', $map) ? intval($map['display_slides']) : 0,
            array_key_exists('display_summaries', $map) ? intval($map['display_summaries']) : 10,
            array_key_exists('display_briefs', $map) ? intval($map['display_briefs']) : 0,
            array_key_exists('select', $map) ? $map['select'] : '',
            array_key_exists('zone', $map) ? $map['zone'] : get_module_zone('news'),
            array_key_exists('select_and', $map) ? $map['select_and'] : '',
            array_key_exists('check', $map) ? ($map['check'] == '1') : true,
            empty($map['render_mode']) ? 'boxes' : $map['render_mode'],
            empty($map['slider_interval']) ? 6000 : intval($map['slider_interval']),
            empty($map['news_articles_per_slide']) ? 3 : intval($map['news_articles_per_slide']),
        ]
PHP;
        $info['special_cache_flags'] = CACHE_AGAINST_DEFAULT | CACHE_AGAINST_PERMISSIVE_GROUPS;
        if (addon_installed('content_privacy')) {
            $info['special_cache_flags'] |= CACHE_AGAINST_MEMBER;
        }
        $info['ttl'] = 60;
        return $info;
    }

    /**
     * Execute the block.
     *
     * @param  array $map A map of parameters
     * @return Tempcode The result of execution
     */
    public function run(array $map) : object
    {
        $error_msg = new Tempcode();
        if (!addon_installed__messaged('news', $error_msg)) {
            return $error_msg;
        }
        if (!addon_installed__messaged('news_shared', $error_msg)) {
            return $error_msg;
        }

        require_lang('cns');
        require_lang('news');
        require_css('news');
        require_code('news');
        require_code('images');

        $block_id = get_block_id($map);

        $check_perms = array_key_exists('check', $map) ? ($map['check'] == '1') : true;

        // Read in parameters
        $display_slides = array_key_exists('display_slides', $map) ? intval($map['display_slides']) : 0;
        $display_summaries = array_key_exists('display_summaries', $map) ? intval($map['display_summaries']) : 10;
        $display_briefs = array_key_exists('display_briefs', $map) ? intval($map['display_briefs']) : 0;
        $zone = isset($map['zone']) ? $map['zone'] : get_module_zone('news');
        $historic = isset($map['historic']) ? $map['historic'] : '';
        $filter = isset($map['filter']) ? $map['filter'] : '';
        $blogs = isset($map['blogs']) ? intval($map['blogs']) : -1;
        $member_based = (isset($map['member_based'])) && ($map['member_based'] == '1');
        $attach_to_url_filter = ((isset($map['attach_to_url_filter']) ? $map['attach_to_url_filter'] : '0') == '1');
        $optimise = (array_key_exists('optimise', $map)) && ($map['optimise'] == '1');
        $render_mode = empty($map['render_mode']) ? 'boxes' : $map['render_mode'];
        $slider_interval = empty($map['slider_interval']) ? 6000 : intval($map['slider_interval']);
        $news_articles_per_slide = empty($map['news_articles_per_slide']) ? 3 : intval($map['news_articles_per_slide']);

        // Pagination
        $start = get_param_integer($block_id . '_start', isset($map['start']) ? intval($map['start']) : 0);
        if ($start != 0) {
            $days = 0;
        }
        $do_pagination = ((isset($map['pagination']) ? $map['pagination'] : '0') == '1');

        // Read in news categories ahead, for performance
        load_news_cat_rows('nc_owner IS NULL');

        // News query
        $select = isset($map['select']) ? $map['select'] : '*';
        $select_and = isset($map['select_and']) ? $map['select_and'] : '';
        $q_filter = '1=1';
        if ($select != '*') {
            $q_filter .= ' AND ' . $this->generate_selectcode_sql($select);
        }
        if (($select_and != '') && ($select_and != '*')) {
            $q_filter .= ' AND ' . $this->generate_selectcode_sql($select_and);
        }
        if ($blogs === 0) {
            if ($q_filter != '') {
                $q_filter .= ' AND ';
            }
            $q_filter .= 'nc_owner IS NULL';
        } elseif ($blogs === 1) {
            if ($q_filter != '') {
                $q_filter .= ' AND ';
            }
            $q_filter .= '(nc_owner IS NOT NULL)';
        }
        if ($blogs != -1) {
            $join = ' LEFT JOIN ' . $GLOBALS['SITE_DB']->get_table_prefix() . 'news_categories c ON c.id=r.news_category';
        } else {
            $join = '';
        }

        // Filtercode
        if ($filter != '') {
            require_code('filtercode');
            list($filter_extra_join, $filter_extra_where) = filtercode_to_sql($GLOBALS['SITE_DB'], parse_filtercode($filter), 'news');
            $join .= implode('', $filter_extra_join);
            $q_filter .= $filter_extra_where;
        }

        // Privacy
        if (addon_installed('content_privacy')) {
            require_code('content_privacy');
            $as_guest = array_key_exists('as_guest', $map) ? ($map['as_guest'] == '1') : false;
            $viewing_member_id = $as_guest ? $GLOBALS['FORUM_DRIVER']->get_guest_id() : null;
            list($privacy_join, $privacy_where) = get_privacy_where_clause('news', 'r', $viewing_member_id);
            $join .= $privacy_join;
            $q_filter .= $privacy_where;
        }

        // Regions
        if (get_option('filter_regions') == '1') {
            require_code('locations');
            $q_filter .= sql_region_filter('news', 'r.id');
        }

        // Permissions
        if ((!$GLOBALS['FORUM_DRIVER']->is_super_admin(get_member())) && ($check_perms)) {
            $q_filter .= get_category_permission_where_clause('news', 'news_category', get_member(), get_permission_where_clause_groups(get_member()), 'r');
        }

        // Read in rows
        $query = 'SELECT COUNT(DISTINCT r.id) FROM ' . get_table_prefix() . 'news r LEFT JOIN ' . $GLOBALS['SITE_DB']->get_table_prefix() . 'news_category_entries d ON d.news_entry=r.id' . $join . ' WHERE ' . $q_filter . ((!has_privilege(get_member(), 'see_not_validated')) ? ' AND validated=1' : '');
        $max_rows = $GLOBALS['SITE_DB']->query_value_if_there($query, false, true);
        if ($historic == '') {
            $query_full = 'SELECT DISTINCT r.* FROM ' . $GLOBALS['SITE_DB']->get_table_prefix() . 'news r LEFT JOIN ' . $GLOBALS['SITE_DB']->get_table_prefix() . 'news_category_entries d ON r.id=d.news_entry' . $join . ' WHERE ' . $q_filter . ((!has_privilege(get_member(), 'see_not_validated')) ? ' AND validated=1' : '') . ' ORDER BY r.date_and_time DESC';
            $rows = $GLOBALS['SITE_DB']->query($query_full, $display_slides + $display_summaries + $display_briefs, $start, false, true, ['title' => 'SHORT_TRANS', 'news' => 'LONG_TRANS', 'news_article' => 'LONG_TRANS']);
        } else {
            $old_limit = cms_extend_time_limit(TIME_LIMIT_EXTEND__SLUGGISH);
            $rows = [];
            $search_start = 0;
            $okayed = 0;
            do {
                $_rows = $GLOBALS['SITE_DB']->query('SELECT DISTINCT r.* FROM ' . $GLOBALS['SITE_DB']->get_table_prefix() . 'news r LEFT JOIN ' . $GLOBALS['SITE_DB']->get_table_prefix() . 'news_category_entries d ON r.id=d.news_entry' . $join . ' WHERE ' . $q_filter . ((!has_privilege(get_member(), 'see_not_validated')) ? ' AND validated=1' : '') . ' ORDER BY r.date_and_time DESC', 200, $search_start, false, true);
                foreach ($_rows as $row) {
                    $ok = false;
                    switch ($historic) {
                        case 'month':
                            if ((date('m', utctime_to_usertime($row['date_and_time'])) == date('m', utctime_to_usertime())) && (date('Y', utctime_to_usertime($row['date_and_time'])) != date('Y', utctime_to_usertime()))) {
                                $ok = true;
                            }
                            break;

                        case 'week':
                            if ((date('W', utctime_to_usertime($row['date_and_time'])) == date('W', utctime_to_usertime())) && (date('Y', utctime_to_usertime($row['date_and_time'])) != date('Y', utctime_to_usertime()))) {
                                $ok = true;
                            }
                            break;

                        case 'day':
                            if ((date('d', utctime_to_usertime($row['date_and_time'])) == date('d', utctime_to_usertime())) && (date('m', utctime_to_usertime($row['date_and_time'])) == date('m', utctime_to_usertime())) && (date('Y', utctime_to_usertime($row['date_and_time'])) != date('Y', utctime_to_usertime()))) {
                                $ok = true;
                            }
                            break;
                    }
                    if ($ok) {
                        $okayed++;

                        if ($okayed < $start) {
                            continue;
                        }

                        if (count($rows) < $display_slides + $display_summaries + $display_briefs) {
                            $rows[] = $row;
                        } else {
                            break 2;
                        }
                    }
                }
                $search_start += 200;
            } while (count($_rows) == 200);
            unset($_rows);
            cms_set_time_limit($old_limit);
        }

        // Split up into display types
        $rows_slides = array_slice($rows, 0, $display_slides);
        $rows_summaries = array_slice($rows, $display_slides, $display_summaries);
        $rows_briefs = array_slice($rows, $display_slides + $display_summaries, $display_briefs);

        // Shared calculations
        $show_in_full = (isset($map['show_in_full'])) && ($map['show_in_full'] == '1');
        $show_author = (addon_installed('authors')) && (!$member_based);
        $prop_url = [];
        if ($attach_to_url_filter) {
            $prop_url += propagate_filtercode();
        }
        if ($select != '*') {
            $prop_url['select'] = $select;
        }
        if (($select_and != '*') && ($select_and != '')) {
            $prop_url['select_and'] = $select_and;
        }
        if ($blogs != -1) {
            $prop_url['blog'] = $blogs;
        }
        $allow_comments_shared = (get_option('is_on_comments') == '1') && (!has_no_forum());
        $base_url = get_base_url();

        // Slides content loop
        $slide_items = [];
        foreach ($rows_slides as $myrow) {
            $slide_items[] = $this->assemble_template_map($myrow, $optimise, false, $show_author, $prop_url, $allow_comments_shared, $base_url, $attach_to_url_filter, $zone, $blogs);
        }
        $slides = new Tempcode();
        $slides_count = 0;
        for ($i = 0; $i < count($slide_items); $i += $news_articles_per_slide) {
            $news_items = [];
            for ($j = 0; $j < $news_articles_per_slide; $j++) {
                if (isset($slide_items[$i + $j])) {
                    $news_items[] = $slide_items[$i + $j];
                }
            }

            $slides->attach(do_template('BLOCK_MAIN_NEWS_SLIDER_SLIDE', [
                '_GUID' => '854c5f329ba048968b307d2944f6c061',
                'BLOCK_ID' => $block_id,
                'ACTIVE' => ($i == 0),
                'NEWS_ITEMS' => $news_items,
            ]));

            $slides_count++;
        }
        if ($slides_count == 0) {
            $slider = new Tempcode();
        } else {
            $slider = do_template('BLOCK_MAIN_NEWS_SLIDER', [
                '_GUID' => '01f5fbd2b0c7c8f249023ecb4254366e',
                'BLOCK_ID' => $block_id,
                'BLOG' => $blogs === 1,
                'SLIDES' => $slides,
                'SLIDES_COUNT' => strval($slides_count),
                'SLIDES_COUNT_ARRAY' => ($slides_count > 1) ? range(1, $slides_count) : null,
                'INTERVAL' => strval($slider_interval),
            ]);
        }

        // Summary content loop
        $summary_content = new Tempcode();
        foreach ($rows_summaries as $myrow) {
            $map2 = $this->assemble_template_map($myrow, $optimise, $show_in_full, $show_author, $prop_url, $allow_comments_shared, $base_url, $attach_to_url_filter, $zone, $blogs);

            $tpl = 'NEWS_BOX';
            if ($render_mode == 'grid') {
                $tpl = 'BLOCK_MAIN_NEWS_GRID_ITEM';
            }

            $summary_content->attach(do_template($tpl, $map2));
        }

        // Brief content loop
        $brief_content = new Tempcode();
        foreach ($rows_briefs as $myrow) {
            $just_news_row = db_map_restrict($myrow, ['id', 'title', 'news', 'news_article']);

            // Basic details
            $date = get_timezoned_date_time_tempcode($myrow['date_and_time']);

            // URL
            $tmp = ['page' => ($zone == '_SELF' && running_script('index')) ? get_page_name() : 'news', 'type' => 'view', 'id' => $myrow['id']] + $prop_url;
            $url = build_url($tmp, $zone);

            // Title
            $title = get_translated_tempcode('news', $just_news_row, 'title');
            $title_plain = get_translated_text($myrow['title']);

            // Render
            $seo_bits = (get_value('disable_tags') === '1') ? ['', ''] : seo_meta_get_for('news', strval($myrow['id']));
            $map2 = ['_GUID' => 'd81bda3a0912a1e708af6bb1f503b296', 'TAGS' => get_loaded_tags('news', explode(',', $seo_bits[0])), 'BLOG' => $blogs === 1, 'ID' => strval($myrow['id']), 'SUBMITTER' => strval($myrow['submitter']), 'DATE' => $date, 'DATE_RAW' => strval($myrow['date_and_time']), 'FULL_URL' => $url, 'NEWS_TITLE_PLAIN' => $title_plain, 'NEWS_TITLE' => $title];
            if (($allow_comments_shared) && ($myrow['allow_comments'] >= 1)) {
                $map2['COMMENT_COUNT'] = '1';
            }
            $brief_content->attach(do_template('NEWS_BRIEF', $map2));
        }

        // Work out management URLs
        $tmp = ['page' => ($zone == '_SELF' && running_script('index')) ? get_page_name() : 'news', 'type' => 'browse'];
        if ($select != '*') {
            $tmp[is_numeric($select) ? 'id' : 'select'] = $select;
        }
        if (($select_and != '*') && ($select_and != '')) {
            $tmp['select_and'] = $select_and;
        }
        if ($blogs != -1) {
            $tmp['blog'] = $blogs;
        }
        $archive_url = build_url($tmp, $zone);

        $submit_url = new Tempcode();
        $management_page = ($blogs === 1) ? 'cms_blogs' : 'cms_news';
        if ((($blogs !== 1) || (has_privilege(get_member(), 'have_personal_category', 'cms_news'))) && (has_actual_page_access(null, $management_page, null, null)) && (has_submit_permission(($blogs === 1) ? 'mid' : 'high', get_member(), get_ip_address(), $management_page))) {
            $map2 = ['page' => $management_page, 'type' => 'add'];
            if (is_numeric($select)) {
                $map2['cat'] = $select; // select news cat by default, if we are only showing one news cat in this block
            } elseif ($select != '*') {
                $pos_a = strpos($select, ',');
                $pos_b = strpos($select, '-');
                if ($pos_a !== false) {
                    $first_cat = substr($select, 0, $pos_a);
                } elseif ($pos_b !== false) {
                    $first_cat = substr($select, 0, $pos_b);
                } else {
                    $first_cat = '';
                }
                if (is_numeric($first_cat)) {
                    $map2['cat'] = $first_cat;
                }
            }
            $map2['redirect'] = protect_url_parameter(SELF_REDIRECT_RIP);
            $submit_url = build_url($map2, get_module_zone($management_page));
        }

        // Block title
        $_title = isset($map['title']) ? protect_from_escaping(escape_html($map['title'])) : do_lang_tempcode(($blogs == 1) ? 'BLOGS_POSTS' : 'NEWS');

        // Feed URLs
        $atom_url = new Tempcode();
        $rss_url = new Tempcode();
        if (addon_installed('syndication_blocks')) {
            if (get_option('is_rss_advertised') == '1') {
                $atom_url = make_string_tempcode(find_script('backend') . '?type=atom&mode=news&select=' . urlencode($select));
                $atom_url->attach(symbol_tempcode('KEEP'));
                $rss_url = make_string_tempcode(find_script('backend') . '?type=rss2&mode=news&select=' . urlencode($select));
                $rss_url->attach(symbol_tempcode('KEEP'));
            }
        }

        // Wipe out management/feed URLs if no links was requested
        if ((isset($map['no_links'])) && ($map['no_links'] == '1')) {
            $submit_url = new Tempcode();
            $archive_url = new Tempcode();
            $atom_url = new Tempcode();
            $rss_url = new Tempcode();
        }

        // No entries?
        if (empty($rows)) {
            if ((!isset($map['render_if_empty'])) || ($map['render_if_empty'] == '0')) {
                return do_template('BLOCK_NO_ENTRIES', [
                    '_GUID' => '9d7065af4dd4026ffb34243fd931f99d',
                    'BLOCK_ID' => $block_id,
                    'HIGH' => false,
                    'TITLE' => $_title,
                    'MESSAGE' => do_lang_tempcode(($blogs == 1) ? 'BLOG_NO_NEWS' : 'NO_NEWS'),
                    'ADD_NAME' => do_lang_tempcode(($blogs == 1) ? 'ADD_NEWS_BLOG' : 'ADD_NEWS'),
                    'SUBMIT_URL' => $submit_url,
                ]);
            }
        }

        // Pagination
        $pagination = null;
        if ($do_pagination) {
            require_code('templates_pagination');
            $pagination = pagination(do_lang_tempcode('NEWS'), $start, $block_id . '_start', $display_slides + $display_summaries + $display_briefs, $block_id . '_max', $max_rows);
        }

        $tpl = 'BLOCK_MAIN_NEWS';
        if ($render_mode == 'grid') {
            $tpl .= '_GRID';
        }

        return do_template($tpl, [
            '_GUID' => '01f5fbd2b0c7c8f249023ecb4254366e',
            'BLOCK_ID' => $block_id,
            'BLOCK_PARAMS' => comma_list_arr_to_str(['block_id' => $block_id] + $map),
            'BLOG' => $blogs === 1,
            'TITLE' => $_title,
            'SLIDER' => $slider,
            'SUMMARY_CONTENT' => $summary_content,
            'BRIEF_CONTENT' => $brief_content,
            'FILTER' => $filter,
            'ARCHIVE_URL' => $archive_url,
            'SUBMIT_URL' => $submit_url,
            'RSS_URL' => $rss_url,
            'ATOM_URL' => $atom_url,
            'PAGINATION' => $pagination,
            'START' => strval($start),
            'MAX' => strval($display_slides + $display_summaries + $display_briefs),
            'START_PARAM' => $block_id . '_start',
            'MAX_PARAM' => $block_id . '_max',
        ]);
    }

    /**
     * Generate a template-ready map for a news article.
     *
     * @param  array $myrow The article row
     * @param  boolean $optimise Whether to optimise the tempcode-output
     * @param  boolean $show_in_full Whether to show the article in full (rather than a summary)
     * @param  boolean $show_author Whether to show the author
     * @param  array $prop_url Shared URL parameters
     * @param  boolean $allow_comments_shared Whether comments are allowed
     * @param  URLPATH $base_url Base URL
     * @param  boolean $attach_to_url_filter Whether to propagate Filtercode in URLs
     * @param  string $zone Zone
     * @param  integer $blogs Whether we are showing blogs (1=only, 0=no, -1=don't care)
     * @return array The template map
     */
    protected function assemble_template_map(array $myrow, bool $optimise, bool $show_in_full, bool $show_author, array $prop_url, bool $allow_comments_shared, string $base_url, bool $attach_to_url_filter, string $zone, int $blogs) : array
    {
        $just_news_row = db_map_restrict($myrow, ['id', 'title', 'news', 'news_article']);

        // Basic details
        $id = $myrow['id'];
        $date = get_timezoned_date_time_tempcode($myrow['date_and_time']);
        $news_title = get_translated_tempcode('news', $just_news_row, 'title');
        $news_title_plain = get_translated_text($myrow['title']);

        // Author
        $author_url = new Tempcode();
        if ($show_author) {
            $url_map = ['page' => 'authors', 'type' => 'browse', 'id' => $myrow['author']];
            if ($attach_to_url_filter) {
                $url_map += propagate_filtercode();
            }
            $author_url = build_url($url_map, get_module_zone('authors'));
        }
        $author = $myrow['author'];

        // Text
        $truncate = false;
        if ($optimise) {
            if ($show_in_full) {
                $news = get_translated_tempcode__and_simplify('news', $just_news_row, 'news_article');
                if ($news->is_empty()) {
                    $news = get_translated_tempcode__and_simplify('news', $just_news_row, 'news');
                }
            } else {
                $news = get_translated_tempcode__and_simplify('news', $just_news_row, 'news');
                if ($news->is_empty()) {
                    $news = get_translated_tempcode__and_simplify('news', $just_news_row, 'news_article');
                    $truncate = true;
                }
            }
        } else {
            if ($show_in_full) {
                $news = get_translated_tempcode('news', $just_news_row, 'news_article');
                if ($news->is_empty()) {
                    $news = get_translated_tempcode('news', $just_news_row, 'news');
                }
            } else {
                $news = get_translated_tempcode('news', $just_news_row, 'news');
                if ($news->is_empty()) {
                    $news = get_translated_tempcode('news', $just_news_row, 'news_article');
                    $truncate = true;
                }
            }
        }

        // URL
        $tmp = ['page' => ($zone == '_SELF' && running_script('index')) ? get_page_name() : 'news', 'type' => 'view', 'id' => $id] + $prop_url;
        $full_url = build_url($tmp, $zone);

        // Category
        $news_cat_row = get_news_cat_row($myrow['news_category']);
        $category = get_translated_text($news_cat_row['nc_title']);
        $category_url = build_url(['page' => 'news', 'type' => 'browse', 'id' => $news_cat_row['id']], $zone);

        if ($myrow['news_image_url'] != '') {
            $rep_image_url = $myrow['news_image_url'];
            if (url_is_local($rep_image_url)) {
                $rep_image_url = $base_url . '/' . $rep_image_url;
            }
        } else {
            $rep_image_url = get_news_category_image_url($news_cat_row['nc_img']);
        }

        // SEO
        $seo_bits = (get_value('disable_tags') === '1') ? ['', ''] : seo_meta_get_for('news', strval($id));

        // Render
        $map2 = [
            'GIVE_CONTEXT' => false,
            'TAGS' => get_loaded_tags('news', explode(',', $seo_bits[0])),
            'ID' => strval($id),
            'TRUNCATE' => $truncate,
            'BLOG' => $blogs === 1,
            'SUBMITTER' => strval($myrow['submitter']),
            'CATEGORY' => $category,
            'CATEGORY_URL' => $category_url,
            '_CATEGORY' => strval($myrow['news_category']),
            'REP_IMAGE_URL' => $rep_image_url,
            'DATE' => $date,
            'DATE_RAW' => strval($myrow['date_and_time']),
            'NEWS_TITLE' => $news_title,
            'NEWS_TITLE_PLAIN' => $news_title_plain,
            'AUTHOR' => $author,
            'AUTHOR_URL' => $author_url,
            'NEWS' => $news,
            'FULL_URL' => $full_url,
        ];
        if (($allow_comments_shared) && ($myrow['allow_comments'] >= 1)) {
            $map2['COMMENT_COUNT'] = '1';
        }

        return $map2;
    }

    /**
     * Generate Selectcode SQL.
     *
     * @param  string $select The Selectcode
     * @return string The SQL
     */
    protected function generate_selectcode_sql(string $select) : string
    {
        require_code('selectcode');
        $selects_1 = selectcode_to_sqlfragment($select, 'r.id', 'news_categories', null, 'r.news_category', 'id');
        $selects_2 = selectcode_to_sqlfragment($select, 'r.id', 'news_categories', null, 'd.news_entry_category', 'id');
        if ((strpos($select, '~') === false) && (strpos($select, '!') === false)) {
            $q_filter = '(' . $selects_1 . ' OR ' . $selects_2 . ')';
        } else {
            $q_filter = '(' . $selects_1 . ' AND (' . $selects_2 . ' OR d.news_entry_category IS NULL))';
        }
        return $q_filter;
    }
}
