<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    forum_blocks
 */

/**
 * Block class.
 */
class Block_side_forum_news
{
    /**
     * Find details of the block.
     *
     * @return ?array Map of block info (null: block is disabled)
     */
    public function info() : ?array
    {
        $info = [];
        $info['author'] = 'Chris Graham';
        $info['organisation'] = 'Composr';
        $info['hacked_by'] = null;
        $info['hack_version'] = null;
        $info['version'] = 2;
        $info['locked'] = false;
        $info['min_cms_version'] = 11.0;
        $info['addon'] = 'forum_blocks';
        $info['parameters'] = ['param', 'forum', 'date_key', 'title', 'check'];
        return $info;
    }

    /**
     * Find caching details for the block.
     *
     * @return ?array Map of cache details (cache_on and ttl) (null: block is disabled)
     */
    public function caching_environment() : ?array
    {
        $info = [];
        $info['cache_on'] = <<<'PHP'
        [
            array_key_exists('title', $map) ? $map['title'] : '',
            array_key_exists('param', $map) ? $map['param'] : 6,
            array_key_exists('forum', $map) ? $map['forum'] : 'Announcements',
            array_key_exists('date_key', $map) ? $map['date_key'] : 'firsttime',
            array_key_exists('check', $map) ? ($map['check'] == '1') : true,
        ]
PHP;
        $info['special_cache_flags'] = CACHE_AGAINST_DEFAULT | CACHE_AGAINST_PERMISSIVE_GROUPS;
        $info['ttl'] = 15;
        return $info;
    }

    /**
     * Execute the block.
     *
     * @param  array $map A map of parameters
     * @return Tempcode The result of execution
     */
    public function run(array $map) : object
    {
        $error_msg = new Tempcode();
        if (!addon_installed__messaged('forum_blocks', $error_msg)) {
            return $error_msg;
        }
        if (!addon_installed__messaged('news_shared', $error_msg)) {
            return $error_msg;
        }

        if (has_no_forum()) {
            return do_template('RED_ALERT', ['_GUID' => 'f1665d0139bf5a40ae38749079b890cd', 'TEXT' => do_lang_tempcode('NO_FORUM_INSTALLED')]);
        }

        require_css('news');
        require_lang('news');

        $block_id = get_block_id($map);

        $check_perms = array_key_exists('check', $map) ? ($map['check'] == '1') : true;

        $limit = empty($map['param']) ? 6 : intval($map['param']);
        $forum_name = array_key_exists('forum', $map) ? $map['forum'] : do_lang('NEWS');

        $date_key = array_key_exists('date_key', $map) ? $map['date_key'] : 'firsttime';

        $archive_url = null;
        $submit_url = new Tempcode();

        $forum_ids = [];
        if ((get_forum_type() == 'cns') && (preg_match('#^[\d\-\*,]*$#', $forum_name) != 0)) {
            require_code('selectcode');
            $forum_names = array_map('strval', selectcode_to_idlist_using_db($forum_name, 'id', 'f_forums', 'f_forums', 'f_parent_forum_id', 'f_parent_forum_id', 'id', true, true, $GLOBALS['FORUM_DB']));
        } else {
            $forum_names = explode(',', $forum_name);
        }
        if (($forum_name == do_lang('NEWS')) && (get_forum_type() == 'cns') && ($GLOBALS['FORUM_DRIVER']->forum_id_from_name($forum_name) === null)) {
            $forum_names = [do_lang('DEFAULT_FORUM_TITLE')];
        }
        foreach ($forum_names as $forum_name) {
            $forum_name = trim($forum_name);

            if ($forum_name == '<announce>') {
                $forum_id = null;
            } else {
                $forum_id = is_numeric($forum_name) ? intval($forum_name) : $GLOBALS['FORUM_DRIVER']->forum_id_from_name($forum_name);
            }

            if ($forum_id !== null) {
                if ((get_forum_type() == 'cns') && ($check_perms)) {
                    if (!has_category_access(get_member(), 'forums', strval($forum_id))) {
                        continue;
                    }
                }

                $forum_ids[$forum_id] = $forum_name;
            }

            if ($forum_id !== null) {
                $forum_ids[$forum_id] = $forum_name;
                if ($archive_url === null) {
                    $archive_url = $GLOBALS['FORUM_DRIVER']->forum_url($forum_id, true); // First forum will count as archive
                    if (get_forum_type() == 'cns') {
                        $submit_url = build_url(['page' => 'topics', 'type' => 'new_topic', 'id' => $forum_id], get_module_zone('topics'));
                    }
                }
            }
        }

        $_title = do_lang_tempcode('NEWS');
        if ((array_key_exists('title', $map)) && ($map['title'] != '')) {
            $_title = protect_from_escaping(escape_html($map['title']));
        }

        $max_rows = 0;
        $topics = $GLOBALS['FORUM_DRIVER']->show_forum_topics($forum_ids, $limit, 0, $max_rows, '', '', false, $date_key);

        $news = [];
        if ($topics !== null) {
            sort_maps_by($topics, $date_key);
            $topics = array_reverse($topics, false);

            foreach ($topics as $topic) {
                $topic_url = $GLOBALS['FORUM_DRIVER']->topic_url($topic['id'], $forum_name, true);
                $title = $topic['title'];
                $date = get_timezoned_date_tempcode($topic[$date_key]);

                $news[] = [
                    'REPLIES' => strval($topic['num']),
                    'FIRSTTIME' => strval($topic['firsttime']),
                    'LASTTIME' => strval($topic['lasttime']),
                    'CLOSED' => strval($topic['closed']),
                    'FIRSTUSERNAME' => $topic['firstusername'],
                    'LASTUSERNAME' => $topic['lastusername'],
                    'FIRSTMEMBERID' => strval($topic['firstmemberid']),
                    'LASTMEMBERID' => strval($topic['lastmemberid']),
                    '_DATE' => strval($topic[$date_key]),
                    'DATE' => $date,
                    'FULL_URL' => $topic_url,
                    'NEWS_TITLE' => escape_html($title),
                ];
            }
        }

        return do_template('BLOCK_SIDE_FORUM_NEWS', [
            '_GUID' => '174fa5ce0d35d9b49dca6347c66494a5',
            'BLOCK_ID' => $block_id,
            'FORUM_NAME' => array_key_exists('forum', $map) ? $map['forum'] : do_lang('NEWS'),
            'TITLE' => $_title,
            'NEWS' => $news,
            'SUBMIT_URL' => $submit_url,
            'ARCHIVE_URL' => ($archive_url === null) ? '' : $archive_url,
        ]);
    }
}
