<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    news
 */

/**
 * Block class.
 */
class Block_side_news_categories
{
    /**
     * Find details of the block.
     *
     * @return ?array Map of block info (null: block is disabled)
     */
    public function info() : ?array
    {
        $info = [];
        $info['author'] = 'Chris Graham';
        $info['organisation'] = 'Composr';
        $info['hacked_by'] = null;
        $info['hack_version'] = null;
        $info['version'] = 2;
        $info['locked'] = false;
        $info['min_cms_version'] = 11.0;
        $info['addon'] = 'news';
        $info['parameters'] = ['select', 'check'];
        return $info;
    }

    /**
     * Find caching details for the block.
     *
     * @return ?array Map of cache details (cache_on and ttl) (null: block is disabled)
     */
    public function caching_environment() : ?array
    {
        $info = [];
        $info['cache_on'] = <<<'PHP'
        [
            array_key_exists('select', $map) ? $map['select'] : '',
            array_key_exists('check', $map) ? ($map['check'] == '1') : true
        ]
PHP;
        $info['special_cache_flags'] = CACHE_AGAINST_DEFAULT | CACHE_AGAINST_PERMISSIVE_GROUPS;
        $info['ttl'] = 60 * 24;
        return $info;
    }

    /**
     * Execute the block.
     *
     * @param  array $map A map of parameters
     * @return Tempcode The result of execution
     */
    public function run(array $map) : object
    {
        $error_msg = new Tempcode();
        if (!addon_installed__messaged('news', $error_msg)) {
            return $error_msg;
        }
        if (!addon_installed__messaged('news_shared', $error_msg)) {
            return $error_msg;
        }

        require_lang('news');

        $block_id = get_block_id($map);

        $check_perms = !array_key_exists('check', $map) || ($map['check'] == '1');

        $select = array_key_exists('select', $map) ? $map['select'] : '';
        if ($select == '') {
            $q_filter = '1=1';
        } else {
            require_code('selectcode');
            $q_filter = selectcode_to_sqlfragment($select, 'r.id', 'news_categories', null, 'r.id', 'id');
        }

        $cnt = $GLOBALS['SITE_DB']->query_select_value('news_categories', 'COUNT(*)', ['nc_owner' => null]);
        if ($cnt > 100) {
            $_categories = $GLOBALS['SITE_DB']->query('SELECT r.* FROM ' . get_table_prefix() . 'news_categories r WHERE nc_owner IS NULL AND EXISTS(SELECT * FROM ' . get_table_prefix() . 'news n WHERE n.news_category=r.id AND n.validated=1) AND ' . $q_filter);
        } else {
            $_categories = $GLOBALS['SITE_DB']->query_select('news_categories r', ['*'], ['nc_owner' => null], ' AND ' . $q_filter);
        }

        $categories = [];
        foreach ($_categories as $category) {
            if ((!$check_perms) || (has_category_access(get_member(), 'news', strval($category['id'])))) {
                $join = ' LEFT JOIN ' . get_table_prefix() . 'news_category_entries d ON d.news_entry=p.id';
                $count = $GLOBALS['SITE_DB']->query_value_if_there('SELECT COUNT(*) FROM ' . get_table_prefix() . 'news p' . $join . ' WHERE validated=1 AND (news_entry_category=' . strval($category['id']) . ' OR news_category=' . strval($category['id']) . ')');
                if ($count > 0) {
                    $category['_nc_title'] = get_translated_text($category['nc_title']);
                    $categories[] = [$category, $count, $category['_nc_title']];
                }
            }
        }
        if (empty($categories)) {
            foreach ($_categories as $category) {
                if ((!$check_perms) || (has_category_access(get_member(), 'news', strval($category['id'])))) {
                    $count = $GLOBALS['SITE_DB']->query_value_if_there('SELECT COUNT(*) FROM ' . get_table_prefix() . 'news p WHERE validated=1 AND news_category=' . strval($category['id']));
                    $category['_nc_title'] = get_translated_text($category['nc_title']);
                    $categories[] = [$category, $count, $category['_nc_title']];
                }
            }
        }

        sort_maps_by($categories, 2);

        $_categories = [];
        foreach ($categories as $_category) {
            list($category, $count) = $_category;
            $url = build_url(['page' => 'news', 'type' => 'browse', 'id' => $category['id']], get_module_zone('news'));
            $name = $category['_nc_title'];
            $_categories[] = [
                'URL' => $url,
                'NAME' => $name,
                '_COUNT' => strval($count),
                'COUNT' => integer_format($count, 0),
            ];
        }

        return do_template('BLOCK_SIDE_NEWS_CATEGORIES', [
            '_GUID' => 'b47a0047247096373e5aa626348c4ebb',
            'BLOCK_ID' => $block_id,
            'CATEGORIES' => $_categories,
            'PRE' => '',
            'POST' => '',
        ]);
    }
}
