<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    language_block
 */

/**
 * Block class.
 */
class Block_top_language
{
    /**
     * Find details of the block.
     *
     * @return ?array Map of block info (null: block is disabled)
     */
    public function info() : ?array
    {
        $info = [];
        $info['author'] = 'Chris Graham';
        $info['organisation'] = 'Composr';
        $info['hacked_by'] = null;
        $info['hack_version'] = null;
        $info['version'] = 1;
        $info['locked'] = false;
        $info['min_cms_version'] = 11.0;
        $info['addon'] = 'language_block';
        $info['parameters'] = [];
        return $info;
    }

    /**
     * Find caching details for the block.
     *
     * @return ?array Map of cache details (cache_on and ttl) (null: block is disabled)
     */
    public function caching_environment() : ?array
    {
        $info = [];
        $info['cache_on'] = <<<'PHP'
        [
            either_param_string('lang', ''),
            user_lang(),
        ]
PHP;
        $info['ttl'] = 60;
        return $info;
    }

    /**
     * Execute the block.
     *
     * @param  array $map A map of parameters
     * @return Tempcode The result of execution
     */
    public function run(array $map) : object
    {
        $error_msg = new Tempcode();
        if (!addon_installed__messaged('language_block', $error_msg)) {
            return $error_msg;
        }

        if (!multi_lang()) {
            return new Tempcode();
        }

        if (get_param_string('lang', '') != '') {
            return new Tempcode(); // Content UI has an active language, so don't provide a selector
        }

        require_code('lang2');

        $block_id = get_block_id($map);

        $langs = find_all_langs();

        foreach (array_keys($langs) as $lang) {
            $langs[$lang] = [
                'FULL_NAME' => lookup_language_full_name($lang),
                'COUNTRY_FLAG' => cms_strtolower_ascii($this->get_lang_country($lang)),
            ];
        }
        sort_maps_by($langs, 'FULL_NAME');

        $current_lang = user_lang();

        return do_template('BLOCK_TOP_LANGUAGE', [
            '_GUID' => '6dd7dd434722d7fd958773bd08e838c7',
            'BLOCK_ID' => $block_id,
            'LANGS' => $langs,
            'CURRENT_LANG_FULL_NAME' => lookup_language_full_name($current_lang),
            'CURRENT_LANG_COUNTRY_FLAG' => cms_strtolower_ascii($this->get_lang_country($current_lang)),
            'CURRENT_LANG' => $current_lang,
        ]);
    }

    /**
     * Execute the block.
     *
     * @param  ID_TEXT $lang_iso ISO language code
     * @return ID_TEXT ISO country code
     */
    protected function get_lang_country(string $lang_iso) : string
    {
        // Mappings between language and primary country, for those where the ISO codes differ
        //  This list is intentionally incomplete.
        //  It is biased towards languages which have many native speakers, and are used by people likely to be accessing a multi-language website.
        // TODO: Automated test to check the accuracy of this?
        static $langs_to_countries = null;
        if ($langs_to_countries === null) {
            $langs_to_countries = [
                'ZH' => 'CN',
                'EN' => (get_option('yeehaw') == '1') ? 'US' : 'GB',
                'AR' => 'SA',
                'JA' => 'JP',
                'JV' => 'ID',
                'MS' => 'MY',
                'VI' => 'VN',
                'KO' => 'KR',
                'FA' => 'IR',
                'UK' => 'UA',
                'PS' => 'AF',

                // Many major native languages of Pakistan
                'UR' => 'PK',
                'PA' => 'PK',
                'SD' => 'PK',

                // Many major native languages of India
                'ML' => 'IN',
                'KN' => 'IN',
                'GU' => 'IN',
                'OR' => 'IN',
                'TA' => 'IN',
                'HI' => 'IN',
                'BN' => 'IN',
                'TE' => 'IN',
                'MR' => 'IN',
            ];
        }

        if (array_key_exists($lang_iso, $langs_to_countries)) {
            $country = $langs_to_countries[$lang_iso];
        } else {
            $country = $lang_iso;
        }
        if (find_theme_image('flags_large/' . cms_strtolower_ascii($country), true) == '') {
            $country = '';
        }

        return $country;
    }
}
