<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    chat
 */

/**
 * Get the list of all available chat sound effects.
 *
 * @param  boolean $only_overridable Map to null if it is not overridable
 * @return array All available sound effects (mapping between base code, and actual code)
 */
function get_effect_set(bool $only_overridable = false) : array
{
    $effects = [
        'message_received' => 'message_received',
        'message_background' => 'message_background',
        'message_initial' => $only_overridable ? null : 'message_initial',
        'message_sent' => $only_overridable ? null : 'message_sent',
        'contact_on' => 'contact_on',
        'contact_off' => 'contact_off',
        'invited' => 'invited',
        'you_connect' => $only_overridable ? null : 'you_connect',
    ];

    return $effects;
}

/**
 * Get a list of template mappings for the current member, between sound effect IDs and the URLs to the mp3 files.
 *
 * @param  boolean $full_urls Whether to use full URLs in the mappings
 * @param  ?MEMBER $for_member Get settings overridden for this specific member (null: global settings)
 * @param  boolean $all_members Get global settings and settings overridden for all members (if this is true we'd expect $for_member to be null)
 * @return array The template mappings
 */
function get_effect_settings(bool $full_urls = false, ?int $for_member = null, bool $all_members = false) : array
{
    static $cache = [];
    if (isset($cache[$full_urls][$for_member][$all_members])) {
        return $cache[$full_urls][$for_member][$all_members];
    }

    $effects = get_effect_set($for_member !== null);

    global $EFFECT_SETTINGS_ROWS;
    if ($EFFECT_SETTINGS_ROWS === null) {
        $EFFECT_SETTINGS_ROWS = collapse_2d_complexity('s_effect_id', 's_url', $GLOBALS['SITE_DB']->query_select('chat_sound_effects', ['s_url', 's_effect_id'], ['s_member' => get_member()]));
    }
    $effect_settings = [];
    if ($all_members) {
        foreach (array_keys($EFFECT_SETTINGS_ROWS) as $effect_id) {
            $matches = [];
            if ((!array_key_exists($effect_id, $effects)) && (preg_match('#^(.*)_(\d+)$#', $effect_id, $matches) != 0) && (array_key_exists($matches[1], $effects))) {
                $effects[$effect_id] = $matches[1];
            }
        }
    }
    foreach ($effects as $effect => $base_effect_code) {
        if ($base_effect_code === null) {
            continue;
        }

        if ($for_member === null) { // Global settings
            if (array_key_exists($effect, $EFFECT_SETTINGS_ROWS)) {
                $member_setting = $EFFECT_SETTINGS_ROWS[$effect];
            } else {
                $member_setting = 'data_custom/sounds/' . $effect . '.mp3';
                if (!file_exists(get_custom_file_base() . '/' . $member_setting)) {
                    $member_setting = 'data/sounds/' . $effect . '.mp3';
                }
                if (!file_exists(get_file_base() . '/' . $member_setting)) {
                    $member_setting = '';
                }
            }
        } else { // Overridden settings
            if (array_key_exists($effect . '_' . strval($for_member), $EFFECT_SETTINGS_ROWS)) {
                $member_setting = $EFFECT_SETTINGS_ROWS[$effect . '_' . strval($for_member)];
            } else {
                $member_setting = '-1';
            }
        }
        $effect_settings[$effect] = [
            'KEY' => $effect,
            'VALUE' => (($full_urls && ($member_setting != '')) ? (((substr($member_setting, 0, 12) == 'data_custom/') ? get_custom_base_url() : get_base_url()) . '/') : '') . $member_setting,
            'EFFECT_TITLE' => do_lang('CHAT_EFFECT_' . $base_effect_code),
        ];
    }

    $cache[$full_urls][$for_member][$all_members] = $effect_settings;

    return $effect_settings;
}
