<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    core_cns
 */

/**
 * Add a usergroup.
 *
 * @param  SHORT_TEXT $name The name of the usergroup
 * @param  BINARY $is_default Whether members are automatically put into the when they join
 * @param  BINARY $is_super_admin Whether members of this usergroup are all super-administrators
 * @param  BINARY $is_super_moderator Whether members of this usergroup are all super-moderators
 * @param  SHORT_TEXT $title The title for primary members of this usergroup that don't have their own title
 * @param  URLPATH $rank_image The rank image for this
 * @param  ?GROUP $promotion_target The that members of this usergroup get promoted to at point threshold (null: no promotion prospects)
 * @param  ?integer $promotion_threshold The point threshold for promotion (null: no promotion prospects)
 * @param  BINARY $promotion_approval Whether usergroup promotion should require manual approval
 * @param  ?MEMBER $group_leader The leader of this usergroup (null: none)
 * @param  ?integer $flood_control_submit_secs The number of seconds that members of this usergroup must endure between submits (group 'best of' applies). 0 means N/A. (null: average for existing usergroups)
 * @param  ?integer $flood_control_access_secs The number of seconds that members of this usergroup must endure between accesses (group 'best of' applies). 0 means N/A. (null: average for existing usergroups)
 * @param  ?integer $max_daily_upload_mb The number of megabytes that members of this usergroup may attach per day (group 'best of' applies). (null: average for existing usergroups)
 * @param  ?integer $max_attachments_per_post The number of attachments that members of this usergroup may attach to something (group 'best of' applies). (null: average for existing usergroups)
 * @param  ?integer $max_avatar_width The maximum avatar width that members of this usergroup may have (group 'best of' applies). (null: average for existing usergroups)
 * @param  ?integer $max_avatar_height The maximum avatar height that members of this usergroup may have (group 'best of' applies). (null: average for existing usergroups)
 * @param  ?integer $max_post_length_comcode The maximum post length that members of this usergroup may make (group 'best of' applies). (null: average for existing usergroups)
 * @param  ?integer $max_sig_length_comcode The maximum signature length that members of this usergroup may make (group 'best of' applies). (null: average for existing usergroups)
 * @param  ?integer $gift_points_base The number of gift points that members of this usergroup start with (group 'best of' applies). (null: average for existing usergroups)
 * @param  ?integer $gift_points_per_day The number of gift points that members of this usergroup get per day (group 'best of' applies). (null: average for existing usergroups)
 * @param  BINARY $enquire_on_new_ips Whether e-mail confirmation is needed for new IP addresses seen for any member of this usergroup (group 'best of' applies)
 * @param  BINARY $is_presented_at_install Whether the usergroup is presented for joining at joining (implies anyone may be in the, but only choosable at joining)
 * @param  BINARY $hidden Whether the name and membership of the is hidden
 * @param  ?integer $order The display order this will be given, relative to other usergroups. Lower numbered usergroups display before higher numbered usergroups (null: next).
 * @param  BINARY $rank_image_pri_only Whether the rank image will not be shown for secondary membership
 * @param  BINARY $open_membership Whether members may join this usergroup without requiring any special permission
 * @param  BINARY $is_private_club Whether this usergroup is a private club. Private clubs may be managed in the CMS zone, and do not have any special permissions - except over their own associated forum.
 * @param  boolean $uniqify Whether to force the title as unique, if there's a conflict
 * @param  boolean $comes_with_permissions Whether permissions should be auto-copied
 * @return AUTO_LINK The ID of the new usergroup
 */
function cns_make_group(string $name, int $is_default = 0, int $is_super_admin = 0, int $is_super_moderator = 0, string $title = '', string $rank_image = '', ?int $promotion_target = null, ?int $promotion_threshold = null, int $promotion_approval = 0, ?int $group_leader = null, ?int $flood_control_submit_secs = null, ?int $flood_control_access_secs = null, ?int $max_daily_upload_mb = null, ?int $max_attachments_per_post = null, ?int $max_avatar_width = null, ?int $max_avatar_height = null, ?int $max_post_length_comcode = null, ?int $max_sig_length_comcode = null, ?int $gift_points_base = null, ?int $gift_points_per_day = null, int $enquire_on_new_ips = 0, int $is_presented_at_install = 0, int $hidden = 0, ?int $order = null, int $rank_image_pri_only = 1, int $open_membership = 0, int $is_private_club = 0, bool $uniqify = false, bool $comes_with_permissions = true) : int
{
    require_code('global4');
    prevent_double_submit('ADD_GROUP', null, $name);

    require_code('form_templates');

    $flood_control_submit_secs = take_param_int_modeavg($flood_control_submit_secs, 'g_flood_control_submit_secs', 'f_groups', 0);
    $flood_control_access_secs = take_param_int_modeavg($flood_control_access_secs, 'g_flood_control_access_secs', 'f_groups', 0);
    $max_daily_upload_mb = take_param_int_modeavg($max_daily_upload_mb, 'g_max_daily_upload_mb', 'f_groups', 70);
    $max_attachments_per_post = take_param_int_modeavg($max_attachments_per_post, 'g_max_attachments_per_post', 'f_groups', 50);
    $max_avatar_width = take_param_int_modeavg($max_avatar_width, 'g_max_avatar_width', 'f_groups', 100);
    $max_avatar_height = take_param_int_modeavg($max_avatar_height, 'g_max_avatar_height', 'f_groups', 100);
    $max_post_length_comcode = take_param_int_modeavg($max_post_length_comcode, 'g_max_post_length_comcode', 'f_groups', 30000);
    $max_sig_length_comcode = take_param_int_modeavg($max_sig_length_comcode, 'g_max_sig_length_comcode', 'f_groups', 700);
    $gift_points_base = take_param_int_modeavg($gift_points_base, 'g_gift_points_base', 'f_groups', 25);
    $gift_points_per_day = take_param_int_modeavg($gift_points_per_day, 'g_gift_points_per_day', 'f_groups', 1);

    if (!running_script('stress_test_loader')) {
        $test = $GLOBALS['FORUM_DB']->query_select_value_if_there('f_groups', 'id', [$GLOBALS['FORUM_DB']->translate_field_ref('g_name') => $name]);
        if ($test !== null) {
            if ($uniqify) {
                $name .= '_' . uniqid('', false);
            } else {
                warn_exit(do_lang_tempcode('ALREADY_EXISTS', escape_html($name)));
            }
        }
    }

    if ($is_super_admin === null) {
        $is_super_admin = 0;
    }
    if ($is_super_moderator === null) {
        $is_super_moderator = 0;
    }

    if (!running_script('stress_test_loader')) {
        if ($order === null) {
            $order = $GLOBALS['FORUM_DB']->query_select_value('f_groups', 'MAX(g_order)');
            if ($order === null) {
                $order = 0;
            } else {
                $order++;
            }
        }
    } else {
        $order = 100;
    }

    $map = [
        'g_is_default' => $is_default,
        'g_is_presented_at_install' => $is_presented_at_install,
        'g_is_super_admin' => $is_super_admin,
        'g_is_super_moderator' => $is_super_moderator,
        'g_group_lead_member' => $group_leader,
        'g_promotion_target_group' => $promotion_target,
        'g_promotion_threshold' => $promotion_threshold,
        'g_promotion_approval' => $promotion_approval,
        'g_flood_control_submit_secs' => $flood_control_submit_secs,
        'g_flood_control_access_secs' => $flood_control_access_secs,
        'g_max_daily_upload_mb' => $max_daily_upload_mb,
        'g_max_attachments_per_post' => $max_attachments_per_post,
        'g_max_avatar_width' => $max_avatar_width,
        'g_max_avatar_height' => $max_avatar_height,
        'g_max_post_length_comcode' => $max_post_length_comcode,
        'g_max_sig_length_comcode' => $max_sig_length_comcode,
        'g_gift_points_base' => $gift_points_base,
        'g_gift_points_per_day' => $gift_points_per_day,
        'g_enquire_on_new_ips' => $enquire_on_new_ips,
        'g_rank_image' => $rank_image,
        'g_hidden' => $hidden,
        'g_order' => $order,
        'g_rank_image_pri_only' => $rank_image_pri_only,
        'g_open_membership' => $open_membership,
        'g_is_private_club' => $is_private_club,
    ];
    $map += insert_lang('g_name', $name, 2, $GLOBALS['FORUM_DB']);
    $map += insert_lang('g_title', $title, 2, $GLOBALS['FORUM_DB']);
    $group_id = $GLOBALS['FORUM_DB']->query_insert('f_groups', $map, true);

    if (((!running_script('install')) || ($name == do_lang('PROBATION'))) && ($is_private_club == 0) && ($comes_with_permissions)) {
        // Copy permissions from members...
        cns_copy_group_permissions($group_id);
    }

    log_it('ADD_GROUP', strval($group_id), $name);

    if ((addon_installed('commandr')) && (!running_script('install')) && (!get_mass_import_mode())) {
        require_code('resource_fs');
        generate_resource_fs_moniker('group', strval($group_id), null, null, true);
    }

    if ($is_private_club == 1) {
        require_code('notifications');
        $subject = do_lang('NEW_CLUB_NOTIFICATION_MAIL_SUBJECT', get_site_name(), $name);
        $view_url = build_url(['page' => 'groups', 'type' => 'view', 'id' => $group_id], get_module_zone('groups'), [], false, false, true);
        $mail = do_notification_lang('NEW_CLUB_NOTIFICATION_MAIL', get_site_name(), comcode_escape($name), [comcode_escape($view_url->evaluate())]);
        dispatch_notification('cns_club', null, $subject, $mail);
    }

    persistent_cache_delete('GROUPS_COUNT');
    persistent_cache_delete('GROUPS_COUNT_PO');
    persistent_cache_delete('GROUPS');
    persistent_cache_delete('GROUPS_PO');
    persistent_cache_delete('SUPER_ADMIN_GROUPS');
    persistent_cache_delete('SUPER_MODERATOR_GROUPS');
    persistent_cache_delete('OPEN_GROUPS');

    global $USERGROUP_LIST_CACHE, $MODERATOR_GROUP_CACHE;
    $USERGROUP_LIST_CACHE = null;
    $MODERATOR_GROUP_CACHE = null;

    require_code('member_mentions');
    dispatch_member_mention_notifications('group', strval($group_id));

    require_code('sitemap_xml');
    notify_sitemap_node_add('_SEARCH:groups:view:' . strval($group_id), null, null, SITEMAP_IMPORTANCE_LOW, 'yearly', $hidden == 0);

    return $group_id;
}

/**
 * Copy permissions from the first default group to another group.
 *
 * @param  AUTO_LINK $group_id The usergroup to copy to
 */
function cns_copy_group_permissions(int $group_id)
{
    require_code('cns_groups');
    $group_members = get_first_default_group();

    $member_access = $GLOBALS['SITE_DB']->query_select('group_privileges', ['*'], ['group_id' => $group_members]);
    foreach ($member_access as &$access) {
        $access['group_id'] = $group_id;
    }
    $GLOBALS['SITE_DB']->query_insert('group_privileges', $GLOBALS['SITE_DB']->bulk_insert_flip($member_access), false, true); // failsafe, in case we have put in some permissions for a group since deleted (can happen during install)
    if (is_on_multi_site_network() && (get_forum_type() == 'cns')) {
        $member_access = $GLOBALS['FORUM_DB']->query_select('group_privileges', ['*'], ['group_id' => $group_members, 'module_the_name' => 'forums']);
        foreach ($member_access as &$access) {
            $access['group_id'] = $group_id;
        }
        $GLOBALS['FORUM_DB']->query_insert('group_privileges', $GLOBALS['SITE_DB']->bulk_insert_flip($member_access), false, true); // failsafe, in case we have put in some permissions for a group since deleted (can happen during install)
    }

    $member_access = $GLOBALS['SITE_DB']->query_select('group_category_access', ['*'], ['group_id' => $group_members]);
    foreach ($member_access as &$access) {
        $access['group_id'] = $group_id;
    }
    $GLOBALS['SITE_DB']->query_insert('group_category_access', $GLOBALS['SITE_DB']->bulk_insert_flip($member_access), false, true); // failsafe, in case we have put in some permissions for a group since deleted (can happen during install)
    if (is_on_multi_site_network() && (get_forum_type() == 'cns')) {
        $member_access = $GLOBALS['FORUM_DB']->query_select('group_category_access', ['*'], ['group_id' => $group_members, 'module_the_name' => 'forums']);
        foreach ($member_access as &$access) {
            $access['group_id'] = $group_id;
        }
        $GLOBALS['FORUM_DB']->query_insert('group_category_access', $GLOBALS['SITE_DB']->bulk_insert_flip($member_access), false, true); // failsafe, in case we have put in some permissions for a group since deleted (can happen during install)
    }

    $member_access = $GLOBALS['SITE_DB']->query_select('group_zone_access', ['*'], ['group_id' => $group_members]);
    foreach ($member_access as &$access) {
        $access['group_id'] = $group_id;
    }

    $GLOBALS['SITE_DB']->query_insert('group_zone_access', $GLOBALS['SITE_DB']->bulk_insert_flip($member_access), false, true); // failsafe, in case we have put in some permissions for a group since deleted (can happen during install)
}

/**
 * Create/change the rank set.
 *
 * @param  string $rank_set The rank set wanted
 * @set none simple fun
 * @return array A list of usergroup IDs
 */
function cns_make_rank_set(string $rank_set) : array
{
    require_lang('cns_ranks');

    if (running_script('install')) {
        $member_group_0 = null;
        $member_group_1 = null;
        $member_group_2 = null;
        $member_group_3 = null;
        $member_group_4 = null;
    } else {
        require_code('cns_groups2');
        $ladder_groups = get_default_rank_ladder_groups();
        $member_group_0 = array_shift($ladder_groups);
        $member_group_1 = array_shift($ladder_groups);
        $member_group_2 = array_shift($ladder_groups);
        $member_group_3 = array_shift($ladder_groups);
        $member_group_4 = array_shift($ladder_groups);

        $GLOBALS['FORUM_DB']->query_update('f_groups', ['g_promotion_target_group' => null]);
    }

    switch ($rank_set) {
        case 'fun':
            $member_group_4 = cns_make_or_edit_group($member_group_4, do_lang('DEFAULT_RANK_fun_4'), 0, 0, 0, do_lang('DESCRIPTION_MEMBERS'), 'cns_rank_images/4');
            $member_group_3 = cns_make_or_edit_group($member_group_3, do_lang('DEFAULT_RANK_fun_3'), 0, 0, 0, do_lang('DESCRIPTION_MEMBERS'), 'cns_rank_images/3', $member_group_4, 10000);
            $member_group_2 = cns_make_or_edit_group($member_group_2, do_lang('DEFAULT_RANK_fun_2'), 0, 0, 0, do_lang('DESCRIPTION_MEMBERS'), 'cns_rank_images/2', $member_group_3, 2500);
            $member_group_1 = cns_make_or_edit_group($member_group_1, do_lang('DEFAULT_RANK_fun_1'), 0, 0, 0, do_lang('DESCRIPTION_MEMBERS'), 'cns_rank_images/1', $member_group_2, 400);
            $member_group_0 = cns_make_or_edit_group($member_group_0, do_lang('DEFAULT_RANK_fun_0'), 1, 0, 0, do_lang('DESCRIPTION_MEMBERS'), 'cns_rank_images/0', $member_group_1, 100);
            return [$member_group_0, $member_group_1, $member_group_2, $member_group_3, $member_group_4];

        case 'simple':
            foreach ([$member_group_4, $member_group_3] as $group_to_delete) {
                if ($group_to_delete !== null) {
                    require_code('cns_groups_action2');
                    cns_delete_group($group_to_delete, $member_group_0);
                }
            }
            $member_group_4 = null;
            $member_group_3 = null;
            $member_group_2 = cns_make_or_edit_group($member_group_2, do_lang('DEFAULT_RANK_simple_2'), 0, 0, 0, do_lang('DESCRIPTION_MEMBERS'), 'icons/tiers/gold', $member_group_3, 10000);
            $member_group_1 = cns_make_or_edit_group($member_group_1, do_lang('DEFAULT_RANK_simple_1'), 0, 0, 0, do_lang('DESCRIPTION_MEMBERS'), 'icons/tiers/silver', $member_group_2, 1000);
            $member_group_0 = cns_make_or_edit_group($member_group_0, do_lang('DEFAULT_RANK_simple_0'), 1, 0, 0, do_lang('DESCRIPTION_MEMBERS'), 'icons/tiers/bronze', $member_group_1, 100);
            return [$member_group_0, $member_group_1, $member_group_2];

        case 'none':
            foreach ([$member_group_4, $member_group_3, $member_group_2, $member_group_1] as $group_to_delete) {
                if ($group_to_delete !== null) {
                    require_code('cns_groups_action2');
                    cns_delete_group($group_to_delete, $member_group_0);
                }
            }
            $member_group_4 = null;
            $member_group_3 = null;
            $member_group_2 = null;
            $member_group_1 = null;
            $member_group_0 = cns_make_or_edit_group($member_group_0, do_lang('MEMBERS'), 1, 0, 0, do_lang('DESCRIPTION_MEMBERS'));
            return [$member_group_0];
    }

    warn_exit(do_lang_tempcode('INTERNAL_ERROR', escape_html('01686b9ed7b15a109174dd9027b71dcb')));
}

/**
 * Add or edit a usergroup with simple settings.
 *
 * @param  ?AUTO_LINK $group_id The ID of the usergroup to edit (null: adding)
 * @param  SHORT_TEXT $name The name of the usergroup
 * @param  BINARY $is_default Whether members are automatically put into the when they join
 * @param  BINARY $is_super_admin Whether members of this usergroup are all super-administrators
 * @param  BINARY $is_super_moderator Whether members of this usergroup are all super-moderators
 * @param  SHORT_TEXT $title The title for primary members of this usergroup that don't have their own title
 * @param  URLPATH $rank_image The rank image for this
 * @param  ?GROUP $promotion_target The that members of this usergroup get promoted to at point threshold (null: no promotion prospects)
 * @param  ?integer $promotion_threshold The point threshold for promotion (null: no promotion prospects)
 * @param  BINARY $promotion_approval Whether usergroup promotion should require manual approval
 * @return AUTO_LINK The ID of the usergroup
 */
function cns_make_or_edit_group(?int $group_id, string $name, int $is_default = 0, int $is_super_admin = 0, int $is_super_moderator = 0, string $title = '', string $rank_image = '', ?int $promotion_target = null, ?int $promotion_threshold = null, int $promotion_approval = 0) : int
{
    if ($group_id === null) {
        $group_id = cns_make_group($name, $is_default, $is_super_admin, $is_super_moderator, $title, $rank_image, $promotion_target, $promotion_threshold, $promotion_approval);
    } else {
        require_code('cns_groups_action2');
        cns_edit_group($group_id, $name, $is_default, $is_super_admin, $is_super_moderator, $title, $rank_image, $promotion_target, $promotion_threshold, $promotion_approval, null, null, null, null, null, null, null, null, null, null, null, 0, 0, 0, null, 1, 0, 0);
    }

    return $group_id;
}
