<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    welcome_emails
 */

/**
 * Generate what we need to send/preview a welcome e-mail.
 *
 * @param  array $mail Welcome-email row
 * @param  ?array $member Member/subscriber row (null: generating a preview: current member/subscriber based around such)
 * @return array A tuple: subject, message, message is-HTML, recipient name
 */
function cns_prepare_welcome_email(array $mail, ?array $member = null) : array
{
    $newsletter_style = (($mail['w_newsletter_id'] !== null) && (addon_installed('newsletter')));

    if ($member === null) {
        $member_row = $GLOBALS['FORUM_DRIVER']->get_member_row(get_member());
        if ($newsletter_style) {
            $member = [
                'id' => -1,
                'email' => $member_row['m_email_address'],
                'join_time' => $member_row['m_join_time'],
                'code_confirm' => 0,
                'the_password' => '',
                'pass_salt' => '',
                'language' => $member_row['m_language'],
                'n_forename' => $member_row['m_username'],
                'n_surname' => '',
            ];
        } else {
            $member = $member_row;
        }
    }

    require_lang('cns');

    // Load up metadata
    if ($newsletter_style) {
        $forename = $member['n_forename'];
        $surname = $member['n_surname'];
        $name = trim($forename . ' ' . $surname);
        require_lang('newsletter');
        if ($name == '') {
            $name = do_lang('NEWSLETTER_SUBSCRIBER_DEFAULT_NAME', get_site_name());
        }
        $email_address = $member['email'];
        $lang = $member['language'];

        $send_id = 'n' . strval($member['id']);
        require_code('newsletter');
        $hash = get_unsubscribe_hash($member['the_password']);
    } else {
        $forename = '';
        $surname = '';
        $name = $GLOBALS['FORUM_DRIVER']->get_displayname($member['m_username']);
        $email_address = $member['m_email_address'];
        $lang = get_lang($member['id']);

        $send_id = 'w' . strval($member['id']);
        $hash = '';
    }

    $subject = get_translated_text($mail['w_subject'], null, $lang);

    $message_raw = get_translated_text($mail['w_text'], null, get_lang($member['id']));

    // Special syntax for referring to specific date-times up to 100 hours into the future
    $matches = [];
    $num_matches = preg_match_all('#\{\{(\d+)\}\}#', $message_raw, $matches);
    for ($i = 0; $i < $num_matches; $i++) {
        $message_raw = str_replace('{{' . strval($i) . '}}', get_timezoned_date(time() + $matches[1][$i] * 60 * 60 * 24), $message_raw);
    }

    // Process with all the newsletter magic, if available
    if (addon_installed('newsletter')) {
        require_code('newsletter');
        $extra_mappings = $member;

        if (get_forum_type() == 'cns') {
            require_code('cns_members');
            $extra_mappings += cns_get_custom_field_mappings($member['id']);
        }

        $is_html = newsletter_is_html();
        if ($is_html) {
            $message_raw = static_evaluate_tempcode(comcode_to_tempcode($message_raw, get_member(), true));
        }

        $message_wrapped = newsletter_prepare($message_raw, $subject, $lang, $forename, $surname, $name, $email_address, $send_id, $hash, @array_map('strval', $extra_mappings));
        $message_final = $message_wrapped;
    } else {
        $is_html = false;
        $_message_final = comcode_to_tempcode($message_raw, null, true);
        $message_final = $_message_final->evaluate($lang);
    }

    return [$subject, $message_final, $is_html, $name];
}
