<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    core
 */

/*
    IMPORTANT: This file is loaded outside the core software, so must work as standalone and not use external APIs.
*/

/**
 * Check the given plain-text maintenance password is valid.
 * This should be handled appropriately in an if guard if $silent_failure could be true.
 *
 * @param  SHORT_TEXT $password_given Given maintenance password
 * @param  boolean $silent_failure Whether to proceed without exiting if the password was wrong
 * @return boolean Whether the password was valid
 */
function check_maintenance_password(string $password_given, bool $silent_failure = false) : bool
{
    if (!_maintenance_password_check__init($silent_failure)) {
        return false;
    }

    global $SITE_INFO;

    $actual_password_hashed = $SITE_INFO['maintenance_password'];
    if (strpos($actual_password_hashed, '$') !== false) {
        $ret = password_verify($password_given, $actual_password_hashed);
        _maintenance_password_check__result($ret, $silent_failure);
        return $ret;
    }
    $salt = '';
    if ((substr($actual_password_hashed, 0, 1) == '!') && (strlen($actual_password_hashed) == 33)) {
        $actual_password_hashed = substr($actual_password_hashed, 1);
        $salt = 'cms';

        // LEGACY
        if ($actual_password_hashed !== md5($password_given . $salt)) {
            $salt = 'ocp';
        }
    }
    $ret = (((strlen($password_given) != 32) && ($actual_password_hashed == $password_given)) || (hash_equals($actual_password_hashed, md5($password_given . $salt))));
    _maintenance_password_check__result($ret, $silent_failure);
    return $ret;
}

/**
 * Check the given hashed maintenance password is valid.
 * This should be handled appropriately in an if guard if $silent_failure could be true.
 *
 * @param  SHORT_TEXT $password_given_hashed Given hashed maintenance password
 * @param  boolean $silent_failure Whether to proceed without exiting if the password was wrong
 * @return boolean Whether it is valid
 */
function check_maintenance_password_from_hash(string $password_given_hashed, bool $silent_failure = false) : bool
{
    if (!_maintenance_password_check__init($silent_failure)) {
        return false;
    }

    global $SITE_INFO;

    $actual_password_hashed = $SITE_INFO['maintenance_password'];

    if ($password_given_hashed === md5($actual_password_hashed)) {
        $ret = true; // LEGACY: Upgrade from v7 where hashed input password given even if plain-text password is in use
        _maintenance_password_check__result($ret, $silent_failure);
        return $ret;
    }

    $ret = hash_equals($password_given_hashed, $actual_password_hashed);
    _maintenance_password_check__result($ret, $silent_failure);
    return $ret;
}

/**
 * Prepare for checking the maintenance password.
 * This call should be done within an if condition and appropriately handled when this returns false (if $silent_failure could be true).
 *
 * @param  boolean $silent_failure Whether to proceed without exiting (false: instead of returning, we will exit if a maintenance password is not defined)
 * @return boolean Whether a maintenance password is available and access should be allowed
 */
function _maintenance_password_check__init(bool $silent_failure = false) : bool
{
    if ((function_exists('php_function_allowed')) && (php_function_allowed('usleep'))) {
        @usleep(500000); // Wait for half a second, to reduce brute force potential
    }

    global $SITE_INFO;

    if (isset($SITE_INFO['admin_password'])) { // LEGACY
        $SITE_INFO['maintenance_password'] = $SITE_INFO['admin_password'];
        unset($SITE_INFO['admin_password']);
    }
    if (isset($SITE_INFO['master_password'])) { // LEGACY: (as of v11)
        $SITE_INFO['maintenance_password'] = $SITE_INFO['master_password'];
        unset($SITE_INFO['master_password']);
    }

    if (!array_key_exists('maintenance_password', $SITE_INFO)) {
        if ($silent_failure) {
            return false;
        }
        exit('Access denied - a maintenance password is not defined in _config.php.');
    }

    return true;
}

/**
 * Log the results of the maintenance password check and exit when applicable.
 *
 * @param  boolean $result Whether login is successful
 * @param  boolean $silent_failure Whether to proceed without exiting if $result was false
 */
function _maintenance_password_check__result(bool $result, bool $silent_failure = false)
{
    $msg = 'administrative script ' . basename($_SERVER['SCRIPT_NAME']);
    if (!empty($_SERVER['REMOTE_ADDR'])) {
        $msg .= ', by IP address ' . $_SERVER['REMOTE_ADDR'];
    }
    if (function_exists('syslog')) {
        if ($result) {
            @syslog(LOG_NOTICE, 'Crypt master: Successfully logged into ' . $msg);
        } else {
            @syslog(LOG_WARNING, 'Crypt master: Incorrect maintenance password given while logging into ' . $msg);
        }
    }

    if (function_exists('error_log')) {
        global $FILE_BASE;
        @ini_set('error_log', $FILE_BASE . '/data_custom/errorlog.php');
        if (!$result) {
            @error_log('Crypt master: WARNING Incorrect maintenance password given while logging into ' . $msg);
        }
    }

    // Actually we want to bail with a clear message to the member about their attempt being logged (to comply with data regulations)
    if (!$result && !$silent_failure) {
        exit('Access denied - Incorrect maintenance password. This attempt has been logged along with the IP address.');
    }
}
