<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    core_forum_drivers
 */

 /**
  * Initialize with constants used in the forum driver.
  */
 function init__forum__mybb() {
    if (!defined('MYBB_USERGROUP_ADMINISTRATORS')) {
        // Define the usergroup IDs of staff in myBB
        define('MYBB_USERGROUP_ADMINISTRATORS', 4);
        define('MYBB_USERGROUP_SUPER_MODERATORS', 3);
        define('MYBB_USERGROUP_MODERATORS', 6);
        define('MYBB_USERGROUP_GUESTS', 1);
    }
 }

/**
 * Forum driver class.
 *
 * @package core_forum_drivers
 */
class Forum_driver_mybb extends Forum_driver_base
{
    /**
     * Check the connected DB is valid for this forum driver.
     *
     * @return boolean Whether it is valid
     */
    public function check_db() : bool
    {
        $test = $this->db->query('SELECT COUNT(*) FROM ' . $this->db->get_table_prefix() . 'users', null, 0, true); // Suppress errors as we're checking to see if the table is there
        return $test !== null;
    }

    /**
     * Get the rows for the top given number of posters on the forum.
     *
     * @param  integer $limit The limit to the number of top posters to fetch
     * @return array The rows for the given number of top posters in the forum
     */
    public function get_top_posters(int $limit) : array
    {
        return $this->db->query('SELECT * FROM ' . $this->db->get_table_prefix() . 'users WHERE uid<>' . strval($this->get_guest_id()) . ' ORDER BY postnum DESC', $limit);
    }

    /**
     * Attempt to to find the member's language from their forum profile. It converts between language-identifiers using a map (lang/map.ini).
     *
     * @param  MEMBER $member_id The member who's language needs to be fetched
     * @return ?LANGUAGE_NAME The member's language (null: unknown)
     */
    public function get_lang(int $member_id) : ?string
    {
        return $this->get_member_row_field($member_id, 'language');
    }

    /**
     * Find the member ID of the forum guest member.
     *
     * @return MEMBER The member ID of the forum guest member
     */
    public function get_guest_id() : int
    {
        return 0;
    }

    /**
     * Get the forums' table prefix for the database.
     *
     * @return string The forum database table prefix
     */
    public function get_drivered_table_prefix() : string
    {
        global $SITE_INFO;
        return $SITE_INFO['mybb_table_prefix'];
    }

    /**
     * Add the specified custom field to the forum (some forums implemented this using proper Custom Profile Fields, others through adding a new field).
     *
     * @param  string $name The name of the new custom field
     * @param  integer $length The length of the new custom field (ignored for Conversr, $type used instead)
     * @param  BINARY $locked Whether the field is locked, e.g. cannot be deleted
     * @param  BINARY $viewable Whether the field is for viewing by the owner and the public (0: cannot be viewed unless the viewing member has view_any_profile_field privilege)
     * @param  BINARY $settable Whether the field is for setting by the owner
     * @param  BINARY $required Whether the field is required
     * @param  string $description Description
     * @param  string $type The field type (it's the same as the software's field types, although we only expect forum drivers to specifically support short_text/long_text/integer/float and for the rest to be mapped to long_text)
     * @param  BINARY $encrypted Whether the field is encrypted
     * @param  ?string $default Default field value (null: standard for field type)
     * @param  SHORT_TEXT $options Field options
     * @param  BINARY $include_in_main_search Whether to include in main keyword search
     * @param  BINARY $allow_template_search Whether to allow template search
     * @param  ID_TEXT $icon An icon to show the CPF with on the member profiles
     * @param  ID_TEXT $section A section to show with on the member-links part of member profiles
     * @param  LONG_TEXT $tempcode This is Tempcode that is used for displaying the field. See the DESCRIPTION_CPF_CODE language string.
     * @param  ID_TEXT $autofill_type Autofill field name from https://html.spec.whatwg.org/multipage/form-control-infrastructure.html#autofill-field
     * @param  ID_TEXT $autofill_hint Autofill hint: '' or 'shipping' or 'billing'
     * @return boolean Whether the custom field was created successfully
     */
    public function install_create_custom_field(string $name, int $length, int $locked = 1, int $viewable = 0, int $settable = 0, int $required = 0, string $description = '', string $type = 'long_text', int $encrypted = 0, ?string $default = null, string $options = '', int $include_in_main_search = 0, int $allow_template_search = 0, string $icon = '', string $section = '', string $tempcode = '', string $autofill_type = '', string $autofill_hint = '') : bool
    {
        list($db_type, $db_default) = $this->remap_cms_field_type_to_db_type($type, $default);
        $query = $this->db->driver->add_table_field__sql($this->db->get_table_prefix() . 'users', 'cms_' . $name, $db_type, $db_default);
        $this->db->query($query, null, 0, true); // Suppress errors in case field already exists
        return true;
    }

    /**
     * Edit the specified custom field in the forum (some forums implemented this using proper Custom Profile Fields, others through adding a new field).
     *
     * @param  string $old_name The name of the existing custom field to edit
     * @param  string $name The new name of the custom field
     * @param  integer $length The length of the custom field (ignored for Conversr, $type used instead)
     * @param  BINARY $locked Whether the field is locked, e.g. cannot be deleted
     * @param  BINARY $viewable Whether the field is for viewing by the owner and the public (0: cannot be viewed unless the viewing member has view_any_profile_field privilege)
     * @param  BINARY $settable Whether the field is for setting by the owner
     * @param  BINARY $required Whether the field is required
     * @param  string $description Description
     * @param  string $type The field type (it's the same as the software's field types, although we only expect forum drivers to specifically support short_text/long_text/integer/float and for the rest to be mapped to long_text)
     * @param  BINARY $encrypted Whether the field is encrypted
     * @param  ?string $default Default field value (null: standard for field type)
     * @param  SHORT_TEXT $options Field options
     * @param  BINARY $include_in_main_search Whether to include in main keyword search
     * @param  BINARY $allow_template_search Whether to allow template search
     * @param  ID_TEXT $icon An icon to show the CPF with on the member profiles
     * @param  ID_TEXT $section A section to show with on the member-links part of member profiles
     * @param  LONG_TEXT $tempcode This is Tempcode that is used for displaying the field. See the DESCRIPTION_CPF_CODE language string.
     * @param  ID_TEXT $autofill_type Autofill field name from https://html.spec.whatwg.org/multipage/form-control-infrastructure.html#autofill-field
     * @param  ID_TEXT $autofill_hint Autofill hint: '' or 'shipping' or 'billing'
     * @return boolean Whether the custom field was edited successfully
     */
    public function install_edit_custom_field(string $old_name, string $name, int $length, int $locked = 1, int $viewable = 0, int $settable = 0, int $required = 0, string $description = '', string $type = 'long_text', int $encrypted = 0, ?string $default = null, string $options = '', int $include_in_main_search = 0, int $allow_template_search = 0, string $icon = '', string $section = '', string $tempcode = '', string $autofill_type = '', string $autofill_hint = '') : bool
    {
        list($db_type) = $this->remap_cms_field_type_to_db_type($type, $default);
        $query = $this->db->driver->alter_table_field($this->db->get_table_prefix() . 'users', 'cms_' . $old_name, $db_type, 'cms_' . $name);
        $this->db->query($query, null, 0, true); // Suppress errors in case field already edited
        return true;
    }

    /**
     * Get an array of attributes to take in from the installer. Almost all forums require a table prefix, which the requirement there-of is defined through this function.
     * The attributes have 4 values in an array:
     * - name, the name of the attribute for _config.php
     * - default, the default value (perhaps obtained through autodetection from forum config)
     * - description, a textual description of the attributes
     * - title, a textual title of the attribute
     *
     * @return array The attributes for the forum
     */
    public function install_specifics() : array
    {
        global $PROBED_FORUM_CONFIG;
        $a = [];
        $a['name'] = 'mybb_table_prefix';
        $a['default'] = array_key_exists('sql_tbl_prefix', $PROBED_FORUM_CONFIG) ? $PROBED_FORUM_CONFIG['sql_tbl_prefix'] : 'mybb_';
        $a['description'] = do_lang('MOST_DEFAULT');
        $a['title'] = 'MyBB ' . do_lang('TABLE_PREFIX');
        return [$a];
    }

    /**
     * Searches for forum auto-config at this path.
     *
     * @param  PATH $path The path in which to search
     * @return boolean Whether the forum auto-config could be found
     */
    public function install_test_load_from(string $path) : bool
    {
        global $PROBED_FORUM_CONFIG;

        if (@file_exists($path . '/inc/config.php')) {
            $config = [];

            @include($path . '/inc/config.php');
            if (array_key_exists('database', $config)) {
                $PROBED_FORUM_CONFIG['sql_host'] = !empty($config['database']['hostname']) ? $config['database']['hostname'] : '';
                $PROBED_FORUM_CONFIG['sql_database'] = !empty($config['database']['database']) ? $config['database']['database'] : '';
                $PROBED_FORUM_CONFIG['sql_user'] = !empty($config['database']['username']) ? $config['database']['username'] : '';
                $PROBED_FORUM_CONFIG['sql_pass_exists'] = !empty($config['database']['password']);
                $PROBED_FORUM_CONFIG['sql_tbl_prefix'] = !@cms_empty_safe($config['database']['table_prefix']) ? $config['database']['table_prefix'] : '';
                $PROBED_FORUM_CONFIG['cookie_member_id'] = 'mybbuser';
            }
            return true;
        }
        return false;
    }

    /**
     * Get an array of paths to search for config at.
     *
     * @return array The paths in which to search for the forum config
     */
    public function install_get_path_search_list() : array
    {
        return [
            'mybb',
            'forums',
            'forum',
            'boards',
            'board',
            '../forums/mybb',
            '../forums',
            '../mybb',
            '../forums',
            '../forum',
            '../boards',
            '../board',
        ];
    }

    /**
     * Get an emoticon chooser template.
     *
     * @param  string $field_name The ID of the form field the emoticon chooser adds to
     * @return Tempcode The emoticon chooser template
     */
    public function get_emoticon_chooser(string $field_name = 'post') : object
    {
        require_code('comcode_compiler');
        $emoticons = $this->db->query_select('smilies', ['*'], ['showclickable' => '1']);
        sort_maps_by($emoticons, 'disporder');
        $em = new Tempcode();
        foreach ($emoticons as $emo) {
            $code = $emo['find'];
            $em->attach(do_template('EMOTICON_CLICK_CODE', ['_GUID' => '89aa93c39b3929b00245981ba632371f', 'FIELD_NAME' => $field_name, 'CODE' => $code, 'IMAGE' => apply_emoticons($code)]));
        }
        return $em;
    }

    /**
     * Pin a topic.
     *
     * @param  AUTO_LINK $topic_id The topic ID
     * @param  boolean $pin True: pin it, False: unpin it
     */
    public function pin_topic(int $topic_id, bool $pin = true)
    {
        $this->db->query_update('threads', ['sticky' => $pin ? 1 : 0], ['tid' => $topic_id], '', 1);
    }

    /**
     * Set a Custom Profile Field's value, if the custom field exists. Only works on specially-named (titled) fields.
     *
     * @param  MEMBER $member_id The member ID
     * @param  string $field The field name (e.g. "firstname" for the CPF with a title of "cms_firstname")
     * @param  string $value The value
     */
    public function set_custom_field(int $member_id, string $field, string $value)
    {
        $this->db->query_update('users', ['cms_' . $field => $value], ['uid' => $member_id], '', null, 0, false, true); // Errors suppressed in case field gone missing
    }

    /**
     * Get Custom Profile Fields values for all 'cms_' prefixed keys.
     *
     * @param  MEMBER $member_id The member ID
     * @return ?array A map of the Custom Profile Fields, key_suffix=>value (null: no fields)
     */
    public function get_custom_fields(int $member_id) : ?array
    {
        $row = $this->get_member_row($member_id);
        if ($row === null) {
            return null;
        }

        $out = [];
        foreach ($row as $attribute => $value) {
            if (substr($attribute, 0, 4) == 'cms_') {
                $out[substr($attribute, 4)] = $value;
            }
        }
        return $out;
    }

    /**
     * From a member row, get the member's primary usergroup.
     *
     * @param  array $r The profile-row
     * @return GROUP The member's primary usergroup
     */
    public function mrow_primary_group(array $r) : int
    {
        return $r['usergroup'];
    }

    /**
     * From a member row, get the member's member ID.
     *
     * @param  array $r The profile-row
     * @return MEMBER The member ID
     */
    public function mrow_member_id(array $r) : int
    {
        return $r['uid'];
    }

    /**
     * From a member row, get the member's last visit date.
     *
     * @param  array $r The profile-row
     * @return TIME The last visit date
     */
    public function mrow_lastvisit(array $r) : int
    {
        return $r['lastvisit'];
    }

    /**
     * From a member row, get the member's name.
     *
     * @param  array $r The profile-row
     * @return string The username
     */
    public function mrow_username(array $r) : string
    {
        return $r['username'];
    }

    /**
     * From a member row, get the member's e-mail address.
     *
     * @param  array $r The profile-row
     * @return SHORT_TEXT The member e-mail address
     */
    public function mrow_email_address(array $r) : string
    {
        return $r['email'];
    }

    /**
     * Get a URL to the specified member's home (control panel).
     *
     * @param  MEMBER $member_id The member ID
     * @return URLPATH The URL to the members home
     */
    public function member_home_url(int $member_id) : string
    {
        return get_forum_base_url() . '/usercp.php';
    }

    /**
     * Get the photo URL for the specified member ID.
     *
     * @param  MEMBER $member_id The member ID
     * @param  boolean $full Get full photo
     * @return URLPATH The URL (blank: none)
     */
    public function get_member_photo_url(int $member_id, bool $full = false) : string
    {
        return '';
    }

    /**
     * Get the avatar URL for the specified member ID.
     *
     * @param  MEMBER $member_id The member ID
     * @param  boolean $fallback_support Whether fallback support should be allowed (passed by reference)
     * @return URLPATH The URL (blank: none)
     */
    protected function _get_member_avatar_url(int $member_id, bool &$fallback_support) : string
    {
        $avatar_path = get_forum_base_url();

        $type = $this->get_member_row_field($member_id, 'avatartype');
        $filename = $this->get_member_row_field($member_id, 'avatar');

        switch (strval($type)) {
            // Type could be: 'gallery', 'upload','remote'
            case 'gallery': // Avatar from Avatars Gallery
                return get_forum_base_url() . '/' . $filename;
            case 'remote': // URL of Remote image Avatar
                return $filename;
            case 'upload': // Uploaded Avatar
                return get_forum_base_url() . preg_replace('/\.\//', '/', $filename);
        }
        return ''; // The avatar is not set
    }

    /**
     * Get a URL to the specified member's profile.
     *
     * @param  MEMBER $member_id The member ID
     * @return URLPATH The URL to the member profile
     */
    protected function _member_profile_url(int $member_id) : string
    {
        return get_forum_base_url() . '/member.php?action=profile&uid=' . strval($member_id);
    }

    /**
     * Get a URL to the registration page (for people to create member accounts).
     *
     * @return URLPATH The URL to the registration page
     */
    protected function _join_url() : string
    {
        return get_forum_base_url() . '/member.php?action=register';
    }

    /**
     * Get a URL to the members-online page.
     *
     * @return URLPATH The URL to the members-online page
     */
    protected function _users_online_url() : string
    {
        return get_forum_base_url() . '/online.php';
    }

    /**
     * Get a URL to send a private/personal message to the given member.
     *
     * @param  MEMBER $member_id The member ID
     * @return URLPATH The URL to the private/personal message page
     */
    protected function _member_pm_url(int $member_id) : string
    {
        return get_forum_base_url() . '/private.php?action=send&uid=' . strval($member_id);
    }

    /**
     * Get a URL to the specified forum.
     *
     * @param  integer $forum_id The forum ID
     * @return URLPATH The URL to the specified forum
     */
    protected function _forum_url(int $forum_id) : string
    {
        return get_forum_base_url() . '/forumdisplay.php?fid=' . strval($forum_id);
    }

    /**
     * Get the forum ID from a forum name. Numeric forum names are considered to be usable as IDs.
     *
     * @param  SHORT_TEXT $forum_name The forum name
     * @return ?integer The forum ID (null: not found)
     */
    public function forum_id_from_name(string $forum_name) : ?int
    {
        return (is_numeric($forum_name)) ? intval($forum_name) : $this->db->query_select_value_if_there('forums', 'fid', ['type' => 'f', 'name' => $forum_name]); //type 'f' - forum, type 'c' - category
    }

    /**
     * Makes a post in the specified forum, in the specified topic according to the given specifications. If the topic doesn't exist, it is created along with a spacer-post.
     * Spacer posts exist in order to allow staff to delete the first true post in a topic. Without spacers, this would not be possible with most forum systems. They also serve to provide meta information on the topic that cannot be encoded in the title (such as a link to the content being commented upon).
     *
     * @param  SHORT_TEXT $forum_name The forum name
     * @param  SHORT_TEXT $topic_identifier The topic identifier (usually <content-type>_<content-id>)
     * @param  MEMBER $member_id The member ID
     * @param  LONG_TEXT $post_title The post title
     * @param  LONG_TEXT $_post The post content in Comcode format
     * @param  string $content_title The topic title; must be same as content title if this is for a comment topic
     * @param  string $topic_identifier_encapsulation_prefix This is put together with the topic identifier to make a more-human-readable topic title or topic description (hopefully the latter and a $content_title title, but only if the forum supports descriptions)
     * @param  ?URLPATH $content_url URL to the content (null: do not make spacer post)
     * @param  ?TIME $time The post time (null: use current time)
     * @param  ?IP $ip The post IP address (null: use current members IP address)
     * @param  ?BINARY $validated Whether the post is validated (null: unknown, find whether it needs to be marked not validated initially). This only works with the Conversr driver.
     * @param  ?BINARY $topic_validated Whether the topic is validated (null: unknown, find whether it needs to be marked not validated initially). This only works with the Conversr driver.
     * @param  boolean $skip_post_checks Whether to skip post checks
     * @param  SHORT_TEXT $poster_name_if_guest The name of the poster
     * @param  ?AUTO_LINK $parent_id ID of post being replied to (null: N/A)
     * @param  boolean $staff_only Whether the reply is only visible to staff
     * @return array Topic ID (may be null), and whether a hidden post has been made
     */
    public function make_post_forum_topic(string $forum_name, string $topic_identifier, int $member_id, string $post_title, string $_post, string $content_title, string $topic_identifier_encapsulation_prefix, ?string $content_url = null, ?int $time = null, ?string $ip = null, ?int $validated = null, ?int $topic_validated = 1, bool $skip_post_checks = false, string $poster_name_if_guest = '', ?int $parent_id = null, bool $staff_only = false) : array
    {
        $__post = comcode_to_tempcode($_post);
        $post = $__post->evaluate();

        if ($time === null) {
            $time = time();
        }
        if ($ip === null) {
            $ip = get_ip_address();
        }
        $forum_id = $this->forum_id_from_name($forum_name);
        if ($forum_id === null) {
            warn_exit(do_lang_tempcode('MISSING_FORUM', escape_html($forum_name)), false, true);
        }

        $username = $this->_get_username($member_id); // needed for the mybb_theads DB table

        $test = $this->db->query_select('forums', ['*'], [], '', 1);
        $fm = array_key_exists('status', $test[0]);

        $topic_id = $this->find_topic_id_for_topic_identifier($forum_name, $topic_identifier);

        $ip_address = $ip;
        $local_ip_address = '127.0.0.1';

        $is_new = ($topic_id === null);
        if ($is_new) {
            $subject = $content_title . ', ' . $topic_identifier_encapsulation_prefix . ': #' . $topic_identifier;
            $topic_id = $this->db->query_insert('threads', ['fid' => $forum_id, 'subject' => $subject, 'username' => $username, 'uid' => $member_id, 'lastposter' => $username, 'lastposteruid' => $member_id, 'visible' => 1, 'dateline' => $time, 'lastpost' => $time], true);
            $home_link = '[url]' . $content_url . '[/url]';
            $this->db->query_insert('posts', ['fid' => $forum_id, 'tid' => $topic_id, 'username' => do_lang('SYSTEM', '', '', '', get_site_default_lang()), 'uid' => $member_id, 'message' => do_lang('SPACER_POST', $home_link, '', '', get_site_default_lang()), 'subject' => $subject, 'dateline' => $time, 'visible' => 1, 'ipaddress' => $ip_address]);
            $this->db->query('UPDATE ' . $this->db->get_table_prefix() . 'forums SET posts=(posts+1), threads=(threads+1) WHERE fid=' . strval($forum_id), 1);
        }

        $GLOBALS['LAST_TOPIC_ID'] = $topic_id;
        $GLOBALS['LAST_TOPIC_IS_NEW'] = $is_new;

        if ($post == '') {
            return [$topic_id, false];
        }

        $this->db->query_insert('posts', ['fid' => $forum_id, 'tid' => $topic_id, 'username' => $username, 'uid' => $member_id, 'message' => $post, 'subject' => $post_title, 'dateline' => $time, 'visible' => 1, 'ipaddress' => $ip_address]);
        $this->db->query('UPDATE ' . $this->db->get_table_prefix() . 'forums SET posts=(posts+1), lastpost=' . strval($time) . ', lastposter=\'' . db_escape_string($username) . '\' WHERE fid=' . strval($forum_id), 1);
        $this->db->query('UPDATE ' . $this->db->get_table_prefix() . 'threads SET lastpost=' . strval($time) . ', lastposter=\'' . db_escape_string($username) . '\', lastposteruid=' . strval($member_id) . ', replies=(replies+1) WHERE tid=' . strval($topic_id), 1);

        return [$topic_id, false];
    }

    /**
     * Get an array of maps for the topic in the given forum.
     *
     * @param  integer $topic_id The topic ID
     * @param  ?integer $count The comment count will be returned here by reference (null: do not gather it)
     * @param  ?integer $max Maximum comments to returned (null: no limit)
     * @param  integer $start Comment to start at
     * @param  boolean $mark_read Whether to mark the topic read (ignored for this forum driver)
     * @param  boolean $reverse Whether to show in reverse
     * @return mixed The array of maps (Each map is: title, message, member, date) (-1 for no such forum, -2 for no such topic)
     */
    public function get_forum_topic_posts(int $topic_id, ?int &$count = null, ?int $max = 100, int $start = 0, bool $mark_read = true, bool $reverse = false)
    {
        if ($topic_id === null) {
            return (-2);
        }

        $order = $reverse ? 'dateline DESC' : 'dateline';
        $rows = $this->db->query('SELECT * FROM ' . $this->db->get_table_prefix() . 'posts WHERE tid=' . strval($topic_id) . ' AND message NOT LIKE \'' . db_encode_like(substr(do_lang('SPACER_POST', '', '', '', get_site_default_lang()), 0, 20) . '%') . '\' ORDER BY ' . $order, $max, $start);
        if ($count !== null) {
            $count = $this->db->query_value_if_there('SELECT COUNT(*) FROM ' . $this->db->get_table_prefix() . 'posts WHERE tid=' . strval($topic_id) . ' AND message NOT LIKE \'' . db_encode_like(substr(do_lang('SPACER_POST', '', '', '', get_site_default_lang()), 0, 20) . '%') . '\'');
        }

        $out = [];
        foreach ($rows as $myrow) {
            $temp = [];
            $temp['title'] = $myrow['subject'];
            if ($temp['title'] === null) {
                $temp['title'] = '';
            }
            push_lax_comcode(true);
            $temp['message'] = $myrow['message'];
            pop_lax_comcode();
            $temp['member'] = $myrow['uid'];
            $temp['date'] = $myrow['dateline'];

            $out[] = $temp;
        }

        return $out;
    }

    /**
     * Get a URL to the specified topic ID. Most forums don't require the second parameter, but some do, so it is required in the interface.
     *
     * @param  integer $topic_id The topic ID
     * @param  string $forum The forum ID
     * @return URLPATH The URL to the topic
     */
    public function topic_url(int $topic_id, string $forum) : string
    {
        return get_forum_base_url() . '/showthread.php?tid=' . strval($topic_id);
    }

    /**
     * Get a URL to the specified post ID.
     *
     * @param  integer $post_id The post ID
     * @param  ?mixed $forum The forum ID (null: private topic)
     * @return URLPATH The URL to the post
     */
    public function post_url(int $post_id, $forum) : string
    {
        $topic_id = $this->db->query_select_value_if_there('posts', 'post_tid', ['pid' => $post_id]);
        if ($topic_id === null) {
            return '';
        }
        $url = get_forum_base_url() . '/showthread.php?tid=' . strval($topic_id) . '&pid=' . strval($post_id) . '#pid' . strval($post_id);
        return $url;
    }

    /**
     * Get the topic ID from a topic identifier in the specified forum. It is used by comment topics, which means that the unique-topic-name assumption holds valid.
     *
     * @param  string $forum The forum name / ID
     * @param  SHORT_TEXT $topic_identifier The topic identifier
     * @return ?integer The topic ID (null: not found)
     */
    public function find_topic_id_for_topic_identifier(string $forum, string $topic_identifier) : ?int
    {
        if (is_integer($forum)) {
            $forum_id = intval($forum);
        } else {
            $forum_id = $this->forum_id_from_name($forum);
            if ($forum_id === null) {
                return null;
            }
        }

        return $this->db->query_value_if_there('SELECT tid FROM ' . $this->db->get_table_prefix() . 'threads WHERE fid=' . strval($forum_id) . ' AND (' . db_string_equal_to('subject', $topic_identifier) . ' OR subject LIKE \'' . db_encode_like('%: #' . $topic_identifier) . '\')');
    }

    /**
     * Get an array of topics in the given forum. Each topic is an array with the following attributes:
     * - id, the topic ID
     * - title, the topic title
     * - lastusername, the username of the last poster
     * - lasttime, the timestamp of the last reply
     * - closed, a Boolean for whether the topic is currently closed or not
     * - firsttitle, the title of the first post
     * - firstpost, the first post (only set if $show_first_posts was true)
     *
     * @param  mixed $name The forum name or forum ID or an array of forum IDs (in such an array the KEYS are forum IDs, and the values ignored)
     * @param  integer $limit The limit
     * @param  integer $start The start position
     * @param  integer $max_rows The total rows (not a parameter: returns by reference)
     * @param  SHORT_TEXT $filter_topic_title The topic title filter (blank: no filter)
     * @param  SHORT_TEXT $filter_topic_description The topic description filter; may apply to the topic title if there is no separate description field with additional wildcarding to match what make_post_forum_topic is doing (blank: no filter)
     * @param  boolean $show_first_posts Whether to show the first posts
     * @param  string $date_key The date key to sort by
     * @set lasttime firsttime
     * @param  boolean $hot Whether to limit to hot topics
     * @param  boolean $only_open Open topics only
     * @return ?array The array of topics (null: error)
     */
    public function show_forum_topics($name, int $limit, int $start, int &$max_rows, string $filter_topic_title = '', string $filter_topic_description = '', bool $show_first_posts = false, string $date_key = 'lasttime', bool $hot = false, bool $only_open = false) : ?array
    {
        // Build forum ID query
        if (is_numeric($name)) { // Forum ID
            $id_list = 'fid=' . strval($name);
        } elseif (!is_array($name)) { // Forum name
            $id = $this->forum_id_from_name($name);
            if ($id === null) {
                return null;
            }
            $id_list = 'fid=' . strval($id);
        } else { // Array of forum IDs
            $id_list = '';
            foreach (array_keys($name) as $id) {
                if ($id === null) {
                    continue;
                }
                if ($id_list != '') {
                    $id_list .= ' OR ';
                }
                $id_list .= 'fid=' . strval($id);
            }
            if ($id_list == '') {
                return null;
            }
        }

        $topic_filter = '';
        if ($filter_topic_title != '') {
            $topic_filter .= ' AND subject LIKE \'' . db_encode_like($filter_topic_title . ', %') . '\'';
        }
        if ($filter_topic_description != '') {
            $topic_filter .= ' AND subject LIKE \'' . db_encode_like('%, ' . $filter_topic_description) . '\'';
        }
        if ($only_open) {
            $topic_filter .= ' AND visible=1 AND ' . db_string_equal_to('closed', '');
        }

        $rows = $this->db->query('SELECT * FROM ' . $this->db->get_table_prefix() . 'threads WHERE (' . $id_list . ')' . $topic_filter . ' ORDER BY ' . (($date_key == 'lasttime') ? 'lastpost' : 'dateline') . ' DESC', $limit, $start);
        $max_rows = $this->db->query_value_if_there('SELECT COUNT(*) FROM ' . $this->db->get_table_prefix() . 'threads WHERE (' . $id_list . ')' . $topic_filter);

        // Generate output
        $out = [];
        foreach ($rows as $i => $r) {
            $out[$i] = [];
            $out[$i]['id'] = $r['tid'];
            $out[$i]['num'] = $r['replies'] + 1;
            $out[$i]['title'] = $r['subject'];
            $out[$i]['description'] = $r['subject'];
            $out[$i]['firsttime'] = $r['dateline'];
            $out[$i]['firstusername'] = $r['username'];
            $out[$i]['lastusername'] = $r['lastposter'];
            $out[$i]['firstmemberid'] = $r['uid'];
            $out[$i]['lastmemberid'] = $r['lastposteruid'];
            $out[$i]['lasttime'] = $r['lastpost'];
            $out[$i]['closed'] = (($r['closed'] == '') && ($r['visible'] == 1)) ? 0 : 1;

            // Get first non-spacer post
            $fp_rows = $this->db->query('SELECT subject,message,uid,dateline FROM ' . $this->db->get_table_prefix() . 'posts p WHERE message NOT LIKE \'' . db_encode_like(substr(do_lang('SPACER_POST', '', '', '', get_site_default_lang()), 0, 20) . '%') . '\' AND tid=' . strval($r['tid']) . ' ORDER BY dateline ASC', 1);

            // Filter topics with no posts
            if (!array_key_exists(0, $fp_rows)) {
                unset($out[$i]);
                continue;
            }

            // Update topic information according to first non-spacer post
            $out[$i]['firsttitle'] = $fp_rows[0]['subject'];
            if ($show_first_posts) {
                $out[$i]['firstpost'] = $fp_rows[0]['message'];
            }

            if ($i == $limit) {
                break;
            }
        }
        if (!empty($out)) {
            return $out;
        }

        return null;
    }

    /**
     * Get an array of members who are in at least one of the given array of groups.
     *
     * @param  array $groups The array of groups
     * @param  ?integer $max Return up to this many entries for primary members and this many entries for secondary members (null: no limit, only use no limit if querying very restricted usergroups!)
     * @param  integer $start Return primary members after this offset and secondary members after this offset
     * @return ?array The array of members (null: no members)
     */
    public function member_group_query(array $groups, ?int $max = null, int $start = 0) : ?array
    {
        $_groups = '';
        foreach ($groups as $group) {
            if ($_groups != '') {
                $_groups .= ' OR ';
            }
            $_groups .= 'u.usergroup=' . strval($group);
        }
        if ($_groups == '') {
            return [];
        }

        return $this->db->query('SELECT * FROM ' . $this->db->get_table_prefix() . 'users u LEFT JOIN ' . $this->db->get_table_prefix() . 'usergroups g ON u.usergroup=g.gid WHERE ' . $_groups . ' ORDER BY u.usergroup ASC', $max, $start, false, true);
    }

    /**
     * Get rows of members before the given one.
     * It cannot be assumed there are no gaps in member IDs, as members may be deleted.
     *
     * @param  MEMBER $member_id The member ID to paginate back from
     * @param  integer $total Number of members to retrieve
     * @return array Member rows
     */
    public function get_previous_members(int $member_id, int $total = 1) : array
    {
        $sql = 'SELECT uid FROM ' . $this->db->get_table_prefix() . 'users WHERE uid<>0';
        $sql .= ' AND uid<' . strval($member_id);
        $sql .= ' ORDER BY uid DESC';
        $rows = $this->db->query($sql, $total);
        return $rows;
    }

    /**
     * Get rows of members after the given one.
     * It cannot be assumed there are no gaps in member IDs, as members may be deleted.
     *
     * @param  ?MEMBER $member_id The member ID to increment (null: find the very first members)
     * @param  integer $total Number of members to retrieve
     * @return array Member rows
     */
    public function get_next_members(?int $member_id, int $total = 1) : array
    {
        $sql = 'SELECT uid FROM ' . $this->db->get_table_prefix() . 'users WHERE uid<>0';
        if ($member_id !== null) {
            $sql .= ' AND uid>' . strval($member_id);
        }
        $sql .= ' ORDER BY uid DESC';
        $rows = $this->db->query($sql, $total);
        return $rows;
    }

    /**
     * Try to find a member with the given IP address.
     *
     * @param  IP $ip The IP address
     * @return array The distinct rows found
     */
    public function probe_ip(string $ip) : array
    {
        return $this->db->query_select('posts', ['DISTINCT uid AS id'], ['ipaddress' => $ip]);
    }

    /**
     * Get the name relating to the specified member ID.
     * If this returns null, then the member has been deleted. Always take potential null output into account.
     *
     * @param  MEMBER $member_id The member ID
     * @return ?SHORT_TEXT The username (null: member deleted)
     */
    protected function _get_username(int $member_id) : ?string
    {
        if ($member_id == $this->get_guest_id()) {
            return do_lang('GUEST');
        }
        return $this->get_member_row_field($member_id, 'username');
    }

    /**
     * Get the e-mail address for the specified member ID.
     *
     * @param  MEMBER $member_id The member ID
     * @return SHORT_TEXT The e-mail address
     */
    protected function _get_member_email_address(int $member_id) : string
    {
        return $this->get_member_row_field($member_id, 'email');
    }

    /**
     * Find if this member may have e-mails sent to them.
     *
     * @param  MEMBER $member_id The member ID
     * @return boolean Whether the member may have e-mails sent to them
     */
    public function get_member_email_allowed(int $member_id) : bool
    {
        $v = $this->get_member_row_field($member_id, 'hideemail');
        if ($v == 0) {
            return true;
        }
        return false;
    }

    /**
     * Get the timestamp of a member's join date.
     *
     * @param  MEMBER $member_id The member ID
     * @return TIME The timestamp
     */
    public function get_member_join_timestamp(int $member_id) : int
    {
        return $this->get_member_row_field($member_id, 'regdate');
    }

    /**
     * Find all members with a name matching the given SQL LIKE string.
     *
     * @param  string $pattern The pattern
     * @param  ?integer $limit Maximum number to return (limits to the most recent active) (null: no limit)
     * @return ?array The array of matched members (null: none found)
     */
    public function get_matching_members(string $pattern, ?int $limit = null) : ?array
    {
        $rows = $this->db->query('SELECT * FROM ' . $this->db->get_table_prefix() . 'users WHERE username LIKE \'' . db_encode_like($pattern) . '\' AND uid<>' . strval($this->get_guest_id()) . ' ORDER BY lastactive DESC', $limit);
        sort_maps_by($rows, 'username');
        return $rows;
    }

    /**
     * Get the given member's post count.
     *
     * @param  MEMBER $member_id The member ID
     * @return integer The post count
     */
    public function get_post_count(int $member_id) : int
    {
        $c = $this->get_member_row_field($member_id, 'postnum');
        if ($c === null) {
            $c = 0;
        }
        return $c;
    }

    /**
     * Get the given member's topic count.
     *
     * @param  MEMBER $member_id The member ID
     * @return integer The topic count
     */
    public function get_topic_count(int $member_id) : int
    {
        return $this->db->query_select_value('threads', 'COUNT(*)', ['uid' => $member_id]);
    }

    /**
     * Find out if the given member ID is banned.
     *
     * @param  MEMBER $member_id The member ID
     * @param  ?ID_TEXT $reasoned_ban Ban reasoning returned by reference (null: none)
     * @return boolean Whether the member is banned
     */
    public function is_banned(int $member_id, ?string &$reasoned_ban = null) : bool
    {
        $rows = $this->db->query('SELECT * FROM ' . $this->db->get_table_prefix() . 'banned WHERE uid=' . strval($member_id));
        if (empty($rows)) {
            return false; // The member was never banned
        }

        $ban_till = null;
        foreach ($rows as $row) {
            // Permanent ban; set the reason and exit immediately
            if (empty($row['lifted'])) {
                $reasoned_ban = $row['reason'];
                return true;
            }

            // Check for temporary bans that are still active; we want the reason for the longest ban in the event multiple temporary bans exist
            if ((($ban_till === null) || ($row['lifted'] > $ban_till)) && ($row['lifted'] > time())) {
                $reasoned_ban = $row['reason'];
                $ban_till = $row['lifted'];
            }
        }

        return ($ban_till !== null);
    }

    /**
     * Find the base URL to the emoticons.
     *
     * @return URLPATH The base URL
     */
    public function get_emo_dir() : string
    {
        //return get_forum_base_url().'/images/smilies/';
        return get_forum_base_url() . '/'; // The database includes "images/smilies" in the path for locally-added smilies.
    }

    /**
     * Get a map between emoticon codes and templates representing the HTML-image-code for this emoticon. The emoticons presented of course depend on the forum involved.
     *
     * @return array The map
     */
    public function find_emoticons() : array
    {
        if ($this->EMOTICON_CACHE !== null) {
            return $this->EMOTICON_CACHE;
        }
        $rows = $this->db->query_select('smilies', ['*']);
        sort_maps_by($rows, 'disporder');

        $this->EMOTICON_CACHE = [];
        foreach ($rows as $myrow) {
            $src = $this->get_emo_dir() . $myrow['image'];
            $this->EMOTICON_CACHE[$myrow['find']] = ['EMOTICON_IMG_CODE_DIR', $src, $myrow['find']];
        }
        return $this->EMOTICON_CACHE;
    }

    /**
     * Find a list of all forum skins (aka themes).
     *
     * @return array The list of skins
     */
    public function get_skin_list() : array
    {
        $table = 'themes';
        $codename = 'name';

        $rows = $this->db->query_select($table, [$codename]);
        return collapse_1d_complexity($codename, $rows);
    }

    /**
     * Try to find the theme that the logged-in/guest member is using, and map it to a site theme.
     * The themes/map.ini file functions to provide this mapping between forum themes, and site themes, and has a slightly different meaning for different forum drivers. For example, some drivers map the forum themes theme directory to the site theme name, while others made the humanly readable name.
     *
     * @param  boolean $skip_member_specific Whether to avoid member-specific lookup (i.e. find via what forum theme is currently configured as the default)
     * @param  ?MEMBER $member_id The member to find for (null: current member)
     * @return ID_TEXT The theme
     */
    public function _get_theme(bool $skip_member_specific = false, ?int $member_id = null) : string
    {
        $def = '';

        // Load in remapper
        require_code('files');
        $map = file_exists(get_file_base() . '/themes/map.ini') ? cms_parse_ini_file_fast(get_file_base() . '/themes/map.ini') : [];

        // Work out
        if (!$skip_member_specific) {
            if ($member_id === null) {
                $member_id = get_member();
            }
            if ($member_id > 0) {
                $skin = $this->get_member_row_field($member_id, 'style');
            } else {
                $skin = 0;
            }
            if ($skin > 0) { // User has a custom theme
                $_user_theme = $this->db->query_select('themes', ['*'], ['tid' => $skin], '', 1);
                $user_theme = (!empty($_user_theme[0]['name'])) ? $_user_theme[0]['name'] : '';

                if ($user_theme !== null) {
                    $def = array_key_exists($user_theme, $map) ? $map[$user_theme] : $user_theme;
                }
            }
        }

        // Look for a skin according to our site name (we bother with this instead of 'default' because the software itself likes to never choose a theme when forum-theme integration is on: all forum [via map] or all site seems cleaner, although it is complex)
        if ((!(strlen($def) > 0)) || (!file_exists(get_custom_file_base() . '/cache/themes/theme' . $skin))) {
            $mybbtheme = $this->db->query_select_value_if_there('themes', 'name', ['name' => get_site_name()]);
            if ($mybbtheme !== null) {
                $def = array_key_exists($mybbtheme, $map) ? $map[$mybbtheme] : $mybbtheme;
            }
        }

        // Default then!
        if ((!(strlen($def) > 0)) || (!file_exists(get_custom_file_base() . '/cache/themes/theme' . $skin))) {
            $def = array_key_exists('default', $map) ? $map['default'] : 'default';
        }

        return $def;
    }

    /**
     * Find if the specified member ID is marked as staff or not.
     *
     * @param  MEMBER $member_id The member ID
     * @return boolean Whether the member is staff
     */
    protected function _is_staff(int $member_id) : bool
    {
        $user_level = $this->get_member_row_field($member_id, 'usergroup');
        if (in_array($user_level, [MYBB_USERGROUP_ADMINISTRATORS, MYBB_USERGROUP_SUPER_MODERATORS, MYBB_USERGROUP_MODERATORS])) {
            return true; // return all administrators + all moderators
        }
        //if ($user_level == MYBB_USERGROUP_ADMINISTRATORS) return true; // This returns only administrators
        return false;
    }

    /**
     * Find if the specified member ID is marked as a super admin or not.
     *
     * @param  MEMBER $member_id The member ID
     * @return boolean Whether the member is a super admin
     */
    protected function _is_super_admin(int $member_id) : bool
    {
        $user_level = $this->get_member_row_field($member_id, 'usergroup');
        if ($user_level == MYBB_USERGROUP_ADMINISTRATORS) {
            return true;
        }
        return false;
    }

    /**
     * Get the number of members currently online on the forums.
     *
     * @return integer The number of members
     */
    public function get_num_users_forums() : int
    {
        return $this->db->query_value_if_there('SELECT COUNT(*) FROM ' . $this->db->get_table_prefix() . 'users WHERE lastactive>' . strval(time() - 60 * intval(get_option('users_online_time'))));
    }

    /**
     * Get the number of members registered on the forum.
     *
     * @return integer The number of members
     */
    public function get_num_members() : int
    {
        return $this->db->query_select_value('users', 'COUNT(*)') - 1;
    }

    /**
     * Get the total topics ever made on the forum.
     *
     * @return integer The number of topics
     */
    public function get_num_topics() : int
    {
        return $this->db->query_select_value('threads', 'COUNT(*)');
    }

    /**
     * Get the total posts ever made on the forum.
     *
     * @return integer The number of posts
     */
    public function get_num_forum_posts() : int
    {
        return $this->db->query_select_value('posts', 'COUNT(*)');
    }

    /**
     * Get the number of new forum posts.
     *
     * @return integer The number of posts
     */
    protected function _get_num_new_forum_posts() : int
    {
        return $this->db->query_value_if_there('SELECT COUNT(*) FROM ' . $this->db->get_table_prefix() . 'posts WHERE dateline>' . strval(time() - 60 * 60 * 24));
    }

    /**
     * Get a member ID from the given member's username.
     *
     * @param  SHORT_TEXT $username The username
     * @return ?MEMBER The member ID (null: not found)
     */
    public function get_member_from_username(string $username) : ?int
    {
        if ($username == do_lang('GUEST')) {
            return $this->get_guest_id();
        }

        return $this->db->query_select_value_if_there('users', 'uid', ['username' => $username]);
    }

    /**
     * Get a member ID from the given member's e-mail address.
     *
     * @param  SHORT_TEXT $email_address The member e-mail address
     * @return ?MEMBER The member ID (null: not found)
     */
    public function get_member_from_email_address(string $email_address) : ?int
    {
        $results = $this->db->query_select('users', ['uid', 'regdate'], ['email' => $email_address], 'ORDER BY regdate DESC', 1);
        return array_key_exists(0, $results) ? $results[0]['uid'] : null;
    }

    /**
     * Get the IDs of the admin groups.
     *
     * @return array The admin group IDs
     */
    protected function _get_super_admin_groups() : array
    {
        $admin_group = $this->db->query_select_value_if_there('usergroups', 'gid', ['title' => 'Administrators'], 'ORDER BY gid DESC');
        if ($admin_group === null) {
            return [];
        }
        return [$admin_group];
    }

    /**
     * Get the IDs of the moderator groups.
     * It should not be assumed that a member only has one group - this depends upon the forum the driver works for. It also does not take the staff site filter into account.
     *
     * @return array The moderator group IDs
     */
    protected function _get_moderator_groups() : array
    {
        $moderator_group = $this->db->query('SELECT gid FROM ' . $this->db->get_table_prefix() . 'usergroups WHERE issupermod=1');

        if ($moderator_group === null) {
            return [];
        }

        $return_moderator_group = [];
        foreach ($moderator_group as $key => $value) {
            if ($value['gid'] !== null) {
                $return_moderator_group[] = $value['gid'];
            }
        }

        return $return_moderator_group;
    }

    /**
     * Get the forum usergroup list.
     *
     * @return array The usergroup list
     */
    protected function _get_usergroup_list() : array
    {
        $results = $this->db->query('SELECT gid,title FROM ' . $this->db->get_table_prefix() . 'usergroups');
        $mod_results = [];
        foreach ($results as $key => $value) {
            $mod_results[] = ['group_id' => $value['gid'], 'group_name' => $value['title']];
        }
        $results2 = collapse_2d_complexity('group_id', 'group_name', $mod_results);
        return $results2;
    }

    /**
     * Get the forum usergroup relating to the specified member ID.
     *
     * @param  MEMBER $member_id The member ID
     * @return array The array of forum usergroups
     */
    protected function _get_members_groups(int $member_id) : array
    {
        if ($member_id == $this->get_guest_id()) {
            return [-1];
        }

        $groups = [];
        $additional_groups = explode(',', $this->get_member_row_field($member_id, 'additionalgroups'));
        foreach ($additional_groups as $key => $value) {
            if (!cms_empty_safe($value)) {
                $groups[] = intval($value);
            }
        }
        $groups[] = $this->get_member_row_field($member_id, 'usergroup');

        return $groups;
    }

    /**
     * Create a member login cookie.
     *
     * @param  MEMBER $member_id The member ID
     * @param  ?SHORT_TEXT $username The username (null: lookup)
     * @param  string $password_raw The password (note this is sometimes not used by forum drivers, as they can also do something with what is already in the database instead)
     */
    public function create_login_cookie(int $member_id, ?string $username, string $password_raw)
    {
        global $SITE_INFO;

        $row = $this->get_member_row($member_id);
        $loginkey = $row['loginkey']; // Used for 'mybbuser' memberid.'_'.'loginkey'

        // Set a User COOKIE
        $member_cookie_name = get_member_cookie();
        cms_setcookie($member_cookie_name, strval($member_id) . '_' . $loginkey, false, true);

        if (substr($member_cookie_name, 0, 5) != 'cms__') {
            $session_id = $this->db->query_select_value_if_there('sessions', 'sid', ['uid' => $member_id]);

            if (!empty($session_id)) {
                $this->db->query_update('sessions', ['time' => time()], ['sid' => $session_id], '', 1);
            } else {
                require_code('crypt');
                $session_id = get_secure_random_string();

                $ip = get_ip_address();
                $packed_ip = inet_pton($ip);
                $hex_ip = 'X\'' . bin2hex($packed_ip) . '\'';
                $this->db->query('INSERT INTO ' . $this->db->get_table_prefix() . 'sessions (sid, uid, time, ip) VALUES (\'' . db_escape_string($session_id) . '\', ' . strval($member_id) . ', ' . strval(time()) . ', ' . $hex_ip . ')');
            }

            // Now lets try and set a COOKIE of MyBB Session ID
            cms_setcookie('sid', $session_id, false, true);
        }
    }

    /**
     * Try and log in using a member cookie.
     * Should only be called if the cookie exists.
     *
     * @return ?array A map of 'id' and 'error'. If 'id' is null, an error occurred and 'error' is set (null: no cookie)
     */
    public function authorise_cookie_login() : ?array
    {
        if ((!isset($_COOKIE[get_member_cookie()])) || (!isset($_COOKIE[get_pass_cookie()]))) {
            return null;
        }

        $username = $_COOKIE[get_member_cookie()];
        $password_hashed = $_COOKIE[get_pass_cookie()];

        return $this->_authorise_login($username, null, $password_hashed, true);
    }

    /**
     * Find if the given member ID and password is valid. If username is null, then the member ID is used instead.
     *
     * @param  ?SHORT_TEXT $username The member username (null: use $member_id)
     * @param  ?MEMBER $member_id The member ID (null: use $username)
     * @param  string $password_raw The raw password
     * @return array A map of 'id' and 'error'. If 'id' is null, an error occurred and 'error' is set
     */
    public function authorise_login(?string $username, ?int $member_id, string $password_raw) : array
    {
        return $this->_authorise_login($username, $member_id, $password_raw, false);
    }

    /**
     * Find if the given member ID and password is valid. If username is null, then the member ID is used instead.
     * All authorisation, cookies, and form-logins, are passed through this function.
     *
     * @param  ?SHORT_TEXT $username The member username (null: use $member_id)
     * @param  ?MEMBER $member_id The member ID (null: use $username)
     * @param  string $password_mixed If $cookie_login is true then this is the value of the password cookie, otherwise it's the password the user tried to log in with
     * @param  boolean $cookie_login Whether this is a cookie login, determines how the hashed password is treated for the value passed in
     * @return array A map of 'id' and 'error'. If 'id' is null, an error occurred and 'error' is set
     */
    protected function _authorise_login(?string $username, ?int $member_id, string $password_mixed, bool $cookie_login = false) : array
    {
        global $SITE_INFO;

        $out = [];
        $out['id'] = null;

        if ($member_id === null) {
            $rows = $this->db->query_select('users', ['*'], ['username' => $username], '', 1);
            if (array_key_exists(0, $rows)) {
                $this->MEMBER_ROWS_CACHED[$rows[0]['uid']] = $rows[0];
            }
        } else {
            $rows = [];
            $rows[0] = $this->get_member_row($member_id);
        }

        if (!array_key_exists(0, $rows) || $rows[0] === null) { // All hands to lifeboats
            $out['error'] = do_lang_tempcode((get_option('login_error_secrecy') == '1') ? 'MEMBER_INVALID_LOGIN' : '_MEMBER_NO_EXIST', $username);
            return $out;
        }
        $row = $rows[0];
        if ($this->is_banned($row['uid'])) { // All hands to the guns
            $out['error'] = do_lang_tempcode('YOU_ARE_BANNED');
            return $out;
        }

        if ($cookie_login) {
            $cookie_loginkey = '';
            $cookie_member = null;
            if (!empty($_COOKIE['mybbuser'])) {
                $cookie_info = [];
                $cookie_info = explode('_', $_COOKIE['mybbuser']);
                $cookie_member = (!@cms_empty_safe($cookie_info[0])) ? intval($cookie_info[0]) : null;
                $cookie_loginkey = (!empty($cookie_info[1])) ? $cookie_info[1] : '';
            }

            $lookup = $this->db->query_select_value_if_there('users', 'uid', ['loginkey' => $cookie_loginkey, 'uid' => $cookie_member]);

            if ($row['uid'] !== $lookup) {
                $out['error'] = do_lang_tempcode((get_option('login_error_secrecy') == '1') ? 'MEMBER_INVALID_LOGIN' : 'MEMBER_BAD_PASSWORD');
                return $out;
            }
        } else {
            if (!hash_equals(md5(md5($row['salt']) . md5($password_mixed)), $row['password'])) {
                $out['error'] = do_lang_tempcode((get_option('login_error_secrecy') == '1') ? 'MEMBER_INVALID_LOGIN' : 'MEMBER_BAD_PASSWORD');
                return $out;
            }
        }

        $out['id'] = $row['uid'];
        return $out;
    }

    /**
     * Salts a password based on a supplied salt.
     *
     * @param  string $password The md5()'ed password
     * @param  string $salt The salt
     * @return string The password hash
     */
    protected function salt_password(string $password, string $salt) : string
    {
        return md5(md5($salt) . $password);
    }

    /**
     * Get a first known IP address of the given member.
     *
     * @param  MEMBER $member_id The member ID
     * @return IP The IP address
     */
    public function get_member_ip(int $member_id) : string
    {
        $ip = $this->db->query_select_value_if_there('users', 'regip', ['uid' => $member_id]);
        if ($ip !== null) {
            $unpacked_ip = inet_ntop($ip);
            return $unpacked_ip;
        }
        return '';
    }

    /**
     * Gets a whole member row from the database.
     *
     * @param  MEMBER $member_id The member ID
     * @return ?array The member row (null: no such member)
     */
    public function get_member_row(int $member_id) : ?array
    {
        if (array_key_exists($member_id, $this->MEMBER_ROWS_CACHED)) {
            return $this->MEMBER_ROWS_CACHED[$member_id];
        }

        $rows = $this->db->query_select('users', ['*'], ['uid' => $member_id], '', 1);
        if ($member_id == $this->get_guest_id()) { // MyBB does not have a guest account; pretend it does
            $rows[0]['username'] = do_lang('GUEST');
            $rows[0]['email'] = '';
            $rows[0]['postnum'] = 0;
            $rows[0]['threadnum'] = 0;
            $rows[0]['avatar'] = '';
            $rows[0]['avatartype'] = 0;
            $rows[0]['usergroup'] = MYBB_USERGROUP_GUESTS;
            $rows[0]['additionalgroups'] = '';
            $rows[0]['displaygroup'] = 0;
            $rows[0]['usertitle'] = '';
            $rows[0]['regdate'] = website_creation_time();
            $rows[0]['lastactive'] = time();
            $rows[0]['lastvisit'] = time();
            $rows[0]['lastpost'] = 0;
            $rows[0]['allownotices'] = 0;
            $rows[0]['hideemail'] = 1;
            $rows[0]['subscriptionmethod'] = 0;
            $rows[0]['invisible'] = 1;
            $rows[0]['receivepms'] = 0;
            $rows[0]['receivefrombuddy'] = 0;
            $rows[0]['pmnotice'] = 0;
            $rows[0]['pmnotify'] = 0;
            $rows[0]['buddyrequestspm'] = 0;
            $rows[0]['buddyrequestsauto'] = 0;
            $rows[0]['showimages'] = 1;
            $rows[0]['showvideos'] = 1;
            $rows[0]['showsigs'] = 1;
            $rows[0]['showavatars'] = 1;
            $rows[0]['showquickreply'] = 1;
            $rows[0]['showredirect'] = 1;
            $rows[0]['ppp'] = 0;
            $rows[0]['tpp'] = 0;
            $rows[0]['daysprune'] = 0;
            $rows[0]['timezone'] = 0;
            $rows[0]['dst'] = 0;
            $rows[0]['dstcorrection'] = 0;
            $rows[0]['style'] = 0;
            $rows[0]['away'] = 0;
            $rows[0]['awaydate'] = 0;
            $rows[0]['referrer'] = 0;
            $rows[0]['referrals'] = 0;
            $rows[0]['reputation'] = 0;
            $rows[0]['regip'] = bin2hex(inet_pton(get_ip_address()));
            $rows[0]['lastip'] = bin2hex(inet_pton(get_ip_address()));
            $rows[0]['timeonline'] = 0;
            $rows[0]['showcodebuttons'] = 1;
            $rows[0]['totalpms'] = 0;
            $rows[0]['unreadpms'] = 0;
            $rows[0]['warningpoints'] = 0;
            $rows[0]['moderateposts'] = 1;
            $rows[0]['moderationtime'] = 0;
            $rows[0]['suspendposting'] = 0;
            $rows[0]['suspensiontime'] = 0;
            $rows[0]['suspendsignature'] = 0;
            $rows[0]['suspendsigtime'] = 0;
            $rows[0]['coppauser'] = 0;
            $rows[0]['classicpostbit'] = 0;
            $rows[0]['loginattempts'] = 0;
            $rows[0]['loginlockoutexpiry'] = 0;
            $rows[0]['usernotes'] = '';
            $rows[0]['sourceeditor'] = 0;
        }
        if (!array_key_exists(0, $rows)) {
            return null;
        }
        $this->MEMBER_ROWS_CACHED[$member_id] = $rows[0];
        return $this->MEMBER_ROWS_CACHED[$member_id];
    }

    /**
     * Gets a named field of a member row from the database.
     *
     * @param  MEMBER $member_id The member ID
     * @param  string $field The field identifier
     * @return mixed The field
     */
    public function get_member_row_field(int $member_id, string $field)
    {
        $row = $this->get_member_row($member_id);
        return ($row === null) ? null : $row[$field];
    }

    /**
     * Custom get member function.
     *
     * @return MEMBER The member or the default guest ID (0)
     */
    public function get_member() : int
    {
        // Get cookie information if available
        $cookie_raw_info = (array_key_exists(get_member_cookie(), $_COOKIE)) ? $_COOKIE[get_member_cookie()] : '';

        $cookie_info = [];
        $cookie_info = explode('_', $cookie_raw_info);
        $cookie_member = (array_key_exists(0, $cookie_info)) ? $cookie_info[0] : '';
        $cookie_loginkey = (array_key_exists(1, $cookie_info)) ? $cookie_info[1] : '';

        if ($cookie_member != '') {
            $row = $this->get_member_row(intval($cookie_member));
            // Is the cookie info correct
            if ($cookie_loginkey == $row['loginkey']) {
                // If it is correct then return the cookie member
                return intval($cookie_member);
            } else {
                // Return the default guest ID, because the login key is not correct
                return $this->get_guest_id();
            }
        }

        // Return the default guest ID, because there is no member cookie information
        return $this->get_guest_id();
    }

    /**
     * Returns the SQL query to use in an ORDER BY when sorting by a member (actually, username, but we only know the ID).
     *
     * @param  ID_TEXT $field The name of the r.* field containing the member ID
     * @param  boolean $use_displayname Whether we want the display name instead of the username
     * @return ?string The SQL query (null: this driver does not support sorting by member)
     */
    public function get_username_sort_query(string $field, bool $use_displayname = false) : ?string
    {
        // No display name support
        return '(SELECT username FROM ' . $this->db->table_prefix . 'users m WHERE m.uid=r.' . $field . ')';
    }
}
