<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    core_forum_drivers
 */

/**
 * Forum driver class.
 *
 * @package core_forum_drivers
 */
class Forum_driver_smf2 extends Forum_driver_base
{
    /**
     * Check the connected DB is valid for this forum driver.
     *
     * @return boolean Whether it is valid
     */
    public function check_db() : bool
    {
        $test = $this->db->query('SELECT COUNT(*) FROM ' . $this->db->get_table_prefix() . 'members', null, 0, true); // Suppress errors as we're checking to see if the table is there
        return $test !== null;
    }

    /**
     * Get the rows for the top given number of posters on the forum.
     *
     * @param  integer $limit The limit to the number of top posters to fetch
     * @return array The rows for the given number of top posters in the forum
     */
    public function get_top_posters(int $limit) : array
    {
        return $this->db->query('SELECT * FROM ' . $this->db->get_table_prefix() . 'members WHERE id_member<>' . strval($this->get_guest_id()) . ' ORDER BY posts DESC', $limit);
    }

    /**
     * Attempt to to find the member's language from their forum profile. It converts between language-identifiers using a map (lang/map.ini).
     *
     * @param  MEMBER $member_id The member who's language needs to be fetched
     * @return ?LANGUAGE_NAME The member's language (null: unknown)
     */
    public function get_lang(int $member_id) : ?string
    {
        return $this->get_member_row_field($member_id, 'lngfile');
    }

    /**
     * Find the member ID of the forum guest member.
     *
     * @return MEMBER The member ID of the forum guest member
     */
    public function get_guest_id() : int
    {
        return 0;
    }

    /**
     * Get the forums' table prefix for the database.
     *
     * @return string The forum database table prefix
     */
    public function get_drivered_table_prefix() : string
    {
        global $SITE_INFO;
        return $SITE_INFO['smf_table_prefix'];
    }

    /**
     * Add the specified custom field to the forum (some forums implemented this using proper Custom Profile Fields, others through adding a new field).
     *
     * @param  string $name The name of the new custom field
     * @param  integer $length The length of the new custom field (ignored for Conversr, $type used instead)
     * @param  BINARY $locked Whether the field is locked, e.g. cannot be deleted
     * @param  BINARY $viewable Whether the field is for viewing by the owner and the public (0: cannot be viewed unless the viewing member has view_any_profile_field privilege)
     * @param  BINARY $settable Whether the field is for setting by the owner
     * @param  BINARY $required Whether the field is required
     * @param  string $description Description
     * @param  string $type The field type (it's the same as the software's field types, although we only expect forum drivers to specifically support short_text/long_text/integer/float and for the rest to be mapped to long_text)
     * @param  BINARY $encrypted Whether the field is encrypted
     * @param  ?string $default Default field value (null: standard for field type)
     * @param  SHORT_TEXT $options Field options
     * @param  BINARY $include_in_main_search Whether to include in main keyword search
     * @param  BINARY $allow_template_search Whether to allow template search
     * @param  ID_TEXT $icon An icon to show the CPF with on the member profiles
     * @param  ID_TEXT $section A section to show with on the member-links part of member profiles
     * @param  LONG_TEXT $tempcode This is Tempcode that is used for displaying the field. See the DESCRIPTION_CPF_CODE language string.
     * @param  ID_TEXT $autofill_type Autofill field name from https://html.spec.whatwg.org/multipage/form-control-infrastructure.html#autofill-field
     * @param  ID_TEXT $autofill_hint Autofill hint: '' or 'shipping' or 'billing'
     * @return boolean Whether the custom field was created successfully
     */
    public function install_create_custom_field(string $name, int $length, int $locked = 1, int $viewable = 0, int $settable = 0, int $required = 0, string $description = '', string $type = 'long_text', int $encrypted = 0, ?string $default = null, string $options = '', int $include_in_main_search = 0, int $allow_template_search = 0, string $icon = '', string $section = '', string $tempcode = '', string $autofill_type = '', string $autofill_hint = '') : bool
    {
        list($db_type, $db_default) = $this->remap_cms_field_type_to_db_type($type, $default);
        $query = $this->db->driver->add_table_field__sql($this->db->get_table_prefix() . 'members', 'cms_' . $name, $db_type, $db_default);
        $this->db->query($query, null, 0, true); // Suppress errors in case field already exists
        return true;
    }

    /**
     * Edit the specified custom field in the forum (some forums implemented this using proper Custom Profile Fields, others through adding a new field).
     *
     * @param  string $old_name The name of the existing custom field to edit
     * @param  string $name The new name of the custom field
     * @param  integer $length The length of the custom field (ignored for Conversr, $type used instead)
     * @param  BINARY $locked Whether the field is locked, e.g. cannot be deleted
     * @param  BINARY $viewable Whether the field is for viewing by the owner and the public (0: cannot be viewed unless the viewing member has view_any_profile_field privilege)
     * @param  BINARY $settable Whether the field is for setting by the owner
     * @param  BINARY $required Whether the field is required
     * @param  string $description Description
     * @param  string $type The field type (it's the same as the software's field types, although we only expect forum drivers to specifically support short_text/long_text/integer/float and for the rest to be mapped to long_text)
     * @param  BINARY $encrypted Whether the field is encrypted
     * @param  ?string $default Default field value (null: standard for field type)
     * @param  SHORT_TEXT $options Field options
     * @param  BINARY $include_in_main_search Whether to include in main keyword search
     * @param  BINARY $allow_template_search Whether to allow template search
     * @param  ID_TEXT $icon An icon to show the CPF with on the member profiles
     * @param  ID_TEXT $section A section to show with on the member-links part of member profiles
     * @param  LONG_TEXT $tempcode This is Tempcode that is used for displaying the field. See the DESCRIPTION_CPF_CODE language string.
     * @param  ID_TEXT $autofill_type Autofill field name from https://html.spec.whatwg.org/multipage/form-control-infrastructure.html#autofill-field
     * @param  ID_TEXT $autofill_hint Autofill hint: '' or 'shipping' or 'billing'
     * @return boolean Whether the custom field was edited successfully
     */
    public function install_edit_custom_field(string $old_name, string $name, int $length, int $locked = 1, int $viewable = 0, int $settable = 0, int $required = 0, string $description = '', string $type = 'long_text', int $encrypted = 0, ?string $default = null, string $options = '', int $include_in_main_search = 0, int $allow_template_search = 0, string $icon = '', string $section = '', string $tempcode = '', string $autofill_type = '', string $autofill_hint = '') : bool
    {
        list($db_type) = $this->remap_cms_field_type_to_db_type($type, $default);
        $query = $this->db->driver->alter_table_field($this->db->get_table_prefix() . 'members', 'cms_' . $old_name, $db_type, 'cms_' . $name);
        $this->db->query($query, null, 0, true); // Suppress errors in case field already edited
        return true;
    }

    /**
     * Get an array of attributes to take in from the installer. Almost all forums require a table prefix, which the requirement there-of is defined through this function.
     * The attributes have 4 values in an array:
     * - name, the name of the attribute for _config.php
     * - default, the default value (perhaps obtained through autodetection from forum config)
     * - description, a textual description of the attributes
     * - title, a textual title of the attribute
     *
     * @return array The attributes for the forum
     */
    public function install_specifics() : array
    {
        global $PROBED_FORUM_CONFIG;
        $a = [];
        $a['name'] = 'smf_table_prefix';
        $a['default'] = array_key_exists('sql_tbl_prefix', $PROBED_FORUM_CONFIG) ? $PROBED_FORUM_CONFIG['sql_tbl_prefix'] : 'smf_';
        $a['description'] = do_lang('MOST_DEFAULT');
        $a['title'] = 'SMF ' . do_lang('TABLE_PREFIX');
        return [$a];
    }

    /**
     * Searches for forum auto-config at this path.
     *
     * @param  PATH $path The path in which to search
     * @return boolean Whether the forum auto-config could be found
     */
    public function install_test_load_from(string $path) : bool
    {
        global $PROBED_FORUM_CONFIG;
        if (@file_exists($path . '/Settings.php')) {
            $db_server = '';
            $db_name = '';
            $db_user = '';
            $db_passwd = '';
            $boardurl = '';
            $db_prefix = '';
            $cookiename = '';
            $auth_secret = '';

            @include($path . '/Settings.php');
            $PROBED_FORUM_CONFIG['sql_host'] = $db_server;
            $PROBED_FORUM_CONFIG['sql_database'] = $db_name;
            $PROBED_FORUM_CONFIG['sql_user'] = $db_user;
            $PROBED_FORUM_CONFIG['sql_pass_exists'] = ($db_passwd != '');
            $PROBED_FORUM_CONFIG['sql_tbl_prefix'] = $db_prefix;
            $PROBED_FORUM_CONFIG['board_url'] = $boardurl;
            $PROBED_FORUM_CONFIG['cookie_member_id'] = $cookiename . ':0';
            $PROBED_FORUM_CONFIG['cookie_member_hash'] = $cookiename . ':1';
            $PROBED_FORUM_CONFIG['set_values'] = [
                ['smf_auth_secret', $auth_secret],
            ];
            return true;
        }
        return false;
    }

    /**
     * Get an array of paths to search for config at.
     *
     * @return array The paths in which to search for the forum config
     */
    public function install_get_path_search_list() : array
    {
        return [
            'smf',
            'SMF',
            'forums',
            'forum',
            'boards',
            'board',
            '../forums/smf',
            '../forums/SMF',
            '../smf',
            '../SMF',
            '../forums',
            '../forum',
            '../boards',
            '../board',
        ];
    }

    /**
     * Get an emoticon chooser template.
     *
     * @param  string $field_name The ID of the form field the emoticon chooser adds to
     * @return Tempcode The emoticon chooser template
     */
    public function get_emoticon_chooser(string $field_name = 'post') : object
    {
        require_code('comcode_compiler');
        $emoticons = $this->db->query_select('smileys', ['*'], ['hidden' => 0]);
        $em = new Tempcode();
        foreach ($emoticons as $emo) {
            $code = $emo['code'];
            $em->attach(do_template('EMOTICON_CLICK_CODE', ['_GUID' => '33e9de26a5fe69b668281f2eeffe06e9', 'FIELD_NAME' => $field_name, 'CODE' => $code, 'IMAGE' => apply_emoticons($code)]));
        }

        return $em;
    }

    /**
     * Pin a topic.
     *
     * @param  AUTO_LINK $topic_id The topic ID
     * @param  boolean $pin True: pin it, False: unpin it
     */
    public function pin_topic(int $topic_id, bool $pin = true)
    {
        $this->db->query_update('topics', ['is_sticky' => $pin ? 1 : 0], ['id_topic' => $topic_id], '', 1);
    }

    /**
     * Set a Custom Profile Field's value, if the custom field exists. Only works on specially-named (titled) fields.
     *
     * @param  MEMBER $member_id The member ID
     * @param  string $field The field name (e.g. "firstname" for the CPF with a title of "cms_firstname")
     * @param  string $value The value
     */
    public function set_custom_field(int $member_id, string $field, string $value)
    {
        $this->db->query_update('members', ['cms_' . $field => $value], ['id_member' => $member_id], '', null, 0, false, true); // Errors suppressed in case field gone missing
    }

    /**
     * Get Custom Profile Fields values for all 'cms_' prefixed keys.
     *
     * @param  MEMBER $member_id The member ID
     * @return ?array A map of the Custom Profile Fields, key_suffix=>value (null: no fields)
     */
    public function get_custom_fields(int $member_id) : ?array
    {
        $row = $this->get_member_row($member_id);
        if ($row === null) {
            return null;
        }

        $out = [];
        foreach ($row as $attribute => $value) {
            if (substr($attribute, 0, 4) == 'cms_') {
                $out[substr($attribute, 4)] = $value;
            }
        }
        return $out;
    }

    /**
     * From a member row, get the member's primary usergroup.
     *
     * @param  array $r The profile-row
     * @return GROUP The member's primary usergroup
     */
    public function mrow_primary_group(array $r) : int
    {
        return $r['id_group'];
    }

    /**
     * From a member row, get the member's member ID.
     *
     * @param  array $r The profile-row
     * @return MEMBER The member ID
     */
    public function mrow_member_id(array $r) : int
    {
        return $r['id_member'];
    }

    /**
     * From a member row, get the member's last visit date.
     *
     * @param  array $r The profile-row
     * @return TIME The last visit date
     */
    public function mrow_lastvisit(array $r) : int
    {
        return $r['last_login'];
    }

    /**
     * From a member row, get the member's name.
     *
     * @param  array $r The profile-row
     * @return string The username
     */
    public function mrow_username(array $r) : string
    {
        return $r['real_name'];
    }

    /**
     * From a member row, get the member's e-mail address.
     *
     * @param  array $r The profile-row
     * @return SHORT_TEXT The member e-mail address
     */
    public function mrow_email_address(array $r) : string
    {
        return $r['email_address'];
    }

    /**
     * Get the photo URL for the specified member ID.
     *
     * @param  MEMBER $member_id The member ID
     * @param  boolean $full Get full photo
     * @return URLPATH The URL (blank: none)
     */
    public function get_member_photo_url(int $member_id, bool $full = false) : string
    {
        return '';
    }

    /**
     * Get the avatar URL for the specified member ID.
     *
     * @param  MEMBER $member_id The member ID
     * @param  boolean $fallback_support Whether fallback support should be allowed (passed by reference)
     * @return URLPATH The URL (blank: none)
     */
    protected function _get_member_avatar_url(int $member_id, bool &$fallback_support) : string
    {
        // Could be remote avatar URL off the bat, in which case $ret is untouched and simply returned at the bottom of the function
        $ret = $this->get_member_row_field($member_id, 'avatar');

        // Uploaded avatars; need to get the attachment
        if ($ret == '') {
            $filename = $this->db->query_select_value_if_there('attachments', 'filename', ['id_member' => $member_id]);
            if ($filename === null) {
                return '';
            }
            return $this->get_setting('custom_avatar_url') . '/' . strval($filename);
        }

        // Gravatar
        if (substr($ret, 0, 11) == 'gravatar://') {
            // Ignore custom Gravatar e-mail addresses for privacy reasons
            return find_script('gravatar') . '?id=' . strval($member_id) . '&from_driver=1';
        }

        // Avatar gallery
        if (url_is_local($ret)) {
            return $this->get_setting('avatar_url') . '/' . $ret;
        }

        return $ret;
    }

    /**
     * Get a URL to the specified member's home (control panel).
     *
     * @param  MEMBER $member_id The member ID
     * @return URLPATH The URL to the members home
     */
    public function member_home_url(int $member_id) : string
    {
        return $this->_member_profile_url($member_id);
    }

    /**
     * Get a URL to the specified member's profile.
     *
     * @param  MEMBER $member_id The member ID
     * @return URLPATH The URL to the member profile
     */
    protected function _member_profile_url(int $member_id) : string
    {
        return get_forum_base_url() . '/index.php?action=profile&u=' . strval($member_id);
    }

    /**
     * Get a URL to the registration page (for people to create member accounts).
     *
     * @return URLPATH The URL to the registration page
     */
    protected function _join_url() : string
    {
        return get_forum_base_url() . '/index.php?action=register';
    }

    /**
     * Get a URL to the members-online page.
     *
     * @return URLPATH The URL to the members-online page
     */
    protected function _users_online_url() : string
    {
        return get_forum_base_url() . '/index.php?action=who';
    }

    /**
     * Get a URL to send a private/personal message to the given member.
     *
     * @param  MEMBER $member_id The member ID
     * @return URLPATH The URL to the private/personal message page
     */
    protected function _member_pm_url(int $member_id) : string
    {
        return get_forum_base_url() . '/index.php?action=pm&sa=send&u=' . strval($member_id);
    }

    /**
     * Get a URL to the specified forum.
     *
     * @param  integer $forum_id The forum ID
     * @return URLPATH The URL to the specified forum
     */
    protected function _forum_url(int $forum_id) : string
    {
        return get_forum_base_url() . '/index.php?board=' . strval($forum_id);
    }

    /**
     * Get the forum ID from a forum name. Numeric forum names are considered to be usable as IDs.
     *
     * @param  SHORT_TEXT $forum_name The forum name
     * @return ?integer The forum ID (null: not found)
     */
    public function forum_id_from_name(string $forum_name) : ?int
    {
        return is_numeric($forum_name) ? intval($forum_name) : $this->db->query_select_value_if_there('boards', 'id_board', ['name' => escape_html($forum_name)]);
    }

    /**
     * Get the topic ID from a topic identifier in the specified forum. It is used by comment topics, which means that the unique-topic-name assumption holds valid.
     *
     * @param  string $forum The forum name / ID
     * @param  SHORT_TEXT $topic_identifier The topic identifier
     * @return ?integer The topic ID (null: not found)
     */
    public function find_topic_id_for_topic_identifier(string $forum, string $topic_identifier) : ?int
    {
        if (is_integer($forum)) {
            $forum_id = intval($forum);
        } else {
            $forum_id = $this->forum_id_from_name($forum);
        }
        return $this->db->query_value_if_there('SELECT t.id_topic FROM ' . $this->db->get_table_prefix() . 'topics t LEFT JOIN ' . $this->db->get_table_prefix() . 'messages p ON t.id_first_msg=p.id_msg WHERE t.id_board=' . strval($forum_id) . ' AND (' . db_string_equal_to('subject', $topic_identifier) . ' OR subject LIKE \'' . db_encode_like('%: #' . $topic_identifier) . '\')');
    }

    /**
     * Makes a post in the specified forum, in the specified topic according to the given specifications. If the topic doesn't exist, it is created along with a spacer-post.
     * Spacer posts exist in order to allow staff to delete the first true post in a topic. Without spacers, this would not be possible with most forum systems. They also serve to provide meta information on the topic that cannot be encoded in the title (such as a link to the content being commented upon).
     *
     * @param  SHORT_TEXT $forum_name The forum name
     * @param  SHORT_TEXT $topic_identifier The topic identifier (usually <content-type>_<content-id>)
     * @param  MEMBER $member_id The member ID
     * @param  LONG_TEXT $post_title The post title
     * @param  LONG_TEXT $post The post content in Comcode format
     * @param  string $content_title The topic title; must be same as content title if this is for a comment topic
     * @param  string $topic_identifier_encapsulation_prefix This is put together with the topic identifier to make a more-human-readable topic title or topic description (hopefully the latter and a $content_title title, but only if the forum supports descriptions)
     * @param  ?URLPATH $content_url URL to the content (null: do not make spacer post)
     * @param  ?TIME $time The post time (null: use current time)
     * @param  ?IP $ip The post IP address (null: use current members IP address)
     * @param  ?BINARY $validated Whether the post is validated (null: unknown, find whether it needs to be marked not validated initially). This only works with the Conversr driver.
     * @param  ?BINARY $topic_validated Whether the topic is validated (null: unknown, find whether it needs to be marked not validated initially). This only works with the Conversr driver.
     * @param  boolean $skip_post_checks Whether to skip post checks
     * @param  SHORT_TEXT $poster_name_if_guest The name of the poster
     * @param  ?AUTO_LINK $parent_id ID of post being replied to (null: N/A)
     * @param  boolean $staff_only Whether the reply is only visible to staff
     * @return array Topic ID (may be null), and whether a hidden post has been made
     */
    public function make_post_forum_topic(string $forum_name, string $topic_identifier, int $member_id, string $post_title, string $post, string $content_title, string $topic_identifier_encapsulation_prefix, ?string $content_url = null, ?int $time = null, ?string $ip = null, ?int $validated = null, ?int $topic_validated = 1, bool $skip_post_checks = false, string $poster_name_if_guest = '', ?int $parent_id = null, bool $staff_only = false) : array
    {
        $post = str_replace("\n", '<br />', $post);

        if ($time === null) {
            $time = time();
        }
        if ($ip === null) {
            $ip = get_ip_address();
        }
        $forum_id = $this->forum_id_from_name($forum_name);
        if ($forum_id === null) {
            warn_exit(do_lang_tempcode('MISSING_FORUM', escape_html($forum_name)), false, true);
        }
        $username = $this->get_username($member_id);
        $topic_id = $this->find_topic_id_for_topic_identifier($forum_name, $topic_identifier);
        $is_new = ($topic_id === null);
        if ($is_new) {
            $topic_id = $this->db->query_insert('topics', ['id_board' => $forum_id, 'id_first_msg' => mt_rand(0, mt_getrandmax()), 'id_last_msg' => mt_rand(0, mt_getrandmax()), 'id_member_started' => $member_id, 'id_member_updated' => $member_id, 'num_replies' => 2], true);
            $home_link = '[url]' . $content_url . '[/url]';
            $post_id = $this->db->query_insert('messages', ['id_topic' => $topic_id, 'id_board' => $forum_id, 'poster_time' => $time, 'id_member' => $this->get_guest_id(), 'subject' => $content_title . ', ' . $topic_identifier_encapsulation_prefix . ': #' . $topic_identifier, 'poster_name' => do_lang('SYSTEM', '', '', '', get_site_default_lang()), 'poster_email' => get_option('staff_address'), 'poster_ip' => '127.0.0.1', 'modified_name' => '', 'body' => do_lang('SPACER_POST', $home_link, '', '', get_site_default_lang())], true);

            $this->db->query('UPDATE ' . $this->db->get_table_prefix() . 'boards SET num_posts=(num_posts+1), num_topics=(num_topics+1) WHERE id_board=' . strval($forum_id), 1);
            $this->db->query_update('topics', ['id_first_msg' => $post_id], ['id_topic' => $topic_id], '', 1);
        } else {
            $post_id = $this->db->query_select_value('messages', 'MIN(id_msg)', ['id_topic' => $topic_id]);
        }

        $GLOBALS['LAST_TOPIC_ID'] = $topic_id;
        $GLOBALS['LAST_TOPIC_IS_NEW'] = $is_new;

        if ($post == '') {
            return [$topic_id, false];
        }

        $email = $this->get_member_email_address($member_id);
        if ($email === null) {
            $email = '';
        }
        $username = $this->get_username($member_id);
        $post_id = $this->db->query_insert('messages', ['id_topic' => $topic_id, 'id_board' => $forum_id, 'poster_time' => $time, 'id_member' => $member_id, 'subject' => $post_title, 'poster_name' => $username, 'poster_email' => $email, 'poster_ip' => $ip, 'body' => $post, 'modified_name' => ''], true);
        $this->db->query('UPDATE ' . $this->db->get_table_prefix() . 'boards SET id_last_msg=' . strval($post_id) . ', num_posts=(num_posts+1) WHERE id_board=' . strval($forum_id), 1); // , last_updated='.strval($time).'
        $this->db->query('UPDATE ' . $this->db->get_table_prefix() . 'topics SET id_last_msg=' . strval($post_id) . ', num_replies=(num_replies+1), id_member_updated=' . strval($member_id) . ' WHERE id_topic=' . strval($topic_id), 1);

        return [$topic_id, false];
    }

    /**
     * Get an array of maps for the topic in the given forum.
     *
     * @param  integer $topic_id The topic ID
     * @param  ?integer $count The comment count will be returned here by reference (null: do not gather it)
     * @param  ?integer $max Maximum comments to returned (null: no limit)
     * @param  integer $start Comment to start at
     * @param  boolean $mark_read Whether to mark the topic read (ignored for this forum driver)
     * @param  boolean $reverse Whether to show in reverse
     * @return mixed The array of maps (Each map is: title, message, member, date) (-1 for no such forum, -2 for no such topic)
     */
    public function get_forum_topic_posts(int $topic_id, ?int &$count = null, ?int $max = 100, int $start = 0, bool $mark_read = true, bool $reverse = false)
    {
        if ($topic_id === null) {
            return (-2);
        }
        $order = $reverse ? 'poster_time DESC' : 'poster_time';
        $rows = $this->db->query('SELECT * FROM ' . $this->db->get_table_prefix() . 'messages WHERE id_topic=' . strval($topic_id) . ' AND body NOT LIKE \'' . db_encode_like(substr(do_lang('SPACER_POST', '', '', '', get_site_default_lang()), 0, 20) . '%') . '\' ORDER BY ' . $order, $max, $start);
        if ($count !== null) {
            $count = $this->db->query_value_if_there('SELECT COUNT(*) FROM ' . $this->db->get_table_prefix() . 'messages WHERE id_topic=' . strval($topic_id) . ' AND body NOT LIKE \'' . db_encode_like(substr(do_lang('SPACER_POST', '', '', '', get_site_default_lang()), 0, 20) . '%') . '\'');
        }
        $out = [];
        foreach ($rows as $myrow) {
            $temp = [];
            $temp['title'] = $myrow['subject'];
            if ($temp['title'] === null) {
                $temp['title'] = '';
            }
            push_lax_comcode(true);
            $temp['message'] = comcode_to_tempcode(str_replace('<br />', "\n", $myrow['body']), $myrow['id_member']);
            pop_lax_comcode();
            $temp['member'] = $myrow['id_member'];
            $temp['date'] = $myrow['poster_time'];

            $out[] = $temp;
        }

        return $out;
    }

    /**
     * Get a URL to the specified topic ID. Most forums don't require the second parameter, but some do, so it is required in the interface.
     *
     * @param  integer $topic_id The topic ID
     * @param  string $forum The forum ID
     * @return URLPATH The URL to the topic
     */
    public function topic_url(int $topic_id, string $forum) : string
    {
        return get_forum_base_url() . '/index.php?topic=' . strval($topic_id) . '.0';
    }

    /**
     * Get a URL to the specified post ID.
     *
     * @param  integer $post_id The post ID
     * @param  ?mixed $forum The forum ID (null: private topic)
     * @return URLPATH The URL to the post
     */
    public function post_url(int $post_id, $forum) : string
    {
        $topic_id = $this->db->query_select_value_if_there('messages', 'id_topic', ['id_msg' => $post_id]);
        if ($topic_id === null) {
            return '?';
        }
        $url = get_forum_base_url() . '/index.php?topic=' . strval($topic_id) . '.msg' . strval($post_id) . '#msg' . strval($post_id);
        return $url;
    }

    /**
     * Get an array of topics in the given forum. Each topic is an array with the following attributes:
     * - id, the topic ID
     * - title, the topic title
     * - lastusername, the username of the last poster
     * - lasttime, the timestamp of the last reply
     * - closed, a Boolean for whether the topic is currently closed or not
     * - firsttitle, the title of the first post
     * - firstpost, the first post (only set if $show_first_posts was true)
     *
     * @param  mixed $name The forum name or forum ID or an array of forum IDs (in such an array the KEYS are forum IDs, and the values ignored)
     * @param  integer $limit The limit
     * @param  integer $start The start position
     * @param  integer $max_rows The total rows (not a parameter: returns by reference)
     * @param  SHORT_TEXT $filter_topic_title The topic title filter (blank: no filter)
     * @param  SHORT_TEXT $filter_topic_description The topic description filter; may apply to the topic title if there is no separate description field with additional wildcarding to match what make_post_forum_topic is doing (blank: no filter)
     * @param  boolean $show_first_posts Whether to show the first posts
     * @param  string $date_key The date key to sort by
     * @set lasttime firsttime
     * @param  boolean $hot Whether to limit to hot topics
     * @param  boolean $only_open Open topics only
     * @return ?array The array of topics (null: error)
     */
    public function show_forum_topics($name, int $limit, int $start, int &$max_rows, string $filter_topic_title = '', string $filter_topic_description = '', bool $show_first_posts = false, string $date_key = 'lasttime', bool $hot = false, bool $only_open = false) : ?array
    {
        // Build forum ID query
        if (is_integer($name)) { // Forum ID
            $id_list = 't.id_board=' . strval($name);
        } elseif (!is_array($name)) { // Forum name
            $id = $this->forum_id_from_name($name);
            if ($id === null) {
                return null;
            }
            $id_list = 't.id_board=' . strval($id);
        } else { // Array of forum IDs
            $id_list = '';
            foreach (array_keys($name) as $id) {
                if ($id_list != '') {
                    $id_list .= ' OR ';
                }
                $id_list .= 't.id_board=' . strval($id);
            }
            if ($id_list == '') {
                return null;
            }
        }

        $topic_filter = '';
        if ($filter_topic_title != '') {
            $topic_filter .= ' AND p.subject LIKE \'' . db_encode_like($filter_topic_title . ', %') . '\'';
        }
        if ($filter_topic_description != '') {
            $topic_filter .= ' AND p.subject LIKE \'' . db_encode_like('%, ' . $filter_topic_description) . '\'';
        }
        if ($only_open) {
            $topic_filter .= ' AND t_locked=0';
        }

        $rows = $this->db->query('SELECT t.num_replies, t.id_topic AS t_id_topic, t.id_member_updated AS t_id_member_updated, t.id_member_started AS t_id_member_started, t.locked AS t_locked, p.subject AS p_subject FROM ' . $this->db->get_table_prefix() . 'topics t LEFT JOIN ' . $this->db->get_table_prefix() . 'messages p ON t.id_first_msg=p.id_msg WHERE (' . $id_list . ')' . $topic_filter . ' ORDER BY ' . (($date_key == 'lasttime') ? 'id_last_msg' : 'id_first_msg') . ' DESC', $limit, $start);
        $max_rows = $this->db->query_value_if_there('SELECT COUNT(*) FROM ' . $this->db->get_table_prefix() . 'topics t LEFT JOIN ' . $this->db->get_table_prefix() . 'messages p ON t.ID_FIRST_MSG=p.ID_MSG WHERE (' . $id_list . ')' . $topic_filter);

        // Generate output
        $out = [];
        foreach ($rows as $i => $r) {
            $out[$i] = [];
            $out[$i]['id'] = $r['t_id_topic'];
            $out[$i]['num'] = $r['num_replies'] + 1;
            $out[$i]['firstusername'] = $this->get_username($r['t_id_member_started']);
            $out[$i]['lastusername'] = $this->get_username($r['t_id_member_updated']);
            $out[$i]['firstmemberid'] = $r['t_id_member_started'];
            $out[$i]['lastmemberid'] = $r['t_id_member_updated'];
            $out[$i]['closed'] = ($r['t_locked'] == 1);
            $out[$i]['title'] = $r['p_subject'];
            $out[$i]['description'] = $r['p_subject'];

            // Get non-spacer post
            $fp_rows = $this->db->query('SELECT subject,poster_time,body,id_member,id_msg FROM ' . $this->db->get_table_prefix() . 'messages WHERE body NOT LIKE \'' . db_encode_like(substr(do_lang('SPACER_POST', '', '', '', get_site_default_lang()), 0, 20) . '%') . '\' AND id_topic=' . strval($out[$i]['id']) . ' ORDER BY id_msg ASC', 1);

            // Filter topics without a post
            if (!array_key_exists(0, $fp_rows)) {
                unset($out[$i]);
                continue;
            }

            // Fix title and time
            $out[$i]['firsttitle'] = $fp_rows[0]['subject'];
            $out[$i]['firsttime'] = $fp_rows[0]['poster_time'];

            // Get last post and adjust last time
            $fp_rows = $this->db->query('SELECT subject,poster_time,body,id_member,id_msg FROM ' . $this->db->get_table_prefix() . 'messages WHERE body NOT LIKE \'' . db_encode_like(substr(do_lang('SPACER_POST', '', '', '', get_site_default_lang()), 0, 20) . '%') . '\' AND id_topic=' . strval($out[$i]['id']) . ' ORDER BY id_msg DESC', 1);
            $out[$i]['lasttime'] = $fp_rows[0]['poster_time'];

            if ($show_first_posts) {
                push_lax_comcode(true);
                $out[$i]['firstpost'] = comcode_to_tempcode(str_replace('<br />', "\n", $fp_rows[0]['body']), $fp_rows[0]['id_member']);
                pop_lax_comcode();
            }
        }
        if (!empty($out)) {
            return $out;
        }
        return null;
    }

    /**
     * Get an array of members who are in at least one of the given array of usergroups.
     *
     * @param  array $groups The array of usergroups
     * @param  ?integer $max Return up to this many entries for primary members and this many entries for secondary members (null: no limit, only use no limit if querying very restricted usergroups!)
     * @param  integer $start Return primary members after this offset and secondary members after this offset
     * @return ?array The array of members (null: no members)
     */
    public function member_group_query(array $groups, ?int $max = null, int $start = 0) : ?array // Doesn't support multi usergroups currently. I don't think it's needed
    {
        $_groups = '';
        foreach ($groups as $group) {
            if ($_groups != '') {
                $_groups .= ' OR ';
            }
            $_groups .= 'id_group=' . strval($group);
        }
        return $this->db->query('SELECT * FROM ' . $this->db->get_table_prefix() . 'members WHERE ' . $_groups . ' ORDER BY id_group,id_member ASC', $max, $start);
    }

    /**
     * Get rows of members before the given one.
     * It cannot be assumed there are no gaps in member IDs, as members may be deleted.
     *
     * @param  MEMBER $member_id The member ID to paginate back from
     * @param  integer $total Number of members to retrieve
     * @return array Member rows
     */
    public function get_previous_members(int $member_id, int $total = 1) : array
    {
        $sql = 'SELECT id_member FROM ' . $this->db->get_table_prefix() . 'members WHERE id_member<>0';
        $sql .= ' AND id_member<' . strval($member_id);
        $sql .= ' ORDER BY id_member DESC';
        $rows = $this->db->query($sql, $total);
        return $rows;
    }

    /**
     * Get rows of members after the given one.
     * It cannot be assumed there are no gaps in member IDs, as members may be deleted.
     *
     * @param  ?MEMBER $member_id The member ID to increment (null: find the very first members)
     * @param  integer $total Number of members to retrieve
     * @return array Member rows
     */
    public function get_next_members(?int $member_id, int $total = 1) : array
    {
        $sql = 'SELECT id_member FROM ' . $this->db->get_table_prefix() . 'members WHERE id_member<>0';
        if ($member_id !== null) {
            $sql .= ' AND id_member>' . strval($member_id);
        }
        $sql .= ' ORDER BY id_member DESC';
        $rows = $this->db->query($sql, $total);
        return $rows;
    }

    /**
     * Try to find a member with the given IP address.
     *
     * @param  IP $ip The IP address
     * @return array The distinct rows found
     */
    public function probe_ip(string $ip) : array
    {
        $a = $this->db->query_select('members', ['DISTINCT id_member AS id'], ['member_ip' => $ip]);
        $b = $this->db->query_select('messages', ['DISTINCT id_member AS id'], ['poster_ip' => $ip]);
        return array_merge($a, $b);
    }

    /**
     * Get the name relating to the specified member ID.
     * If this returns null, then the member has been deleted. Always take potential null output into account.
     *
     * @param  MEMBER $member_id The member ID
     * @return ?SHORT_TEXT The username (null: member deleted)
     */
    protected function _get_username(int $member_id) : ?string
    {
        if ($member_id == $this->get_guest_id()) {
            return do_lang('GUEST');
        }
        return $this->get_member_row_field($member_id, 'real_name');
    }

    /**
     * Get the display name of a username.
     * If no display name generator is configured, this will be the same as the username.
     *
     * @param  ID_TEXT $username The username
     * @return SHORT_TEXT The display name
     */
    public function get_displayname(string $username) : string
    {
        return $this->db->query_select_value_if_there('members', 'real_name', ['member_name' => $username]);
    }

    /**
     * Get the e-mail address for the specified member ID.
     *
     * @param  MEMBER $member_id The member ID
     * @return SHORT_TEXT The e-mail address
     */
    protected function _get_member_email_address(int $member_id) : string
    {
        return $this->get_member_row_field($member_id, 'email_address');
    }

    /**
     * Find if this member may have e-mails sent to them.
     *
     * @param  MEMBER $member_id The member ID
     * @return boolean Whether the member may have e-mails sent to them
     */
    public function get_member_email_allowed(int $member_id) : bool
    {
        $r = $this->get_member_row($member_id);
        $v = array_key_exists('im_email_notify', $r) ? $r['im_email_notify'] : $r['pm_email_notify'];
        if ($v == 1) {
            return true;
        }
        return false;
    }

    /**
     * Get the timestamp of a member's join date.
     *
     * @param  MEMBER $member_id The member ID
     * @return TIME The timestamp
     */
    public function get_member_join_timestamp(int $member_id) : int
    {
        return $this->get_member_row_field($member_id, 'date_registered');
    }

    /**
     * Find all members with a name matching the given SQL LIKE string.
     *
     * @param  string $pattern The pattern
     * @param  ?integer $limit Maximum number to return (limits to the most recent active) (null: no limit)
     * @return ?array The array of matched members (null: none found)
     */
    public function get_matching_members(string $pattern, ?int $limit = null) : ?array
    {
        $rows = $this->db->query('SELECT * FROM ' . $this->db->get_table_prefix() . 'members WHERE real_name LIKE \'' . db_encode_like($pattern) . '\'' . ' ORDER BY last_login DESC', $limit);
        sort_maps_by($rows, 'real_name');
        return $rows;
    }

    /**
     * Get the given member's post count.
     *
     * @param  MEMBER $member_id The member ID
     * @return integer The post count
     */
    public function get_post_count(int $member_id) : int
    {
        $c = $this->get_member_row_field($member_id, 'posts');
        if ($c === null) {
            $c = 0;
        }
        return $c;
    }

    /**
     * Get the given member's topic count.
     *
     * @param  MEMBER $member_id The member ID
     * @return integer The topic count
     */
    public function get_topic_count(int $member_id) : int
    {
        return $this->db->query_select_value('topics', 'COUNT(*)', ['id_member_started' => $member_id]);
    }

    /**
     * Find out if the given member ID is banned.
     *
     * @param  MEMBER $member_id The member ID
     * @param  ?ID_TEXT $reasoned_ban Ban reasoning returned by reference (null: none)
     * @return boolean Whether the member is banned
     */
    public function is_banned(int $member_id, ?string &$reasoned_ban = null) : bool
    {
        $rows = $this->db->query('SELECT id_ban FROM ' . $this->db->get_table_prefix() . 'banned WHERE id_member=' . strval($member_id) . ' AND expire_time IS NULL OR expire_time>' . strval(time()), null, 0, true); // Suppress errors in case of other table name
        if ($rows === null) {
            $rows = $this->db->query('SELECT id_ban FROM ' . $this->db->get_table_prefix() . 'ban_items i LEFT JOIN ' . $this->db->get_table_prefix() . 'ban_groups g on i.id_ban_group=g.id_ban_group WHERE id_member=' . strval($member_id) . ' AND expire_time IS NULL OR expire_time>' . strval(time()));
        }
        return !empty($rows);
    }

    /**
     * Get an SMF setting.
     *
     * @param  ID_TEXT $setting The name of the setting
     * @param  boolean $allow_missing Whether to allow a missing setting without throwing an error
     * @return ?string The setting value (null: Setting does not exist and $allow_setting is true)
     */
    public function get_setting(string $setting, bool $allow_missing = false) : ?string
    {
        static $SETTINGS_CACHE = [];

        if (isset($SETTINGS_CACHE[$setting])) {
            return $SETTINGS_CACHE[$setting];
        }
        if ($allow_missing) {
            $SETTINGS_CACHE[$setting] = $this->db->query_select_value_if_there('settings', 'value', ['variable' => $setting]);
        } else {
            $SETTINGS_CACHE[$setting] = $this->db->query_select_value('settings', 'value', ['variable' => $setting]);
        }
        return $SETTINGS_CACHE[$setting];
    }

    /**
     * Find the base URL to the emoticons.
     *
     * @return URLPATH The base URL
     */
    public function get_emo_dir() : string
    {
        $url = $this->get_setting('smileys_url');

        // If the setting is not specified, we have to bail on error since get_emo_dir cannot return null.
        if ($url === null) {
            warn_exit(do_lang_tempcode('INTERNAL_ERROR', escape_html('0999ee4a2de05d188be3d080ad912e4f')));
        }

        return $url . '/';
    }

    /**
     * Get a map between emoticon codes and templates representing the HTML-image-code for this emoticon. The emoticons presented of course depend on the forum involved.
     *
     * @return array The map
     */
    public function find_emoticons() : array
    {
        if ($this->EMOTICON_CACHE !== null) {
            return $this->EMOTICON_CACHE;
        }
        $rows = $this->db->query('SELECT * FROM ' . $this->db->get_table_prefix() . 'smiley_files sf JOIN ' . $this->db->get_table_prefix() . 'smileys s ON s.id_smiley=sf.id_smiley ORDER BY sf.smiley_set');
        $this->EMOTICON_CACHE = [];
        foreach ($rows as $myrow) {
            $set = $myrow['smiley_set'];
            $src = $myrow['filename'];
            if (url_is_local($src)) {
                $src = $this->get_emo_dir() . $set . '/' . $src;
            }
            $this->EMOTICON_CACHE[$myrow['code']] = ['EMOTICON_IMG_CODE_DIR', $src, $myrow['code']];
        }
        return $this->EMOTICON_CACHE;
    }

    /**
     * Find a list of all forum skins (aka themes).
     *
     * @return array The list of skins
     */
    public function get_skin_list() : array
    {
        $rows = $this->db->query_select('themes', ['value'], ['variable' => 'name']);
        return collapse_1d_complexity('value', $rows);
    }

    /**
     * Try to find the theme that the logged-in/guest member is using, and map it to a site theme.
     * The themes/map.ini file functions to provide this mapping between forum themes, and site themes, and has a slightly different meaning for different forum drivers. For example, some drivers map the forum themes theme directory to the site theme name, while others made the humanly readable name.
     *
     * @param  boolean $skip_member_specific Whether to avoid member-specific lookup (i.e. find via what forum theme is currently configured as the default)
     * @param  ?MEMBER $member_id The member to find for (null: current member)
     * @return ID_TEXT The theme
     */
    public function _get_theme(bool $skip_member_specific = false, ?int $member_id = null) : string
    {
        $def = '';

        // Load in remapper
        require_code('files');
        $map = file_exists(get_file_base() . '/themes/map.ini') ? cms_parse_ini_file_fast(get_file_base() . '/themes/map.ini') : [];

        if (!$skip_member_specific) {
            // Work out
            if ($member_id === null) {
                $member_id = get_member();
            }
            if ($member_id > 0) {
                $skin = $this->get_member_row_field($member_id, 'id_theme');
            } else {
                $skin = 0;
            }
            if ($skin > 0) { // User has a custom theme
                $obb = $this->db->query_select_value('themes', 'value', ['variable' => 'name', 'id_theme' => $skin]);
                $def = array_key_exists($obb, $map) ? $map[$obb] : $obb;
            }
        }

        // Look for a skin according to our site name (we bother with this instead of 'default' because the software itself likes to never choose a theme when forum-theme integration is on: all forum [via map] or all site seems cleaner, although it is complex)
        if ((!(strlen($def) > 0)) || (!file_exists(get_custom_file_base() . '/themes/' . $def))) {
            $obb = $this->db->query_select_value_if_there('themes', 'value', ['variable' => 'name', 'value' => get_site_name()]);
            if ($obb !== null) {
                $def = array_key_exists($obb, $map) ? $map[$obb] : $obb;
            }
        }

        // Default then!
        if ((!(strlen($def) > 0)) || (!file_exists(get_custom_file_base() . '/themes/' . $def))) {
            $def = array_key_exists('default', $map) ? $map['default'] : 'default';
        }

        return $def;
    }

    /**
     * Find if the specified member ID is marked as staff or not.
     *
     * @param  MEMBER $member_id The member ID
     * @return boolean Whether the member is staff
     */
    protected function _is_staff(int $member_id) : bool
    {
        $usergroups = $this->get_members_groups($member_id);
        return ((in_array(1, $usergroups)) || (in_array(2, $usergroups)));
    }

    /**
     * Find if the specified member ID is marked as a super admin or not.
     *
     * @param  MEMBER $member_id The member ID
     * @return boolean Whether the member is a super admin
     */
    protected function _is_super_admin(int $member_id) : bool
    {
        $usergroups = $this->get_members_groups($member_id);
        return (in_array(1, $usergroups));
    }

    /**
     * Get the number of members currently online on the forums.
     *
     * @return integer The number of members
     */
    public function get_num_users_forums() : int
    {
        return $this->db->query_value_if_there('SELECT COUNT(DISTINCT session_id) FROM ' . $this->db->get_table_prefix() . 'sessions WHERE last_update>' . strval(time() - 60 * intval(get_option('users_online_time'))));
    }

    /**
     * Get the number of members registered on the forum.
     *
     * @return integer The number of members
     */
    public function get_num_members() : int
    {
        return $this->db->query_select_value('members', 'COUNT(*)');
    }

    /**
     * Get the total topics ever made on the forum.
     *
     * @return integer The number of topics
     */
    public function get_num_topics() : int
    {
        return $this->db->query_select_value('topics', 'COUNT(*)');
    }

    /**
     * Get the total posts ever made on the forum.
     *
     * @return integer The number of posts
     */
    public function get_num_forum_posts() : int
    {
        return $this->db->query_select_value('messages', 'COUNT(*)');
    }

    /**
     * Get the number of new forum posts.
     *
     * @return integer The number of posts
     */
    protected function _get_num_new_forum_posts() : int
    {
        return $this->db->query_value_if_there('SELECT COUNT(*) FROM ' . $this->db->get_table_prefix() . 'messages WHERE poster_time>' . strval(time() - 60 * 60 * 24));
    }

    /**
     * Get a member ID from the given member's username.
     *
     * @param  SHORT_TEXT $username The username
     * @return ?MEMBER The member ID (null: not found)
     */
    public function get_member_from_username(string $username) : ?int
    {
        if ($username == do_lang('GUEST')) {
            return $this->get_guest_id();
        }

        return $this->db->query_select_value_if_there('members', 'id_member', ['real_name' => $username]);
    }

    /**
     * Get a member ID from the given member's e-mail address.
     *
     * @param  SHORT_TEXT $email_address The member e-mail address
     * @return ?MEMBER The member ID (null: not found)
     */
    public function get_member_from_email_address(string $email_address) : ?int
    {
        $results = $this->db->query_select('members', ['id_member', 'date_registered'], ['email_address' => $email_address], 'ORDER BY date_registered DESC', 1);
        return array_key_exists(0, $results) ? $results[0]['id_member'] : null;
    }

    /**
     * Get the IDs of the admin usergroups.
     *
     * @return array The admin usergroup ids
     */
    protected function _get_super_admin_groups() : array
    {
        return [1];
    }

    /**
     * Get the IDs of the moderator usergroups.
     * It should not be assumed that a member only has one usergroup - this depends upon the forum the driver works for. It also does not take the staff site filter into account.
     *
     * @return array The moderator usergroup ids
     */
    protected function _get_moderator_groups() : array
    {
        return [2];
    }

    /**
     * Get the forum usergroup list.
     *
     * @return array The usergroup list
     */
    protected function _get_usergroup_list() : array
    {
        return [0 => do_lang('GUESTS')] + collapse_2d_complexity('id_group', 'group_name', $this->db->query_select('membergroups', ['id_group', 'group_name']));
    }

    /**
     * Get the forum usergroup relating to the specified member ID.
     *
     * @param  MEMBER $member_id The member ID
     * @return array The array of forum usergroups
     */
    protected function _get_members_groups(int $member_id) : array
    {
        if ($member_id == $this->get_guest_id()) {
            return [0];
        }

        $additional = $this->get_member_row_field($member_id, 'additional_groups');
        if ($additional != '') {
            $usergroups = explode(',', $additional);
        } else {
            $usergroups = [];
        }
        $primary_group = $this->get_member_row_field($member_id, 'id_group');
        if (!empty($primary_group)) {
            $usergroups[] = $primary_group;
        }
        $usergroups[] = $this->get_member_row_field($member_id, 'id_post_group');
        return $usergroups;
    }

    /**
     * Create a member login cookie for SMF 2.1.
     *
     * @param  MEMBER $member_id The member ID
     * @param  ?SHORT_TEXT $username The username (null: lookup)
     * @param  string $password_raw The password (note this is sometimes not used by forum drivers, as they can also do something with what is already in the database instead)
     */
    public function create_login_cookie(int $member_id, ?string $username, string $password_raw)
    {
        return;

        // TODO: #5258 Not implemented. Currently, making this work for SMF would make for very fragile code. Consider SSI or REST instead.

        /*
        list($stub,) = explode(':', get_member_cookie());

        // Create our cookie
        $row = $this->get_member_row($member_id);
        $_password = $this->cookie_hash_salt($row['passwd'], $row['password_salt']);
        $expiry_time = (time() + intval(get_cookie_days()) * 24 * 60 * 60); // Actually we must use what SMF uses for cookie time.
        $local_cookie_domain = '';
        $global_cookie_domain = '';
        $use_local_cookies = $this->get_setting('localCookies', true);
        $use_global_cookies = $this->get_setting('globalCookies', true);

        // if use_local_cookies then we need to get the local cookie domain. Same for global cookies.

        // Note: SMF also encodes third-party plugin data into their cookie. This is not possible to do for Composr.

        $data = json_encode([$member_id, $_password, $expiry_time, $local_cookie_domain, $global_cookie_domain]);

        cms_setcookie($stub, $data, false, true);
        */
    }

    /**
     * Try and log in using a member cookie.
     * Should only be called if the cookie exists.
     *
     * @return ?array A map of 'id' and 'error'. If 'id' is null, an error occurred and 'error' is set (null: no cookie)
     */
    public function authorise_cookie_login() : ?array
    {
        if ((!isset($_COOKIE[get_member_cookie()])) || (!isset($_COOKIE[get_pass_cookie()]))) {
            return null;
        }

        $username = $_COOKIE[get_member_cookie()];
        $password_hashed = $_COOKIE[get_pass_cookie()];

        return $this->_authorise_login($username, null, $password_hashed, true);
    }

    /**
     * Find if the given member ID and password is valid. If username is null, then the member ID is used instead.
     *
     * @param  ?SHORT_TEXT $username The member username (null: use $member_id)
     * @param  ?MEMBER $member_id The member ID (null: use $username)
     * @param  string $password_raw The raw password
     * @return array A map of 'id' and 'error'. If 'id' is null, an error occurred and 'error' is set
     */
    public function authorise_login(?string $username, ?int $member_id, string $password_raw) : array
    {
        return $this->_authorise_login($username, $member_id, $password_raw, false);
    }

    /**
     * Find if the given member ID and password is valid. If username is null, then the member ID is used instead.
     * All authorisation, cookies, and form-logins, are passed through this function.
     *
     * @param  ?SHORT_TEXT $username The member username (null: use $member_id)
     * @param  ?MEMBER $member_id The member ID (null: use $username)
     * @param  string $password_mixed If $cookie_login is true then this is the value of the password cookie, otherwise it's the password the user tried to log in with
     * @param  boolean $cookie_login Whether this is a cookie login, determines how the hashed password is treated for the value passed in
     * @return array A map of 'id' and 'error'. If 'id' is null, an error occurred and 'error' is set
     */
    protected function _authorise_login(?string $username, ?int $member_id, string $password_mixed, bool $cookie_login = false) : array
    {
        $out = [];
        $out['id'] = null;

        if ($member_id === null) {
            $rows = $this->db->query_select('members', ['*'], ['member_name' => $username], '', 1);
            if (array_key_exists(0, $rows)) {
                $this->MEMBER_ROWS_CACHED[$rows[0]['id_member']] = $rows[0];
            } else {
                $rows = $this->db->query_select('members', ['*'], ['real_name' => $username], '', 1);
                if (array_key_exists(0, $rows)) {
                    $this->MEMBER_ROWS_CACHED[$rows[0]['id_member']] = $rows[0];
                }
            }
        } else {
            $rows = [];
            $rows[0] = $this->get_member_row($member_id);
        }

        if (!array_key_exists(0, $rows) || $rows[0] === null) { // All hands to lifeboats
            $out['error'] = do_lang_tempcode((get_option('login_error_secrecy') == '1') ? 'MEMBER_INVALID_LOGIN' : '_MEMBER_NO_EXIST', $username);
            return $out;
        }
        $row = $rows[0];
        if ($this->is_banned($row['id_member'])) { // All hands to the guns
            $out['error'] = do_lang_tempcode('YOU_ARE_BANNED');
            return $out;
        }

        // Verify password below
        if (substr($row['passwd'], 0, 4) == '$2y$') { // SMF 2 password
            $GLOBALS['SMF_PW_STYLE'] = 2.0;
            if ($cookie_login) {
                // TODO: Cookie login probably does not work #5258
                $cookie_hash = $this->cookie_hash_salt($row['passwd'], $row['password_salt']);
                $passes = hash_equals($cookie_hash, $password_mixed);
            } else {
                $passes = password_verify(cms_strtolower_ascii($username) . $password_mixed, $row['passwd']);
            }
        } else { // SMF 1 and 1.1 password
            $GLOBALS['SMF_PW_STYLE'] = array_key_exists('pm_ignore_list', $row) ? 1.1 : 1.0;
            $password_hashed = $cookie_login ? $password_mixed : $this->_legacy_password_hash($password_mixed, $row['passwd']);
            $bits = explode('::', $password_hashed);
            if (!array_key_exists(1, $bits)) {
                $bits[1] = $bits[0];
            }
            if ($cookie_login) {
                if ($GLOBALS['SMF_PW_STYLE'] == 1.1) {
                    $passes = hash_equals(sha1($row['passwd'] . $row['password_salt']), $bits[1]);
                } else {
                    $passes = hash_equals($this->_legacy_password_hash($row['passwd'], 'ys', true), $bits[0]);
                }
            } else {
                if ($GLOBALS['SMF_PW_STYLE'] == 1.1) {
                    $passes = hash_equals($row['passwd'], $bits[1]);
                } else {
                    $passes = hash_equals($row['passwd'], $bits[0]);
                }
            }
        }

        if (!$passes) {
            $out['error'] = do_lang_tempcode((get_option('login_error_secrecy') == '1') ? 'MEMBER_INVALID_LOGIN' : 'MEMBER_BAD_PASSWORD');
            return $out;
        }

        if (substr(get_member_cookie(), 0, 5) != 'cms__') {
            require_code('users_active_actions');
            cms_eatcookie('PHPSESSID');
        }

        $out['id'] = $row['id_member'];

        return $out;
    }

    /**
     * LEGACY: SMF hashing for SMF 1 and 1.1.
     *
     * @param  string $password_raw The value to hash
     * @param  string $key Special key to hash with
     * @param  boolean $smf1_only Whether to just use original SMF hashing
     * @return string The hashed data
     * @ignore
     */
    protected function _legacy_password_hash(string $password_raw, string $key, bool $smf1_only = false) : string
    {
        $key = cms_strtolower_ascii($key);
        $new_key = str_pad((strlen($key) <= 64) ? $key : pack('H*', md5($key)), 64, chr(0x00));

        $a = md5(($new_key ^ str_repeat(chr(0x5c), 64)) . pack('H*', md5(($new_key ^ str_repeat(chr(0x36), 64)) . $password_raw))); // SMF 1.0 style
        if ($smf1_only) {
            return $a;
        }
        $b = sha1($key . $password_raw); // SMF 1.1 style

        return $a . '::' . $b;
    }

    /**
     * Hashes password with salt and authentication secret for use in the SMF 2 cookie.
     *
     * @param  string $password The password
     * @param  string $salt The salt
     * @return string The hashed password
     */
    protected function cookie_hash_salt(string $password, string $salt) : string
    {
        // Append the salt to get a user-specific authentication secret.
        $secret_key = get_value('smf_auth_secret') . $salt;

        // Now use that to generate an HMAC of the password.
        return hash_hmac('sha512', $password, $secret_key);
    }

    /**
     * Get a first known IP address of the given member.
     *
     * @param  MEMBER $member_id The member ID
     * @return IP The IP address
     */
    public function get_member_ip(int $member_id) : string
    {
        return $this->get_member_row_field($member_id, 'member_ip');
    }

    /**
     * Gets a whole member row from the database.
     *
     * @param  MEMBER $member_id The member ID
     * @return ?array The member row (null: no such member)
     */
    public function get_member_row(int $member_id) : ?array
    {
        if (array_key_exists($member_id, $this->MEMBER_ROWS_CACHED)) {
            return $this->MEMBER_ROWS_CACHED[$member_id];
        }

        $rows = $this->db->query_select('members', ['*'], ['id_member' => $member_id], '', 1);
        if ($member_id == $this->get_guest_id()) {
            $rows[0]['member_name'] = do_lang('GUEST');
            $rows[0]['real_name'] = do_lang('GUEST');
            $rows[0]['email_address'] = null;
            $rows[0]['im_email_notify'] = 0;
            $rows[0]['pm_email_notify'] = 0;
            $rows[0]['date_registered'] = time();
            $rows[0]['posts'] = 0;
            $rows[0]['avatar'] = '';
            $rows[0]['id_theme'] = null;
            $rows[0]['id_group'] = 0;
            $rows[0]['additional_groups'] = '';
            $rows[0]['lngfile'] = null;
        }
        if (!array_key_exists(0, $rows)) {
            return null;
        }
        $this->MEMBER_ROWS_CACHED[$member_id] = $rows[0];
        return $this->MEMBER_ROWS_CACHED[$member_id];
    }

    /**
     * Gets a named field of a member row from the database.
     *
     * @param  MEMBER $member_id The member ID
     * @param  string $field The field identifier
     * @return mixed The field
     */
    public function get_member_row_field(int $member_id, string $field)
    {
        $row = $this->get_member_row($member_id);
        return ($row === null) ? null : $row[$field];
    }

    /**
     * Add a member to a usergroup.
     *
     * @param  MEMBER $member_id The member ID
     * @param  GROUP $group_id The usergroup ID
     */
    public function add_member_to_group(int $member_id, int $group_id)
    {
        $additional = $this->get_member_row_field($member_id, 'additional_groups');
        if ($additional != '') {
            $usergroups = explode(',', $additional);
        } else {
            $usergroups = [];
        }

        $usergroups[] = strval($group_id);
        $this->db->query_update('members', ['additional_groups' => implode(',', $usergroups)], ['id_member' => $member_id]);
    }

    /**
     * Remove a member from a usergroup.
     *
     * @param  MEMBER $member_id The member ID
     * @param  GROUP $group_id The usergroup ID
     */
    public function remove_member_from_group(int $member_id, int $group_id)
    {
        $additional = $this->get_member_row_field($member_id, 'additional_groups');
        if ($additional != '') {
            $usergroups = explode(',', $additional);
        } else {
            $usergroups = [];
        }
        $usergroups = array_diff($usergroups, [strval($group_id)]);
        $this->db->query_update('members', ['additional_groups' => implode(',', $usergroups)], ['id_member' => $member_id]);
    }

    /**
     * Returns the SQL query to use in an ORDER BY when sorting by a member (actually, username, but we only know the ID).
     *
     * @param  ID_TEXT $field The name of the r.* field containing the member ID
     * @param  boolean $use_displayname Whether we want the display name instead of the username
     * @return ?string The SQL query (null: this driver does not support sorting by member)
     */
    public function get_username_sort_query(string $field, bool $use_displayname = false) : ?string
    {
        // No support for display name
        return '(SELECT member_name FROM ' . $this->db->table_prefix . 'members m WHERE m.id_member=r.' . $field . ')';
    }
}
