<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    core_cns
 */

/**
 * Hook class.
 */
class Hook_sw_core_cns
{
    /**
     * Run function for features in the setup wizard.
     *
     * @return array Current settings
     */
    public function get_current_settings() : array
    {
        $settings = [];

        if (get_forum_type() != 'cns') {
            return $settings;
        }

        if (!is_on_multi_site_network()) {
            require_code('cns_groups2');
            $ladder_groups = get_default_rank_ladder_groups();
            switch (count($ladder_groups)) {
                case 5:
                    $settings['rank_set'] = 'fun';
                    break;
                case 3:
                    $settings['rank_set'] = 'simple';
                    break;
                case 1:
                default:
                    $settings['rank_set'] = 'none';
                    break;
            }

            $settings['have_default_full_emoticon_set'] = has_predefined_content('core_cns', 'have_default_full_emoticon_set') ? '1' : '0';

            $have_default_cpf_set = false;
            if (has_predefined_content('core_cns', 'about')) {
                $have_default_cpf_set = true;
            }
            if (has_predefined_content('core_cns', 'interests')) {
                $have_default_cpf_set = true;
            }
            if (has_predefined_content('core_cns', 'occupation')) {
                $have_default_cpf_set = true;
            }
            if (has_predefined_content('core_cns', 'staff_notes')) {
                $have_default_cpf_set = true;
            }
            $settings['have_default_cpf_set'] = $have_default_cpf_set ? '1' : '0';
        }

        return $settings;
    }

    /**
     * Run function for features in the setup wizard.
     *
     * @param  array $field_defaults Default values for the fields, from the install-profile
     * @return array A pair: Input fields, Hidden fields
     */
    public function get_fields(array $field_defaults) : array
    {
        if (!addon_installed('cns_forum') || get_forum_type() != 'cns' || post_param_integer('addon_cns_forum', null) === 0) {
            return [new Tempcode(), new Tempcode()];
        }

        $current_settings = $this->get_current_settings();
        $field_defaults += $current_settings; // $field_defaults will take precedence, due to how "+" operator works in PHP

        require_lang('cns');
        $fields = new Tempcode();

        if (!is_on_multi_site_network()) {
            require_lang('cns_ranks');
            $rank_set_options = new Tempcode();
            foreach (['none', 'simple', 'fun'] as $rank_set) {
                $rank_set_options->attach(form_input_list_entry($rank_set, $rank_set == $field_defaults['rank_set'], do_lang_tempcode('RANK_SET_' . $rank_set)));
            }
            $fields->attach(form_input_list(do_lang_tempcode('RANK_SET'), do_lang_tempcode('DESCRIPTION_RANK_SET'), 'rank_set', $rank_set_options));

            require_code('cns_general_action2');
            $fields->attach(form_input_tick(do_lang_tempcode('HAVE_DEFAULT_FULL_EMOTICON_SET'), do_lang_tempcode('DESCRIPTION_HAVE_DEFAULT_FULL_EMOTICON_SET', escape_html(integer_format(cns_get_num_emoticons_on_disk()))), 'have_default_full_emoticon_set', $field_defaults['have_default_full_emoticon_set'] == '1'));

            $fields->attach(form_input_tick(do_lang_tempcode('HAVE_DEFAULT_CPF_SET'), do_lang_tempcode('DESCRIPTION_HAVE_DEFAULT_CPF_SET'), 'have_default_cpf_set', $field_defaults['have_default_cpf_set'] == '1'));
        }

        return [$fields, new Tempcode()];
    }

    /**
     * Run function for setting features from the setup wizard.
     */
    public function set_fields()
    {
        if (!addon_installed('cns_forum') || get_forum_type() != 'cns' || post_param_integer('addon_cns_forum', null) === 0) {
            return;
        }

        if (!is_on_multi_site_network()) {
            $rank_set = post_param_string('rank_set', null);
            if ($rank_set !== null) {
                require_code('cns_groups_action');
                require_lang('cns');
                cns_make_rank_set($rank_set);
            }

            install_predefined_content('core_cns', [
               'have_default_full_emoticon_set' => (post_param_integer('have_default_full_emoticon_set', 0) == 1),
               'about' => (post_param_integer('have_default_cpf_set', 0) == 1),
               'interests' => (post_param_integer('have_default_cpf_set', 0) == 1),
               'occupation' => (post_param_integer('have_default_cpf_set', 0) == 1),
               'staff_notes' => (post_param_integer('have_default_cpf_set', 0) == 1),
            ]);
        }
    }

    /**
     * Run function for blocks in the setup wizard.
     *
     * @return array A map between block names and pairs (BLOCK_POSITION_* constants for what is supported, then a BLOCK_POSITION_* constant for what is the default)
     */
    public function get_blocks() : array
    {
        if (get_forum_type() == 'cns') {
            return ['side_cns_private_topics' => [BLOCK_POSITION_PANEL, null]];
        }
        return [];
    }
}
