<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    stats
 */

/**
 * Hook class.
 */
class Hook_admin_stats_content extends CMSStatsProvider
{
    /**
     * Find metadata about stats graphs that are provided by this stats hook.
     *
     * @param  boolean $for_kpi Whether this is for setting up a KPI
     * @return ?array Map of metadata (null: hook is disabled)
     */
    public function info(bool $for_kpi = false) : ?array
    {
        return [
            'content_views' => [
                'label' => do_lang_tempcode('CONTENT_VIEWS'),
                'category' => 'hits',
                'filters' => [
                    'content_views__content_type' => new CMSStatsListFilter('content_views__content_type', do_lang_tempcode('CONTENT_TYPE'), $this->find_all_content_types()),
                ],
                'pivot' => null,
            ],
            'content_views_per_content_day' => [
                'label' => do_lang_tempcode('CONTENT_VIEWS_PER_CONTENT_DAY'),
                'category' => 'hits',
                'filters' => [
                    'content_views__content_type' => new CMSStatsListFilter('content_views__content_type', do_lang_tempcode('CONTENT_TYPE'), $this->find_all_content_types()),
                ],
                'pivot' => null,
            ],
        ];
    }

    /**
     * Preprocess raw data in the database into something we can efficiently draw graphs/conclusions from.
     * This is for flat and timeless data.
     *
     * @param  TIME $start_time Start timestamp
     * @param  TIME $end_time End timestamp
     * @param  array $data_buckets Map of data buckets; a map of bucket name to nested maps
     */
    public function preprocess_raw_data_flat(int $start_time, int $end_time, array &$data_buckets)
    {
        $server_timezone = get_server_timezone();

        $date_pivots = $this->get_date_pivots();

        $limit_per_content_type = 100;

        $data_buckets['content_views'] = [];
        $data_buckets['content_views_per_content_day'] = [];

        require_code('content');
        $cma_hooks = find_all_hook_obs('systems', 'content_meta_aware', 'Hook_content_meta_aware_');
        $content_types = [];
        foreach ($cma_hooks as $content_type => $hook_ob) {
            $info = $hook_ob->info();
            if (($info !== null) && ($info['views_field'] !== null) && ($info['id_field'] !== null) && ($info['title_field'] !== null)) {
                $db = $info['db'];
                $table = $info['table'];
                $views_field = $info['views_field'];
                $add_time_field = $info['add_time_field'];

                $select = [];
                append_content_select_for_fields($select, $info, ['title', 'id', 'views', 'add_time', 'category']);
                $rows = $db->query_select($table, $select, [], 'ORDER BY ' . $views_field . ' DESC', $limit_per_content_type);
                foreach ($rows as $row) {
                    $title = $hook_ob->get_title($row);
                    $id = $hook_ob->get_id_string($row);
                    $views = $row[$views_field];

                    $data_buckets['content_views'][$content_type][$title . ' (' . $hook_ob->get_id_string($row) . ')'] = $views;

                    if ($add_time_field !== null) {
                        $add_time = $row[$add_time_field];
                        if ($add_time !== null) {
                            $days = floatval(time() - $add_time + 1/*prevent divide by zero errors*/) / floatval(60 * 60 * 24);
                            $data_buckets['content_views_per_content_day'][$content_type][$title . ' (' . $id . ')'] = floatval($views) / $days;
                        }
                    }
                }
            }
        }
    }

    /**
     * Generate final data from preprocessed data.
     *
     * @param  string $bucket Data bucket we want data for
     * @param  string $pivot Pivot value
     * @param  array $filters Map of filters (including pivot if applicable)
     * @return ?array Final data in standardised map format (null: could not generate)
     */
    public function generate_final_data(string $bucket, string $pivot, array $filters) : ?array
    {
        $data = [];

        $content_types = $this->find_all_content_types();

        $where = [
            'p_bucket' => $bucket,
        ];
        $_data = $GLOBALS['SITE_DB']->query_select_value_if_there('stats_preprocessed_flat', 'p_data', $where);
        if ($_data !== null) {
            $__data = @unserialize($_data);
            if ($__data !== false) {
                foreach ($__data as $content_type => $_) {
                    if ((!empty($filters[$bucket . '__content_type'])) && ($filters[$bucket . '__content_type'] != $content_type)) {
                        continue;
                    }

                    $prefix = array_key_exists($content_type, $content_types) ? $content_types[$content_type] : $content_type;

                    foreach ($_ as $title => $total) {
                        $data[$prefix . ': ' . $title] = $total;
                    }
                }
            }
        }

        switch ($bucket) {
            case 'content_views':
                $y_label = do_lang_tempcode('VIEWS');
                break;

            case 'content_views_per_content_day':
                $y_label = do_lang_tempcode('CONTENT_VIEWS_PER_CONTENT_DAY');
                break;

            default:
                $y_label = new Tempcode();
                fatal_exit(do_lang_tempcode('INTERNAL_ERROR', escape_html('cf65ff8b3d3b5bb7b30f64a747fd796b')));
        }

        return [
            'type' => self::GRAPH_BAR_CHART,
            'data' => $data,
            'x_axis_label' => do_lang_tempcode('TIME_IN_TIMEZONE', escape_html(make_nice_timezone_name(get_site_timezone()))),
            'y_axis_label' => $y_label,
            'limit_bars' => true,
        ];
    }
}
