<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    stats
 */

/**
 * Hook class.
 */
class Hook_admin_stats_emails extends CMSStatsProvider
{
    /**
     * Find metadata about stats graphs that are provided by this stats hook.
     *
     * @param  boolean $for_kpi Whether this is for setting up a KPI
     * @return ?array Map of metadata (null: hook is disabled)
     */
    public function info(bool $for_kpi = false) : ?array
    {
        require_lang('mail');
        require_lang('email_log');

        return [
            'emails_sent' => [
                'label' => do_lang_tempcode('EMAIL_LOG'),
                'category' => 'server_performance',
                'filters' => [
                    'emails_sent__month_range' => new CMSStatsDateMonthRangeFilter('emails_sent__month_range', do_lang_tempcode('DATE_RANGE'), null, $for_kpi),
                ],
                'pivot' => new CMSStatsDatePivot('emails_sent__pivot', $this->get_date_pivots(!$for_kpi)),
            ],
            'unsubscribed_emails' => [
                'label' => do_lang_tempcode('UNSUBSCRIBED_EMAILS'),
                'category' => 'conversions',
                'filters' => [
                    'unsubscribed_emails__month_range' => new CMSStatsDateMonthRangeFilter('unsubscribed_emails__month_range', do_lang_tempcode('DATE_RANGE'), null, $for_kpi),
                ],
                'pivot' => new CMSStatsDatePivot('unsubscribed_emails__pivot', $this->get_date_pivots(!$for_kpi)),
            ],
        ];
    }

    /**
     * Preprocess raw data in the database into something we can efficiently draw graphs/conclusions from.
     *
     * @param  TIME $start_time Start timestamp
     * @param  TIME $end_time End timestamp
     * @param  array $data_buckets Map of data buckets; a map of bucket name to nested maps with the following maps in sequence: 'month', 'pivot', 'value' (then further map data) ; extended and returned by reference
     */
    public function preprocess_raw_data(int $start_time, int $end_time, array &$data_buckets)
    {
        require_code('temporal');

        $server_timezone = get_server_timezone();
        $date_pivots = $this->get_date_pivots();

        /* e-mails sent */

        $max = 1000;
        $start = 0;

        $query = 'SELECT m_date_and_time FROM ' . get_table_prefix() . 'logged_mail_messages WHERE ';
        $query .= 'm_date_and_time>=' . strval($start_time) . ' AND ';
        $query .= 'm_date_and_time<=' . strval($end_time);
        $query .= ' ORDER BY m_date_and_time';
        do {
            $rows = $GLOBALS['SITE_DB']->query($query, $max, $start);
            foreach ($rows as $row) {
                $timestamp = $row['m_date_and_time'];
                $timestamp = tz_time($timestamp, $server_timezone);

                $month = to_epoch_interval_index($timestamp, 'months');

                foreach (array_keys($date_pivots) as $pivot) {
                    $pivot_value = $this->calculate_date_pivot_value($pivot, $timestamp);

                    if (!isset($data_buckets['emails_sent'][$month][$pivot][$pivot_value])) {
                        $data_buckets['emails_sent'][$month][$pivot][$pivot_value] = 0;
                    }
                    $data_buckets['emails_sent'][$month][$pivot][$pivot_value]++;
                }
            }

            $start += $max;
        } while (!empty($rows));

        /* unsubscribed e-mails */

        $max = 1000;
        $start = 0;

        $query = 'SELECT b_time FROM ' . get_table_prefix() . 'unsubscribed_emails WHERE ';
        $query .= 'b_time>=' . strval($start_time) . ' AND ';
        $query .= 'b_time<=' . strval($end_time);
        $query .= ' ORDER BY b_time';
        do {
            $rows = $GLOBALS['SITE_DB']->query($query, $max, $start);
            foreach ($rows as $row) {
                $timestamp = $row['b_time'];
                $timestamp = tz_time($timestamp, $server_timezone);

                $month = to_epoch_interval_index($timestamp, 'months');

                foreach (array_keys($date_pivots) as $pivot) {
                    $pivot_value = $this->calculate_date_pivot_value($pivot, $timestamp);

                    if (!isset($data_buckets['unsubscribed_emails'][$month][$pivot][$pivot_value])) {
                        $data_buckets['unsubscribed_emails'][$month][$pivot][$pivot_value] = 0;
                    }
                    $data_buckets['unsubscribed_emails'][$month][$pivot][$pivot_value]++;
                }
            }

            $start += $max;
        } while (!empty($rows));
    }

    /**
     * Generate final data from preprocessed data.
     *
     * @param  string $bucket Data bucket we want data for
     * @param  string $pivot Pivot value
     * @param  array $filters Map of filters (including pivot if applicable)
     * @return ?array Final data in standardised map format (null: could not generate)
     */
    public function generate_final_data(string $bucket, string $pivot, array $filters) : ?array
    {
        $range = $this->convert_month_range_filter_to_pair($filters[$bucket . '__month_range']);

        $data = $this->fill_data_by_date_pivots($pivot, $range[0], $range[1]);

        $where = [
            'p_bucket' => $bucket,
            'p_pivot' => $pivot,
        ];
        $extra = '';
        $extra .= ' AND p_month>=' . strval($range[0]);
        $extra .= ' AND p_month<=' . strval($range[1]);
        $data_rows = $GLOBALS['SITE_DB']->query_select('stats_preprocessed', ['p_data'], $where, $extra);
        foreach ($data_rows as $data_row) {
            $_data = @unserialize($data_row['p_data']);
            foreach ($_data as $pivot_value => $num_emails_sent) {
                $pivot_value = $this->make_date_pivot_value_nice($pivot, $pivot_value);

                $data[$pivot_value] = $num_emails_sent;
            }
        }

        switch ($bucket) {
            case 'emails_sent':
                return [
                    'type' => null,
                    'data' => $data,
                    'x_axis_label' => do_lang_tempcode('TIME_IN_TIMEZONE', escape_html(make_nice_timezone_name(get_site_timezone()))),
                    'y_axis_label' => do_lang_tempcode('EMAILS_SENT'),
                ];

            case 'unsubscribed_emails':
                return [
                    'type' => null,
                    'data' => $data,
                    'x_axis_label' => do_lang_tempcode('TIME_IN_TIMEZONE', escape_html(make_nice_timezone_name(get_site_timezone()))),
                    'y_axis_label' => do_lang_tempcode('UNSUBSCRIBED_EMAILS'),
                ];
        }
    }
}
