<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    cns_clubs
 */

/**
 * Hook class.
 */
class Hook_search_cns_clubs extends FieldsSearchHook
{
    /**
     * Find details for this search hook.
     *
     * @param  boolean $check_permissions Whether to check permissions
     * @param  ?MEMBER $member_id The member ID to check with (null: current member)
     * @return ~?array Map of search hook details (null: hook is disabled) (false: access denied)
     */
    public function info(bool $check_permissions = true, ?int $member_id = null)
    {
        if (!addon_installed('cns_clubs')) {
            return null;
        }

        if (get_forum_type() != 'cns') {
            return null;
        }

        if ($member_id === null) {
            $member_id = get_member();
        }

        if ($check_permissions) {
            if (!has_actual_page_access($member_id, 'groups')) {
                return false;
            }
        }

        if ($GLOBALS['FORUM_DB']->query_select_value('f_groups', 'COUNT(*)', ['g_is_private_club' => 1]) == 0) {
            return null;
        }

        require_lang('cns');

        $info = [];
        $info['lang'] = do_lang_tempcode('CLUBS');
        $info['default'] = (get_option('search_cns_clubs') == '1');
        $info['extra_sort_fields'] = $this->_get_extra_sort_fields('_group');

        $info['permissions'] = [
            [
                'type' => 'zone',
                'zone_name' => get_module_zone('groups'),
            ],
            [
                'type' => 'page',
                'zone_name' => get_module_zone('groups'),
                'page_name' => 'groups',
            ],
        ];

        return $info;
    }

    /**
     * Get a list of extra fields to ask for.
     *
     * @return ?array A list of maps specifying extra fields (null: no tree)
     */
    public function get_fields() : ?array
    {
        return $this->_get_fields('_group');
    }

    /**
     * Run function for search results.
     *
     * @param  string $search_query Search query
     * @param  string $content_where WHERE clause that selects the content according to the search query; passed in addition to $search_query to avoid unnecessary reparsing.  ? refers to the yet-unknown field name (blank: full-text search)
     * @param  string $where_clause Initial WHERE clause that already takes $search_under into account (should be nothing else unless it is guaranteed hook will use the global get_search_rows function)
     * @param  string $search_under Comma-separated list of categories to search under
     * @param  boolean $only_search_meta Whether to only do a META (tags) search
     * @param  boolean $only_titles Whether only to search titles (as opposed to both titles and content)
     * @param  integer $max Start position in total results
     * @param  integer $start Maximum results to return in total
     * @param  string $sort The sort type (gets remapped to a field in this function)
     * @param  ID_TEXT $direction Order direction
     * @param  SHORT_TEXT $author Username/Author to match for
     * @param  ?MEMBER $author_id Member-ID to match for (null: unknown)
     * @param  mixed $cutoff Cutoff date (TIME or a pair representing the range or null)
     * @return array List of maps (template, orderer)
     */
    public function run(string $search_query, string $content_where, string $where_clause, string $search_under, bool $only_search_meta, bool $only_titles, int $max, int $start, string $sort, string $direction, string $author, ?int $author_id, $cutoff) : array
    {
        $remapped_orderer = '';
        switch ($sort) {
            case 'title':
                $remapped_orderer = 'g_name';
                break;
        }

        require_lang('cns');

        // Calculate our where clause (search)
        $sq = build_search_submitter_clauses('g_group_lead_member', $author_id, $author);
        if ($sq === null) {
            return [];
        } else {
            $where_clause .= $sq;
        }

        $where_clause .= ' AND ';
        $where_clause .= 'g_hidden=0 AND g_is_private_club=1';

        $table = 'f_groups r';
        $trans_fields = ['r.g_name' => 'SHORT_TRANS', 'r.g_title' => 'SHORT_TRANS'];
        $nontrans_fields = [];
        $this->_get_search_parameterisation_advanced_for_content_type('_group', $table, $where_clause, $trans_fields, $nontrans_fields);

        // Calculate and perform query
        $rows = get_search_rows(null, 'id', $search_query, $content_where, $where_clause, $only_search_meta, $only_titles, $max, $start, $remapped_orderer, $direction, $table, 'r.*', $trans_fields, $nontrans_fields);

        $out = [];
        foreach ($rows as $i => $row) {
            $out[$i]['data'] = $row;
            unset($rows[$i]);
            if (($remapped_orderer != '') && (array_key_exists($remapped_orderer, $row))) {
                $out[$i]['orderer'] = $row[$remapped_orderer];
            }
        }

        return $out;
    }

    /**
     * Run function for rendering a search result.
     *
     * @param  array $row The data row stored when we retrieved the result
     * @return Tempcode The output
     */
    public function render(array $row) : object
    {
        $leader = $GLOBALS['FORUM_DRIVER']->member_profile_hyperlink($row['g_group_lead_member']);

        require_code('cns_groups');
        $group_name = cns_get_group_name($row['id']);

        require_code('cns_groups2');
        $num_members = cns_get_group_members_raw_count($row['id'], false, false, true, false);

        $title = do_lang('CONTENT_IS_OF_TYPE', do_lang('CLUB'), $group_name);

        $summary = do_lang_tempcode(($row['g_open_membership'] == 1) ? 'CLUB_WITH_MEMBERS_OPEN' : 'CLUB_WITH_MEMBERS_APPROVAL', escape_html($group_name), escape_html(integer_format($num_members, 0)), $leader);

        $url = build_url(['page' => 'groups', 'type' => 'view', 'id' => $row['id']], get_module_zone('groups'));

        return do_template('SIMPLE_PREVIEW_BOX', [
            '_GUID' => '2f7814a2e1f868d2ac73fba69f3aeee1',
            'ID' => strval($row['id']),
            'TITLE' => $title,
            'SUMMARY' => $summary,
            'URL' => $url,
            'RESOURCE_TYPE' => 'group',
        ]);
    }
}
