<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    cns_forum
 */

/**
 * Hook class.
 */
class Hook_search_cns_posts extends FieldsSearchHook
{
    /**
     * Find details for this search hook.
     *
     * @param  boolean $check_permissions Whether to check permissions
     * @param  ?MEMBER $member_id The member ID to check with (null: current member)
     * @return ~?array Map of search hook details (null: hook is disabled) (false: access denied)
     */
    public function info(bool $check_permissions = true, ?int $member_id = null)
    {
        if (!addon_installed('cns_forum')) {
            return null;
        }

        if (get_forum_type() != 'cns') {
            return null;
        }

        if ($member_id === null) {
            $member_id = get_member();
        }

        if ($check_permissions) {
            if (!has_actual_page_access($member_id, 'topicview')) {
                return false;
            }
        }

        if ($GLOBALS['FORUM_DRIVER']->get_num_forum_posts() == 0) {
            return null;
        }

        require_lang('cns');

        $info = [];
        $info['lang'] = do_lang_tempcode('FORUM_POSTS');
        $info['default'] = (get_option('search_cns_posts') == '1');
        $info['special_on'] = [];
        $info['special_off'] = ['open' => do_lang_tempcode('POST_SEARCH_OPEN'), 'closed' => do_lang_tempcode('POST_SEARCH_CLOSED'), 'pinned' => do_lang_tempcode('POST_SEARCH_PINNED')];
        if ((has_privilege($member_id, 'see_not_validated')) && (addon_installed('validation'))) {
            $info['special_off']['not_validated'] = do_lang_tempcode('POST_SEARCH_NOT_VALIDATED');
        }
        if (can_use_fast_custom_index('cns_posts')) {
            $info['special_on']['starter'] = do_lang_tempcode('POST_SEARCH_STARTER');
        } else {
            $info['special_off']['starter'] = do_lang_tempcode('POST_SEARCH_STARTER');
        }
        $info['category'] = 'p_cache_forum_id';
        $info['integer_category'] = true;
        $info['extra_sort_fields'] = $this->_get_extra_sort_fields('_post');
        $info['user_label'] = do_lang_tempcode('USERNAME');

        $info['permissions'] = [
            [
                'type' => 'zone',
                'zone_name' => get_module_zone('topicview'),
            ],
            [
                'type' => 'page',
                'zone_name' => get_module_zone('topicview'),
                'page_name' => 'topicview',
            ],
        ];

        // Change search_under_children to search_under
        $search_under_children = get_param_string('search_under_children', null);
        if (is_numeric($search_under_children)) {
            require_code('cns_forums');
            require_code('cns_forums2');
            list(, $search_under) = cns_get_forum_tree(null, intval($search_under_children), '', null, null, true, 1);
            //$_GET['search_under'] = $search_under . (($search_under == '') ? '' : ',') . $search_under_children;
            $_GET['search_under'] = $search_under_children . (($search_under == '') ? '' : (',' . $search_under));
        }

        return $info;
    }

    /**
     * Empty the fast custom index.
     */
    public function empty_index()
    {
        $GLOBALS['FORUM_DB']->query_delete('f_posts_fulltext_index');
    }

    /**
     * Perform indexing using the fast custom index.
     *
     * @param  ?TIME $since Only index records newer than this (null: no limit)
     * @param  ?integer $total_singular_ngram_tokens Write into a count of singular ngrams (typically, words) in here (null: do not count)
     * @param  ?array $statistics_map Write into this map of singular ngram (typically, words) to number of occurrences (null: do not maintain a map)
     */
    public function index_for_search(?int $since = null, ?int &$total_singular_ngram_tokens = null, ?array &$statistics_map = null)
    {
        $engine = new Fast_custom_index();

        $index_table = 'f_posts_fulltext_index';
        $clean_scan = ($GLOBALS['FORUM_DB']->query_select_value_if_there($index_table, 'i_ngram') === null);

        $has_custom_fields = ($GLOBALS['FORUM_DB']->query_select_value_if_there('catalogue_fields', 'id', ['c_name' => '_post']) !== null);

        $fields_to_index = [
            'p_title' => APPEARANCE_CONTEXT_TITLE,
            'p_post' => APPEARANCE_CONTEXT_BODY,
        ];
        $key_transfer_map = [
            'id' => 'i_post_id',
        ];
        $filter_field_transfer_map = [
            'p_time' => 'i_add_time',
            'p_cache_forum_id' => 'i_forum_id',
            'p_posting_member' => 'i_posting_member',
            't_is_open' => 'i_open',
            't_pinned' => 'i_pinned',
            'i_starter' => 'i_starter',
        ];

        $db = $GLOBALS['FORUM_DB'];

        // A way to force-resume where we left off, if we're debugging our way through
        if (get_value('fast_custom_index__startup_hack', '0') == '1') {
            $last_post_id = $db->query_select_value_if_there('f_posts_fulltext_index', 'MAX(i_post_id)');
            if ($last_post_id !== null) {
                $_since = $db->query_select_value_if_there('f_posts', 'p_time', ['id' => $last_post_id]);
                if ($_since !== null) {
                    $since = $_since;
                }
            }
        }


        $lang_fields = find_lang_fields('f_posts', 'p');

        $sql = 'SELECT p.id,p.p_time,p.p_last_edit_time,p.p_posting_member,p.p_title,p.p_post,p.p_cache_forum_id,t_is_open,t_pinned,t_cache_first_post_id FROM ' . $db->get_table_prefix() . 'f_posts p JOIN ' . $db->get_table_prefix() . 'f_topics t ON p.p_topic_id=t.id';
        $sql .= ' WHERE p_cache_forum_id IS NOT NULL';
        $since_clause = $engine->generate_since_where_clause($db, $index_table, ['p_time' => false, 'p_last_edit_time' => true], $since, $statistics_map);
        $sql .= $since_clause;

        $max_post_length = intval(get_option('fast_custom_index__max_post_length'));
        if ($max_post_length > 0) {
            $sql .= ' AND ' . db_function('LENGTH', [$GLOBALS['FORUM_DB']->translate_field_ref('p_post')]) . '<' . strval($max_post_length);
        }

        $max = 100;
        $start_id = -1;
        do {
            $rows = $db->query($sql . ' AND p.id>' . strval($start_id) . ' ORDER BY p.id', $max, 0, false, false, $lang_fields);
            foreach ($rows as $row) {
                $content_fields = $row + ['i_starter' => ($row['t_cache_first_post_id'] == $row['id']) ? 1 : 0];

                if ($has_custom_fields) {
                    $ce_id = $GLOBALS['SITE_DB']->query_select_value_if_there('catalogue_entry_linkage', 'catalogue_entry_id', ['content_type' => 'post', 'content_id' => strval($row['id'])]);
                    if ($ce_id !== null) {
                        $engine->get_content_fields_from_catalogue_entry($content_fields, $fields_to_index, '_post', $ce_id);
                    }
                }

                $engine->index_for_search($db, $index_table, $content_fields, $fields_to_index, $key_transfer_map, $filter_field_transfer_map, $total_singular_ngram_tokens, $statistics_map, null, $clean_scan);

                $start_id = $row['id'];
            }
        } while (!empty($rows));
    }

    /**
     * Get details for an ajax-tree-list of entries for the content covered by this search hook.
     *
     * @return array A pair: the hook, and the options
     */
    public function ajax_tree() : array
    {
        return ['choose_forum', ['compound_list' => true]];
    }

    /**
     * Get a list of extra fields to ask for.
     *
     * @return ?array A list of maps specifying extra fields (null: no tree)
     */
    public function get_fields() : ?array
    {
        return $this->_get_fields('_post');
    }

    /**
     * Run function for search results.
     *
     * @param  string $search_query Search query
     * @param  string $content_where WHERE clause that selects the content according to the search query; passed in addition to $search_query to avoid unnecessary reparsing.  ? refers to the yet-unknown field name (blank: full-text search)
     * @param  string $where_clause Initial WHERE clause that already takes $search_under into account (should be nothing else unless it is guaranteed hook will use the global get_search_rows function)
     * @param  string $search_under Comma-separated list of categories to search under
     * @param  boolean $only_search_meta Whether to only do a META (tags) search
     * @param  boolean $only_titles Whether only to search titles (as opposed to both titles and content)
     * @param  integer $max Start position in total results
     * @param  integer $start Maximum results to return in total
     * @param  string $sort The sort type (gets remapped to a field in this function)
     * @param  ID_TEXT $direction Order direction
     * @param  SHORT_TEXT $author Username/Author to match for
     * @param  ?MEMBER $author_id Member-ID to match for (null: unknown)
     * @param  mixed $cutoff Cutoff date (TIME or a pair representing the range or null)
     * @return array List of maps (template, orderer)
     */
    public function run(string $search_query, string $content_where, string $where_clause, string $search_under, bool $only_search_meta, bool $only_titles, int $max, int $start, string $sort, string $direction, string $author, ?int $author_id, $cutoff) : array
    {
        if (in_array($search_query, [
            do_lang('POSTS_WITHIN_TOPIC'),
            do_lang('SEARCH_POSTS_WITHIN_TOPIC'),
            do_lang('SEARCH_FORUM_POSTS'),
            do_lang('_SEARCH_PRIVATE_TOPICS'),
        ])) {
            return []; // Search placeholder label, not real search
        }

        if (get_forum_type() != 'cns') {
            return [];
        }
        require_code('cns_forums');
        require_code('cns_posts');
        require_css('cns');

        $remapped_orderer = '';
        switch ($sort) {
            case 'title':
                $remapped_orderer = 'p_title';
                break;

            case 'add_date':
                $remapped_orderer = 'p_time';
                break;
        }

        require_lang('cns');

        // Calculate and perform query
        $permissions_module = 'forums';
        $db = $GLOBALS['FORUM_DB'];
        $index_table = 'f_posts_fulltext_index';
        if (can_use_fast_custom_index('cns_posts', $db, $index_table, $search_query, Fast_custom_index::active_search_has_special_filtering() || $cutoff !== null || $author != '' || ($search_under != '-1' && $search_under != '!') || get_param_integer('option_tick_cns_posts_starter', 0) == 1)) {
            // This search hook implements the fast custom index, which we use where possible...

            $table = 'f_posts r';

            // Calculate our where clause (search)
            $where_clause = '';
            $extra_join_clause = '';
            $sq = build_search_submitter_clauses('ixxx.i_posting_member', $author_id, $author);
            $use_simple_index = !$only_titles && !$only_search_meta; // MySQL may pick an index that can't do sorting, requiring a lot of disk I/O, so we will guide it to one that can if we know it can; if there is extra filtering likely to narrow down results a lot to make sorting less of a perf hit then we let MySQL choose
            if ($sq === null) {
                return [];
            } elseif ($sq != '') {
                $extra_join_clause .= $sq;
                $use_simple_index = false;
            }
            $this->_handle_date_check($cutoff, 'ixxx.i_add_time', $extra_join_clause);
            if (get_param_integer('option_tick_cns_posts_not_validated', 0) == 1) {
                $where_clause .= ' AND ';
                $where_clause .= 'r.p_validated=0';
            }
            if (get_param_integer('option_tick_cns_posts_open', 0) == 1) {
                $extra_join_clause .= ' AND ';
                $extra_join_clause .= 'ixxx.i_open=1';
                $use_simple_index = false;
            }
            if (get_param_integer('option_tick_cns_posts_closed', 0) == 1) {
                $extra_join_clause .= ' AND ';
                $extra_join_clause .= 'ixxx.i_open=0';
                $use_simple_index = false;
            }
            if (get_param_integer('option_tick_cns_posts_pinned', 0) == 1) {
                $extra_join_clause .= ' AND ';
                $extra_join_clause .= 'ixxx.i_pinned=1';
                $use_simple_index = false;
            }
            if (get_param_integer('option_tick_cns_posts_starter', 0) == 1) {
                $extra_join_clause .= ' AND ';
                $extra_join_clause .= 'ixxx.i_starter=1';
                $use_simple_index = false;
            }

            // Category filter
            if (($search_under != '!') && ($search_under != '-1')) {
                $cats = explode(',', $search_under);
                $_cats = [];
                foreach ($cats as $i => $cat) {
                    if (trim($cat) == '') {
                        continue;
                    }

                    $_cats[] = strval(intval($cat));
                }
                if (!empty($_cats)) {
                    $extra_join_clause .= ' AND ixxx.i_forum_id IN (' . implode(',', $_cats) . ')';
                }

                if (count($_cats) <= 1) {
                    $use_simple_index = false;
                }
            }

            $where_clause .= ' AND ';
            $where_clause .= '(p_whisper_to_member IS NULL';
            if (!is_guest()) {
                $where_clause .= ' OR p_whisper_to_member=' . strval(get_member()) . ' OR p_posting_member=' . strval(get_member());
            }
            $where_clause .= ')';
            if ((!has_privilege(get_member(), 'see_not_validated')) && (addon_installed('validation'))) {
                $where_clause .= ' AND ';
                $where_clause .= 'p_validated=1';
            }

            $engine = new Fast_custom_index();

            if (Fast_custom_index::active_search_has_special_filtering()) {
                $trans_fields = [];
                $nontrans_fields = [];
                $this->_get_search_parameterisation_advanced_for_content_type('_post', $table, $where_clause, $trans_fields, $nontrans_fields);
                // ^ Nothing done with trans_fields and nontrans_fields
            }

            $key_transfer_map = ['id' => 'i_post_id'];
            $index_permissions_field = 'i_forum_id';
            $rows = $engine->get_search_rows($db, $index_table, $db->get_table_prefix() . $table, $key_transfer_map, $where_clause, $extra_join_clause, $search_query, $only_search_meta, $only_titles, $max, $start, $remapped_orderer, $direction, $permissions_module, $index_permissions_field, false, $use_simple_index ? 'main_19' : null);
        } else {
            // Calculate our where clause (search)
            $sq = build_search_submitter_clauses('p_posting_member', $author_id, $author);
            if ($sq === null) {
                return [];
            } else {
                $where_clause .= $sq;
            }
            $this->_handle_date_check($cutoff, 'p_time', $where_clause);
            if (get_param_integer('option_tick_cns_posts_not_validated', 0) == 1) {
                $where_clause .= ' AND ';
                $where_clause .= 'r.p_validated=0';
            }
            if (get_param_integer('option_tick_cns_posts_open', 0) == 1) {
                $where_clause .= ' AND ';
                $where_clause .= 's.t_is_open=1';
            }
            if (get_param_integer('option_tick_cns_posts_closed', 0) == 1) {
                $where_clause .= ' AND ';
                $where_clause .= 's.t_is_open=0';
            }
            if (get_param_integer('option_tick_cns_posts_pinned', 0) == 1) {
                $where_clause .= ' AND ';
                $where_clause .= 's.t_pinned=1';
            }
            if (get_param_integer('option_tick_cns_posts_starter', 0) == 1) {
                $where_clause .= ' AND ';
                $where_clause .= 's.t_cache_first_post_id=r.id';
            }
            $where_clause .= ' AND ';
            $where_clause .= 'r.p_cache_forum_id IS NOT NULL AND (r.p_whisper_to_member IS NULL';
            if (!is_guest()) {
                $where_clause .= ' OR (r.p_whisper_to_member=' . strval(get_member()) . ' OR r.p_posting_member=' . strval(get_member()) . ')';
            }
            $where_clause .= ')';

            if ((!has_privilege(get_member(), 'see_not_validated')) && (addon_installed('validation'))) {
                $where_clause .= ' AND ';
                $where_clause .= 'r.p_validated=1';
            }

            $table = 'f_posts r JOIN ' . $GLOBALS['FORUM_DB']->get_table_prefix() . 'f_topics s ON r.p_topic_id=s.id';
            $trans_fields = ['!' => '!', 'r.p_post' => 'LONG_TRANS__COMCODE'];
            $nontrans_fields = ['r.p_title'/*,'s.t_description' Performance problem due to how full text works*/];
            $this->_get_search_parameterisation_advanced_for_content_type('_post', $table, $where_clause, $trans_fields, $nontrans_fields);

            $rows = get_search_rows(null, 'id', $search_query, $content_where, $where_clause, $only_search_meta, $only_titles, $max, $start, $remapped_orderer, $direction, $table, 'r.*,t_forum_id,t_cache_first_title', $trans_fields, $nontrans_fields, $permissions_module, 's.t_forum_id');
        }

        $out = [];
        foreach ($rows as $i => $row) {
            $out[$i]['data'] = $row;
            unset($rows[$i]);
            if (($remapped_orderer != '') && (array_key_exists($remapped_orderer, $row))) {
                $out[$i]['orderer'] = $row[$remapped_orderer];
            } elseif (strpos($remapped_orderer, '_rating:') !== false) {
                $out[$i]['orderer'] = $row[$remapped_orderer];
            }
        }

        return $out;
    }

    /**
     * Run function for rendering a search result.
     *
     * @param  array $row The data row stored when we retrieved the result
     * @return Tempcode The output
     */
    public function render(array $row) : object
    {
        global $SEARCH_QUERY_TERMS;
        $highlight_bits = ($SEARCH_QUERY_TERMS === null) ? [] : $SEARCH_QUERY_TERMS;
        push_lax_comcode(true);
        $summary = get_translated_text($row['p_post']);
        $text_summary_h = comcode_to_tempcode($summary, null, false, null, null, COMCODE_NORMAL, $highlight_bits);
        pop_lax_comcode();
        $text_summary = generate_text_summary($text_summary_h->evaluate(), $highlight_bits);

        require_code('cns_posts2');
        return render_post_box($row, false, true, true, null, '', protect_from_escaping($text_summary));
    }
}
