<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    downloads
 */

/**
 * Hook class.
 */
class Hook_search_downloads extends FieldsSearchHook
{
    /**
     * Find details for this search hook.
     *
     * @param  boolean $check_permissions Whether to check permissions
     * @param  ?MEMBER $member_id The member ID to check with (null: current member)
     * @return ~?array Map of search hook details (null: hook is disabled) (false: access denied)
     */
    public function info(bool $check_permissions = true, ?int $member_id = null)
    {
        if (!addon_installed('downloads')) {
            return null;
        }

        if ($member_id === null) {
            $member_id = get_member();
        }

        if ($check_permissions) {
            if (!has_actual_page_access($member_id, 'downloads')) {
                return false;
            }
        }

        if ($GLOBALS['SITE_DB']->query_select_value('download_downloads', 'COUNT(*)') == 0) {
            return null;
        }

        require_lang('downloads');

        $info = [];
        $info['lang'] = do_lang_tempcode('SECTION_DOWNLOADS');
        $info['default'] = (get_option('search_downloads') == '1');
        $info['category'] = 'category_id';
        $info['integer_category'] = true;
        $info['extra_sort_fields'] = ['file_size' => do_lang_tempcode('FILE_SIZE')] + $this->_get_extra_sort_fields('_download');

        $info['permissions'] = [
            [
                'type' => 'zone',
                'zone_name' => get_module_zone('downloads'),
            ],
            [
                'type' => 'page',
                'zone_name' => get_module_zone('downloads'),
                'page_name' => 'downloads',
            ],
        ];

        return $info;
    }

    /**
     * Get details for an ajax-tree-list of entries for the content covered by this search hook.
     *
     * @return array A pair: the hook, and the options
     */
    public function ajax_tree() : array
    {
        return ['choose_download_category', ['compound_list' => true]];
    }

    /**
     * Get a list of extra fields to ask for.
     *
     * @return ?array A list of maps specifying extra fields (null: no tree)
     */
    public function get_fields() : ?array
    {
        return $this->_get_fields('_download');
    }

    /**
     * Run function for search results.
     *
     * @param  string $search_query Search query
     * @param  string $content_where WHERE clause that selects the content according to the search query; passed in addition to $search_query to avoid unnecessary reparsing.  ? refers to the yet-unknown field name (blank: full-text search)
     * @param  string $where_clause Initial WHERE clause that already takes $search_under into account (should be nothing else unless it is guaranteed hook will use the global get_search_rows function)
     * @param  string $search_under Comma-separated list of categories to search under
     * @param  boolean $only_search_meta Whether to only do a META (tags) search
     * @param  boolean $only_titles Whether only to search titles (as opposed to both titles and content)
     * @param  integer $max Start position in total results
     * @param  integer $start Maximum results to return in total
     * @param  string $sort The sort type (gets remapped to a field in this function)
     * @param  ID_TEXT $direction Order direction
     * @param  SHORT_TEXT $author Username/Author to match for
     * @param  ?MEMBER $author_id Member-ID to match for (null: unknown)
     * @param  mixed $cutoff Cutoff date (TIME or a pair representing the range or null)
     * @return array List of maps (template, orderer)
     */
    public function run(string $search_query, string $content_where, string $where_clause, string $search_under, bool $only_search_meta, bool $only_titles, int $max, int $start, string $sort, string $direction, string $author, ?int $author_id, $cutoff) : array
    {
        $remapped_orderer = '';
        switch ($sort) {
            case 'average_rating':
            case 'compound_rating':
                $remapped_orderer = $sort . ':downloads:id';
                break;

            case 'title':
                $remapped_orderer = 'name';
                break;

            case 'add_date':
                $remapped_orderer = 'add_date';
                break;

            case 'file_size':
                $remapped_orderer = $sort;
                break;
        }

        require_code('downloads');
        require_lang('downloads');
        require_css('downloads');

        // Calculate our where clause (search)
        $sq = build_search_submitter_clauses('submitter', $author_id, $author, 'author');
        if ($sq === null) {
            return [];
        } else {
            $where_clause .= $sq;
        }
        $this->_handle_date_check($cutoff, 'add_date', $where_clause);

        if ((!has_privilege(get_member(), 'see_not_validated')) && (addon_installed('validation'))) {
            $where_clause .= ' AND ';
            $where_clause .= 'validated=1';
        }

        $privacy_join = '';
        if (addon_installed('content_privacy')) {
            require_code('content_privacy');
            list($privacy_join, $privacy_where) = get_privacy_where_clause('download', 'r');
            $where_clause .= $privacy_where;
        }

        $table = 'download_downloads r';
        $trans_fields = ['r.name' => 'SHORT_TRANS', 'r.the_description' => 'LONG_TRANS__COMCODE', 'r.additional_details' => 'LONG_TRANS__COMCODE'];
        $nontrans_fields = ['r.original_filename', 'r.download_data_mash'];
        $this->_get_search_parameterisation_advanced_for_content_type('_download', $table, $where_clause, $trans_fields, $nontrans_fields);

        // Calculate and perform query
        $rows = get_search_rows('downloads_download', 'id', $search_query, $content_where, $where_clause, $only_search_meta, $only_titles, $max, $start, $remapped_orderer, $direction, $table . $privacy_join, 'r.*', $trans_fields, $nontrans_fields, 'downloads', 'category_id');

        $out = [];
        foreach ($rows as $i => $row) {
            $out[$i]['data'] = $row;
            unset($rows[$i]);
            if (($remapped_orderer != '') && (array_key_exists($remapped_orderer, $row))) {
                $out[$i]['orderer'] = $row[$remapped_orderer];
            } elseif (strpos($remapped_orderer, '_rating:') !== false) {
                $out[$i]['orderer'] = $row[$remapped_orderer];
            }
        }

        return $out;
    }

    /**
     * Run function for rendering a search result.
     *
     * @param  array $row The data row stored when we retrieved the result
     * @return Tempcode The output
     */
    public function render(array $row) : object
    {
        global $SEARCH_QUERY_TERMS;
        $highlight_bits = ($SEARCH_QUERY_TERMS === null) ? [] : $SEARCH_QUERY_TERMS;

        $_text_summary = get_translated_text($row['the_description']);
        push_lax_comcode(true);
        $text_summary_h = comcode_to_tempcode($_text_summary, null, false, null, null, COMCODE_NORMAL, $highlight_bits);
        pop_lax_comcode();
        $cnt = 0;
        $text_summary = generate_text_summary($text_summary_h->evaluate(), $highlight_bits, $cnt);

        if ((get_value('highlight_extracted_file_data') === '1') && ($cnt == 0) && (array_key_exists(0, $highlight_bits)) && ($row['download_data_mash'] != '')) {
            $pos = strpos($row['download_data_mash'], $highlight_bits[0]) - 1000;
            $mash_portion = substr($row['download_data_mash'], $pos, 10000);
            $_text_summary = trim(cms_preg_replace_safe('#\s+#', ' ', $mash_portion));
            $text_summary = generate_text_summary($_text_summary, $highlight_bits);
        } else {
            $_text_summary = get_translated_text($row['the_description']);
            push_lax_comcode(true);
            $text_summary_h = comcode_to_tempcode($_text_summary, null, false, null, null, COMCODE_NORMAL, $highlight_bits);
            pop_lax_comcode();
            $text_summary = generate_text_summary($text_summary_h->evaluate(), $highlight_bits);
        }

        return render_download_box($row, true, true, null, protect_from_escaping($text_summary));
    }
}
