<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    wiki
 */

/**
 * Hook class.
 */
class Hook_actionlog_wiki extends Hook_actionlog
{
    /**
     * Get details of action log entry types handled by this hook.
     *
     * @return array Map of handler data in standard format
     */
    public function get_handlers() : array
    {
        if (!addon_installed('wiki')) {
            return [];
        }

        require_lang('wiki');

        return [
            'WIKI_ADD_PAGE' => [
                'flags' => ACTIONLOG_FLAGS_NONE,
                'cma_hook' => 'wiki_page',
                'identifier_index' => 0,
                'written_context_index' => 1,
                'followup_page_links' => [
                    '_WIKI_PAGE' => '_SEARCH:wiki:browse:{ID}',
                ],
            ],
            'WIKI_EDIT_PAGE' => [
                'flags' => ACTIONLOG_FLAGS_NONE,
                'cma_hook' => 'wiki_page',
                'identifier_index' => 0,
                'written_context_index' => 1,
                'followup_page_links' => [
                    '_WIKI_PAGE' => '_SEARCH:wiki:browse:{ID}',
                ],
            ],
            'WIKI_DELETE_PAGE' => [
                'flags' => ACTIONLOG_FLAGS_NONE,
                'cma_hook' => 'wiki_page',
                'identifier_index' => 0,
                'written_context_index' => 1,
                'followup_page_links' => [
                    'WIKI_HOME' => '_SEARCH:wiki',
                ],
            ],
            'WIKI_EDIT_TREE' => [
                'flags' => ACTIONLOG_FLAGS_NONE,
                'cma_hook' => 'wiki_page',
                'identifier_index' => 0,
                'written_context_index' => 1,
                'followup_page_links' => [
                    '_WIKI_PAGE' => '_SEARCH:wiki:browse:{ID}',
                ],
            ],
            'WIKI_MAKE_POST' => [
                'flags' => ACTIONLOG_FLAGS_NONE,
                'cma_hook' => 'wiki_post',
                'identifier_index' => 0,
                'written_context_index' => null,
                'followup_page_links' => [
                    '_WIKI_PAGE' => '_SEARCH:wiki:browse:{1}',
                    'WIKI_EDIT_POST' => '_SEARCH:wiki:post:post_id={ID}',
                ],
            ],
            'WIKI_EDIT_POST' => [
                'flags' => ACTIONLOG_FLAGS_NONE,
                'cma_hook' => 'wiki_post',
                'identifier_index' => 0,
                'written_context_index' => null,
                'followup_page_links' => [
                    '_WIKI_PAGE' => '_SEARCH:wiki:browse:{1}',
                    'WIKI_EDIT_POST' => '_SEARCH:wiki:post:post_id={ID}',
                ],
            ],
            'WIKI_DELETE_POST' => [
                'flags' => ACTIONLOG_FLAGS_NONE,
                'cma_hook' => 'wiki_post',
                'identifier_index' => 0,
                'written_context_index' => null,
                'followup_page_links' => [
                    '_WIKI_PAGE' => '_SEARCH:wiki:browse:{1}',
                ],
            ],
            'MERGE_WIKI_POSTS' => [
                'flags' => ACTIONLOG_FLAGS_NONE,
                'cma_hook' => 'wiki_page',
                'identifier_index' => 0,
                'written_context_index' => 1,
                'followup_page_links' => [
                    '_WIKI_PAGE' => '_SEARCH:wiki:browse:{ID}',
                ],
            ],
            'WIKI_MOVE_POST' => [
                'flags' => ACTIONLOG_FLAGS_NONE,
                'cma_hook' => 'wiki_post',
                'identifier_index' => 0,
                'written_context_index' => null,
                'followup_page_links' => [
                    '_WIKI_PAGE' => '_SEARCH:wiki:browse:{1}',
                    'WIKI_EDIT_POST' => '_SEARCH:wiki:post:post_id={ID}',
                ],
            ],
        ];
    }

    /**
     * Get written context for an action log entry handled by this hook.
     *
     * @param  array $actionlog_row Action log row
     * @param  array $handler_data Handler data
     * @param  ?string $identifier Identifier (null: none)
     * @return ?string Written context (null: none)
     */
    protected function get_written_context(array $actionlog_row, array $handler_data, ?string $identifier) : ?string
    {
        switch ($actionlog_row['the_type']) {
            case 'WIKI_MAKE_POST':
            case 'WIKI_DELETE_POST':
                $_page_title = $GLOBALS['SITE_DB']->query_select_value_if_there('wiki_pages', 'title', ['id' => intval($actionlog_row['param_b'])]);
                if ($_page_title === null) {
                    $page_title = '#' . $actionlog_row['param_b'];
                } else {
                    $page_title = get_translated_text($_page_title);
                }

                $written_context = do_lang('SOMETHING_IN', '#' . $actionlog_row['param_a'], $page_title);
                return $written_context;

            case 'WIKI_EDIT_POST':
                $member_id = $GLOBALS['SITE_DB']->query_select_value_if_there('wiki_posts', 'member_id', ['id' => intval($actionlog_row['param_a'])]);
                if ($member_id !== null) {
                    $username = $GLOBALS['FORUM_DRIVER']->get_username($member_id);
                    if ($username === null) {
                        $username = '#' . strval($member_id);
                    }
                } else {
                    $username = do_lang('UNKNOWN');
                }

                $_page_title = $GLOBALS['SITE_DB']->query_select_value_if_there('wiki_pages', 'title', ['id' => intval($actionlog_row['param_b'])]);
                if ($_page_title === null) {
                    $page_title = '#' . $actionlog_row['param_b'];
                } else {
                    $page_title = get_translated_text($_page_title);
                }

                require_lang('wiki');
                $written_context = do_lang('POST_BY_X_IN_Y', $username, $page_title);
                return $written_context;

            case 'WIKI_MOVE_POST':
                $_page_title = $GLOBALS['SITE_DB']->query_select_value_if_there('wiki_pages', 'title', ['id' => intval($actionlog_row['param_b'])]);
                if ($_page_title === null) {
                    $page_title = '#' . $actionlog_row['param_a'];
                } else {
                    $page_title = get_translated_text($_page_title);
                }

                $written_context = do_lang('SOMETHING_TO', '#' . $actionlog_row['param_a'], $page_title);
                return $written_context;
        }

        return parent::get_written_context($actionlog_row, $handler_data, $identifier);
    }
}
