<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    cns_cpfs
 */

/**
 * Hook class.
 */
class Hook_addon_registry_cns_cpfs
{
    /**
     * Get a list of file permissions to set.
     *
     * @param  boolean $runtime Whether to include wildcards represented runtime-created chmoddable files
     * @return array File permissions to set
     */
    public function get_chmod_array(bool $runtime = false) : array
    {
        $chmod = [];
        if ($runtime) {
            $chmod = array_merge(
                $chmod,
                [
                    'uploads/cns_cpf_upload/*',
                ]
            );
        }
        $chmod = array_merge(
            $chmod,
            [
                'uploads/cns_cpf_upload',
            ]
        );
        return $chmod;
    }

    /**
     * Get the version of the software this addon is for (used in generating the TAR filename).
     *
     * @return float Version number
     */
    public function get_version() : float
    {
        return cms_version_number();
    }

    /**
     * Get the minimum required version of the website software needed to use this addon.
     *
     * @return float Minimum required website software version
     */
    public function get_min_cms_version() : float
    {
        return 11.0;
    }

    /**
     * Get the maximum compatible version of the website software to use this addon.
     *
     * @return ?float Maximum compatible website software version (null: no maximum version currently)
     */
    public function get_max_cms_version() : ?float
    {
        return 11.9;
    }

    /**
     * Get the addon category.
     *
     * @return string The category
     */
    public function get_category() : string
    {
        return 'Community';
    }

    /**
     * Get the description of the addon.
     *
     * @return string Description of the addon
     */
    public function get_description() : string
    {
        return 'Custom Profile Fields, so members may save additional details. If this is uninstalled any existing Custom Profile Fields will remain in the system.';
    }

    /**
     * Get a list of tutorials that apply to this addon.
     *
     * @return array List of tutorials
     */
    public function get_applicable_tutorials() : array
    {
        return [
            'tut_adv_members',
        ];
    }

    /**
     * Get a mapping of dependency types.
     *
     * @return array A structure specifying dependency information
     */
    public function get_dependencies() : array
    {
        return [
            'requires' => [],
            'recommends' => [],
            'conflicts_with' => [],
        ];
    }

    /**
     * Explicitly say which icon should be used.
     *
     * @return URLPATH Icon
     */
    public function get_default_icon() : string
    {
        return 'themes/default/images/icons/menu/adminzone/tools/users/custom_profile_fields.svg';
    }

    /**
     * Get a list of files that belong to this addon.
     *
     * @return array List of files
     */
    public function get_file_list() : array
    {
        return [
            'adminzone/pages/modules/admin_cns_customprofilefields.php',
            'lang/EN/cns_privacy.ini',
            'sources/hooks/systems/actionlog/cns_cpfs.php',
            'sources/hooks/systems/addon_registry/cns_cpfs.php',
            'sources/hooks/systems/commandr_fs/cpfs.php',
            'sources/hooks/systems/commandr_fs_extended_member/cpf_perms.php',
            'sources/hooks/systems/page_groupings/cns_cpfs.php',
            'sources/hooks/systems/profiles_tabs_edit/privacy.php',
            'sources/hooks/systems/resource_meta_aware/cpf.php',
            'themes/default/images/icons/menu/adminzone/tools/users/custom_profile_fields.svg',
            'themes/default/images/icons_monochrome/menu/adminzone/tools/users/custom_profile_fields.svg',
            'themes/default/javascript/cns_cpfs.js',
            'themes/default/templates/CNS_CPF_PERMISSIONS_TAB.tpl',
            'themes/default/templates/CNS_CPF_STATS_SCREEN.tpl',
            'uploads/cns_cpf_upload/.htaccess',
            'uploads/cns_cpf_upload/index.html',
        ];
    }

    /**
     * Get mapping between template names and the method of this class that can render a preview of them.
     *
     * @return array The mapping
     */
    public function tpl_previews() : array
    {
        return [
            'templates/CNS_CPF_PERMISSIONS_TAB.tpl' => 'cns_cpf_permissions_tab',
            'templates/CNS_CPF_STATS_SCREEN.tpl' => 'administrative__cns_cpf_stats_screen',
        ];
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__cns_cpf_permissions_tab() : object
    {
        return lorem_globalise(do_lorem_template('CNS_CPF_PERMISSIONS_TAB', [
            'FIELDS' => placeholder_fields(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__administrative__cns_cpf_stats_screen() : object
    {
        $lines = [];
        foreach (placeholder_array() as $value) {
            $lines[] = [
                '_CNT' => placeholder_number(),
                'CNT' => placeholder_number(),
                'VAL' => lorem_phrase(),
            ];
        }

        return lorem_globalise(do_lorem_template('CNS_CPF_STATS_SCREEN', [
            'TITLE' => lorem_screen_title(),
            'STATS' => $lines,
        ]), null, '', true);
    }
}
