<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    cns_forum
 */

/**
 * Hook class.
 */
class Hook_addon_registry_cns_forum
{
    /**
     * Get a list of file permissions to set.
     *
     * @param  boolean $runtime Whether to include wildcards represented runtime-created chmoddable files
     * @return array File permissions to set
     */
    public function get_chmod_array(bool $runtime = false) : array
    {
        $chmod = [];
        if ($runtime) {
            $chmod = array_merge(
                $chmod,
                [
                    'forum/pages/comcode_custom/**/*.txt',
                    'forum/pages/html_custom/**/*.htm',
                ]
            );
        }
        $chmod = array_merge(
            $chmod,
            [
                'forum/pages/comcode_custom/**',
                'forum/pages/html_custom/**',
            ]
        );
        return $chmod;
    }

    /**
     * Get the version of the software this addon is for (used in generating the TAR filename).
     *
     * @return float Version number
     */
    public function get_version() : float
    {
        return cms_version_number();
    }

    /**
     * Get the minimum required version of the website software needed to use this addon.
     *
     * @return float Minimum required website software version
     */
    public function get_min_cms_version() : float
    {
        return 11.0;
    }

    /**
     * Get the maximum compatible version of the website software to use this addon.
     *
     * @return ?float Maximum compatible website software version (null: no maximum version currently)
     */
    public function get_max_cms_version() : ?float
    {
        return 11.9;
    }

    /**
     * Get the addon category.
     *
     * @return string The category
     */
    public function get_category() : string
    {
        return 'Community';
    }

    /**
     * Get the description of the addon.
     *
     * @return string Description of the addon
     */
    public function get_description() : string
    {
        return 'The Conversr forum- a modern advanced forum for members to interact on.';
    }

    /**
     * Get a list of tutorials that apply to this addon.
     *
     * @return array List of tutorials
     */
    public function get_applicable_tutorials() : array
    {
        return [
            'tut_forums',
            'tut_adv_forums',
            'tut_forum_tracking',
            'tut_moderation',
            'tut_information',
        ];
    }

    /**
     * Get a mapping of dependency types.
     *
     * @return array A structure specifying dependency information
     */
    public function get_dependencies() : array
    {
        return [
            'requires' => [],
            'recommends' => [],
            'conflicts_with' => [],
        ];
    }

    /**
     * Explicitly say which icon should be used.
     *
     * @return URLPATH Icon
     */
    public function get_default_icon() : string
    {
        return 'themes/default/images/icons/menu/social/forum/forums.svg';
    }

    /**
     * Get a list of files that belong to this addon.
     *
     * @return array List of files
     */
    public function get_file_list() : array
    {
        return [
            'adminzone/pages/modules/admin_cns_forum_groupings.php',
            'adminzone/pages/modules/admin_cns_forums.php',
            'forum/index.php',
            'forum/pages/comcode/.htaccess',
            'forum/pages/comcode/EN/.htaccess',
            'forum/pages/comcode/EN/index.html',
            'forum/pages/comcode/EN/panel_right.txt',
            'forum/pages/comcode/index.html',
            'forum/pages/comcode_custom/.htaccess',
            'forum/pages/comcode_custom/EN/.htaccess',
            'forum/pages/comcode_custom/EN/index.html',
            'forum/pages/comcode_custom/index.html',
            'forum/pages/html/.htaccess',
            'forum/pages/html/EN/.htaccess',
            'forum/pages/html/EN/index.html',
            'forum/pages/html/index.html',
            'forum/pages/html_custom/EN/.htaccess',
            'forum/pages/html_custom/EN/index.html',
            'forum/pages/html_custom/index.html',
            'forum/pages/index.html',
            'forum/pages/minimodules/.htaccess',
            'forum/pages/minimodules/index.html',
            'forum/pages/minimodules_custom/.htaccess',
            'forum/pages/minimodules_custom/index.html',
            'forum/pages/modules/.htaccess',
            'forum/pages/modules/forumview.php',
            'forum/pages/modules/index.html',
            'forum/pages/modules/topics.php',
            'forum/pages/modules/topicview.php',
            'forum/pages/modules/vforums.php',
            'forum/pages/modules_custom/.htaccess',
            'forum/pages/modules_custom/index.html',
            'forum/rules.php',
            'sources/blocks/main_bottom_bar.php',
            'sources/blocks/main_cns_involved_topics.php',
            'sources/blocks/main_member_bar.php',
            'sources/blocks/side_cns_private_topics.php',
            'sources/cns_forum_email_integration.php',
            'sources/cns_forumview.php',
            'sources/cns_forumview_pt.php',
            'sources/cns_topicview.php',
            'sources/hooks/modules/admin_import_types/cns_forum.php',
            'sources/hooks/modules/admin_newsletter/cns_forumview.php',
            'sources/hooks/modules/admin_stats/cns_forum.php',
            'sources/hooks/modules/admin_themewizard/cns_forum.php',
            'sources/hooks/modules/search/cns_own_pt.php',
            'sources/hooks/modules/search/cns_posts.php',
            'sources/hooks/modules/search/cns_within_topic.php',
            'sources/hooks/systems/actionlog/cns_forum.php',
            'sources/hooks/systems/addon_registry/cns_forum.php',
            'sources/hooks/systems/ajax_tree/choose_forum.php',
            'sources/hooks/systems/ajax_tree/choose_topic.php',
            'sources/hooks/systems/attachments/cns_post.php',
            'sources/hooks/systems/cleanup/cns.php',
            'sources/hooks/systems/cleanup/cns_topics.php',
            'sources/hooks/systems/commandr_fs/forum_groupings.php',
            'sources/hooks/systems/commandr_fs/forums.php',
            'sources/hooks/systems/commandr_scheduled/publish_post.php',
            'sources/hooks/systems/commandr_scheduled/publish_topic.php',
            'sources/hooks/systems/config/delete_time_limit.php',
            'sources/hooks/systems/config/delete_trashed_pts.php',
            'sources/hooks/systems/config/edit_time_limit.php',
            'sources/hooks/systems/config/enable_add_topic_btn_in_topic.php',
            'sources/hooks/systems/config/enable_forum_dupe_buttons.php',
            'sources/hooks/systems/config/enable_mark_forum_read.php',
            'sources/hooks/systems/config/enable_mark_topic_unread.php',
            'sources/hooks/systems/config/enable_multi_quote.php',
            'sources/hooks/systems/config/enable_poll_point_weighting.php',
            'sources/hooks/systems/config/enable_post_emphasis.php',
            'sources/hooks/systems/config/enable_pt_filtering.php',
            'sources/hooks/systems/config/enable_pt_restrict.php',
            'sources/hooks/systems/config/force_guest_names.php',
            'sources/hooks/systems/config/forum_posts_per_page.php',
            'sources/hooks/systems/config/forum_topics_per_page.php',
            'sources/hooks/systems/config/inline_pp_advertise.php',
            'sources/hooks/systems/config/intro_forum_id.php',
            'sources/hooks/systems/config/is_on_anonymous_posts.php',
            'sources/hooks/systems/config/is_on_post_titles.php',
            'sources/hooks/systems/config/is_on_show_online.php',
            'sources/hooks/systems/config/is_on_topic_descriptions.php',
            'sources/hooks/systems/config/is_on_topic_emoticons.php',
            'sources/hooks/systems/config/mailing_list_style_default.php',
            'sources/hooks/systems/config/max_forum_detail.php',
            'sources/hooks/systems/config/max_forum_inspect.php',
            'sources/hooks/systems/config/overt_whisper_suggestion.php',
            'sources/hooks/systems/config/points_posting.php',
            'sources/hooks/systems/config/points_voting_cns.php',
            'sources/hooks/systems/config/post_read_history_days.php',
            'sources/hooks/systems/config/private_topics_per_page.php',
            'sources/hooks/systems/config/search_cns_own_pt.php',
            'sources/hooks/systems/config/search_cns_posts.php',
            'sources/hooks/systems/config/seq_post_ids.php',
            'sources/hooks/systems/config/smart_topic_notification_default.php',
            'sources/hooks/systems/config/sound_enabled_default.php',
            'sources/hooks/systems/config/threaded_buttons.php',
            'sources/hooks/systems/config/topic_polls_weighting_base.php',
            'sources/hooks/systems/config/topic_polls_weighting_ceiling.php',
            'sources/hooks/systems/config/topic_polls_weighting_multiplier.php',
            'sources/hooks/systems/config/topic_polls_weighting_offset.php',
            'sources/hooks/systems/content_meta_aware/forum.php',
            'sources/hooks/systems/cron/cns_forum_email_integration.php',
            'sources/hooks/systems/member_boxes/.htaccess',
            'sources/hooks/systems/member_boxes/index.html',
            'sources/hooks/systems/notifications/cns_pts.php',
            'sources/hooks/systems/notifications/cns_topic.php',
            'sources/hooks/systems/page_groupings/cns_forum.php',
            'sources/hooks/systems/points/post__add.php',
            'sources/hooks/systems/points/topic_poll__vote.php',
            'sources/hooks/systems/preview/cns_post.php',
            'sources/hooks/systems/profiles_tabs/posts.php',
            'sources/hooks/systems/profiles_tabs/pts.php',
            'sources/hooks/systems/resource_meta_aware/forum_grouping.php',
            'sources/hooks/systems/rss/cns_forumview.php',
            'sources/hooks/systems/rss/cns_polls.php',
            'sources/hooks/systems/rss/cns_private_topics.php',
            'sources/hooks/systems/rss/cns_topicview.php',
            'sources/hooks/systems/rss/cns_unread_topics.php',
            'sources/hooks/systems/sitemap/forum.php',
            'sources/hooks/systems/sitemap/topic.php',
            'sources/hooks/systems/tasks/cns_recache.php',
            'sources/hooks/systems/tasks/cns_topics_recache.php',
            'sources/hooks/systems/tasks/export_voting_power.php',
            'sources/hooks/systems/tasks/notify_topics_moved.php',
            'sources_custom/hooks/systems/member_boxes/.htaccess',
            'sources_custom/hooks/systems/member_boxes/index.html',
            'themes/default/images/icons/buttons/add_topic.svg',
            'themes/default/images/icons/buttons/forum.svg',
            'themes/default/images/icons/buttons/linear.svg',
            'themes/default/images/icons/buttons/mark_read.svg',
            'themes/default/images/icons/buttons/mark_unread.svg',
            'themes/default/images/icons/buttons/new_quote.svg',
            'themes/default/images/icons/buttons/threaded.svg',
            'themes/default/images/icons/buttons/whisper.svg',
            'themes/default/images/icons/cns_general/new_posts.svg',
            'themes/default/images/icons/cns_general/new_posts_redirect.svg',
            'themes/default/images/icons/cns_general/no_new_posts.svg',
            'themes/default/images/icons/cns_general/no_new_posts_redirect.svg',
            'themes/default/images/icons/cns_general/redirect.svg',
            'themes/default/images/icons/cns_topic_modifiers/announcement.svg',
            'themes/default/images/icons/cns_topic_modifiers/closed.svg',
            'themes/default/images/icons/cns_topic_modifiers/hot.svg',
            'themes/default/images/icons/cns_topic_modifiers/index.html',
            'themes/default/images/icons/cns_topic_modifiers/involved.svg',
            'themes/default/images/icons/cns_topic_modifiers/not_validated.svg',
            'themes/default/images/icons/cns_topic_modifiers/pinned.svg',
            'themes/default/images/icons/cns_topic_modifiers/poll.svg',
            'themes/default/images/icons/cns_topic_modifiers/unread.svg',
            'themes/default/images/icons/menu/adminzone/structure/forum/index.html',
            'themes/default/images/icons/menu/social/forum/forums.svg',
            'themes/default/images/icons/menu/social/forum/index.html',
            'themes/default/images/icons/menu/social/forum/inline_personal_posts.svg',
            'themes/default/images/icons/menu/social/forum/vforums/index.html',
            'themes/default/images/icons/menu/social/forum/vforums/involved_topics.svg',
            'themes/default/images/icons/menu/social/forum/vforums/posts_since_last_visit.svg',
            'themes/default/images/icons/menu/social/forum/vforums/recently_read_topics.svg',
            'themes/default/images/icons/menu/social/forum/vforums/unanswered_topics.svg',
            'themes/default/images/icons/menu/social/forum/vforums/unread_topics.svg',
            'themes/default/images/icons_monochrome/buttons/add_topic.svg',
            'themes/default/images/icons_monochrome/buttons/forum.svg',
            'themes/default/images/icons_monochrome/buttons/linear.svg',
            'themes/default/images/icons_monochrome/buttons/mark_read.svg',
            'themes/default/images/icons_monochrome/buttons/mark_unread.svg',
            'themes/default/images/icons_monochrome/buttons/new_quote.svg',
            'themes/default/images/icons_monochrome/buttons/threaded.svg',
            'themes/default/images/icons_monochrome/buttons/whisper.svg',
            'themes/default/images/icons_monochrome/cns_general/new_posts.svg',
            'themes/default/images/icons_monochrome/cns_general/new_posts_redirect.svg',
            'themes/default/images/icons_monochrome/cns_general/no_new_posts.svg',
            'themes/default/images/icons_monochrome/cns_general/no_new_posts_redirect.svg',
            'themes/default/images/icons_monochrome/cns_general/redirect.svg',
            'themes/default/images/icons_monochrome/cns_topic_modifiers/announcement.svg',
            'themes/default/images/icons_monochrome/cns_topic_modifiers/closed.svg',
            'themes/default/images/icons_monochrome/cns_topic_modifiers/hot.svg',
            'themes/default/images/icons_monochrome/cns_topic_modifiers/index.html',
            'themes/default/images/icons_monochrome/cns_topic_modifiers/involved.svg',
            'themes/default/images/icons_monochrome/cns_topic_modifiers/not_validated.svg',
            'themes/default/images/icons_monochrome/cns_topic_modifiers/pinned.svg',
            'themes/default/images/icons_monochrome/cns_topic_modifiers/poll.svg',
            'themes/default/images/icons_monochrome/cns_topic_modifiers/unread.svg',
            'themes/default/images/icons_monochrome/menu/adminzone/structure/forum/index.html',
            'themes/default/images/icons_monochrome/menu/social/forum/forums.svg',
            'themes/default/images/icons_monochrome/menu/social/forum/index.html',
            'themes/default/images/icons_monochrome/menu/social/forum/inline_personal_posts.svg',
            'themes/default/images/icons_monochrome/menu/social/forum/vforums/index.html',
            'themes/default/images/icons_monochrome/menu/social/forum/vforums/involved_topics.svg',
            'themes/default/images/icons_monochrome/menu/social/forum/vforums/posts_since_last_visit.svg',
            'themes/default/images/icons_monochrome/menu/social/forum/vforums/recently_read_topics.svg',
            'themes/default/images/icons_monochrome/menu/social/forum/vforums/unanswered_topics.svg',
            'themes/default/images/icons_monochrome/menu/social/forum/vforums/unread_topics.svg',
            'themes/default/javascript/cns_forum.js',
            'themes/default/templates/BLOCK_MAIN_BOTTOM_BAR.tpl',
            'themes/default/templates/BLOCK_MAIN_CNS_INVOLVED_TOPICS.tpl',
            'themes/default/templates/BLOCK_MAIN_MEMBER_BAR.tpl',
            'themes/default/templates/BLOCK_SIDE_CNS_PRIVATE_TOPICS.tpl',
            'themes/default/templates/CNS_EDIT_FORUM_SCREEN.tpl',
            'themes/default/templates/CNS_EDIT_FORUM_SCREEN_FORUM.tpl',
            'themes/default/templates/CNS_EDIT_FORUM_SCREEN_GROUPING.tpl',
            'themes/default/templates/CNS_FORUM.tpl',
            'themes/default/templates/CNS_FORUM_GROUPING.tpl',
            'themes/default/templates/CNS_FORUM_INTRO_QUESTION_POPUP.tpl',
            'themes/default/templates/CNS_FORUM_INTRO_QUESTION_SCREEN.tpl',
            'themes/default/templates/CNS_FORUM_IN_GROUPING.tpl',
            'themes/default/templates/CNS_FORUM_LATEST.tpl',
            'themes/default/templates/CNS_FORUM_LIST_LINE.tpl',
            'themes/default/templates/CNS_FORUM_SCREEN.tpl',
            'themes/default/templates/CNS_FORUM_TOPIC_LIST_LINE.tpl',
            'themes/default/templates/CNS_FORUM_TOPIC_ROW.tpl',
            'themes/default/templates/CNS_FORUM_TOPIC_ROW_LAST_POST.tpl',
            'themes/default/templates/CNS_FORUM_TOPIC_WRAPPER.tpl',
            'themes/default/templates/CNS_GUEST_BAR.tpl',
            'themes/default/templates/CNS_GUEST_DETAILS.tpl',
            'themes/default/templates/CNS_MEMBER_BAR.tpl',
            'themes/default/templates/CNS_MEMBER_PROFILE_POSTS.tpl',
            'themes/default/templates/CNS_MEMBER_PROFILE_PTS.tpl',
            'themes/default/templates/CNS_MEMBER_PT_RULES_SCREEN.tpl',
            'themes/default/templates/CNS_PINNED_DIVIDER.tpl',
            'themes/default/templates/CNS_POSTER_GUEST.tpl',
            'themes/default/templates/CNS_POSTER_MEMBER.tpl',
            'themes/default/templates/CNS_POSTING_SCREEN_POSTS.tpl',
            'themes/default/templates/CNS_POST_BOX.tpl',
            'themes/default/templates/CNS_PRIVATE_TOPIC_LINK.tpl',
            'themes/default/templates/CNS_PT_BETWEEN.tpl',
            'themes/default/templates/CNS_PT_FILTERS.tpl',
            'themes/default/templates/CNS_TOPIC_FIRST_UNREAD.tpl',
            'themes/default/templates/CNS_TOPIC_MARKER.tpl',
            'themes/default/templates/CNS_TOPIC_POLL.tpl',
            'themes/default/templates/CNS_TOPIC_POLL_ANSWER_RADIO.tpl',
            'themes/default/templates/CNS_TOPIC_POLL_ANSWER_RESULTS.tpl',
            'themes/default/templates/CNS_TOPIC_POLL_ANSWER_TICK.tpl',
            'themes/default/templates/CNS_TOPIC_POLL_BUTTON.tpl',
            'themes/default/templates/CNS_TOPIC_POLL_BUTTON_RESULTS.tpl',
            'themes/default/templates/CNS_TOPIC_POLL_VIEW_RESULTS.tpl',
            'themes/default/templates/CNS_TOPIC_POLL_VOTERS_SCREEN.tpl',
            'themes/default/templates/CNS_TOPIC_POLL_VOTING_POWER.tpl',
            'themes/default/templates/CNS_TOPIC_POST.tpl',
            'themes/default/templates/CNS_TOPIC_POST_LAST_EDITED.tpl',
            'themes/default/templates/CNS_TOPIC_SCREEN.tpl',
            'themes/default/templates/CNS_VFORUM_FILTERING.tpl',
            'themes/default/templates/CNS_VFORUM_SCREEN.tpl',
            'themes/default/templates/CNS_WHISPER_CHOICE_SCREEN.tpl',
            'themes/default/templates/MEMBER_BAR_SEARCH.tpl',
            'themes/default/text/CNS_POST_FROM_MAILING_LIST.txt',
            'themes/default/text/CNS_QUOTE_FCOMCODE.txt',
        ];
    }

    /**
     * Get mapping between template names and the method of this class that can render a preview of them.
     *
     * @return array The mapping
     */
    public function tpl_previews() : array
    {
        return [
            'templates/CNS_TOPIC_POST_AVATAR.tpl' => 'cns_topic_not_voted_checkboxes_wrap',
            'templates/CNS_FORUM_TOPIC_WRAPPER.tpl' => 'cns_forum',
            'templates/CNS_EDIT_FORUM_SCREEN_GROUPING.tpl' => 'administrative__cns_edit_forum_screen',
            'templates/CNS_USER_MEMBER.tpl' => 'cns_topic_not_voted_checkboxes_wrap',
            'templates/CNS_MEMBER_BOX.tpl' => 'cns_topic_not_voted_checkboxes_wrap',
            'templates/CNS_EDIT_FORUM_SCREEN_FORUM.tpl' => 'administrative__cns_edit_forum_screen',
            'templates/CNS_EDIT_FORUM_SCREEN.tpl' => 'administrative__cns_edit_forum_screen',
            'templates/CNS_RANK_IMAGE.tpl' => 'cns_rank_image',
            'templates/CNS_POST_BOX.tpl' => 'cns_isolated_post',
            'templates/BLOCK_MAIN_CNS_INVOLVED_TOPICS.tpl' => 'block_main_cns_involved_topics',
            'templates/BLOCK_SIDE_CNS_PRIVATE_TOPICS.tpl' => 'block_side_cns_private_topics',
            'templates/CNS_FORUM_TOPIC_ROW_LAST_POST.tpl' => 'cns_forum',
            'templates/CNS_PT_BETWEEN.tpl' => 'cns_forum',
            'templates/CNS_TOPIC_MARKER.tpl' => 'cns_forum',
            'templates/CNS_FORUM_TOPIC_ROW.tpl' => 'cns_forum',
            'templates/CNS_FORUM_TOPIC_LIST_LINE.tpl' => 'cns_forum_topic_list_line',
            'templates/CNS_FORUM_LIST_LINE.tpl' => 'cns_forum_list_line',
            'templates/CNS_FORUM_INTRO_QUESTION_POPUP.tpl' => 'cns_forum_intro_question_popup',
            'templates/CNS_MEMBER_BOX_CUSTOM_FIELD.tpl' => 'cns_topic_not_voted_checkboxes_wrap',
            'templates/CNS_POSTER_MEMBER.tpl' => 'cns_topic_not_voted_checkboxes_wrap',
            'templates/CNS_GUEST_DETAILS.tpl' => 'cns_topic_not_voted_checkboxes_wrap',
            'templates/CNS_POSTER_GUEST.tpl' => 'cns_isolated_post',
            'templates/CNS_TOPIC_POST_LAST_EDITED.tpl' => 'cns_topic_not_voted_checkboxes_wrap',
            'templates/CNS_TOPIC_POST.tpl' => 'cns_topic_not_voted_checkboxes_wrap',
            'templates/CNS_MEMBER_BAR.tpl' => 'block_main_member_bar_member',
            'templates/MEMBER_BAR_SEARCH.tpl' => 'block_main_member_bar_search',
            'templates/CNS_GUEST_BAR.tpl' => 'block_main_member_bar_guest',
            'templates/BLOCK_MAIN_MEMBER_BAR.tpl' => 'block_main_member_bar_member',
            'templates/BLOCK_MAIN_BOTTOM_BAR.tpl' => 'block_main_bottom_bar',
            'templates/CNS_FORUM_INTRO_QUESTION_SCREEN.tpl' => 'cns_forum_intro_question_screen',
            'templates/CNS_FORUM_LATEST.tpl' => 'cns_forum',
            'templates/CNS_FORUM_IN_GROUPING.tpl' => 'cns_forum',
            'templates/CNS_FORUM_GROUPING.tpl' => 'cns_forum',
            'templates/CNS_PINNED_DIVIDER.tpl' => 'cns_forum',
            'templates/CNS_FORUM.tpl' => 'cns_forum',
            'templates/CNS_FORUM_SCREEN.tpl' => 'cns_forum',
            'templates/CNS_VFORUM_SCREEN.tpl' => 'cns_vforum',
            'templates/CNS_WHISPER_CHOICE_SCREEN.tpl' => 'cns_whisper_choice_screen',
            'text/CNS_QUOTE_FCOMCODE.txt' => 'cns_quote_fcomcode',
            'templates/CNS_MEMBER_PT_RULES_SCREEN.tpl' => 'cns_member_pt_rules_screen',
            'templates/CNS_POSTING_SCREEN_POSTS.tpl' => 'cns_posting_screen_posts',
            'templates/CNS_TOPIC_FIRST_UNREAD.tpl' => 'cns_topic_not_voted_checkboxes_wrap',
            'templates/CNS_TOPIC_POLL_BUTTON.tpl' => 'cns_topic_not_voted_checkboxes_wrap',
            'templates/CNS_TOPIC_POLL_BUTTON_RESULTS.tpl' => 'cns_topic_voted_wrap',
            'templates/CNS_TOPIC_POLL_ANSWER_RESULTS.tpl' => 'cns_topic_voted_wrap',
            'templates/CNS_TOPIC_POLL_ANSWER_TICK.tpl' => 'cns_topic_not_voted_checkboxes_wrap',
            'templates/CNS_TOPIC_POLL_ANSWER_RADIO.tpl' => 'cns_topic_not_voted_radio_wrap',
            'templates/CNS_TOPIC_POLL.tpl' => 'cns_topic_not_voted_checkboxes_wrap',
            'templates/CNS_TOPIC_SCREEN.tpl' => 'cns_topic_not_voted_checkboxes_wrap',
            'templates/NOTIFICATION_BUTTONS.tpl' => 'cns_topic_not_voted_checkboxes_wrap',
            'templates/RESULTS_TABLE_MEMBER_CELL.tpl' => 'cns_topic_poll_voters',
            'templates/CNS_TOPIC_POLL_VOTERS_SCREEN.tpl' => 'cns_topic_poll_voters',
            'templates/CNS_TOPIC_POLL_VOTING_POWER.tpl' => 'cns_topic_poll_voters',
            'templates/HELP_ICON_PHRASE.tpl' => 'cns_topic_poll_voters',
            'templates/CNS_TOPIC_POLL_VIEW_RESULTS.tpl' => 'cns_topic_voted_wrap',
            'templates/CNS_PRIVATE_TOPIC_LINK.tpl' => 'cns_private_topic_link',
            'templates/CNS_PT_FILTERS.tpl' => 'cns_forum',
            'templates/CNS_MEMBER_PROFILE_POSTS.tpl' => 'cns_member_profile_posts',
            'templates/CNS_MEMBER_PROFILE_PTS.tpl' => 'cns_member_profile_pts',
            'templates/CNS_VFORUM_FILTERING.tpl' => 'cns_vforum_filtering',
            'text/CNS_POST_FROM_MAILING_LIST.txt' => 'cns_post_from_mailing_list',
        ];
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__cns_vforum_filtering() : object
    {
        require_lang('cns');

        $filtering = do_lorem_template('CNS_VFORUM_FILTERING', []);

        $content = do_lorem_template('CNS_FORUM', [
            'BREADCRUMBS' => placeholder_breadcrumbs(),
            'FILTERS' => '',
            'FILTERING' => $filtering,
            'FORUM_NAME' => lorem_phrase(),
            'STARTER_TITLE' => lorem_phrase(),
            'BUTTONS' => '',
            'TOPIC_WRAPPER' => '',
            'FORUM_GROUPINGS' => '',
            'ID' => placeholder_numeric_id(),
            'DESCRIPTION' => lorem_phrase(),
            'PARENT_FORUM' => '',
        ]);

        $members_viewing = new Tempcode();
        foreach (placeholder_array() as $_k => $_v) {
            $members_viewing->attach(do_lorem_template('CNS_USER_MEMBER', [
                'FIRST' => true,
                'PROFILE_URL' => placeholder_url(),
                'USERNAME' => lorem_word(),
                'MEMBER_ID' => placeholder_numeric_id(),
                'AT' => lorem_phrase(),
                'COLOUR' => lorem_word(),
            ]));
        }

        return lorem_globalise(do_lorem_template('CNS_FORUM_SCREEN', [
            'TITLE' => lorem_screen_title(),
            'CONTENT' => $content,
            '_NUM_GUESTS' => placeholder_number(),
            '_NUM_MEMBERS' => placeholder_number(),
            'NUM_GUESTS' => placeholder_number(),
            'NUM_MEMBERS' => placeholder_number(),
            'MEMBERS_VIEWING' => $members_viewing,
            'ID' => placeholder_numeric_id(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__cns_private_topic_link() : object
    {
        require_css('cns');

        return lorem_globalise(do_lorem_template('CNS_PRIVATE_TOPIC_LINK', [
            'TOPIC_URL' => placeholder_url(),
            'TITLE' => lorem_phrase(),
            'DATE' => placeholder_date(),
            'DATE_RAW' => placeholder_date_raw(),
            'LAST_POST_BY_POSTER_URL' => placeholder_url(),
            'LAST_POST_BY_USERNAME' => lorem_word(),
            'WITH_POSTER_URL' => placeholder_url(),
            'WITH_USERNAME' => lorem_word(),
            'WITH_POSTER_ID' => placeholder_first_admin_id(),
            'BY_POSTER_URL' => placeholder_url(),
            'BY_USERNAME' => lorem_word(),
            'BY_POSTER_ID' => placeholder_first_admin_id(),
            'TO_POSTER_URL' => placeholder_url(),
            'TO_USERNAME' => lorem_word(),
            'TO_POSTER_ID' => placeholder_first_admin_id(),
            '_NUM_POSTS' => placeholder_number(),
            'NUM_POSTS' => placeholder_number(),
            'HAS_READ' => false,
            'LAST_POST_BY_POSTER_ID' => placeholder_first_admin_id(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__administrative__cns_edit_forum_screen() : object
    {
        require_css('cns');
        require_css('cns_editor');

        require_lang('cns');

        $orderings = '<option>1</option><option>2</option><option>3</option>'; // XHTMLXHTML

        $orderings = '<label for="order_' . strval(1) . '">' . do_lang('SORT') . ' <select id="order_' . strval(1) . '" name="order_' . strval(1) . '">' . $orderings . '</select></label>';

        $forum_groupings = do_lorem_template('CNS_EDIT_FORUM_SCREEN_GROUPING', [
            'ORDERINGS' => $orderings,
            'GROUPING' => lorem_word_2(),
            'SUBFORUMS' => lorem_phrase(),
        ]);

        $root_forum = do_lorem_template('CNS_EDIT_FORUM_SCREEN_FORUM', [
            'ID' => placeholder_numeric_id(),
            'ORDERINGS' => lorem_phrase(),
            'FORUM_GROUPINGS' => $forum_groupings,
            'CLASS' => lorem_phrase(),
            'FORUM' => lorem_phrase(),
            'VIEW_URL' => placeholder_url(),
            'EDIT_URL' => placeholder_url(),
        ]);

        return lorem_globalise(do_lorem_template('CNS_EDIT_FORUM_SCREEN', [
            'REORDER_URL' => placeholder_url(),
            'TITLE' => lorem_screen_title(),
            'ROOT_FORUM' => $root_forum,
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__cns_rank_image() : object
    {
        require_css('cns');

        return lorem_globalise(do_lorem_template('CNS_RANK_IMAGE', [
            'GROUP_NAME' => lorem_word(),
            'USERNAME' => lorem_word(),
            'IMG' => 'cns_rank_images/' . placeholder_img_code('cns_rank_images'),
            'IS_LEADER' => lorem_phrase(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__cns_isolated_post() : object
    {
        require_css('cns');

        require_lang('cns');

        $first_unread = do_lorem_template('CNS_TOPIC_FIRST_UNREAD');

        $last_edited = do_lorem_template('CNS_TOPIC_POST_LAST_EDITED', [
            'LAST_EDIT_DATE_RAW' => placeholder_date_raw(),
            'LAST_EDIT_DATE' => placeholder_date(),
            'LAST_EDIT_PROFILE_URL' => placeholder_url(),
            'LAST_EDIT_USERNAME' => lorem_word(),
        ]);

        $custom_fields = do_lorem_template('CNS_MEMBER_BOX_CUSTOM_FIELD', [
            'NAME' => lorem_phrase(),
            'RAW' => placeholder_ip(),
            'VALUE' => placeholder_ip(),
            'MEMBER_ID' => placeholder_numeric_id(),
        ]);

        $poster_details = do_lorem_template('CNS_GUEST_DETAILS', [
            'CUSTOM_FIELDS' => $custom_fields,
        ]);

        $poster = do_lorem_template('CNS_POSTER_GUEST', [
            'LOOKUP_IP_URL' => placeholder_url(),
            'POSTER_DETAILS' => $poster_details,
            'POSTER_USERNAME' => lorem_word(),
        ]);

        $post_avatar = do_lorem_template('CNS_TOPIC_POST_AVATAR', [
            'AVATAR' => placeholder_image_url(),
        ]);

        $rank_images = do_lorem_template('CNS_RANK_IMAGE', [
            'GROUP_NAME' => lorem_phrase(),
            'USERNAME' => lorem_word(),
            'IMG' => 'cns_rank_images/' . placeholder_img_code('cns_rank_images'),
            'IS_LEADER' => lorem_phrase(),
        ]);

        $buttons = do_lorem_template('BUTTON_SCREEN_ITEM', [
            'REL' => lorem_word(),
            'IMMEDIATE' => false,
            'IMG' => 'buttons/proceed',
            'URL' => placeholder_url(),
            'FULL_TITLE' => lorem_phrase(),
            'TITLE' => lorem_word(),
        ]);

        $map = [
            'ID' => placeholder_numeric_id(),
            'GIVE_CONTEXT' => false,
            'TOPIC_FIRST_POST_ID' => placeholder_numeric_id(),
            'TOPIC_FIRST_POSTER' => lorem_phrase(),
            'POST_ID' => placeholder_numeric_id(),
            'URL' => placeholder_url(),
            'CLASS' => lorem_phrase(),
            'EMPHASIS' => lorem_phrase(),
            'FIRST_UNREAD' => $first_unread,
            'POSTER_TITLE' => lorem_phrase(),
            'POST_TITLE' => lorem_phrase(),
            'POST_DATE_RAW' => placeholder_date_raw(),
            'POST_DATE' => placeholder_date(),
            'POST' => lorem_phrase(),
            'TOPIC_ID' => placeholder_numeric_id(),
            'LAST_EDITED_RAW' => lorem_phrase(),
            'LAST_EDITED' => $last_edited,
            'POSTER_ID' => placeholder_numeric_id(),
            'POSTER' => $poster,
            'POSTER_DETAILS' => $poster_details,
            'POST_AVATAR' => $post_avatar,
            'RANK_IMAGES' => $rank_images,
            'BUTTONS' => $buttons,
            'SIGNATURE' => lorem_phrase(),
            'NOT_VALIDATED' => lorem_phrase(),
            'DESCRIPTION' => lorem_phrase(),
        ];
        $post = do_lorem_template('CNS_TOPIC_POST', $map);

        return lorem_globalise(do_lorem_template('CNS_POST_BOX', [
            'GIVE_CONTEXT' => true,
            'ID' => placeholder_numeric_id(),
            'POST' => $post,
            'URL' => placeholder_url(),
            'BREADCRUMBS' => lorem_phrase(),
            'ACTUAL_POST' => lorem_paragraph_html(),
        ] + $map + ['ACTUAL_POST' => $post]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__block_main_cns_involved_topics() : object
    {
        require_css('cns');
        require_lang('cns');

        $topics = new Tempcode();

        foreach (placeholder_array() as $i => $v) {
            $part = new Tempcode();
            $part->attach(do_lorem_template('RESULTS_LAUNCHER_PAGE_NUMBER_LINK', [
                'TITLE' => lorem_phrase(),
                'URL' => placeholder_url(),
                'P' => strval(1)
            ]));
            $part->attach(do_lorem_template('RESULTS_LAUNCHER_CONTINUE', [
                'TITLE' => lorem_phrase(),
                'MAX' => strval(3),
                '_NUM_PAGES' => strval(1),
                'NUM_PAGES' => integer_format(1),
                'URL_STUB' => placeholder_url(),
            ]));
            $pages = do_lorem_template('RESULTS_LAUNCHER_WRAP', ['PART' => $part]);

            $marker = new Tempcode();
            if ($i == 2) {
                $marker = do_template('CNS_TOPIC_MARKER', ['_GUID' => '64f4d9a4d92869b2a7cee169e6494301', 'ID' => strval($i)]);
            }

            $topic_row_links = [];
            if ($i == 1) {
                $topic_row_links[] = ['URL' => placeholder_url(), 'IMG' => 'unread', 'ALT' => lorem_phrase()];
            }

            $topic_row_modifiers = [];
            if ($i != 1) {
                $topic_row_modifiers[] = ['IMG' => 'hot', 'ALT' => lorem_phrase()];
            }

            $poster = do_lorem_template('CNS_USER_MEMBER', [
                'FIRST' => true,
                'PROFILE_URL' => placeholder_url(),
                'USERNAME' => lorem_word(),
                'MEMBER_ID' => placeholder_first_admin_id(),
            ]);

            $last_post = do_lorem_template('CNS_FORUM_TOPIC_ROW_LAST_POST', [
                'ID' => placeholder_numeric_id(),
                'DATE_RAW' => placeholder_date(),
                'DATE' => placeholder_date(),
                'POSTER' => $poster,
                'LAST_URL' => placeholder_url(),
            ]);

            $topics->attach(do_lorem_template('CNS_FORUM_TOPIC_ROW', [
                'BREADCRUMBS' => placeholder_breadcrumbs(),
                'RAW_TIME' => placeholder_date(),
                'UNREAD' => ($i == 1),
                'ID' => strval($i),
                'FORUM_ID' => placeholder_numeric_id(),
                'HOVER' => placeholder_date(),
                'PAGES' => $pages,
                'MARKER' => $marker,
                'TOPIC_ROW_LINKS' => $topic_row_links,
                'TOPIC_ROW_MODIFIERS' => $topic_row_modifiers,
                '_TOPIC_ROW_MODIFIERS' => [],
                'POST' => lorem_paragraph_html(),
                'EMOTICON' => new Tempcode(),
                'DESCRIPTION' => lorem_sentence(),
                'URL' => placeholder_url(),
                'TITLE' => lorem_phrase(),
                '_POSTER' => placeholder_first_admin_id(),
                'POSTER' => $poster,
                '_NUM_POSTS' => strval(1),
                'NUM_POSTS' => integer_format(1, 0),
                '_NUM_VIEWS' => strval(1),
                'NUM_VIEWS' => integer_format(1, 0),
                'LAST_POST' => $last_post,
            ]));
        }

        $topics = do_lorem_template('CNS_FORUM_TOPIC_WRAPPER', [
            'TYPE' => 'main_cns_involved_topics',
            'SORT' => '',
            'MAX' => '3',
            'MAY_CHANGE_MAX' => false,
            'BREADCRUMBS' => placeholder_breadcrumbs(),
            'ACTION_URL' => placeholder_url(),
            'BUTTONS' => '',
            'STARTER_TITLE' => '',
            'FORUM_NAME' => lorem_phrase(),
            'TOPICS' => $topics,
            'PAGINATION' => placeholder_pagination(),
            'MODERATOR_ACTIONS' => new Tempcode(),
            'ID' => null,
        ]);

        return lorem_globalise(do_lorem_template('BLOCK_MAIN_CNS_INVOLVED_TOPICS', [
            'BLOCK_ID' => lorem_word(),
            'TOPICS' => $topics,
            'BLOCK_PARAMS' => '',

            'START' => '0',
            'MAX' => '3',
            'START_PARAM' => 'x_start',
            'MAX_PARAM' => 'x_max',
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__block_side_cns_private_topics() : object
    {
        require_css('cns');

        require_lang('cns');

        $out = new Tempcode();
        foreach (placeholder_array() as $k => $v) {
            $out->attach(do_lorem_template('CNS_PRIVATE_TOPIC_LINK', [
                'TOPIC_URL' => placeholder_url(),
                'TITLE' => lorem_phrase(),
                'DATE' => placeholder_date(),
                'DATE_RAW' => placeholder_date_raw(),
                'LAST_POST_BY_POSTER_URL' => placeholder_url(),
                'LAST_POST_BY_USERNAME' => lorem_word(),
                'WITH_POSTER_URL' => placeholder_url(),
                'WITH_USERNAME' => lorem_word(),
                'WITH_POSTER_ID' => placeholder_first_admin_id(),
                'BY_POSTER_URL' => placeholder_url(),
                'BY_USERNAME' => lorem_word(),
                'BY_POSTER_ID' => placeholder_first_admin_id(),
                'TO_POSTER_URL' => placeholder_url(),
                'TO_USERNAME' => lorem_word(),
                'TO_POSTER_ID' => placeholder_first_admin_id(),
                '_NUM_POSTS' => placeholder_number(),
                'NUM_POSTS' => placeholder_number(),
                'HAS_READ' => true,
                'LAST_POST_BY_POSTER_ID' => placeholder_first_admin_id(),
            ]));
        }

        return lorem_globalise(do_lorem_template('BLOCK_SIDE_CNS_PRIVATE_TOPICS', [
            'BLOCK_ID' => lorem_word(),
            'SEND_URL' => placeholder_url(),
            'VIEW_URL' => placeholder_url(),
            'CONTENT' => $out,
            'FORUM_NAME' => lorem_phrase(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__cns_forum_topic_list_line() : object
    {
        require_css('cns');

        return lorem_globalise(do_lorem_template('CNS_FORUM_TOPIC_LIST_LINE', [
            'PRE' => lorem_phrase(),
            'TOPIC_TITLE' => lorem_phrase(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__cns_forum_list_line() : object
    {
        require_css('cns');

        return lorem_globalise(do_lorem_template('CNS_FORUM_LIST_LINE', [
            'PRE' => lorem_phrase(),
            'NAME' => lorem_word(),
            'CAT_BIT' => lorem_phrase(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__cns_forum_intro_question_popup() : object
    {
        require_css('cns');

        require_lang('cns');

        return lorem_globalise(do_lorem_template('CNS_FORUM_INTRO_QUESTION_POPUP', [
            'QUESTION' => lorem_phrase(),
            'ANSWER' => lorem_phrase(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__block_main_member_bar_guest() : object
    {
        require_css('cns');
        require_css('cns_header');
        require_lang('cns');

        $bar = do_lorem_template('CNS_GUEST_BAR', [
            'LOGIN_URL' => placeholder_url(),
            'JOIN_URL' => placeholder_url(),
            'FULL_LOGIN_URL' => placeholder_url(),
            'NEW_POSTS_URL' => placeholder_url(),
            'UNANSWERED_TOPICS_URL' => placeholder_url(),
        ]);

        return lorem_globalise(do_lorem_template('BLOCK_MAIN_MEMBER_BAR', [
            'BLOCK_ID' => lorem_word(),
            'BAR' => $bar,
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__block_main_member_bar_search() : object
    {
        require_css('cns');

        require_lang('cns');

        return lorem_globalise(do_template('MEMBER_BAR_SEARCH', []), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__block_main_member_bar_member() : object
    {
        require_css('cns');
        require_css('cns_header');
        require_lang('cns');

        $details = new Tempcode();
        $details->attach(do_lorem_template('BLOCK_SIDE_PERSONAL_STATS_LINE', [
            'KEY' => lorem_word(),
            'VALUE' => placeholder_number(),
            'RAW_VALUE' => placeholder_number(),
        ]));

        $links = new Tempcode();
        $links->attach(do_lorem_template('BLOCK_SIDE_PERSONAL_STATS_LINK', [
            'NAME' => lorem_word(),
            'URL' => placeholder_url(),
            'REL' => 'me',
        ]));

        $bar = do_template('CNS_MEMBER_BAR', [
            '_GUID' => 'd26f142a850c232ef14ec6de627cef4a',
            'AVATAR_URL' => placeholder_image_url(),
            'PROFILE_URL' => placeholder_url(),
            'USERNAME' => lorem_phrase(),
            'LOGOUT_URL' => placeholder_url(),
            '_NUM_POINTS_ADVANCE' => placeholder_number(),
            '_NUM_POINTS' => placeholder_number(),
            '_NUM_POSTS' => placeholder_number(),
            'NUM_POINTS_ADVANCE' => placeholder_number(),
            'NUM_POINTS' => placeholder_number(),
            'NUM_POSTS' => placeholder_number(),
            'PRIMARY_GROUP' => lorem_phrase(),
            'LAST_VISIT_DATE_RAW' => placeholder_date_raw(),
            'LAST_VISIT_DATE' => placeholder_date(),
            'PRIVATE_TOPIC_URL' => placeholder_url(),
            'NEW_POSTS_URL' => placeholder_url(),
            'UNREAD_TOPICS_URL' => placeholder_url(),
            'RECENTLY_READ_URL' => placeholder_url(),
            'INLINE_PERSONAL_POSTS_URL' => placeholder_url(),
            'UNANSWERED_TOPICS_URL' => placeholder_url(),
            'INVOLVED_TOPICS_URL' => placeholder_url(),
            'PT_EXTRA' => '',
            '_NUM_UNREAD_PTS' => placeholder_number(),
            '_NEW_TOPICS' => placeholder_number(),
            '_NEW_POSTS' => placeholder_number(),
            'NUM_UNREAD_PTS' => placeholder_number(),
            'NEW_TOPICS' => placeholder_number(),
            'NEW_POSTS' => placeholder_number(),
            'MAX_AVATAR_HEIGHT' => placeholder_number(),
            'LINKS' => $links,
            'LINKS_ECOMMERCE' => $links,
            'DETAILS' => $details,
        ]);

        return lorem_globalise(do_lorem_template('BLOCK_MAIN_MEMBER_BAR', [
            'BLOCK_ID' => lorem_word(),
            'BAR' => $bar,
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__block_main_bottom_bar() : object
    {
        require_css('cns');
        require_css('cns_footer');
        require_lang('cns');

        $birthdays = [];
        foreach (placeholder_array() as $k => $v) {
            $birthdays[] = [
                'AGE' => placeholder_number(),
                'PROFILE_URL' => placeholder_url(),
                'USERNAME' => lorem_word(),
                'BIRTHDAY_URL' => placeholder_url(),
            ];
        }

        $groups = [];
        foreach (placeholder_array() as $k => $v) {
            $groups[] = [
                'GID' => placeholder_numeric_id(),
                'GCOLOUR' => lorem_word(),
                'GTITLE' => $v,
            ];
        }

        return lorem_globalise(do_lorem_template('BLOCK_MAIN_BOTTOM_BAR', [
            'BLOCK_ID' => lorem_word(),
            'NEWEST_MEMBER_PROFILE_URL' => placeholder_url(),
            'NEWEST_MEMBER_USERNAME' => lorem_word(),
            '_NUM_MEMBERS' => placeholder_number(),
            '_NUM_TOPICS' => placeholder_number(),
            '_NUM_POSTS' => placeholder_number(),
            '_NUM_GUESTS_ONLINE' => placeholder_number(),
            '_NUM_MEMBERS_ONLINE' => placeholder_number(),
            'NUM_MEMBERS' => placeholder_number(),
            'NUM_TOPICS' => placeholder_number(),
            'NUM_POSTS' => placeholder_number(),
            'NUM_GUESTS_ONLINE' => placeholder_number(),
            'NUM_MEMBERS_ONLINE' => placeholder_number(),
            'BIRTHDAYS' => $birthdays,
            'USERS_ONLINE' => lorem_phrase(),
            'USERS_ONLINE_URL' => placeholder_url(),
            'GROUPS' => $groups,
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__cns_forum_intro_question_screen() : object
    {
        require_css('cns');

        require_lang('cns');

        return lorem_globalise(do_lorem_template('CNS_FORUM_INTRO_QUESTION_SCREEN', [
            'ANSWER' => lorem_phrase(),
            'TITLE' => lorem_screen_title(),
            'URL' => placeholder_url(),
            'QUESTION' => lorem_phrase(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__cns_forum() : object
    {
        require_css('cns');
        require_lang('cns');

        $buttons = new Tempcode();
        foreach (placeholder_array() as $k => $v) {
            $buttons->attach(do_lorem_template('BUTTON_SCREEN', [
                'REL' => lorem_word(),
                'URL' => placeholder_url(),
                'IMG' => 'buttons/proceed',
                'TITLE' => lorem_word(),
                'IMMEDIATE' => false,
            ]));
        }

        $topics = new Tempcode();
        foreach (placeholder_array() as $k => $v) {
            $marker = do_lorem_template('CNS_TOPIC_MARKER', [
                'ID' => placeholder_numeric_id() . strval($k),
            ]);

            $topic_row_links = [];
            $topic_row_links[] = [
                'URL' => placeholder_url(),
                'IMG' => placeholder_img_code('icons/cns_topic_modifiers'),
                'ALT' => lorem_phrase(),
            ];

            $topic_row_modifiers = [];
            $topic_row_modifiers[] = [
                'IMG' => placeholder_img_code('icons/cns_topic_modifiers'),
                'ALT' => lorem_phrase(),
            ];

            $b = do_lorem_template('CNS_USER_MEMBER', [
                'FIRST' => true,
                'COLOUR' => lorem_word(),
                'PROFILE_URL' => placeholder_url(),
                'USERNAME' => lorem_word(),
                'MEMBER_ID' => placeholder_numeric_id(),
                'AT' => lorem_phrase(),
            ]);

            $poster = do_lorem_template('CNS_PT_BETWEEN', [
                'A' => lorem_phrase(),
                'B' => $b,
            ]);

            $last_post = do_lorem_template('CNS_FORUM_TOPIC_ROW_LAST_POST', [
                'ID' => placeholder_numeric_id(),
                'DATE_RAW' => placeholder_date_raw(),
                'DATE' => placeholder_date(),
                'POSTER' => lorem_phrase(),
                'LAST_URL' => placeholder_url(),
            ]);

            $topics->attach(do_lorem_template('CNS_FORUM_TOPIC_ROW', [
                'BREADCRUMBS' => placeholder_breadcrumbs(),
                'RAW_TIME' => placeholder_date_raw(),
                'UNREAD' => lorem_phrase(),
                'ID' => placeholder_random_id(),
                'FORUM_ID' => placeholder_numeric_id(),
                'HOVER' => lorem_phrase(),
                'PAGES' => lorem_phrase(),
                'MARKER' => $marker,
                'TOPIC_ROW_LINKS' => $topic_row_links,
                'TOPIC_ROW_MODIFIERS' => $topic_row_modifiers,
                '_TOPIC_ROW_MODIFIERS' => [],
                'POST' => lorem_phrase(),
                'EMOTICON' => 'cns_emoticons/' . placeholder_img_code('cns_emoticons'),
                'DESCRIPTION' => lorem_paragraph(),
                'URL' => placeholder_url(),
                'TITLE' => lorem_phrase(),
                'POSTER' => $poster,
                '_POSTER' => placeholder_numeric_id(),
                '_NUM_POSTS' => placeholder_number(),
                '_NUM_VIEWS' => placeholder_number(),
                'NUM_POSTS' => placeholder_number(),
                'NUM_VIEWS' => placeholder_number(),
                'LAST_POST' => $last_post,
            ]));
        }

        $forum_groupings = new Tempcode();
        foreach (placeholder_array(1) as $k => $v) {
            $forums = new Tempcode();
            foreach (placeholder_array() as $_k => $_v) {
                $poster = do_lorem_template('CNS_USER_MEMBER', [
                    'FIRST' => true,
                    'USERNAME' => lorem_word(),
                    'MEMBER_ID' => placeholder_numeric_id(),
                    'PROFILE_URL' => placeholder_url(),
                    'AT' => lorem_phrase(),
                ]);

                $latest = do_lorem_template('CNS_FORUM_LATEST', [
                    'DATE' => placeholder_date(),
                    'DATE_RAW' => placeholder_date_raw(),
                    'TOPIC_URL' => placeholder_url(),
                    'TOPIC_TITLE' => lorem_phrase(),
                    'POSTER' => $poster,
                    'MEMBER_ID' => placeholder_numeric_id(),
                    'ID' => placeholder_numeric_id(),
                ]);

                $forums->attach(do_lorem_template('CNS_FORUM_IN_GROUPING', [
                    'ID' => placeholder_random_id(),
                    'NEW_POST_OR_NOT' => placeholder_img_code('icons/cns_general'),
                    'LANG_NEW_POST_OR_NOT' => lorem_word(),
                    'FORUM_NAME' => lorem_phrase(),
                    'FORUM_URL' => placeholder_url(),
                    'DESCRIPTION' => lorem_phrase(),
                    'NUM_POSTS' => placeholder_number(),
                    'NUM_TOPICS' => placeholder_number(),
                    'LATEST' => $latest,
                    'SUBFORUMS' => lorem_phrase(),
                    'EDIT_URL' => placeholder_url(),
                    'FORUM_RULES_URL' => placeholder_url(),
                    'INTRO_QUESTION_URL' => placeholder_url(),
                ]));
                if ($_k == 1) {
                    $forums->attach(do_lorem_template('CNS_PINNED_DIVIDER', []));
                }
            }

            $forum_groupings->attach(do_lorem_template('CNS_FORUM_GROUPING', [
                'GROUPING_ID' => placeholder_random_id(),
                'EXPAND_TYPE' => 'expand',
                'DISPLAY' => 'block',
                'GROUPING_TITLE' => lorem_phrase(),
                'GROUPING_DESCRIPTION' => lorem_phrase(),
                'FORUMS' => $forums,
            ]));
        }

        $filters = do_lorem_template('CNS_PT_FILTERS', [
            'FILTERS' => [],
            'RESET_URL' => placeholder_url(),
        ]);

        $pagination = placeholder_pagination();
        $topic_wrapper = do_lorem_template('CNS_FORUM_TOPIC_WRAPPER', [
            'TYPE' => 'browse',
            'MAX' => lorem_phrase(),
            'SORT' => lorem_phrase(),
            'MAY_CHANGE_MAX' => lorem_phrase(),
            'BUTTONS' => $buttons,
            'STARTER_TITLE' => lorem_phrase(),
            'PAGINATION' => $pagination,
            'MODERATOR_ACTIONS' => placeholder_options(),
            'ACTION_URL' => placeholder_url(),
            'TOPICS' => $topics,
            'FORUM_NAME' => lorem_word(),
            'BREADCRUMBS' => placeholder_breadcrumbs(),
            'ID' => placeholder_numeric_id(),
        ]);
        $content = do_lorem_template('CNS_FORUM', [
            'BREADCRUMBS' => placeholder_breadcrumbs(),
            'FILTERS' => $filters,
            'FORUM_NAME' => lorem_phrase(),
            'STARTER_TITLE' => lorem_phrase(),
            'BUTTONS' => $buttons,
            'TOPIC_WRAPPER' => $topic_wrapper,
            'FORUM_GROUPINGS' => $forum_groupings,
            'ID' => placeholder_numeric_id(),
            'DESCRIPTION' => lorem_phrase(),
            'MAIL_EMAIL_ADDRESS' => 'example@example.com',
            'PARENT_FORUM' => '',
        ]);

        $members_viewing = new Tempcode();
        foreach (placeholder_array() as $_k => $_v) {
            $members_viewing->attach(do_lorem_template('CNS_USER_MEMBER', [
                'FIRST' => true,
                'PROFILE_URL' => placeholder_url(),
                'USERNAME' => lorem_word(),
                'MEMBER_ID' => placeholder_numeric_id(),
                'AT' => lorem_phrase(),
                'COLOUR' => lorem_word(),
            ]));
        }

        return lorem_globalise(do_lorem_template('CNS_FORUM_SCREEN', [
            'TITLE' => lorem_screen_title(),
            'CONTENT' => $content,
            '_NUM_GUESTS' => placeholder_number(),
            '_NUM_MEMBERS' => placeholder_number(),
            'NUM_GUESTS' => placeholder_number(),
            'NUM_MEMBERS' => placeholder_number(),
            'MEMBERS_VIEWING' => $members_viewing,
            'ID' => placeholder_numeric_id(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__cns_vforum() : object
    {
        require_css('cns');

        require_lang('cns');

        return lorem_globalise(do_lorem_template('CNS_VFORUM_SCREEN', [
            'TITLE' => lorem_screen_title(),
            'CONTENT' => lorem_phrase(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__cns_whisper_choice_screen() : object
    {
        require_css('cns');

        require_lang('cns');

        return lorem_globalise(do_lorem_template('CNS_WHISPER_CHOICE_SCREEN', [
            'URL' => placeholder_url(),
            'TITLE' => lorem_screen_title(),
            'USERNAME' => lorem_word(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__cns_quote_fcomcode() : object
    {
        require_css('cns');

        return lorem_globalise(do_lorem_template('CNS_QUOTE_FCOMCODE', [
            'ID' => placeholder_numeric_id(),
            'TITLE' => lorem_phrase(),
            'POST' => lorem_phrase(),
            'BY' => lorem_phrase(),
            'BY_ID' => placeholder_numeric_id(),
        ], null, false, null, '.txt', 'text'), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__cns_member_pt_rules_screen() : object
    {
        require_css('cns');

        require_lang('cns');

        return lorem_globalise(do_lorem_template('CNS_MEMBER_PT_RULES_SCREEN', [
            'TITLE' => lorem_screen_title(),
            'USERNAME' => lorem_word(),
            'MEMBER_ID' => placeholder_numeric_id(),
            'URL' => placeholder_url(),
            'RULES' => lorem_phrase(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__cns_posting_screen_posts() : object
    {
        require_css('cns');
        require_lang('cns');

        $avatar = do_template('CNS_TOPIC_POST_AVATAR', ['_GUID' => 'acb4f6efb8b1ae27ae719c99ddcf8057', 'AVATAR' => placeholder_avatar()]);

        $last_edited = do_template('CNS_TOPIC_POST_LAST_EDITED', [
            '_GUID' => 'b75e04134d53797e2e91236536fa3ecd',
            'LAST_EDIT_DATE_RAW' => placeholder_date(),
            'LAST_EDIT_DATE' => placeholder_date(),
            'LAST_EDIT_PROFILE_URL' => placeholder_url(),
            'LAST_EDIT_USERNAME' => lorem_word(),
        ]);

        $map = [
            'ID' => placeholder_numeric_id(),
            'GIVE_CONTEXT' => true,
            'TOPIC_FIRST_POST_ID' => '',
            'TOPIC_FIRST_POSTER' => '',
            'POST_ID' => placeholder_random_id(),
            'URL' => placeholder_url(),
            'CLASS' => '',
            'EMPHASIS' => new Tempcode(),
            'FIRST_UNREAD' => '',
            'POSTER_TITLE' => lorem_phrase(),
            'POST_TITLE' => lorem_phrase(),
            'POST_DATE_RAW' => placeholder_date(),
            'POST_DATE' => placeholder_date(),
            'POST' => lorem_paragraph(),
            'TOPIC_ID' => placeholder_numeric_id(),
            'LAST_EDITED_RAW' => placeholder_date(),
            'LAST_EDITED' => $last_edited,
            'POSTER_ID' => placeholder_first_admin_id(),
            'POSTER' => lorem_word(),
            'POSTER_DETAILS' => new Tempcode(),
            'POST_AVATAR' => $avatar,
            'RANK_IMAGES' => new Tempcode(),
            'BUTTONS' => '',
            'SIGNATURE' => '',
            'NOT_VALIDATED' => '',
            'DESCRIPTION' => '',
            'PREVIEWING' => true,
            'RATING' => new Tempcode(),
        ];
        $_post = do_template('CNS_TOPIC_POST', $map);

        $posts = do_template('CNS_POST_BOX', [
            '_GUID' => '98382ac3c88efdcebe7d3491bf70866e',
            'BREADCRUMBS' => placeholder_breadcrumbs(),
            'POST' => $_post,
        ] + $map + ['ACTUAL_POST' => lorem_paragraph()]);

        return lorem_globalise(do_lorem_template('CNS_POSTING_SCREEN_POSTS', [
            'POSTS' => $posts,
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__cns_member_profile_posts() : object
    {
        require_css('cns');

        require_lang('cns');

        $topics = new Tempcode();
        if (addon_installed('cns_forum')) {
            foreach (placeholder_array() as $k => $v) {
                $marker = do_lorem_template('CNS_TOPIC_MARKER', [
                    'ID' => placeholder_numeric_id() . strval($k),
                ]);

                $topic_row_links = [];
                $topic_row_links[] = [
                    'URL' => placeholder_url(),
                    'IMG' => placeholder_img_code('icons/cns_topic_modifiers'),
                    'ALT' => lorem_phrase(),
                ];

                $topic_row_modifiers = [];
                $topic_row_modifiers[] = [
                    'IMG' => placeholder_img_code('icons/cns_topic_modifiers'),
                    'ALT' => lorem_phrase(),
                ];

                $b = do_lorem_template('CNS_USER_MEMBER', [
                    'FIRST' => true,
                    'COLOUR' => lorem_word(),
                    'PROFILE_URL' => placeholder_url(),
                    'USERNAME' => lorem_word(),
                    'MEMBER_ID' => placeholder_numeric_id(),
                    'AT' => lorem_phrase(),
                ]);

                $poster = do_lorem_template('CNS_PT_BETWEEN', [
                    'A' => lorem_phrase(),
                    'B' => $b,
                ]);

                $last_post = do_lorem_template('CNS_FORUM_TOPIC_ROW_LAST_POST', [
                    'ID' => placeholder_numeric_id(),
                    'DATE_RAW' => placeholder_date_raw(),
                    'DATE' => placeholder_date(),
                    'POSTER' => lorem_phrase(),
                    'LAST_URL' => placeholder_url(),
                ]);

                $topics->attach(do_lorem_template('CNS_FORUM_TOPIC_ROW', [
                    'BREADCRUMBS' => placeholder_breadcrumbs(),
                    'RAW_TIME' => placeholder_date_raw(),
                    'UNREAD' => lorem_phrase(),
                    'ID' => placeholder_numeric_id(),
                    'FORUM_ID' => placeholder_numeric_id(),
                    'HOVER' => lorem_phrase(),
                    'PAGES' => lorem_phrase(),
                    'MARKER' => $marker,
                    'TOPIC_ROW_LINKS' => $topic_row_links,
                    'TOPIC_ROW_MODIFIERS' => $topic_row_modifiers,
                    '_TOPIC_ROW_MODIFIERS' => [lorem_word()],
                    'POST' => lorem_phrase(),
                    'EMOTICON' => 'cns_emoticons/' . placeholder_img_code('cns_emoticons'),
                    'DESCRIPTION' => lorem_paragraph(),
                    'URL' => placeholder_url(),
                    'TITLE' => lorem_phrase(),
                    'POSTER' => $poster,
                    '_POSTER' => placeholder_numeric_id(),
                    '_NUM_POSTS' => placeholder_number(),
                    '_NUM_VIEWS' => placeholder_number(),
                    'NUM_POSTS' => placeholder_number(),
                    'NUM_VIEWS' => placeholder_number(),
                    'LAST_POST' => $last_post,
                ]));
            }
        }

        $pagination = placeholder_pagination();

        $buttons = lorem_phrase();

        $topic_wrapper = do_lorem_template('CNS_FORUM_TOPIC_WRAPPER', [
            'TYPE' => '',
            'MAX' => lorem_phrase(),
            'SORT' => lorem_phrase(),
            'MAY_CHANGE_MAX' => lorem_phrase(),
            'BUTTONS' => $buttons,
            'STARTER_TITLE' => lorem_phrase(),
            'PAGINATION' => $pagination,
            'MODERATOR_ACTIONS' => placeholder_options(),
            'ACTION_URL' => placeholder_url(),
            'TOPICS' => $topics,
            'FORUM_NAME' => lorem_phrase(),
            'BREADCRUMBS' => placeholder_breadcrumbs(),
            'ID' => placeholder_numeric_id(),
        ]);

        return lorem_globalise(do_lorem_template('CNS_MEMBER_PROFILE_POSTS', [
            'MEMBER_ID' => placeholder_numeric_id(),
            'TOPICS' => $topic_wrapper,
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__cns_member_profile_pts() : object
    {
        require_css('cns');
        require_lang('cns');

        $buttons = new Tempcode();
        foreach (placeholder_array() as $k => $v) {
            $buttons->attach(do_lorem_template('BUTTON_SCREEN', [
                'REL' => lorem_word(),
                'URL' => placeholder_url(),
                'IMG' => 'buttons/proceed',
                'TITLE' => lorem_word(),
                'IMMEDIATE' => false,
            ]));
        }

        $topics = new Tempcode();
        foreach (placeholder_array() as $k => $v) {
            $marker = do_lorem_template('CNS_TOPIC_MARKER', [
                'ID' => placeholder_numeric_id() . strval($k),
            ]);

            $topic_row_links = [];
            $topic_row_links[] = [
                'URL' => placeholder_url(),
                'IMG' => placeholder_img_code('icons/cns_topic_modifiers'),
                'ALT' => lorem_phrase(),
            ];

            $topic_row_modifiers = [];
            $topic_row_modifiers[] = [
                'IMG' => placeholder_img_code('icons/cns_topic_modifiers'),
                'ALT' => lorem_phrase(),
            ];

            $b = do_lorem_template('CNS_USER_MEMBER', [
                'FIRST' => true,
                'COLOUR' => lorem_word(),
                'PROFILE_URL' => placeholder_url(),
                'USERNAME' => lorem_word(),
                'MEMBER_ID' => placeholder_numeric_id(),
                'AT' => lorem_phrase(),
            ]);

            $poster = do_lorem_template('CNS_PT_BETWEEN', [
                'A' => lorem_phrase(),
                'B' => $b,
            ]);

            $last_post = do_lorem_template('CNS_FORUM_TOPIC_ROW_LAST_POST', [
                'ID' => placeholder_numeric_id(),
                'DATE_RAW' => placeholder_date_raw(),
                'DATE' => placeholder_date(),
                'POSTER' => lorem_phrase(),
                'LAST_URL' => placeholder_url(),
            ]);

            $topics->attach(do_lorem_template('CNS_FORUM_TOPIC_ROW', [
                'BREADCRUMBS' => placeholder_breadcrumbs(),
                'RAW_TIME' => placeholder_date_raw(),
                'UNREAD' => lorem_phrase(),
                'ID' => placeholder_random_id(),
                'FORUM_ID' => placeholder_numeric_id(),
                'HOVER' => lorem_phrase(),
                'PAGES' => lorem_phrase(),
                'MARKER' => $marker,
                'TOPIC_ROW_LINKS' => $topic_row_links,
                'TOPIC_ROW_MODIFIERS' => $topic_row_modifiers,
                '_TOPIC_ROW_MODIFIERS' => [],
                'POST' => lorem_phrase(),
                'EMOTICON' => 'cns_emoticons/' . placeholder_img_code('cns_emoticons'),
                'DESCRIPTION' => lorem_paragraph(),
                'URL' => placeholder_url(),
                'TITLE' => lorem_phrase(),
                'POSTER' => $poster,
                '_POSTER' => placeholder_numeric_id(),
                '_NUM_POSTS' => placeholder_number(),
                '_NUM_VIEWS' => placeholder_number(),
                'NUM_POSTS' => placeholder_number(),
                'NUM_VIEWS' => placeholder_number(),
                'LAST_POST' => $last_post,
            ]));
        }

        $forum_groupings = new Tempcode();

        $filters = do_lorem_template('CNS_PT_FILTERS', [
            'FILTERS' => [],
            'RESET_URL' => placeholder_url(),
        ]);

        $pagination = placeholder_pagination();
        $topic_wrapper = do_lorem_template('CNS_FORUM_TOPIC_WRAPPER', [
            'TYPE' => 'browse',
            'MAX' => lorem_phrase(),
            'SORT' => lorem_phrase(),
            'MAY_CHANGE_MAX' => lorem_phrase(),
            'BUTTONS' => $buttons,
            'STARTER_TITLE' => lorem_phrase(),
            'PAGINATION' => $pagination,
            'MODERATOR_ACTIONS' => placeholder_options(),
            'ACTION_URL' => placeholder_url(),
            'TOPICS' => $topics,
            'FORUM_NAME' => lorem_word(),
            'BREADCRUMBS' => placeholder_breadcrumbs(),
            'ID' => placeholder_numeric_id(),
        ]);
        $content = do_lorem_template('CNS_FORUM', [
            'BREADCRUMBS' => placeholder_breadcrumbs(),
            'FILTERS' => $filters,
            'FORUM_NAME' => lorem_phrase(),
            'STARTER_TITLE' => lorem_phrase(),
            'BUTTONS' => $buttons,
            'TOPIC_WRAPPER' => $topic_wrapper,
            'FORUM_GROUPINGS' => $forum_groupings,
            'ID' => placeholder_numeric_id(),
            'DESCRIPTION' => lorem_phrase(),
            'MAIL_EMAIL_ADDRESS' => 'example@example.com',
            'PARENT_FORUM' => '',
        ]);

        return lorem_globalise(do_template('CNS_MEMBER_PROFILE_PTS', [
            '_GUID' => '1bcb43f6ded79efdc08cde4ee80be15d',
            'CONTENT' => $content,
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__cns_topic_not_voted_checkboxes_wrap() : object
    {
        return $this->_tpl_preview__cns_topic_screen(false, true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__cns_topic_not_voted_radio_wrap() : object
    {
        return $this->_tpl_preview__cns_topic_screen(false, false);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__cns_topic_voted_wrap() : object
    {
        return $this->_tpl_preview__cns_topic_screen(true, false);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @param  boolean $voted Whether to show for a has-voted-poll or not
     * @param  boolean $checkboxes Whether to show checkboxes (as oppose to radio buttons). Irrelevant if !$voted.
     * @return Tempcode Preview
     */
    public function _tpl_preview__cns_topic_screen(bool $voted, bool $checkboxes) : object
    {
        require_css('cns');

        require_lang('cns');
        require_lang('cns_polls');

        require_javascript('editing');

        $warning_details = do_lorem_template('WARNING_BOX', [
            'WARNING' => lorem_phrase(),
        ]);

        foreach (placeholder_array() as $k => $v) {
            $members_viewing = new Tempcode();
            foreach (placeholder_array() as $_k => $_v) {
                $members_viewing->attach(do_lorem_template('CNS_USER_MEMBER', [
                    'FIRST' => $members_viewing->is_empty(),
                    'PROFILE_URL' => placeholder_url(),
                    'USERNAME' => lorem_word(),
                    'MEMBER_ID' => placeholder_numeric_id(),
                    'AT' => lorem_phrase(),
                    'COLOUR' => lorem_word(),
                ]));
            }

            $pagination = placeholder_pagination();

            $poll = new Tempcode();

            if (!$voted) {
                $answers = new Tempcode();
                foreach (placeholder_array() as $_k => $_v) {
                    $answers->attach(do_lorem_template($checkboxes ? 'CNS_TOPIC_POLL_ANSWER_TICK' : 'CNS_TOPIC_POLL_ANSWER_RADIO', [
                        'REAL_BUTTON' => '',
                        'ID' => placeholder_random_id(),
                        'ANSWER' => lorem_phrase(),
                        'I' => strval($_k),
                        'DISABLE_ANSWERS' => false
                    ]));
                }

                $buttons = do_lorem_template('CNS_TOPIC_POLL_BUTTON', [
                    'RESULTS_URL' => placeholder_url(),
                    'CAN_VIEW_RESULTS' => true,
                    'VOTE_WILL_FORFEIGHT' => true,
                    'VOTE_URL' => placeholder_url(),
                ]);

                $poll->attach(do_lorem_template('CNS_TOPIC_POLL', [
                    'ID' => placeholder_random_id(),
                    'QUESTION' => lorem_phrase(),
                    'PRIVATE' => true,
                    'ANSWERS' => $answers,
                    'BUTTONS' => $buttons,
                    'VOTE_URL' => placeholder_url(),
                    '_MINIMUM_SELECTIONS' => placeholder_number(),
                    '_MAXIMUM_SELECTIONS' => placeholder_number(),
                    'MINIMUM_SELECTIONS' => placeholder_number(),
                    'MAXIMUM_SELECTIONS' => placeholder_number(),
                    '_TOTAL_VOTES' => placeholder_number(),
                    'TOTAL_VOTES' => placeholder_number(),
                    'CLOSING_TIME' => placeholder_date_raw(),
                    'VOTES_REVEALED' => true,
                    'SHOW_BUTTONS' => true,
                    'FOOTER_MESSAGE' => new Tempcode(),
                    'REVOKE_URL' => placeholder_url(),
                    'NOT_VOTED' => true,
                    'IS_OPEN' => true,
                ]));
            } else {
                $answers = new Tempcode();
                $answers->attach(do_lorem_template('CNS_TOPIC_POLL_ANSWER_RESULTS', [
                    'ID' => placeholder_random_id(),
                    '_NUM_VOTES' => '10',
                    '_TOTAL_VOTES' => '45',
                    'NUM_VOTES' => '10',
                    'VOTING_POWER' => '30',
                    'TOTAL_VOTES' => '30',
                    'TOTAL_VOTING_POWER' => '90',
                    'POINT_WEIGHTING' => placeholder_number(),
                    'WIDTH' => '30',
                    'ANSWER' => lorem_phrase(),
                    'I' => '0',
                    'VOTERS_URL' => placeholder_url(),
                ]));
                $answers->attach(do_lorem_template('CNS_TOPIC_POLL_ANSWER_RESULTS', [
                    'ID' => placeholder_random_id(),
                    '_NUM_VOTES' => '15',
                    '_TOTAL_VOTES' => '45',
                    'NUM_VOTES' => '15',
                    'VOTING_POWER' => '45',
                    'TOTAL_VOTES' => '45',
                    'TOTAL_VOTING_POWER' => '135',
                    'POINT_WEIGHTING' => placeholder_number(),
                    'WIDTH' => '45',
                    'ANSWER' => lorem_phrase(),
                    'I' => '1',
                    'VOTERS_URL' => placeholder_url(),
                ]));
                $answers->attach(do_lorem_template('CNS_TOPIC_POLL_ANSWER_RESULTS', [
                    'ID' => placeholder_random_id(),
                    '_NUM_VOTES' => '20',
                    '_TOTAL_VOTES' => '45',
                    'NUM_VOTES' => '20',
                    'VOTING_POWER' => '60',
                    'TOTAL_VOTES' => '60',
                    'TOTAL_VOTING_POWER' => '180',
                    'POINT_WEIGHTING' => placeholder_number(),
                    'WIDTH' => '60',
                    'ANSWER' => lorem_phrase(),
                    'I' => '2',
                    'VOTERS_URL' => placeholder_url(),
                ]));

                $buttons = do_lorem_template('CNS_TOPIC_POLL_BUTTON_RESULTS', [
                    'REVOKE_URL' => placeholder_url(),
                ]);

                $poll->attach(do_lorem_template('CNS_TOPIC_POLL_VIEW_RESULTS', [
                    'ID' => placeholder_random_id(),
                    'PRIVATE' => true,
                    'QUESTION' => lorem_phrase(),
                    'ANSWERS' => $answers,
                    'BUTTONS' => $buttons,
                    'VOTE_URL' => placeholder_url(),
                    '_MINIMUM_SELECTIONS' => placeholder_number(),
                    '_MAXIMUM_SELECTIONS' => placeholder_number(),
                    'MINIMUM_SELECTIONS' => placeholder_number(),
                    'MAXIMUM_SELECTIONS' => placeholder_number(),
                    '_TOTAL_VOTES' => placeholder_number(),
                    'TOTAL_VOTES' => placeholder_number(),
                    'TOTAL_VOTING_POWER' => placeholder_number(),
                    'POINT_WEIGHTING' => placeholder_number(),
                    'CLOSING_TIME' => placeholder_date_raw(),
                    'VOTES_REVEALED' => true,
                    'SHOW_BUTTONS' => true,
                    'FOOTER_MESSAGE' => new Tempcode(),
                    'REVOKE_URL' => placeholder_url(),
                    'NOT_VOTED' => true,
                    'IS_OPEN' => true,
                ]));
            }


            // Buttons
            $buttons = new Tempcode();
            foreach (placeholder_array(1) as $_k => $_v) {
                $buttons->attach(do_lorem_template('BUTTON_SCREEN', [
                    'REL' => lorem_word(),
                    'IMMEDIATE' => null,
                    'URL' => placeholder_url(),
                    'IMG' => 'buttons/proceed',
                    'TITLE' => lorem_word(),
                ]));
            }

            // Posts...

            $posts = new Tempcode();

            $first_unread = do_lorem_template('CNS_TOPIC_FIRST_UNREAD');

            $last_edited = do_lorem_template('CNS_TOPIC_POST_LAST_EDITED', [
                'LAST_EDIT_DATE_RAW' => placeholder_date_raw(),
                'LAST_EDIT_DATE' => placeholder_date(),
                'LAST_EDIT_PROFILE_URL' => placeholder_url(),
                'LAST_EDIT_USERNAME' => lorem_word(),
            ]);

            $custom_fields = do_lorem_template('CNS_MEMBER_BOX_CUSTOM_FIELD', [
                'NAME' => lorem_phrase(),
                'RAW' => placeholder_ip(),
                'VALUE' => placeholder_ip(),
                'MEMBER_ID' => placeholder_numeric_id(),
            ]);
            $poster_details = do_lorem_template('CNS_GUEST_DETAILS', [
                'CUSTOM_FIELDS' => $custom_fields,
            ]);

            $map = [
                'GIVE_CONTEXT' => false,
                'MEMBER_ID' => placeholder_numeric_id(),
                'USERNAME' => lorem_phrase(),
                '_POSTS' => placeholder_number(),
                'POSTS' => placeholder_number(),
                '_POINTS' => placeholder_number(),
                'POINTS' => placeholder_number(),
                'JOIN_DATE_RAW' => placeholder_date_raw(),
                'JOIN_DATE' => placeholder_date(),
                'PRIMARY_GROUP_NAME' => lorem_phrase(),
                'SECONDARY_GROUPS' => [
                    lorem_word(),
                ],
                'CUSTOM_FIELDS' => lorem_phrase(),
                'ONLINE' => false,
                'AVATAR_URL' => placeholder_image_url(),
                'DOB_LABEL' => do_lang('DATE_OF_BIRTH'),
                'DOB' => placeholder_date(),
                '_DOB' => placeholder_date_raw(),
                '_DOB_CENSORED' => placeholder_date_raw(),
            ];
            if (addon_installed('cns_warnings')) {
                $map['_NUM_WARNINGS'] = placeholder_number();
                $map['NUM_WARNINGS'] = placeholder_number();
            }
            $poster_details_mem = do_lorem_template('CNS_MEMBER_BOX', $map);

            $poster = do_lorem_template('CNS_POSTER_MEMBER', [
                'ID' => placeholder_random_id(),
                'POSTER_DETAILS' => $poster_details_mem,
                'PROFILE_URL' => placeholder_url(),
                'POSTER_USERNAME' => lorem_word(),
                'HIGHLIGHT_NAME' => false,
                'ONLINE' => false,
            ]);

            $post_avatar = do_lorem_template('CNS_TOPIC_POST_AVATAR', [
                'AVATAR' => placeholder_image_url(),
            ]);

            $rank_images = do_lorem_template('CNS_RANK_IMAGE', [
                'GROUP_NAME' => lorem_phrase(),
                'USERNAME' => lorem_word(),
                'IMG' => 'cns_rank_images/' . placeholder_img_code('cns_rank_images'),
                'IS_LEADER' => lorem_phrase(),
            ]);

            // Buttons
            $buttons = new Tempcode();
            foreach (placeholder_array(1) as $_k => $_v) {
                $buttons->attach(do_lorem_template('BUTTON_SCREEN_ITEM', [
                    'REL' => lorem_word(),
                    'IMMEDIATE' => '',
                    'URL' => placeholder_url(),
                    'IMG' => 'buttons/proceed',
                    'FULL_TITLE' => lorem_phrase(),
                    'TITLE' => lorem_word(),
                ]));
            }

            $posts->attach(do_lorem_template('CNS_TOPIC_POST', [
                'ID' => placeholder_random_id(),
                'GIVE_CONTEXT' => false,
                'TOPIC_FIRST_POST_ID' => placeholder_numeric_id(),
                'TOPIC_FIRST_POSTER' => lorem_phrase(),
                'POST_ID' => placeholder_random_id(),
                'FORUM_ID' => placeholder_numeric_id(),
                'URL' => placeholder_url(),
                'CLASS' => lorem_phrase(),
                'EMPHASIS' => lorem_phrase(),
                'FIRST_UNREAD' => $first_unread,
                'POSTER_TITLE' => lorem_word(),
                'POST_TITLE' => lorem_word(),
                'POST_DATE_RAW' => placeholder_date_raw(),
                'POST_DATE' => placeholder_date(),
                'POST' => lorem_phrase(),
                'TOPIC_ID' => placeholder_numeric_id(),
                'LAST_EDITED_RAW' => lorem_phrase(),
                'LAST_EDITED' => $last_edited,
                'POSTER_ID' => placeholder_numeric_id(),
                'POSTER' => $poster,
                'POSTER_DETAILS' => $poster_details,
                'POST_AVATAR' => $post_avatar,
                'RANK_IMAGES' => $rank_images,
                'BUTTONS' => $buttons,
                'SIGNATURE' => lorem_phrase(),
                'NOT_VALIDATED' => lorem_phrase(),
                'DESCRIPTION' => lorem_phrase(),
            ]));
        }

        // Buttons
        $buttons = new Tempcode();
        foreach (placeholder_array(1) as $k => $v) {
            $buttons->attach(do_lorem_template('BUTTON_SCREEN', [
                'REL' => lorem_word(),
                'IMMEDIATE' => '',
                'URL' => placeholder_url(),
                'IMG' => 'buttons/proceed',
                'TITLE' => lorem_word(),
            ]));
        }

        return lorem_globalise(do_lorem_template('CNS_TOPIC_SCREEN', [
            'TITLE' => lorem_screen_title(),
            'THREADED' => false,
            'ID' => placeholder_numeric_id(),
            '_TITLE' => lorem_phrase(),
            'WARNING_DETAILS' => $warning_details,
            'MAX' => lorem_phrase(),
            'ACTION_URL' => placeholder_url(),
            '_NUM_GUESTS' => placeholder_number(),
            '_NUM_MEMBERS' => placeholder_number(),
            'NUM_GUESTS' => placeholder_number(),
            'NUM_MEMBERS' => placeholder_number(),
            'MEMBERS_VIEWING' => $members_viewing,
            'PAGINATION' => $pagination,
            'MODERATOR_ACTIONS' => placeholder_options(),
            'MARKED_POST_ACTIONS' => placeholder_options(),
            'QUICK_REPLY' => placeholder_comments_form(false, true),
            'BREADCRUMBS' => placeholder_breadcrumbs(),
            'POLL' => $poll,
            'BUTTON_SCREENS' => $buttons,
            'POSTS' => $posts,
            'MAY_CHANGE_MAX' => lorem_word(),
            'LAST_POSTER' => lorem_word(),
            'FORUM_ID' => placeholder_numeric_id(),
            'TICKET_FORUM' => false,
            'SERIALIZED_OPTIONS' => '',
            'HASH' => '',
            'IS_ALREADY_READ' => true,
            'MAY_DOUBLE_POST' => true,
        ]), null, '', true);
    }

    /**
     * Uninstall default content.
     */
    public function uninstall_test_content()
    {
        if (get_forum_type() != 'cns') {
            return;
        }

        require_code('cns_topics_action');
        require_code('cns_topics_action2');
        require_code('cns_posts_action');
        require_code('cns_posts_action2');
        require_code('cns_forums');
        require_code('cns_topics');
        require_lang('cns');

        // Normal topic
        $to_delete = $GLOBALS['FORUM_DB']->query_select('f_topics', ['id'], ['t_cache_first_title' => lorem_phrase()]);
        foreach ($to_delete as $record) {
            cns_delete_topic($record['id']);
        }

        // Closed topic
        $to_delete = $GLOBALS['FORUM_DB']->query_select('f_topics', ['id'], ['t_cache_first_title' => lorem_phrase() . ' (' . do_lang('MODIFIER_closed') . ')']);
        foreach ($to_delete as $record) {
            cns_delete_topic($record['id']);
        }

        // Not validated topic
        $to_delete = $GLOBALS['FORUM_DB']->query_select('f_topics', ['id'], ['t_cache_first_title' => lorem_phrase() . ' (' . do_lang('MODIFIER_not_validated') . ')']);
        foreach ($to_delete as $record) {
            cns_delete_topic($record['id']);
        }

        // Pinned topic
        $to_delete = $GLOBALS['FORUM_DB']->query_select('f_topics', ['id'], ['t_cache_first_title' => lorem_phrase() . ' (' . do_lang('MODIFIER_pinned') . ')']);
        foreach ($to_delete as $record) {
            cns_delete_topic($record['id']);
        }

        // Announcement topic
        $to_delete = $GLOBALS['FORUM_DB']->query_select('f_topics', ['id'], ['t_cache_first_title' => lorem_phrase() . ' (' . do_lang('MODIFIER_announcement') . ')']);
        foreach ($to_delete as $record) {
            cns_delete_topic($record['id']);
        }

        // Poll topic
        require_code('cns_polls_action');

        $to_delete = $GLOBALS['FORUM_DB']->query_select('f_topics', ['id'], ['t_cache_first_title' => lorem_phrase() . ' (' . do_lang('MODIFIER_poll') . ')']);
        foreach ($to_delete as $record) {
            cns_delete_topic($record['id']);
        }
    }

    /**
     * Install default content.
     */
    public function install_test_content()
    {
        if (get_forum_type() != 'cns') {
            return;
        }

        require_code('cns_topics_action');
        require_code('cns_topics_action2');
        require_code('cns_posts_action');
        require_code('cns_posts_action2');
        require_code('cns_forums');
        require_code('cns_topics');
        require_lang('cns');

        // Normal topic
        $topic_id = cns_make_topic(db_get_first_id(), lorem_phrase(), 'cns_emoticons/lol', 1, 1, 0, 0, null, null, false);
        cns_make_post($topic_id, lorem_phrase(), lorem_chunk(), 0, true, 1, 0, null, null, null, null, null, null, null, false, true, null, true, lorem_phrase());
        cns_make_post($topic_id, lorem_phrase(), lorem_chunk(), 0, false, 1, 0, null, null, null, db_get_first_id() + 2, get_member(), null, null, false, true, null, true, lorem_phrase()); // With inline PP

        // Closed topic
        $topic_id = cns_make_topic(db_get_first_id(), lorem_phrase(), '', 1, 0, 0, 0, null, null, false);
        cns_make_post($topic_id, lorem_phrase(), lorem_chunk(), 0, true, 1, 0, null, null, null, null, null, null, null, false, true, null, true, lorem_phrase() . ' (' . do_lang('MODIFIER_closed') . ')');
        cns_make_post($topic_id, lorem_phrase(), lorem_chunk(), 0, true, 1, 1, null, null, null, null, null, null, null, false, true, null, true, lorem_phrase() . ' (' . do_lang('MODIFIER_closed') . ')'); // With emphasised post

        // Not validated topic
        $topic_id = cns_make_topic(db_get_first_id(), lorem_phrase(), '', 0, 1, 0, 0, null, null, false);
        cns_make_post($topic_id, lorem_phrase(), lorem_chunk(), 0, true, 1, 0, null, null, null, null, null, null, null, false, true, null, true, lorem_phrase() . ' (' . do_lang('MODIFIER_not_validated') . ')');

        // Pinned topic
        $topic_id = cns_make_topic(db_get_first_id(), lorem_phrase(), '', 1, 1, 1, 0, null, null, false);
        cns_make_post($topic_id, lorem_phrase(), lorem_chunk(), 0, true, 1, 0, null, null, null, null, null, null, null, false, true, null, true, lorem_phrase() . ' (' . do_lang('MODIFIER_pinned') . ')');

        // Announcement topic
        $topic_id = cns_make_topic(db_get_first_id(), lorem_phrase(), '', 1, 1, 1, 1, null, null, false);
        cns_make_post($topic_id, lorem_phrase(), lorem_chunk(), 0, true, 1, 0, null, null, null, null, null, null, null, false, true, null, true, lorem_phrase() . ' (' . do_lang('MODIFIER_announcement') . ')');

        require_code('cns_polls_action');

        // Poll topic
        $topic_id = cns_make_topic(db_get_first_id(), lorem_phrase(), '', 1, 1, 0, 0, null, null, false);
        cns_make_post($topic_id, lorem_phrase(), lorem_chunk(), 0, true, 1, 0, null, null, null, null, null, null, null, false, true, null, true, lorem_phrase() . ' (' . do_lang('MODIFIER_poll') . ')');
        cns_make_poll($topic_id, lorem_phrase(), 0, 1, 1, 1, 0, [lorem_phrase() . ' 1', lorem_phrase() . ' 2'], 0, 0, 1, 1, false);

        // In "General Chat"
        $general_chat_forum_id = $GLOBALS['FORUM_DRIVER']->forum_id_from_name(do_lang('DEFAULT_FORUM_TITLE'));
        if ($general_chat_forum_id !== null) {
            $topic_id = cns_make_topic($general_chat_forum_id, lorem_phrase(), '', 1, 1, 0, 0, null, null, false);
            cns_make_post($topic_id, lorem_phrase(), lorem_chunk(), 0, true, 1, 0, null, null, null, null, null, null, null, false, true, null, true, lorem_phrase());
        }

        // Private topic
        $test_member_id = $GLOBALS['FORUM_DRIVER']->get_member_from_username('test');
        if ($test_member_id !== null) {
            $topic_id = cns_make_topic(null, lorem_phrase(), '', 1, 1, 0, 0, $test_member_id, get_member(), false);
            cns_make_post($topic_id, lorem_phrase(), lorem_chunk(), 0, true, 1, 0, null, null, null, $test_member_id, null, null, null, false, true, null, true, lorem_phrase());
        }
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__cns_post_from_mailing_list() : object
    {
        require_lang('cns');
        require_lang('cns_mailinglists');

        return lorem_globalise(do_lorem_template('CNS_POST_FROM_MAILING_LIST', [
            'UNCONFIRMED_MEMBER_NOTICE' => true,
            'POST' => lorem_paragraph(),
            'USERNAME' => lorem_phrase(),
        ], null, false, null, '.txt', 'text'), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__cns_topic_poll_voters() : object
    {
        require_code('cns_polls');
        require_code('templates_results_table');
        require_lang('cns_polls');

        $field_titles = [
            do_lang_tempcode('DATE_TIME'),
            do_lang_tempcode('MEMBER'),
            protect_from_escaping(do_lorem_template('HELP_ICON_PHRASE', [
                'LABEL' => do_lang_tempcode('VOTING_POWER'),
                'TOOLTIP' => do_lang_tempcode('DESCRIPTION_VOTING_POWER'),
            ])),
            do_lang_tempcode('ANSWER')
        ];
        $sortables = [
            'pv_date_time' => do_lang_tempcode('DATE_TIME'),
            'pv_member_id' => do_lang_tempcode('MEMBER'),
            'voting_power' => do_lang_tempcode('VOTING_POWER'),
            'answer' => do_lang_tempcode('ANSWER')
        ];
        $header_row = results_header_row($field_titles, $sortables);
        $footer_row = results_footer_row([
            '',
            '',
            do_lang_tempcode('TOTAL_VOTING_POWER', placeholder_number()),
            ''
        ]);

        $member = do_lorem_template('RESULTS_TABLE_MEMBER_CELL', [
            'PROFILE_URL' => placeholder_avatar(),
            'USERNAME' => lorem_phrase(),
            'MEMBER_ID' => placeholder_numeric_id()
        ]);

        $_rows = new Tempcode();
        for ($i = 0; $i < 3; $i++) {
            $_rows->attach(results_entry([
                placeholder_date(),
                $member,
                do_lorem_template('CNS_TOPIC_POLL_VOTING_POWER', [
                    'GUID' => '690423ddee434df79de7bead4fb5e79e',
                    'EQUATION' => with_whitespace(do_lang_tempcode('VOTING_POWER_EQUATION', 'maximumVotingPower', 'offset', ['multiplier', 'points', 'logBase'])),
                    'EQUATION_WITH_NUMBERS' => do_lang_tempcode('VOTING_POWER_EQUATION', placeholder_number(), placeholder_number(), [placeholder_number(), placeholder_number(), placeholder_number()]),
                    'VOTING_POWER' => placeholder_number()
                ]),
                lorem_phrase()
            ], true));
        }

        $table = results_table(lorem_phrase(), 0, 'start', 3, 'max', 3, $header_row, $_rows, $sortables, null, null, 'sort', null, [], null, null, '866198fc95db4ad3abc322d9ad144875', false, null, false, false, $footer_row);

        return lorem_globalise(do_lorem_template('CNS_TOPIC_POLL_VOTERS_SCREEN', [
            '_GUID' => 'd75c813e372c3ca8d1204609e54c9d65',
            'TABLE' => $table,
            'TITLE' => lorem_phrase(),
            'PAGINATION' => placeholder_pagination(),
        ]), null, '', true);
    }
}
