<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    core_abstract_components
 */

/**
 * Hook class.
 */
class Hook_addon_registry_core_abstract_components
{
    /**
     * Get a list of file permissions to set.
     *
     * @param  boolean $runtime Whether to include wildcards represented runtime-created chmoddable files
     * @return array File permissions to set
     */
    public function get_chmod_array(bool $runtime = false) : array
    {
        return [];
    }

    /**
     * Get the version of the software this addon is for (used in generating the TAR filename).
     *
     * @return float Version number
     */
    public function get_version() : float
    {
        return cms_version_number();
    }

    /**
     * Get the minimum required version of the website software needed to use this addon.
     *
     * @return float Minimum required website software version
     */
    public function get_min_cms_version() : float
    {
        return 11.0;
    }

    /**
     * Get the maximum compatible version of the website software to use this addon.
     *
     * @return ?float Maximum compatible website software version (null: no maximum version currently)
     */
    public function get_max_cms_version() : ?float
    {
        return 11.9;
    }

    /**
     * Get the addon category.
     *
     * @return string The category
     */
    public function get_category() : string
    {
        return 'Architecture';
    }

    /**
     * Get the description of the addon.
     *
     * @return string Description of the addon
     */
    public function get_description() : string
    {
        return 'Core rendering functionality.';
    }

    /**
     * Get a list of tutorials that apply to this addon.
     *
     * @return array List of tutorials
     */
    public function get_applicable_tutorials() : array
    {
        return [
            'tut_themes',
        ];
    }

    /**
     * Get a mapping of dependency types.
     *
     * @return array A structure specifying dependency information
     */
    public function get_dependencies() : array
    {
        return [
            'requires' => [],
            'recommends' => [],
            'conflicts_with' => [],
        ];
    }

    /**
     * Explicitly say which icon should be used.
     *
     * @return URLPATH Icon
     */
    public function get_default_icon() : string
    {
        return 'themes/default/images/icons/admin/component.svg';
    }

    /**
     * Get a list of files that belong to this addon.
     *
     * @return array List of files
     */
    public function get_file_list() : array
    {
        return [
            'data/change_detection.php',
            'data/edit_ping.php',
            'data/fractional_edit.php',
            'sources/graphs.php',
            'sources/hooks/systems/addon_registry/core_abstract_components.php',
            'sources/hooks/systems/change_detection/.htaccess',
            'sources/hooks/systems/change_detection/index.html',
            'sources_custom/hooks/systems/change_detection/.htaccess',
            'sources_custom/hooks/systems/change_detection/index.html',
            'themes/default/css/graphs.css',
            'themes/default/javascript/charts.js',
            'themes/default/javascript/fractional_edit.js',
            'themes/default/templates/BLOCK_TOOLTIP.tpl',
            'themes/default/templates/BUTTON_SCREEN.tpl',
            'themes/default/templates/BUTTON_SCREEN_ITEM.tpl',
            'themes/default/templates/FRACTIONAL_EDIT.tpl',
            'themes/default/templates/GRAPH_BAR_CHART.tpl',
            'themes/default/templates/GRAPH_BUBBLE_BAR_CHART.tpl',
            'themes/default/templates/GRAPH_LINE_CHART.tpl',
            'themes/default/templates/GRAPH_PIE_CHART.tpl',
            'themes/default/templates/GRAPH_SCATTER_DIAGRAM.tpl',
            'themes/default/templates/GRAPH_STACKED_BAR_CHART.tpl',
            'themes/default/templates/HANDLE_CONFLICT_RESOLUTION.tpl',
            'themes/default/templates/IMG_THUMB.tpl',
            'themes/default/templates/INLINE_TOOLTIP.tpl',
            'themes/default/templates/POST.tpl',
            'themes/default/templates/POST_CHILD_LOAD_LINK.tpl',
            'themes/default/templates/STAFF_ACTIONS.tpl',
            'themes/default/templates/STANDARDBOX_accordion.tpl',
            'themes/default/templates/STANDARDBOX_accordion_wrap.tpl',
            'themes/default/templates/STANDARDBOX_default.tpl',
        ];
    }

    /**
     * Get mapping between template names and the method of this class that can render a preview of them.
     *
     * @return array The mapping
     */
    public function tpl_previews() : array
    {
        return [
            'templates/BUTTON_SCREEN_ITEM.tpl' => 'button_screen_item',
            'templates/FRACTIONAL_EDIT.tpl' => 'administrative__fractional_edit',
            'templates/INLINE_TOOLTIP.tpl' => 'inline_tooltip',
            'templates/BLOCK_TOOLTIP.tpl' => 'block_tooltip',
            'templates/IMG_THUMB.tpl' => 'img_thumb',
            'templates/BUTTON_SCREEN.tpl' => 'button_screen',
            'templates/STANDARDBOX_default.tpl' => 'standardbox_default',
            'templates/STANDARDBOX_accordion.tpl' => 'standardbox_accordion',
            'templates/STANDARDBOX_accordion_wrap.tpl' => 'standardbox_accordion',
            'templates/HANDLE_CONFLICT_RESOLUTION.tpl' => 'administrative__handle_conflict_resolution',
            'templates/STAFF_ACTIONS.tpl' => 'staff_actions',
            'templates/GRAPH_SCATTER_DIAGRAM.tpl' => 'graph_scatter_diagram',
            'templates/GRAPH_LINE_CHART.tpl' => 'graph_line_chart',
            'templates/GRAPH_PIE_CHART.tpl' => 'graph_pie_chart',
            'templates/GRAPH_BAR_CHART.tpl' => 'graph_bar_chart',
            'templates/GRAPH_STACKED_BAR_CHART.tpl' => 'graph_stacked_bar_chart',
            'templates/GRAPH_BUBBLE_BAR_CHART.tpl' => 'graph_bubble_bar_chart',
        ];
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__staff_actions() : object
    {
        return lorem_globalise(do_lorem_template('STAFF_ACTIONS', [
            '1_TITLE' => lorem_phrase(),
            '1_URL' => placeholder_url(),
            '2_TITLE' => lorem_phrase(),
            '2_URL' => placeholder_url(),
            '3_TITLE' => lorem_phrase(),
            '3_URL' => placeholder_url(),
            '4_TITLE' => lorem_phrase(),
            '4_URL' => placeholder_url(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__button_screen_item() : object
    {
        return lorem_globalise(do_lorem_template('BUTTON_SCREEN_ITEM', [
            'REL' => lorem_phrase(),
            'IMMEDIATE' => lorem_phrase(),
            'URL' => placeholder_url(),
            'TITLE' => lorem_word(),
            'FULL_TITLE' => lorem_phrase(),
            'IMG' => 'admin/edit',
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__administrative__fractional_edit() : object
    {
        return lorem_globalise(do_lorem_template('FRACTIONAL_EDIT', [
            'VALUE' => lorem_phrase(),
            'URL' => placeholder_url(),
            'EDIT_TEXT' => lorem_sentence_html(),
            'EDIT_PARAM_NAME' => lorem_word(),
            'EXPLICIT_EDITING_LINKS' => true,
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__inline_tooltip() : object
    {
        return lorem_globalise(do_lorem_template('INLINE_TOOLTIP', [
            'LABEL' => lorem_sentence_html(),
            'TOOLTIP' => lorem_sentence_html(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__block_tooltip() : object
    {
        return lorem_globalise(do_lorem_template('BLOCK_TOOLTIP', [
            'LABEL' => lorem_phrase(),
            'TOOLTIP' => lorem_phrase(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__img_thumb() : object
    {
        return lorem_globalise(do_lorem_template('IMG_THUMB', [
            'JS_TOOLTIP' => lorem_phrase(),
            'CAPTION' => lorem_phrase(),
            'URL' => placeholder_image_url(),
            'WIDTH' => placeholder_number(),
            'HEIGHT' => placeholder_number(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__button_screen() : object
    {
        return lorem_globalise(do_lorem_template('BUTTON_SCREEN', [
            'IMMEDIATE' => true,
            'URL' => placeholder_url(),
            'TITLE' => lorem_word(),
            'IMG' => 'buttons/proceed',
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__standardbox_default() : object
    {
        return $this->_tpl_preview__standardbox('default');
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__standardbox_accordion() : object
    {
        $links = [];
        foreach (placeholder_array() as $k => $v) {
            $links[] = placeholder_link();
        }

        $meta = [];
        foreach (placeholder_array() as $k => $v) {
            $meta[] = [
                'KEY' => strval($k),
                'VALUE' => $v,
            ];
        }

        $boxes = new Tempcode();
        $box = do_lorem_template('STANDARDBOX_accordion', [
            'CONTENT' => lorem_sentence(),
            'LINKS' => $links,
            'META' => $meta,
            'OPTIONS' => placeholder_array(),
            'TITLE' => lorem_phrase(),
            'TOP_LINKS' => placeholder_link(),
            'WIDTH' => '',
        ]);
        $boxes->attach($box);
        $box = do_lorem_template('STANDARDBOX_accordion', [
            'CONTENT' => lorem_sentence(),
            'LINKS' => $links,
            'META' => $meta,
            'OPTIONS' => placeholder_array(),
            'TITLE' => lorem_phrase(),
            'TOP_LINKS' => placeholder_link(),
            'WIDTH' => '',
        ]);
        $boxes->attach($box);

        $wrapped = do_lorem_template('STANDARDBOX_accordion_wrap', [
            'CONTENT' => $boxes,
            'TITLE' => '',
        ]);

        return lorem_globalise($wrapped, null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @param  string $type View type
     * @return Tempcode Preview
     */
    public function _tpl_preview__standardbox(string $type) : object
    {
        $links = [];
        foreach (placeholder_array() as $k => $v) {
            $links[] = placeholder_link();
        }

        $meta = [];
        foreach (placeholder_array() as $k => $v) {
            $meta[] = [
                'KEY' => strval($k),
                'VALUE' => $v,
            ];
        }

        $box = do_lorem_template('STANDARDBOX_' . $type, [
            'CONTENT' => lorem_sentence(),
            'LINKS' => $links,
            'META' => $meta,
            'OPTIONS' => placeholder_array(),
            'TITLE' => '',
            'TOP_LINKS' => placeholder_link(),
            'WIDTH' => '',
        ]);

        return lorem_globalise($box, null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__administrative__handle_conflict_resolution() : object
    {
        return lorem_globalise(do_lorem_template('HANDLE_CONFLICT_RESOLUTION', []));
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__graph_scatter_diagram() : object
    {
        $datapoints = [
            [
                'x' => 1,
                'y' => 1,
                'tooltip' => lorem_sentence(),
            ],
            [
                'x' => 3,
                'y' => 2,
                'tooltip' => lorem_sentence(),
            ],
        ];

        $_datapoints = [];
        foreach ($datapoints as $p) {
            $_datapoints[] = [
                'X' => strval($p['x']),
                'Y' => strval($p['y']),
                'R' => '1',
                'TOOLTIP' => $p['tooltip'],
            ];
        }

        $datasets = [];
        $datasets[] = [
            'DATAPOINTS' => $_datapoints,
            'CATEGORY' => '',
            'COLOR' => '#FF0000',
        ];

        return lorem_globalise(do_lorem_template('GRAPH_SCATTER_DIAGRAM', [
            'ID' => lorem_word(),
            'WIDTH' => '500px',
            'HEIGHT' => '500px',
            'X_AXIS_LABEL' => lorem_phrase(),
            'Y_AXIS_LABEL' => lorem_phrase(),
            'DATASETS' => $datasets,
            'BEGIN_AT_ZERO' => true,
            'BUBBLE' => false,
            'CLAMP_Y_AXIS' => false,
            'MAX' => null,
            'LOGARITHMIC' => false,
            'HAS_WORDWRAP' => false,
        ]));
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__graph_line_chart() : object
    {
        $datasets = [
            [
                'datapoints' => [
                    [
                        'value' => 1,
                        'tooltip' => lorem_sentence(),
                    ],
                    [
                        'value' => 3,
                        'tooltip' => lorem_sentence(),
                    ],
                ],
            ],
        ];

        $_datasets = [];
        foreach ($datasets as $i => $dataset) {
            $datapoints = [];
            foreach ($dataset['datapoints'] as $p) {
                $datapoints[] = [
                    'VALUE' => strval($p['value']),
                    'TOOLTIP' => $p['tooltip'],
                ];
            }

            $_datasets[] = [
                'LABEL' => lorem_phrase(),
                'COLOR' => '#FF0000',
                'DATAPOINTS' => $datapoints,
            ];
        }

        return lorem_globalise(do_lorem_template('GRAPH_LINE_CHART', [
            'ID' => lorem_word(),
            'WIDTH' => '500px',
            'HEIGHT' => '500px',
            'X_LABELS' => [lorem_word(), lorem_word()],
            'X_AXIS_LABEL' => lorem_phrase(),
            'Y_AXIS_LABEL' => lorem_phrase(),
            'DATASETS' => $_datasets,
            'BEGIN_AT_ZERO' => true,
            'SHOW_DATA_LABELS' => true,
            'FILL' => false,
            'CLAMP_Y_AXIS' => false,
            'MAX' => null,
            'LOGARITHMIC' => false,
            'HAS_WORDWRAP' => false,
        ]));
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__graph_pie_chart() : object
    {
        $datapoints = [
            [
                'label' => lorem_phrase(),
                'value' => 1,
                'tooltip' => lorem_sentence(),
            ],
            [
                'label' => lorem_phrase(),
                'value' => 3,
                'tooltip' => lorem_sentence(),
            ],
        ];

        $_datapoints = [];
        foreach ($datapoints as $x => $p) {
            $_datapoints[] = [
                'LABEL' => $p['label'],
                'VALUE' => strval($p['value']),
                'TOOLTIP' => $p['tooltip'],
                'COLOR' => '#FF0000',
            ];
        }

        return lorem_globalise(do_lorem_template('GRAPH_PIE_CHART', [
            'ID' => lorem_word(),
            'WIDTH' => '500px',
            'HEIGHT' => '500px',
            'DATAPOINTS' => $_datapoints,
            'SHOW_DATA_LABELS' => true,
            'DOUGHNUT' => false,
            'HAS_WORDWRAP' => false,
        ]));
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__graph_bar_chart() : object
    {
        $datapoints = [
            [
                'label' => lorem_phrase(),
                'value' => 1,
                'tooltip' => lorem_sentence(),
            ],
            [
                'label' => lorem_phrase(),
                'value' => 3,
                'tooltip' => lorem_sentence(),
            ],
        ];

        $_datapoints = [];
        foreach ($datapoints as $x => $p) {
            $_datapoints[] = [
                'LABEL' => $p['label'],
                'VALUE' => strval($p['value']),
                'TOOLTIP' => $p['tooltip'],
                'COLOR' => '#FF0000',
            ];
        }

        return lorem_globalise(do_lorem_template('GRAPH_BAR_CHART', [
            'ID' => lorem_word(),
            'WIDTH' => '500px',
            'HEIGHT' => '500px',
            'X_AXIS_LABEL' => lorem_phrase(),
            'Y_AXIS_LABEL' => lorem_phrase(),
            'DATAPOINTS' => $_datapoints,
            'BEGIN_AT_ZERO' => true,
            'SHOW_DATA_LABELS' => true,
            'HORIZONTAL' => false,
            'CLAMP_Y_AXIS' => false,
            'MAX' => null,
            'LOGARITHMIC' => false,
            'HAS_WORDWRAP' => false,
        ]));
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__graph_stacked_bar_chart() : object
    {
        $datapoints_a = [
            [
                'label' => lorem_phrase(),
                'value' => 1,
                'tooltip' => lorem_sentence(),
            ],
            [
                'label' => lorem_phrase(),
                'value' => 3,
                'tooltip' => lorem_sentence(),
            ],
        ];
        $datapoints_b = [
            [
                'label' => lorem_phrase(),
                'value' => 2,
                'tooltip' => lorem_sentence(),
            ],
            [
                'label' => lorem_phrase(),
                'value' => 2,
                'tooltip' => lorem_sentence(),
            ],
        ];

        $_datapoints_a = [];
        foreach ($datapoints_a as $x => $p) {
            $_datapoints_a[] = [
                'LABEL' => $p['label'],
                'VALUE' => strval($p['value']),
                'TOOLTIP' => $p['tooltip'],
                'COLOR' => '#FF0000',
            ];
        }
        $_datapoints_b = [];
        foreach ($datapoints_b as $x => $p) {
            $_datapoints_b[] = [
                'LABEL' => $p['label'],
                'VALUE' => strval($p['value']),
                'TOOLTIP' => $p['tooltip'],
                'COLOR' => '#00FF00',
            ];
        }

        $_datasets = [];
        $_datasets[] = [
            'LABEL' => 'A',
            'COLOR' => '#FF0000',
            'DATAPOINTS' => $_datapoints_a,
        ];
        $_datasets[] = [
            'LABEL' => 'B',
            'COLOR' => '#00FF00',
            'DATAPOINTS' => $_datapoints_b,
        ];

        $_labels = [
            ['LABEL' => 'A'],
            ['LABEL' => 'B'],
        ];

        return lorem_globalise(do_lorem_template('GRAPH_STACKED_BAR_CHART', [
            'ID' => lorem_word(),
            'WIDTH' => '500px',
            'HEIGHT' => '500px',
            'X_AXIS_LABEL' => lorem_phrase(),
            'Y_AXIS_LABEL' => lorem_phrase(),
            'DATASETS' => $_datasets,
            'LABELS' => $_labels,
            'BEGIN_AT_ZERO' => true,
            'SHOW_DATA_LABELS' => true,
            'HORIZONTAL' => false,
            'STACKED' => true,
            'CLAMP_Y_AXIS' => false,
            'MAX' => null,
            'LOGARITHMIC' => false,
        ]));
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__graph_bubble_bar_chart() : object
    {
        $datapoints_a = [
            [
                'label' => lorem_phrase(),
                'value' => 1,
                'tooltip' => lorem_sentence(),
            ],
            [
                'label' => lorem_phrase(),
                'value' => 3,
                'tooltip' => lorem_sentence(),
            ],
        ];
        $datapoints_b = [
            [
                'label' => lorem_phrase(),
                'value' => 2,
                'tooltip' => lorem_sentence(),
            ],
            [
                'label' => lorem_phrase(),
                'value' => 2,
                'tooltip' => lorem_sentence(),
            ],
        ];

        $_datapoints_a = [];
        foreach ($datapoints_a as $x => $p) {
            $_datapoints_a[] = [
                'LABEL' => $p['label'],
                'VALUE' => strval($p['value']),
                'TOOLTIP' => $p['tooltip'],
            ];
        }
        $_datapoints_b = [];
        foreach ($datapoints_b as $x => $p) {
            $_datapoints_b[] = [
                'LABEL' => $p['label'],
                'VALUE' => strval($p['value']),
                'TOOLTIP' => $p['tooltip'],
            ];
        }

        $_datasets = [];
        $_datasets[] = [
            'Y_LABEL' => 'A',
            'DATAPOINTS' => $_datapoints_a,
        ];
        $_datasets[] = [
            'Y_LABEL' => 'B',
            'DATAPOINTS' => $_datapoints_b,
        ];

        $_labels = [
            'Alpha',
            'Beta',
        ];

        return lorem_globalise(do_lorem_template('GRAPH_BUBBLE_BAR_CHART', [
            'ID' => lorem_word(),
            'WIDTH' => '500px',
            'HEIGHT' => '500px',
            'X_AXIS_LABEL' => 'X axis',
            'Y_AXIS_LABEL' => 'Y axis',
            'Z_AXIS_LABEL' => 'Z axis',
            'TITLE' => lorem_phrase(),
            'LABELS' => $_labels,
            'DATASETS' => $_datasets,
            'SHOW_DATA_LABELS' => true,
            'COLOR' => '#FF0000',
            'MIN' => '1',
            'MAX' => '5',
            'HAS_WORDWRAP' => false,
        ]));
    }
}
