<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    core_notifications
 */

/**
 * Hook class.
 */
class Hook_addon_registry_core_notifications
{
    /**
     * Get a list of file permissions to set.
     *
     * @param  boolean $runtime Whether to include wildcards represented runtime-created chmoddable files
     * @return array File permissions to set
     */
    public function get_chmod_array(bool $runtime = false) : array
    {
        return [];
    }

    /**
     * Get the version of the software this addon is for (used in generating the TAR filename).
     *
     * @return float Version number
     */
    public function get_version() : float
    {
        return cms_version_number();
    }

    /**
     * Get the minimum required version of the website software needed to use this addon.
     *
     * @return float Minimum required website software version
     */
    public function get_min_cms_version() : float
    {
        return 11.0;
    }

    /**
     * Get the maximum compatible version of the website software to use this addon.
     *
     * @return ?float Maximum compatible website software version (null: no maximum version currently)
     */
    public function get_max_cms_version() : ?float
    {
        return 11.9;
    }

    /**
     * Get the addon category.
     *
     * @return string The category
     */
    public function get_category() : string
    {
        return 'Architecture';
    }

    /**
     * Get the description of the addon.
     *
     * @return string Description of the addon
     */
    public function get_description() : string
    {
        return 'Sends out action-triggered notifications to members listening to them.';
    }

    /**
     * Get a list of tutorials that apply to this addon.
     *
     * @return array List of tutorials
     */
    public function get_applicable_tutorials() : array
    {
        return [
            'tut_notifications',
        ];
    }

    /**
     * Get a mapping of dependency types.
     *
     * @return array A structure specifying dependency information
     */
    public function get_dependencies() : array
    {
        return [
            'requires' => [],
            'recommends' => [],
            'conflicts_with' => [],
        ];
    }

    /**
     * Explicitly say which icon should be used.
     *
     * @return URLPATH Icon
     */
    public function get_default_icon() : string
    {
        return 'themes/default/images/icons/tool_buttons/notifications2.svg';
    }

    /**
     * Get a list of files that belong to this addon.
     *
     * @return array List of files
     */
    public function get_file_list() : array
    {
        return [
            'adminzone/pages/modules/admin_notifications.php',
            'data/notifications.php',
            'data_custom/modules/web_notifications/.htaccess',
            'data_custom/modules/web_notifications/index.html',
            'lang/EN/notifications.ini',
            'site/pages/modules/notifications.php',
            'sources/blocks/top_notifications.php',
            'sources/hooks/systems/addon_registry/core_notifications.php',
            'sources/hooks/systems/commandr_fs_extended_config/notification_lockdown.php',
            'sources/hooks/systems/commandr_fs_extended_member/notifications_enabled.php',
            'sources/hooks/systems/config/allow_auto_notifications.php',
            'sources/hooks/systems/config/block_top_notifications.php',
            'sources/hooks/systems/config/notification_desktop_alerts.php',
            'sources/hooks/systems/config/notification_enable_digests.php',
            'sources/hooks/systems/config/notification_enable_private_topics.php',
            'sources/hooks/systems/config/notification_keep_days.php',
            'sources/hooks/systems/config/notification_poll_frequency.php',
            'sources/hooks/systems/config/pt_notifications_as_web.php',
            'sources/hooks/systems/config/web_notifications_enabled.php',
            'sources/hooks/systems/cron/notification_digests.php',
            'sources/hooks/systems/notification_types_extended/.htaccess',
            'sources/hooks/systems/notification_types_extended/index.html',
            'sources/hooks/systems/notifications/.htaccess',
            'sources/hooks/systems/notifications/index.html',
            'sources/hooks/systems/page_groupings/notifications.php',
            'sources/hooks/systems/profiles_tabs_edit/notifications.php',
            'sources/hooks/systems/rss/web_notifications.php',
            'sources/hooks/systems/startup/notification_poller_init.php',
            'sources/hooks/systems/tasks/dispatch_notification.php',
            'sources/notification_poller.php',
            'sources/notifications.php',
            'sources/notifications2.php',
            'sources_custom/hooks/systems/notification_types_extended/.htaccess',
            'sources_custom/hooks/systems/notification_types_extended/index.html',
            'sources_custom/hooks/systems/notifications/.htaccess',
            'sources_custom/hooks/systems/notifications/index.html',
            'themes/default/css/notifications.css',
            'themes/default/images/icons/buttons/notifications_disable.svg',
            'themes/default/images/icons/buttons/notifications_enable.svg',
            'themes/default/images/icons/menu/adminzone/setup/notifications.svg',
            'themes/default/images/icons/tool_buttons/notifications.svg',
            'themes/default/images/icons/tool_buttons/notifications2.svg',
            'themes/default/images/icons_monochrome/buttons/notifications_disable.svg',
            'themes/default/images/icons_monochrome/buttons/notifications_enable.svg',
            'themes/default/images/icons_monochrome/menu/adminzone/setup/notifications.svg',
            'themes/default/images/icons_monochrome/tool_buttons/notifications.svg',
            'themes/default/images/icons_monochrome/tool_buttons/notifications2.svg',
            'themes/default/images/notifications/index.html',
            'themes/default/images/notifications/notifications.ico',
            'themes/default/javascript/core_notifications.js',
            'themes/default/templates/BLOCK_TOP_NOTIFICATIONS.tpl',
            'themes/default/templates/NOTIFICATIONS_MANAGE.tpl',
            'themes/default/templates/NOTIFICATIONS_MANAGE_ADVANCED_SCREEN.tpl',
            'themes/default/templates/NOTIFICATIONS_MANAGE_SCREEN.tpl',
            'themes/default/templates/NOTIFICATIONS_TREE.tpl',
            'themes/default/templates/NOTIFICATION_BROWSE_SCREEN.tpl',
            'themes/default/templates/NOTIFICATION_BUTTONS.tpl',
            'themes/default/templates/NOTIFICATION_POLLER.tpl',
            'themes/default/templates/NOTIFICATION_PT_DESKTOP.tpl',
            'themes/default/templates/NOTIFICATION_TYPES.tpl',
            'themes/default/templates/NOTIFICATION_VIEW_SCREEN.tpl',
            'themes/default/templates/NOTIFICATION_WEB.tpl',
            'themes/default/templates/NOTIFICATION_WEB_DESKTOP.tpl',
        ];
    }

    /**
     * Get mapping between template names and the method of this class that can render a preview of them.
     *
     * @return array The mapping
     */
    public function tpl_previews() : array
    {
        return [
            'templates/NOTIFICATIONS_MANAGE.tpl' => 'notifications_regular',
            'templates/NOTIFICATIONS_MANAGE_SCREEN.tpl' => 'notifications_regular',
            'templates/NOTIFICATIONS_MANAGE_ADVANCED_SCREEN.tpl' => 'notifications_advanced',
            'templates/NOTIFICATIONS_TREE.tpl' => 'notifications_advanced',
            'templates/NOTIFICATION_TYPES.tpl' => 'notifications_regular',
            'templates/NOTIFICATION_WEB.tpl' => 'notification_web',
            'templates/NOTIFICATION_WEB_DESKTOP.tpl' => 'notification_web_desktop',
            'templates/NOTIFICATION_PT_DESKTOP.tpl' => 'notification_pt_desktop',
            'templates/BLOCK_TOP_NOTIFICATIONS.tpl' => 'block_top_notifications',
            'templates/NOTIFICATION_POLLER.tpl' => 'notification_poller',
            'templates/NOTIFICATION_BROWSE_SCREEN.tpl' => 'notification_browse_screen',
            'templates/NOTIFICATION_VIEW_SCREEN.tpl' => 'notification_view_screen',
        ];
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__notification_browse_screen() : object
    {
        $notifications = new Tempcode();
        $notifications->attach(
            do_lorem_template('NOTIFICATION_WEB', [
                'ID' => placeholder_numeric_id(),
                'SUBJECT' => lorem_phrase(),
                'MESSAGE' => lorem_paragraph_html(),
                'FROM_USERNAME' => lorem_phrase(),
                'FROM_MEMBER_ID' => placeholder_numeric_id(),
                'FROM_URL' => placeholder_url(),
                'FROM_AVATAR_URL' => placeholder_image_url(),
                'PRIORITY' => '3',
                '_TIME' => placeholder_date_raw(),
                'DATE' => placeholder_date(),
                'NOTIFICATION_CODE' => placeholder_codename(),
                'CODE_CATEGORY' => placeholder_codename(),
                'HAS_READ' => false,
            ])
        );

        return lorem_globalise(do_lorem_template('NOTIFICATION_BROWSE_SCREEN', [
            'TITLE' => lorem_screen_title(),
            'NOTIFICATIONS' => $notifications,
            'PAGINATION' => placeholder_pagination(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__notification_view_screen() : object
    {
        return lorem_globalise(do_lorem_template('NOTIFICATION_VIEW_SCREEN', [
            'TITLE' => lorem_screen_title(),
            'ID' => placeholder_numeric_id(),
            'SUBJECT' => lorem_phrase(),
            'MESSAGE' => lorem_paragraph_html(),
            'FROM_USERNAME' => lorem_phrase(),
            'FROM_MEMBER_ID' => placeholder_numeric_id(),
            'FROM_URL' => placeholder_url(),
            'FROM_AVATAR_URL' => placeholder_image_url(),
            'PRIORITY' => '3',
            '_TIME' => placeholder_date_raw(),
            'DATE' => placeholder_date(),
            'NOTIFICATION_CODE' => placeholder_codename(),
            'CODE_CATEGORY' => placeholder_codename(),
            'HAS_READ' => true,
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__notification_poller() : object
    {
        return lorem_globalise(do_lorem_template('NOTIFICATION_POLLER', []), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__notification_web() : object
    {
        return lorem_globalise(do_lorem_template('NOTIFICATION_WEB', [
            'ID' => placeholder_numeric_id(),
            'SUBJECT' => lorem_phrase(),
            'MESSAGE' => lorem_paragraph_html(),
            'FROM_USERNAME' => lorem_phrase(),
            'FROM_MEMBER_ID' => placeholder_numeric_id(),
            'FROM_URL' => placeholder_url(),
            'FROM_AVATAR_URL' => placeholder_image_url(),
            'PRIORITY' => '3',
            '_TIME' => placeholder_date_raw(),
            'DATE' => placeholder_date(),
            'NOTIFICATION_CODE' => placeholder_codename(),
            'CODE_CATEGORY' => placeholder_codename(),
            'HAS_READ' => true,
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__notification_web_desktop() : object
    {
        return lorem_globalise(do_lorem_template('NOTIFICATION_WEB_DESKTOP', [
            'ID' => placeholder_numeric_id(),
            'SUBJECT' => lorem_phrase(),
            'MESSAGE' => lorem_paragraph_html(),
            'FROM_USERNAME' => lorem_phrase(),
            'FROM_MEMBER_ID' => placeholder_numeric_id(),
            'FROM_URL' => placeholder_url(),
            'FROM_AVATAR_URL' => placeholder_image_url(),
            'PRIORITY' => '3',
            '_TIME' => placeholder_date_raw(),
            'DATE' => placeholder_date(),
            'NOTIFICATION_CODE' => placeholder_codename(),
            'CODE_CATEGORY' => placeholder_codename(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__notification_pt_desktop() : object
    {
        return lorem_globalise(do_lorem_template('NOTIFICATION_PT_DESKTOP', [
            'ID' => placeholder_numeric_id(),
            'SUBJECT' => lorem_phrase(),
            'MESSAGE' => lorem_paragraph_html(),
            'FROM_USERNAME' => lorem_phrase(),
            'FROM_MEMBER_ID' => placeholder_numeric_id(),
            'URL' => placeholder_url(),
            'FROM_AVATAR_URL' => placeholder_image_url(),
            '_TIME' => placeholder_date_raw(),
            'DATE' => placeholder_date(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__block_top_notifications() : object
    {
        return lorem_globalise(do_lorem_template('BLOCK_TOP_NOTIFICATIONS', [
            'BLOCK_ID' => lorem_word(),
            'NUM_UNREAD_WEB_NOTIFICATIONS' => placeholder_number(),
            'NUM_UNREAD_PTS' => placeholder_number(),
            'NOTIFICATIONS' => '',
            'PTS' => '',
            'MAX' => '5',
        ]), null, '', false);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__notifications_regular() : object
    {
        require_css('notifications');
        require_lang('cns_mailinglists');

        $notification_types = [];
        $notification_types[] = [
            'NTYPE' => placeholder_codename(),
            'LABEL' => lorem_phrase(),
            'CHECKED' => true,
            'RAW' => placeholder_number(),
            'AVAILABLE' => true,
            'SCOPE' => placeholder_codename(),
        ];
        $notification_types_titles = [];
        $notification_types_titles[] = [
            'NTYPE' => placeholder_codename(),
            'LABEL' => lorem_phrase(),
            'RAW' => placeholder_number(),
        ];
        $notification_types_tempcode = do_template('NOTIFICATION_TYPES', ['_GUID' => '2fbe39a0988eddb728117197f83688f3', 'NOTIFICATION_TYPES' => $notification_types]);
        $notification_code_map = [
            'NOTIFICATION_CODE' => placeholder_codename(),
            'NOTIFICATION_LABEL' => lorem_phrase(),
            'NOTIFICATION_TYPES' => $notification_types_tempcode,
            'SUPPORTS_CATEGORIES' => true,
        ];
        $notification_sections = [];
        $notification_sections[lorem_phrase()] = [
            'NOTIFICATION_SECTION' => lorem_phrase(),
            'NOTIFICATION_CODES' => [
                $notification_code_map,
            ],
        ];
        $interface = do_lorem_template('NOTIFICATIONS_MANAGE', [
            'COLOR' => 'FFFFFF',
            'INTRO' => lorem_paragraph(),
            'NOTIFICATION_TYPES_TITLES' => $notification_types_titles,
            'NOTIFICATION_SECTIONS' => $notification_sections,
            'AUTO_NOTIFICATION_CONTRIB_CONTENT' => false,
            'SMART_TOPIC_NOTIFICATION' => false,
            'MAILING_LIST_STYLE' => false,
            'MAILING_LIST_STYLE_DESCRIPTION' => do_lang_tempcode('DESCRIPTION_MAILING_LIST_STYLE', ''),
            'MEMBER_ID' => placeholder_numeric_id(),
            'ADVANCED_COLUMN' => true,
            'SHOW_PRIVILEGES' => false,
        ]);

        return lorem_globalise(do_lorem_template('NOTIFICATIONS_MANAGE_SCREEN', [
            'TITLE' => lorem_screen_title(),
            'INTERFACE' => $interface,
            'ACTION_URL' => get_self_url(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__notifications_advanced() : object
    {
        require_css('notifications');

        $notification_types = [];
        $notification_types[] = [
            'NTYPE' => placeholder_codename(),
            'LABEL' => lorem_phrase(),
            'CHECKED' => true,
            'RAW' => placeholder_number(),
            'AVAILABLE' => true,
            'SCOPE' => placeholder_codename(),
        ];
        $notification_categories = [];
        $notification_categories[] = [
            'NUM_CHILDREN' => '0',
            'DEPTH' => '0',
            'NOTIFICATION_CATEGORY' => placeholder_codename(),
            'NOTIFICATION_TYPES' => $notification_types,
            'CATEGORY_TITLE' => lorem_phrase(),
            'CHECKED' => true,
            'CHILDREN' => '',
        ];
        $tree = do_lorem_template('NOTIFICATIONS_TREE', [
            'NOTIFICATION_CODE' => placeholder_codename(),
            'NOTIFICATION_CATEGORIES' => $notification_categories,
        ]);
        $notification_types_titles = [];
        $notification_types_titles[] = [
            'NTYPE' => placeholder_codename(),
            'LABEL' => lorem_phrase(),
            'RAW' => placeholder_number(),
        ];

        return lorem_globalise(do_lorem_template('NOTIFICATIONS_MANAGE_ADVANCED_SCREEN', [
            'TITLE' => lorem_screen_title(),
            '_TITLE' => lorem_phrase(),
            'COLOR' => 'FFFFFF',
            'ACTION_URL' => placeholder_url(),
            'NOTIFICATION_TYPES_TITLES' => $notification_types_titles,
            'TREE' => $tree,
            'NOTIFICATION_CODE' => placeholder_codename(),
        ]), null, '', true);
    }
}
