<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    core_zone_editor
 */

/**
 * Hook class.
 */
class Hook_addon_registry_core_zone_editor
{
    /**
     * Get a list of file permissions to set.
     *
     * @param  boolean $runtime Whether to include wildcards represented runtime-created chmoddable files
     * @return array File permissions to set
     */
    public function get_chmod_array(bool $runtime = false) : array
    {
        return [];
    }

    /**
     * Get the version of the software this addon is for (used in generating the TAR filename).
     *
     * @return float Version number
     */
    public function get_version() : float
    {
        return cms_version_number();
    }

    /**
     * Get the minimum required version of the website software needed to use this addon.
     *
     * @return float Minimum required website software version
     */
    public function get_min_cms_version() : float
    {
        return 11.0;
    }

    /**
     * Get the maximum compatible version of the website software to use this addon.
     *
     * @return ?float Maximum compatible website software version (null: no maximum version currently)
     */
    public function get_max_cms_version() : ?float
    {
        return 11.9;
    }

    /**
     * Get the addon category.
     *
     * @return string The category
     */
    public function get_category() : string
    {
        return 'Admin Utilities';
    }

    /**
     * Get the description of the addon.
     *
     * @return string Description of the addon
     */
    public function get_description() : string
    {
        return 'Manage zones (sub-sites).';
    }

    /**
     * Get a list of tutorials that apply to this addon.
     *
     * @return array List of tutorials
     */
    public function get_applicable_tutorials() : array
    {
        return [
            'tut_structure',
            'tut_adv_structure',
        ];
    }

    /**
     * Get a mapping of dependency types.
     *
     * @return array A structure specifying dependency information
     */
    public function get_dependencies() : array
    {
        return [
            'requires' => [],
            'recommends' => [],
            'conflicts_with' => [],
        ];
    }

    /**
     * Explicitly say which icon should be used.
     *
     * @return URLPATH Icon
     */
    public function get_default_icon() : string
    {
        return 'themes/default/images/icons/menu/adminzone/structure/zones/zone_editor.svg';
    }

    /**
     * Get a list of files that belong to this addon.
     *
     * @return array List of files
     */
    public function get_file_list() : array
    {
        return [
            'adminzone/pages/modules/admin_zones.php',
            'sources/hooks/systems/addon_registry/core_zone_editor.php',
            'sources/hooks/systems/resource_meta_aware/zone.php',
            'sources/hooks/systems/snippets/exists_zone.php',
            'themes/default/css/zone_editor.css',
            'themes/default/images/icons/menu/adminzone/structure/zones/zone_editor.svg',
            'themes/default/images/icons_monochrome/menu/adminzone/structure/zones/zone_editor.svg',
            'themes/default/javascript/core_zone_editor.js',
            'themes/default/templates/ZONE_EDITOR_PANEL.tpl',
            'themes/default/templates/ZONE_EDITOR_SCREEN.tpl',
        ];
    }

    /**
     * Get mapping between template names and the method of this class that can render a preview of them.
     *
     * @return array The mapping
     */
    public function tpl_previews() : array
    {
        return [
            'templates/ZONE_EDITOR_PANEL.tpl' => 'administrative__zone_editor_screen',
            'templates/ZONE_EDITOR_SCREEN.tpl' => 'administrative__zone_editor_screen',
        ];
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__administrative__zone_editor_screen() : object
    {
        require_lang('zones');

        $comcode_editor = do_lorem_template('COMCODE_EDITOR_BUTTON', [
            'DIVIDER' => true,
            'FIELD_NAME' => lorem_word(),
            'LABEL' => lorem_word(),
            'TITLE' => lorem_word(),
            'B' => 'block',
            'IS_POSTING_FIELD' => false,
        ]);

        $editor = [];
        foreach (['panel_left', 'panel_middle', 'panel_right'] as $i => $for) {
            $editor[$for] = do_lorem_template('ZONE_EDITOR_PANEL', [
                'CLASS' => '',
                'ZONES' => '',
                'CURRENT_ZONE' => '',
                'ZONE' => '',
                'COMCODE' => lorem_phrase(),
                'PREVIEW' => lorem_paragraph_html(),
                'ID' => $for,
                'IS_PANEL' => true,
                'TYPE' => lorem_phrase(),
                'EDIT_URL' => placeholder_url(),
                'SETTINGS' => null,
                'COMCODE_EDITOR' => $comcode_editor,
            ]);
        }

        return lorem_globalise(do_lorem_template('ZONE_EDITOR_SCREEN', [
            'PING_URL' => placeholder_url(),
            'WARNING_DETAILS' => '',
            'TITLE' => lorem_screen_title(),
            'ID' => '',
            'LANG' => fallback_lang(),
            'URL' => placeholder_url(),
            'LEFT_EDITOR' => $editor['panel_left'],
            'RIGHT_EDITOR' => $editor['panel_right'],
            'MIDDLE_EDITOR' => $editor['panel_middle'],
            'PANEL_TOP_EDIT_URL' => placeholder_url(),
            'PANEL_BOTTOM_EDIT_URL' => placeholder_url(),
        ]), null, '', true);
    }
}
