<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    galleries
 */

/**
 * Hook class.
 */
class Hook_addon_registry_galleries
{
    /**
     * Get a list of file permissions to set.
     *
     * @param  boolean $runtime Whether to include wildcards represented runtime-created chmoddable files
     * @return array File permissions to set
     */
    public function get_chmod_array(bool $runtime = false) : array
    {
        $chmod = [];
        if ($runtime) {
            $chmod = array_merge(
                $chmod,
                [
                    'uploads/galleries/*',
                    'uploads/galleries_thumbs/*',
                    'uploads/watermarks/*',
                ]
            );
        }
        $chmod = array_merge(
            $chmod,
            [
                'uploads/galleries',
                'uploads/galleries_thumbs',
                'uploads/watermarks',
            ]
        );
        return $chmod;
    }

    /**
     * Get the version of the software this addon is for (used in generating the TAR filename).
     *
     * @return float Version number
     */
    public function get_version() : float
    {
        return cms_version_number();
    }

    /**
     * Get the minimum required version of the website software needed to use this addon.
     *
     * @return float Minimum required website software version
     */
    public function get_min_cms_version() : float
    {
        return 11.0;
    }

    /**
     * Get the maximum compatible version of the website software to use this addon.
     *
     * @return ?float Maximum compatible website software version (null: no maximum version currently)
     */
    public function get_max_cms_version() : ?float
    {
        return 11.9;
    }

    /**
     * Get the addon category.
     *
     * @return string The category
     */
    public function get_category() : string
    {
        return 'Information Display';
    }

    /**
     * Get the description of the addon.
     *
     * @return string Description of the addon
     */
    public function get_description() : string
    {
        return 'Galleries, including support for video galleries, and member personal galleries.';
    }

    /**
     * Get a list of tutorials that apply to this addon.
     *
     * @return array List of tutorials
     */
    public function get_applicable_tutorials() : array
    {
        return [
            'tut_galleries',
            'tut_adv_galleries',
        ];
    }

    /**
     * Get a mapping of dependency types.
     *
     * @return array A structure specifying dependency information
     */
    public function get_dependencies() : array
    {
        return [
            'requires' => [],
            'recommends' => [],
            'conflicts_with' => [],
        ];
    }

    /**
     * Explicitly say which icon should be used.
     *
     * @return URLPATH Icon
     */
    public function get_default_icon() : string
    {
        return 'themes/default/images/icons/menu/rich_content/galleries.svg';
    }

    /**
     * Get a list of files that belong to this addon.
     *
     * @return array List of files
     */
    public function get_file_list() : array
    {
        return [
            'cms/pages/modules/cms_galleries.php',
            'data/images/homepage_hero_slider/bastei_bridge.jpg',
            'data/images/homepage_hero_slider/index.html',
            'data/images/homepage_hero_slider/rustic.jpg',
            'data/images/homepage_hero_slider/waterfall.jpg',
            'lang/EN/galleries.ini',
            'lang/EN/homepage_hero_slider.ini',
            'site/download_gallery.php',
            'site/pages/modules/galleries.php',
            'site/show_gallery_video.php',
            'sources/blocks/main_gallery_embed.php',
            'sources/blocks/main_gallery_mosaic.php',
            'sources/blocks/main_hero_slider.php',
            'sources/blocks/main_image_fader.php',
            'sources/blocks/main_personal_galleries_list.php',
            'sources/blocks/side_galleries.php',
            'sources/galleries.php',
            'sources/galleries2.php',
            'sources/galleries3.php',
            'sources/hooks/blocks/side_stats/galleries.php',
            'sources/hooks/modules/admin_import/galleries.php',
            'sources/hooks/modules/admin_import_types/galleries.php',
            'sources/hooks/modules/admin_newsletter/galleries.php',
            'sources/hooks/modules/admin_setupwizard/galleries.php',
            'sources/hooks/modules/admin_setupwizard_installprofiles/portfolio.php',
            'sources/hooks/modules/galleries_users/.htaccess',
            'sources/hooks/modules/galleries_users/index.html',
            'sources/hooks/modules/search/galleries.php',
            'sources/hooks/modules/search/images.php',
            'sources/hooks/modules/search/videos.php',
            'sources/hooks/systems/actionlog/galleries.php',
            'sources/hooks/systems/addon_registry/galleries.php',
            'sources/hooks/systems/ajax_tree/choose_gallery.php',
            'sources/hooks/systems/ajax_tree/choose_image.php',
            'sources/hooks/systems/ajax_tree/choose_video.php',
            'sources/hooks/systems/block_ui_renderers/galleries.php',
            'sources/hooks/systems/commandr_fs/galleries.php',
            'sources/hooks/systems/config/default_video_height.php',
            'sources/hooks/systems/config/default_video_width.php',
            'sources/hooks/systems/config/enable_ecards.php',
            'sources/hooks/systems/config/ffmpeg_path.php',
            'sources/hooks/systems/config/galleries_show_stats_count_galleries.php',
            'sources/hooks/systems/config/galleries_show_stats_count_images.php',
            'sources/hooks/systems/config/galleries_show_stats_count_videos.php',
            'sources/hooks/systems/config/galleries_sort_order.php',
            'sources/hooks/systems/config/galleries_subcat_narrowin.php',
            'sources/hooks/systems/config/gallery_entries_carousel_per_page.php',
            'sources/hooks/systems/config/gallery_entries_grid_per_page.php',
            'sources/hooks/systems/config/gallery_feedback_fields.php',
            'sources/hooks/systems/config/gallery_media_default_sort_order.php',
            'sources/hooks/systems/config/gallery_media_title_required.php',
            'sources/hooks/systems/config/gallery_member_synced.php',
            'sources/hooks/systems/config/gallery_mode_is.php',
            'sources/hooks/systems/config/gallery_permissions.php',
            'sources/hooks/systems/config/gallery_rep_image.php',
            'sources/hooks/systems/config/gallery_selectors.php',
            'sources/hooks/systems/config/gallery_watermarks.php',
            'sources/hooks/systems/config/homepage_slider_cycle_duration.php',
            'sources/hooks/systems/config/homepage_slider_effect.php',
            'sources/hooks/systems/config/homepage_slider_enabled.php',
            'sources/hooks/systems/config/homepage_slider_fullscreen.php',
            'sources/hooks/systems/config/homepage_slider_gallery.php',
            'sources/hooks/systems/config/manual_gallery_codename.php',
            'sources/hooks/systems/config/manual_gallery_media_types.php',
            'sources/hooks/systems/config/manual_gallery_parent.php',
            'sources/hooks/systems/config/max_personal_gallery_images_high.php',
            'sources/hooks/systems/config/max_personal_gallery_images_low.php',
            'sources/hooks/systems/config/max_personal_gallery_videos_high.php',
            'sources/hooks/systems/config/max_personal_gallery_videos_low.php',
            'sources/hooks/systems/config/maximum_image_size.php',
            'sources/hooks/systems/config/personal_under_members.php',
            'sources/hooks/systems/config/points_ADD_IMAGE.php',
            'sources/hooks/systems/config/points_ADD_VIDEO.php',
            'sources/hooks/systems/config/reverse_thumb_order.php',
            'sources/hooks/systems/config/search_galleries.php',
            'sources/hooks/systems/config/search_images.php',
            'sources/hooks/systems/config/search_videos.php',
            'sources/hooks/systems/config/show_empty_galleries.php',
            'sources/hooks/systems/config/show_gallery_counts.php',
            'sources/hooks/systems/config/subgallery_link_limit.php',
            'sources/hooks/systems/config/video_height_setting.php',
            'sources/hooks/systems/config/video_width_setting.php',
            'sources/hooks/systems/config_categories/gallery.php',
            'sources/hooks/systems/content_meta_aware/gallery.php',
            'sources/hooks/systems/content_meta_aware/image.php',
            'sources/hooks/systems/content_meta_aware/video.php',
            'sources/hooks/systems/notifications/gallery_entry.php',
            'sources/hooks/systems/page_groupings/galleries.php',
            'sources/hooks/systems/points/image__add.php',
            'sources/hooks/systems/points/video__add.php',
            'sources/hooks/systems/preview/image.php',
            'sources/hooks/systems/preview/video.php',
            'sources/hooks/systems/privacy/galleries.php',
            'sources/hooks/systems/profiles_tabs/galleries.php',
            'sources/hooks/systems/reorganise_uploads/galleries.php',
            'sources/hooks/systems/rss/galleries.php',
            'sources/hooks/systems/sitemap/gallery.php',
            'sources/hooks/systems/sitemap/image.php',
            'sources/hooks/systems/sitemap/video.php',
            'sources/hooks/systems/snippets/exists_gallery.php',
            'sources/hooks/systems/symbols/GALLERY_VIDEO_FOR_URL.php',
            'sources/hooks/systems/tasks/delete_gallery_media.php',
            'sources/hooks/systems/tasks/download_gallery.php',
            'sources/hooks/systems/tasks/import_gallery_media.php',
            'sources/hooks/systems/tasks/import_orphaned_gallery_media.php',
            'sources/hooks/systems/trackback/images.php',
            'sources/hooks/systems/trackback/videos.php',
            'sources_custom/hooks/modules/galleries_users/.htaccess',
            'sources_custom/hooks/modules/galleries_users/index.html',
            'themes/default/css/galleries.css',
            'themes/default/css/widget_glide.css',
            'themes/default/images/audio_thumb.png',
            'themes/default/images/icons/buttons/slideshow.svg',
            'themes/default/images/icons/menu/cms/galleries/add_one_image.svg',
            'themes/default/images/icons/menu/cms/galleries/add_one_multimedia.svg',
            'themes/default/images/icons/menu/cms/galleries/add_one_video.svg',
            'themes/default/images/icons/menu/cms/galleries/edit_one_image.svg',
            'themes/default/images/icons/menu/cms/galleries/edit_one_multimedia.svg',
            'themes/default/images/icons/menu/cms/galleries/edit_one_video.svg',
            'themes/default/images/icons/menu/cms/galleries/index.html',
            'themes/default/images/icons/menu/rich_content/galleries.svg',
            'themes/default/images/icons_monochrome/buttons/slideshow.svg',
            'themes/default/images/icons_monochrome/menu/cms/galleries/add_one_image.svg',
            'themes/default/images/icons_monochrome/menu/cms/galleries/add_one_multimedia.svg',
            'themes/default/images/icons_monochrome/menu/cms/galleries/add_one_video.svg',
            'themes/default/images/icons_monochrome/menu/cms/galleries/edit_one_image.svg',
            'themes/default/images/icons_monochrome/menu/cms/galleries/edit_one_multimedia.svg',
            'themes/default/images/icons_monochrome/menu/cms/galleries/edit_one_video.svg',
            'themes/default/images/icons_monochrome/menu/cms/galleries/index.html',
            'themes/default/images/icons_monochrome/menu/rich_content/galleries.svg',
            'themes/default/images/video_thumb.png',
            'themes/default/javascript/galleries.js',
            'themes/default/javascript/glide.js',
            'themes/default/javascript/masonry.js',
            'themes/default/templates/BLOCK_MAIN_GALLERY_EMBED.tpl',
            'themes/default/templates/BLOCK_MAIN_GALLERY_MOSAIC.tpl',
            'themes/default/templates/BLOCK_MAIN_HERO_SLIDER.tpl',
            'themes/default/templates/BLOCK_MAIN_IMAGE_FADER.tpl',
            'themes/default/templates/BLOCK_MAIN_PERSONAL_GALLERIES_LIST.tpl',
            'themes/default/templates/BLOCK_SIDE_GALLERIES.tpl',
            'themes/default/templates/BLOCK_SIDE_GALLERIES_LINE.tpl',
            'themes/default/templates/BLOCK_SIDE_GALLERIES_LINE_CONTAINER.tpl',
            'themes/default/templates/BLOCK_SIDE_GALLERIES_LINE_DEPTH.tpl',
            'themes/default/templates/CNS_MEMBER_PROFILE_GALLERIES.tpl',
            'themes/default/templates/GALLERY_BOX.tpl',
            'themes/default/templates/GALLERY_CAROUSEL_ENTRY.tpl',
            'themes/default/templates/GALLERY_CAROUSEL_MODE_IMAGE.tpl',
            'themes/default/templates/GALLERY_CAROUSEL_MODE_SCREEN.tpl',
            'themes/default/templates/GALLERY_CAROUSEL_MODE_VIDEO.tpl',
            'themes/default/templates/GALLERY_ENTRY_LIST_LINE.tpl',
            'themes/default/templates/GALLERY_ENTRY_SCREEN.tpl',
            'themes/default/templates/GALLERY_ENTRY_WRAP.tpl',
            'themes/default/templates/GALLERY_GRID_MODE_SCREEN.tpl',
            'themes/default/templates/GALLERY_HOMEPAGE_HERO_SLIDE.tpl',
            'themes/default/templates/GALLERY_IMAGE.tpl',
            'themes/default/templates/GALLERY_IMAGE_BOX.tpl',
            'themes/default/templates/GALLERY_IMPORT_SCREEN.tpl',
            'themes/default/templates/GALLERY_MOSAIC_IMAGE.tpl',
            'themes/default/templates/GALLERY_MOSAIC_MODE_SCREEN.tpl',
            'themes/default/templates/GALLERY_MOSAIC_VIDEO.tpl',
            'themes/default/templates/GALLERY_NAV.tpl',
            'themes/default/templates/GALLERY_POPULAR.tpl',
            'themes/default/templates/GALLERY_SLIDESHOW_CAROUSEL_ENTRY.tpl',
            'themes/default/templates/GALLERY_SLIDESHOW_SCREEN.tpl',
            'themes/default/templates/GALLERY_VIDEO.tpl',
            'themes/default/templates/GALLERY_VIDEO_BOX.tpl',
            'themes/default/templates/GALLERY_VIDEO_INFO.tpl',
            'uploads/galleries/.htaccess',
            'uploads/galleries/index.html',
            'uploads/galleries_thumbs/.htaccess',
            'uploads/galleries_thumbs/index.html',
            'uploads/watermarks/.htaccess',
            'uploads/watermarks/index.html',
        ];
    }

    /**
     * Get mapping between template names and the method of this class that can render a preview of them.
     *
     * @return array The mapping
     */
    public function tpl_previews() : array
    {
        return [
            'templates/BLOCK_MAIN_IMAGE_FADER.tpl' => 'block_main_image_fader',
            'templates/GALLERY_IMPORT_SCREEN.tpl' => 'administrative__gallery_import_screen',
            'templates/GALLERY_POPULAR.tpl' => 'gallery_popular',
            'templates/GALLERY_IMAGE.tpl' => 'block_main_gallery_embed',
            'templates/GALLERY_ENTRY_WRAP.tpl' => 'block_main_gallery_embed',
            'templates/GALLERY_VIDEO.tpl' => 'block_main_gallery_embed',
            'templates/BLOCK_MAIN_GALLERY_EMBED.tpl' => 'block_main_gallery_embed',
            'templates/MASS_SELECT_MARKER.tpl' => 'block_main_gallery_embed',
            'templates/MASS_SELECT_DELETE_FORM.tpl' => 'block_main_gallery_embed',
            'templates/BLOCK_SIDE_GALLERIES_LINE_DEPTH.tpl' => 'block_side_galleries',
            'templates/BLOCK_SIDE_GALLERIES_LINE.tpl' => 'block_side_galleries',
            'templates/BLOCK_SIDE_GALLERIES_LINE_CONTAINER.tpl' => 'block_side_galleries',
            'templates/BLOCK_SIDE_GALLERIES.tpl' => 'block_side_galleries',
            'templates/GALLERY_VIDEO_INFO.tpl' => 'gallery_video_info',
            'templates/GALLERY_BOX.tpl' => 'gallery_gallery_box',
            'templates/GALLERY_ENTRY_LIST_LINE.tpl' => 'gallery_entry_list_line',
            'templates/GALLERY_CAROUSEL_MODE_IMAGE.tpl' => 'gallery_carousel_mode_image',
            'templates/GALLERY_CAROUSEL_MODE_VIDEO.tpl' => 'gallery_carousel_mode_video',
            'templates/GALLERY_CAROUSEL_ENTRY.tpl' => 'gallery_carousel_mode_image',
            'templates/GALLERY_CAROUSEL_MODE_SCREEN.tpl' => 'gallery_carousel_mode_image',
            'templates/GALLERY_GRID_MODE_SCREEN.tpl' => 'gallery_grid_mode_screen',
            'templates/GALLERY_ENTRY_SCREEN.tpl' => 'gallery_entry_screen',
            'templates/GALLERY_NAV.tpl' => 'gallery_entry_screen',
            'templates/CNS_MEMBER_PROFILE_GALLERIES.tpl' => 'cns_member_profile_galleries',
            'templates/BLOCK_MAIN_PERSONAL_GALLERIES_LIST.tpl' => 'cns_member_profile_galleries',
            'templates/GALLERY_VIDEO_BOX.tpl' => 'gallery_video_box',
            'templates/GALLERY_IMAGE_BOX.tpl' => 'gallery_image_box',
            'templates/BLOCK_MAIN_HERO_SLIDER.tpl' => 'block_main_hero_slider',
            'templates/GALLERY_HOMEPAGE_HERO_SLIDE.tpl' => 'block_main_hero_slider',
            'templates/BLOCK_MAIN_GALLERY_MOSAIC.tpl' => 'block_main_gallery_mosaic',
            'templates/GALLERY_MOSAIC_IMAGE.tpl' => 'block_main_gallery_mosaic',
            'templates/GALLERY_MOSAIC_VIDEO.tpl' => 'block_main_gallery_mosaic',
            'templates/GALLERY_MOSAIC_MODE_SCREEN.tpl' => 'gallery_mosaic_mode_screen',
            'templates/GALLERY_SLIDESHOW_CAROUSEL_ENTRY.tpl' => 'gallery_slideshow_screen',
            'templates/GALLERY_SLIDESHOW_SCREEN.tpl' => 'gallery_slideshow_screen',
        ];
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__gallery_image_box() : object
    {
        return lorem_globalise(do_lorem_template('GALLERY_IMAGE_BOX', [
            'TITLE' => lorem_phrase(),
            'BREADCRUMBS' => lorem_phrase(),
            'ADD_DATE_RAW' => placeholder_date_raw(),
            'ID' => placeholder_numeric_id(),
            'NOTES' => '',
            'GALLERY_TITLE' => lorem_phrase(),
            'CAT' => placeholder_codename(),
            'VIEWS' => placeholder_number(),
            'URL' => placeholder_url(),
            'IMAGE_URL' => placeholder_image_url(),
            'DESCRIPTION' => lorem_paragraph_html(),
            'GIVE_CONTEXT' => true,
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__gallery_video_box() : object
    {
        return lorem_globalise(do_lorem_template('GALLERY_VIDEO_BOX', [
            'TITLE' => lorem_phrase(),
            'BREADCRUMBS' => lorem_phrase(),
            'ADD_DATE_RAW' => placeholder_date_raw(),
            'ID' => placeholder_numeric_id(),
            'NOTES' => '',
            'GALLERY_TITLE' => lorem_phrase(),
            'CAT' => placeholder_codename(),
            'VIEWS' => placeholder_number(),
            'URL' => placeholder_url(),
            'VIDEO_URL' => placeholder_url(),
            'DESCRIPTION' => lorem_paragraph_html(),
            'THUMB_URL' => placeholder_image_url(),
            'VIDEO_WIDTH' => placeholder_number(),
            'VIDEO_HEIGHT' => placeholder_number(),
            'VIDEO_LENGTH' => placeholder_number(),
            'CLOSED_CAPTIONS_URL' => placeholder_url(),
            'GIVE_CONTEXT' => true,
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__cns_member_profile_galleries() : object
    {
        $gallery = do_lorem_template('GALLERY_BOX', [
            'GIVE_CONTEXT' => false,
            'REP_IMAGE_URL' => placeholder_image_url(),
            'NUM_VIDEOS' => placeholder_number(),
            'NUM_IMAGES' => placeholder_number(),
            'NUM_CHILDREN' => placeholder_number(),
            'ID' => placeholder_codename(),
            'LANG' => lorem_word(),
            'ADD_DATE_RAW' => placeholder_date_raw(),
            'ADD_DATE' => placeholder_date(),
            'MEMBER_INFO' => placeholder_member_box(),
            'URL' => placeholder_url(),
            'TITLE' => lorem_phrase(),
            'DESCRIPTION' => lorem_paragraph(),
            'COMMENT_COUNT' => placeholder_number(),
        ]);

        $galleries = do_lorem_template('BLOCK_MAIN_PERSONAL_GALLERIES_LIST', [
            'BLOCK_ID' => lorem_word(),

            'GALLERIES' => $gallery,
            'PAGINATION' => placeholder_pagination(),
            'BLOCK_PARAMS' => '',
            'ADD_GALLERY_URL' => placeholder_url(),
            'ADD_IMAGE_URL' => placeholder_url(),
            'ADD_VIDEO_URL' => placeholder_url(),
            'MEMBER_ID' => placeholder_numeric_id(),

            'START' => '0',
            'MAX' => '10',
            'START_PARAM' => 'x_start',
            'MAX_PARAM' => 'x_max',
        ]);

        return lorem_globalise(do_lorem_template('CNS_MEMBER_PROFILE_GALLERIES', [
            'MEMBER_ID' => placeholder_numeric_id(),
            'GALLERIES' => $galleries,
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__block_main_image_fader() : object
    {
        return lorem_globalise(do_lorem_template('BLOCK_MAIN_IMAGE_FADER', [
            'BLOCK_ID' => lorem_word(),
            'GALLERY_URL' => placeholder_url(),
            'FIRST_URL' => placeholder_image_url(),
            'PREVIOUS_URL' => placeholder_image_url(),
            'NEXT_URL' => placeholder_image_url(),
            'TITLES' => [],
            'IMAGES' => [],
            'HTML' => [],
            'MILL' => '3000',
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__administrative__gallery_import_screen() : object
    {
        require_css('forms');

        require_lang('dearchive');

        // Need to create the form fields (instead of using placeholder_form()) because javascript is using a field called 'files' (list type).
        $fields = new Tempcode();
        $orphaned_content = new Tempcode();
        $orphaned_content->attach(do_lorem_template('FORM_SCREEN_INPUT_LIST_ENTRY', [
            'SELECTED' => false,
            'DISABLED' => false,
            'CLASS' => '',
            'NAME' => 'test',
            'TEXT' => 'test',
        ]));
        $input = do_lorem_template('FORM_SCREEN_INPUT_LIST', [
            'TABINDEX' => placeholder_number(),
            'REQUIRED' => '-required',
            'NAME' => 'files',
            'CONTENT' => $orphaned_content,
            'INLINE_LIST' => true,
            'SIZE' => '9',
        ]);
        $fields->attach(do_lorem_template('FORM_SCREEN_FIELD', [
            'REQUIRED' => true,
            'SKIP_LABEL' => false,
            'NAME' => 'files',
            'PRETTY_NAME' => lorem_phrase(),
            'DESCRIPTION' => lorem_paragraph_html(),
            'DESCRIPTION_SIDE' => '',
            'INPUT' => $input,
            'COMCODE' => new Tempcode(),
        ]));

        $form = do_lorem_template('FORM', [
            'TABINDEX' => placeholder_number(),
            'FIELDS' => $fields,
            'SUBMIT_ICON' => 'admin/import',
            'SUBMIT_NAME' => lorem_word(),
            'SUBMIT_BUTTON_CLASS' => null,
            'URL' => placeholder_url(),
            'TEXT' => lorem_phrase(),
            'HIDDEN' => '',
            'JS_FUNCTION_CALLS' => [],
            'SKIP_REQUIRED' => null,
            'SECONDARY_FORM' => false,
            'SUPPORT_AUTOSAVE' => null,
            'ANALYTIC_EVENT_CATEGORY' => null,
            'MODSECURITY_WORKAROUND' => null,
            'SKIP_WEBSTANDARDS' => null,
            'GET' => null,
            'NO_SIZING' => null,
            'PREVIEW' => null,
            'TITLE' => null,
        ]);

        return lorem_globalise(do_lorem_template('GALLERY_IMPORT_SCREEN', [
            'TITLE' => lorem_screen_title(),
            'FORM2' => placeholder_form(),
            'FORM' => $form,
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__gallery_popular() : object
    {
        return lorem_globalise(do_lorem_template('GALLERY_POPULAR', [
            'CHILDREN' => lorem_sentence_html(),
            'CAT' => 'root',
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__block_main_gallery_embed() : object
    {
        // Create the 'GALLERY_ENTRY_WRAP' template
        $entry = new Tempcode();
        $map = [
            'MEDIA_TYPE' => 'image',
            'TITLE' => lorem_phrase(),
            'DESCRIPTION' => lorem_paragraph_html(),
            'ID' => placeholder_numeric_id(),
            'FILE_SIZE' => lorem_word(),
            'SUBMITTER' => placeholder_numeric_id(),
            'IMAGE_URL' => placeholder_image_url(),
            'CAT' => lorem_word(),
            'VIEW_URL' => placeholder_url(),
            'ADD_DATE_RAW' => placeholder_date_raw(),
            'EDIT_DATE_RAW' => placeholder_date_raw(),
            'ADD_DATE' => placeholder_date(),
            'EDIT_DATE' => placeholder_date(),
            'VIEWS' => placeholder_number(),
            '_EDIT_URL' => placeholder_url(),
            'COMMENT_COUNT' => true,
        ];
        $entry = do_lorem_template('GALLERY_IMAGE', $map);
        $entries = new Tempcode();
        $entries->attach(do_lorem_template('GALLERY_ENTRY_WRAP', ['ENTRY' => $entry] + $map));

        $video_details = do_lorem_template('GALLERY_VIDEO_INFO', [
            'HEIGHT' => placeholder_number(),
            'WIDTH' => placeholder_number(),
            'LENGTH' => placeholder_number(),
            'CLOSED_CAPTIONS_URL' => placeholder_url(),
        ]);
        $map = [
            'MEDIA_TYPE' => 'video',
            'TITLE' => lorem_phrase(),
            'DESCRIPTION' => lorem_phrase(),
            'ID' => placeholder_numeric_id(),
            'FILE_SIZE' => lorem_word(),
            'SUBMITTER' => placeholder_numeric_id(),
            'FULL_URL' => placeholder_url(),
            'THUMB_URL' => placeholder_image_url(),
            'CAT' => lorem_word(),
            'VIEW_URL' => placeholder_url(),
            'VIDEO_DETAILS' => $video_details,
            'ADD_DATE_RAW' => placeholder_date_raw(),
            'EDIT_DATE_RAW' => placeholder_date_raw(),
            'ADD_DATE' => placeholder_date(),
            'EDIT_DATE' => placeholder_date(),
            '_EDIT_URL' => placeholder_url(),
            'VIEWS' => placeholder_number(),
            'COMMENT_COUNT' => true,
        ];
        $entry = do_lorem_template('GALLERY_VIDEO', $map);
        $entries->attach(do_lorem_template('GALLERY_ENTRY_WRAP', ['ENTRY' => $entry] + $map));

        // Create 'BLOCK_MAIN_GALLERY_EMBED' with 'GALLERY_ENTRY_WRAP' as sub-template
        return lorem_globalise(do_lorem_template('BLOCK_MAIN_GALLERY_EMBED', [
            'BLOCK_ID' => lorem_word(),
            'TITLE' => lorem_phrase(),
            'CAT' => placeholder_codename(),
            'ENTRIES' => $entries,
            'PAGINATION' => placeholder_pagination(),
            'BLOCK_PARAMS' => '',

            'START' => '0',
            'MAX' => '10',
            'START_PARAM' => 'x_start',
            'MAX_PARAM' => 'x_max',
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__block_side_galleries() : object
    {
        $content = new Tempcode();
        foreach (placeholder_array() as $k => $v) {
            $out = new Tempcode();
            foreach (placeholder_array() as $_k => $_v) {
                $out->attach(do_lorem_template('BLOCK_SIDE_GALLERIES_LINE_DEPTH', ['TITLE' => lorem_phrase(), 'URL' => placeholder_url()]));
            }
            $out->attach(do_lorem_template('BLOCK_SIDE_GALLERIES_LINE', ['TITLE' => lorem_phrase(), 'URL' => placeholder_url()]));

            $content->attach(do_lorem_template('BLOCK_SIDE_GALLERIES_LINE_CONTAINER', ['ID' => placeholder_random_id(), 'CAPTION' => lorem_phrase(), 'CONTENTS' => $out]));
        }

        return lorem_globalise(do_lorem_template('BLOCK_SIDE_GALLERIES', [
            'BLOCK_ID' => lorem_word(),
            'DEPTH' => true,
            'CONTENT' => $content,
            'TITLE' => lorem_phrase(),
            'ID' => placeholder_codename(),
        ]), null, '', true);
    }

    /**placeholder_date_raw()
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__gallery_video_info() : object
    {
        return lorem_globalise(do_lorem_template('GALLERY_VIDEO_INFO', [
            'HEIGHT' => placeholder_number(),
            'WIDTH' => placeholder_number(),
            'LENGTH' => placeholder_number(),
            'CLOSED_CAPTIONS_URL' => placeholder_url(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__gallery_entry_list_line() : object
    {
        return lorem_globalise(do_lorem_template('GALLERY_ENTRY_LIST_LINE', [
            'BREADCRUMBS' => lorem_phrase(),
            'URL' => placeholder_url(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__gallery_carousel_mode_video() : object
    {
        require_javascript('editing');

        $video = do_lorem_template('GALLERY_CAROUSEL_MODE_VIDEO', [
            'MAIN' => lorem_phrase(),
            '_TITLE' => lorem_phrase(),
            'DESCRIPTION' => lorem_paragraph_html(),
            'CAT' => placeholder_codename(),
            'RATING_DETAILS' => placeholder_rating('video'),
            'THUMB_URL' => placeholder_image_url(),
            'FULL_URL' => placeholder_url(),
            'ID' => placeholder_numeric_id(),
            'VIEWS' => lorem_phrase(),
            'ADD_DATE_RAW' => placeholder_date_raw(),
            'EDIT_DATE_RAW' => placeholder_date_raw(),
            'SUBMITTER' => placeholder_numeric_id(),
            'VIDEO_PLAYER' => placeholder_image(),
            'VIDEO_DETAILS' => lorem_phrase(),
            'VIEW_URL' => placeholder_url(),
            'EDIT_URL' => placeholder_url(),
            'COMMENT_COUNT' => true,
        ]);

        $entries = new Tempcode();
        foreach (placeholder_array(10) as $k => $v) {
            $entries->attach(do_lorem_template('GALLERY_CAROUSEL_ENTRY', [
                'DESCRIPTION' => lorem_paragraph_html(),
                '_TITLE' => lorem_phrase(),
                'ID' => strval($k),
                'VIEWS' => placeholder_number(),
                'ADD_DATE_RAW' => placeholder_date_raw(),
                'EDIT_DATE_RAW' => placeholder_date_raw(),
                'SUBMITTER' => placeholder_numeric_id(),
                'CLASS' => lorem_word(),
                'IMAGE_URL' => placeholder_image_url(),
                'VIEW_URL' => placeholder_url(),
                'VIEW_URL_2' => placeholder_url(),
                'TYPE' => lorem_word(),
                '_EDIT_URL' => placeholder_url(),
                'CAT' => placeholder_codename(),
            ]));
        }

        return lorem_globalise(do_lorem_template('GALLERY_CAROUSEL_MODE_SCREEN', [
            '_TITLE' => lorem_phrase(),
            'VIEW_URL' => placeholder_url(),
            'FULL_URL' => placeholder_url(),
            'PROBE_TYPE' => lorem_phrase(),
            'ENTRY_TITLE' => lorem_phrase(),
            'ENTRY_DESCRIPTION' => lorem_paragraph_html(),
            'ENTRY_SUBMITTER' => placeholder_numeric_id(),
            'ENTRY_VIEWS' => placeholder_number(),
            'ENTRY_ADD_DATE_RAW' => placeholder_date_raw(),
            'ENTRY_EDIT_DATE_RAW' => placeholder_date_raw(),
            'ENTRY_TRACKBACK_DETAILS' => placeholder_trackbacks_wrapper(),
            'ENTRY_RATING_DETAILS' => placeholder_rating('image'),
            'ENTRY_COMMENT_DETAILS' => placeholder_comments(placeholder_comments_form()),
            'ENTRY_EDIT_URL' => placeholder_url(),
            'WARNING_DETAILS' => '',
            'TAGS' => placeholder_tags(),
            'RATING_DETAILS' => placeholder_rating('gallery'),
            'COMMENT_DETAILS' => placeholder_comments(placeholder_comments_form()),
            'REP_IMAGE_URL' => placeholder_image_url(),
            'TITLE' => lorem_screen_title(),
            'DESCRIPTION' => lorem_paragraph_html(),
            'CURRENT_ENTRY' => $video,
            'ENTRIES' => $entries,
            'EDIT_URL' => placeholder_url(),
            'ADD_GALLERY_URL' => placeholder_url(),
            'IMAGE_URL' => placeholder_image_url(),
            'MEMBER_ID' => placeholder_numeric_id(),
            'VIDEO_URL' => placeholder_url(),
            'MAY_DOWNLOAD' => lorem_phrase(),
            'CAT' => placeholder_codename(),
            'FIRST_ENTRY_ID' => placeholder_numeric_id(),
            'SORTING' => lorem_phrase(),
            'ACCEPT_IMAGES' => true,
            'ACCEPT_VIDEOS' => true,
            'VIEWS' => placeholder_number(),
            'OWNER' => null,
            'PROBE_ID' => placeholder_numeric_id(),
            'GALLERY_SORT' => 'title ASC',
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__gallery_carousel_mode_image() : object
    {
        require_javascript('editing');

        $image = do_lorem_template('GALLERY_CAROUSEL_MODE_IMAGE', [
            'MAIN' => lorem_phrase(),
            '_TITLE' => lorem_phrase(),
            'DESCRIPTION' => lorem_paragraph_html(),
            'FILE_SIZE' => placeholder_filesize(),
            'RATING_DETAILS' => placeholder_rating('image'),
            'CAT' => placeholder_codename(),
            'IMAGE_URL' => placeholder_image_url(),
            'ID' => placeholder_numeric_id(),
            'VIEWS' => lorem_phrase(),
            'ADD_DATE_RAW' => placeholder_date_raw(),
            'EDIT_DATE_RAW' => placeholder_date_raw(),
            'SUBMITTER' => placeholder_numeric_id(),
            'VIEW_URL' => placeholder_url(),
            'EDIT_URL' => placeholder_url(),
            'COMMENT_COUNT' => true,
        ]);

        $entries = new Tempcode();
        foreach (placeholder_array(10) as $k => $v) {
            $entries->attach(do_lorem_template('GALLERY_CAROUSEL_ENTRY', [
                'DESCRIPTION' => lorem_paragraph_html(),
                '_TITLE' => lorem_phrase(),
                'ID' => strval($k),
                'VIEWS' => placeholder_number(),
                'ADD_DATE_RAW' => placeholder_date_raw(),
                'EDIT_DATE_RAW' => placeholder_date_raw(),
                'SUBMITTER' => placeholder_numeric_id(),
                'CLASS' => lorem_word(),
                'IMAGE_URL' => placeholder_image_url(),
                'VIEW_URL' => placeholder_url(),
                'VIEW_URL_2' => placeholder_url(),
                'TYPE' => lorem_word(),
                '_EDIT_URL' => placeholder_url(),
                'CAT' => placeholder_codename(),
            ]));
        }

        return lorem_globalise(do_lorem_template('GALLERY_CAROUSEL_MODE_SCREEN', [
            '_TITLE' => lorem_phrase(),
            'VIEW_URL' => placeholder_url(),
            'FULL_URL' => placeholder_url(),
            'PROBE_TYPE' => lorem_phrase(),
            'ENTRY_VIEWS' => placeholder_number(),
            'ENTRY_ADD_DATE_RAW' => placeholder_date_raw(),
            'ENTRY_EDIT_DATE_RAW' => placeholder_date_raw(),
            'ENTRY_TRACKBACK_DETAILS' => placeholder_trackbacks_wrapper(),
            'ENTRY_RATING_DETAILS' => placeholder_rating('video'),
            'ENTRY_COMMENT_DETAILS' => placeholder_comments(placeholder_comments_form(false)),
            'ENTRY_EDIT_URL' => placeholder_url(),
            'WARNING_DETAILS' => '',
            'TAGS' => placeholder_tags(),
            'RATING_DETAILS' => placeholder_rating('gallery'),
            'COMMENT_DETAILS' => placeholder_comments(placeholder_comments_form(false)),
            'REP_IMAGE_URL' => placeholder_image_url(),
            'TITLE' => lorem_screen_title(),
            'DESCRIPTION' => lorem_paragraph_html(),
            'CURRENT_ENTRY' => $image,
            'ENTRIES' => $entries,
            'EDIT_URL' => placeholder_url(),
            'ADD_GALLERY_URL' => placeholder_url(),
            'IMAGE_URL' => placeholder_image_url(),
            'MEMBER_ID' => placeholder_numeric_id(),
            'VIDEO_URL' => placeholder_url(),
            'MAY_DOWNLOAD' => lorem_phrase(),
            'CAT' => placeholder_codename(),
            'FIRST_ENTRY_ID' => placeholder_numeric_id(),
            'SORTING' => lorem_phrase(),
            'ACCEPT_IMAGES' => true,
            'ACCEPT_VIDEOS' => true,
            'VIEWS' => placeholder_number(),
            'OWNER' => null,
            'PROBE_ID' => placeholder_numeric_id(),
            'ENTRY_TITLE' => lorem_phrase(),
            'ENTRY_DESCRIPTION' => lorem_paragraph_html(),
            'ENTRY_SUBMITTER' => placeholder_numeric_id(),
            'GALLERY_SORT' => 'title ASC',
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__gallery_gallery_box() : object
    {
        return lorem_globalise(do_lorem_template('GALLERY_BOX', [
            'GIVE_CONTEXT' => false,
            'REP_IMAGE_URL' => placeholder_image_url(),
            'NUM_VIDEOS' => placeholder_number(),
            'NUM_IMAGES' => placeholder_number(),
            'NUM_CHILDREN' => placeholder_number(),
            'ID' => placeholder_codename(),
            'LANG' => lorem_word(),
            'ADD_DATE_RAW' => placeholder_date_raw(),
            'ADD_DATE' => placeholder_date(),
            'MEMBER_INFO' => placeholder_member_box(),
            'URL' => placeholder_url(),
            'TITLE' => lorem_phrase(),
            'DESCRIPTION' => lorem_paragraph_html(),
            'COMMENT_COUNT' => placeholder_number(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__gallery_grid_mode_screen() : object
    {
        require_javascript('editing');

        return lorem_globalise(do_lorem_template('GALLERY_GRID_MODE_SCREEN', [
            '_TITLE' => lorem_phrase(),
            'TAGS' => placeholder_tags(),
            'CAT' => lorem_word(),
            'RATING_DETAILS' => placeholder_rating('gallery'),
            'COMMENT_DETAILS' => placeholder_comments(placeholder_comments_form(false)),
            'ADD_GALLERY_URL' => placeholder_url(),
            'EDIT_URL' => placeholder_url(),
            'TITLE' => lorem_screen_title(),
            'DESCRIPTION' => lorem_paragraph_html(),
            'IMAGE_URL' => placeholder_url(),
            'VIDEO_URL' => placeholder_url(),
            'MAY_DOWNLOAD' => lorem_phrase(),
            'MEMBER_ID' => placeholder_numeric_id(),
            'ACCEPT_IMAGES' => true,
            'ACCEPT_VIDEOS' => true,
            'VIEWS' => placeholder_number(),
            'OWNER' => placeholder_numeric_id(),
            'CAT_SELECT' => '*',
            'DAYS' => '',
            'IMAGE_SELECT' => '*',
            'VIDEO_SELECT' => '*',
            'FILTER' => '',
            'GALLERY_SORT' => 'title ASC',
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__gallery_entry_screen() : object
    {
        require_javascript('editing');

        $nav = do_lorem_template('GALLERY_NAV', [
            'BACK_URL' => placeholder_url(),
            'NEXT_URL' => placeholder_url(),
            'BACK_IMAGE_URL' => placeholder_image_url(),
            'BACK_ENTRY_TITLE' => lorem_phrase(),
            'NEXT_IMAGE_URL' => placeholder_image_url(),
            'NEXT_ENTRY_TITLE' => lorem_phrase(),
            '_X' => placeholder_number(),
            '_N' => placeholder_number(),
            'X' => placeholder_number(),
            'N' => placeholder_number(),
            'SLIDESHOW_URL' => placeholder_url(),
            'SLIDESHOW_NEXT_URL' => placeholder_url(),
            'SLIDESHOW_PREVIOUS_URL' => placeholder_url(),
            'MORE_URL' => placeholder_url(),
            'CATEGORY_NAME' => lorem_word(),
        ]);

        return lorem_globalise(do_lorem_template('GALLERY_ENTRY_SCREEN', [
            'CAT' => placeholder_codename(),
            'MEDIA_TYPE' => 'video',
            'ID' => placeholder_numeric_id(),
            'TRUE_GALLERY_TITLE' => lorem_phrase(),
            'E_TITLE' => lorem_phrase(),
            'GALLERY_TITLE' => lorem_phrase(),
            'MEMBER_ID' => placeholder_numeric_id(),
            'TAGS' => placeholder_tags(),
            'TITLE' => lorem_screen_title(),
            'SUBMITTER' => placeholder_numeric_id(),
            'URL' => placeholder_image_url(), // Raw URL
            '_X' => placeholder_number(),
            'X' => placeholder_number(),
            '_N' => placeholder_number(),
            'N' => placeholder_number(),
            '_VIEWS' => placeholder_number(),
            'VIEWS' => placeholder_number(),
            'ADD_DATE_RAW' => placeholder_date_raw(),
            'EDIT_DATE_RAW' => placeholder_date_raw(),
            'ADD_DATE' => placeholder_date(),
            'EDIT_DATE' => placeholder_date(),
            'RATING_DETAILS' => placeholder_rating('video'),
            'TRACKBACK_DETAILS' => placeholder_trackbacks_wrapper(),
            'COMMENT_DETAILS' => placeholder_comments(placeholder_comments_form(false)),
            'EDIT_URL' => placeholder_url(),
            'IMAGE_URL' => placeholder_image_url(), // Thumbnail URL
            'NAV' => $nav,
            'DESCRIPTION' => lorem_phrase(),
            'VIDEO' => null,
            'WARNING_DETAILS' => '',
            'VIDEO_DETAILS' => null,
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__block_main_hero_slider() : object
    {
        $content_html = do_lorem_template('GALLERY_HOMEPAGE_HERO_SLIDE', [
            'HEADLINE' => lorem_phrase(),
            'SUBLINE' => lorem_phrase(),
            'TEXT' => lorem_chunk_html(),
            'LINK1_URL' => placeholder_url(),
            'LINK2_URL' => placeholder_url(),
            'LINK1_TEXT' => lorem_phrase(),
            'LINK2_TEXT' => lorem_phrase(),
        ]);

        $items = [
            [
                'BACKGROUND_ID' => placeholder_codename(),
                'BACKGROUND_TYPE' => lorem_word(),
                'BACKGROUND_URL' => placeholder_url(),
                'BACKGROUND_IMAGE_URL' => placeholder_image_url(),
                'CONTENT_HTML' => $content_html,
                'EDIT_URL' => placeholder_url(),
            ],
        ];

        return lorem_globalise(do_lorem_template('BLOCK_MAIN_HERO_SLIDER', [
            'BLOCK_ID' => placeholder_codename(),
            'EFFECT' => 'slide',
            'FULLSCREEN' => '1',
            'SHOW_INDICATORS' => '1',
            'SHOW_SCROLL_DOWN' => '0',
            'INTERVAL' => '0',
            'ITEMS' => $items,
            'HAS_MULTIPLE_ITEMS' => (count($items) > 1) ? '1' : '0',
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__block_main_gallery_mosaic() : object
    {
        $entries = new Tempcode();

        foreach (placeholder_array() as $k => $v) {
            $map = [
                'RATING_DETAILS' => placeholder_rating('image', 'RATING_INLINE_STATIC'),
                'TITLE' => lorem_phrase(),
                'DESCRIPTION' => lorem_paragraph_html(),
                'ID' => placeholder_random_id(),
                'MEDIA_TYPE' => 'image',
                'FILE_SIZE' => lorem_word(),
                'SUBMITTER' => placeholder_numeric_id(),
                'CAT' => lorem_word(),
                'IMAGE_URL' => placeholder_image_url(),
                'VIEW_URL' => placeholder_url(),
                'VIEWS' => placeholder_number(),
                'ADD_DATE_RAW' => placeholder_date_raw(),
                'EDIT_DATE_RAW' => placeholder_date_raw(),
                '_EDIT_URL' => placeholder_url(),
                'COMMENT_COUNT' => true,
            ];
            $entries->attach(do_lorem_template('GALLERY_MOSAIC_IMAGE', $map));
        }


        $video_details = do_lorem_template('GALLERY_VIDEO_INFO', [
            'HEIGHT' => placeholder_number(),
            'WIDTH' => placeholder_number(),
            'LENGTH' => placeholder_number(),
            'CLOSED_CAPTIONS_URL' => placeholder_url(),
        ]);

        foreach (placeholder_array() as $k => $v) {
            $map = [
                'RATING_DETAILS' => placeholder_rating('video', 'RATING_INLINE_STATIC'),
                'TITLE' => lorem_phrase(),
                'DESCRIPTION' => lorem_paragraph_html(),
                'ID' => placeholder_random_id(),
                'MEDIA_TYPE' => 'video',
                'FILE_SIZE' => lorem_word(),
                'SUBMITTER' => placeholder_numeric_id(),
                'FULL_URL' => placeholder_url(),
                'CAT' => lorem_word(),
                'THUMB_URL' => placeholder_image_url(),
                'VIEW_URL' => placeholder_url(),
                'VIDEO_DETAILS' => $video_details,
                'VIEWS' => placeholder_number(),
                'ADD_DATE_RAW' => placeholder_date_raw(),
                'EDIT_DATE_RAW' => placeholder_date_raw(),
                '_EDIT_URL' => placeholder_url(),
                'COMMENT_COUNT' => true,
            ];
            $entries->attach(do_lorem_template('GALLERY_MOSAIC_VIDEO', $map));
        }

        return lorem_globalise(do_lorem_template('BLOCK_MAIN_GALLERY_MOSAIC', [
            'BLOCK_ID' => placeholder_codename(),
            'BLOCK_PARAMS' => '',
            'SLIDESHOW_URL' => null,
            'PAGINATION' => placeholder_pagination(),
            'ENTRIES' => $entries,
            'START' => strval(0),
            'MAX' => strval(10),
            'START_PARAM' => placeholder_codename() . '_start',
            'MAX_PARAM' => placeholder_codename() . '_max',
            'EXTRA_GET_PARAMS' => null,
            'TITLE' => lorem_phrase(),
            'CAT' => lorem_word(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__gallery_mosaic_mode_screen() : object
    {
        return lorem_globalise(do_lorem_template('GALLERY_MOSAIC_MODE_SCREEN', [
            '_TITLE' => lorem_phrase(),
            'MEMBER_ID' => '',
            'TAGS' => placeholder_tags(),
            'CAT' => '',
            'RATING_DETAILS' => placeholder_rating('gallery', 'RATING_INLINE_STATIC'),
            'COMMENT_DETAILS' => placeholder_comments(placeholder_comments_form(false)),
            'ADD_GALLERY_URL' => placeholder_url(),
            'EDIT_URL' => placeholder_url(),
            'TITLE' => lorem_screen_title(),
            'DESCRIPTION' => lorem_paragraph_html(),
            'IMAGE_URL' => placeholder_url(),
            'VIDEO_URL' => placeholder_url(),
            'MAY_DOWNLOAD' => true,
            'ACCEPT_IMAGES' => true,
            'ACCEPT_VIDEOS' => true,
            'VIEWS' => strval(123),
            'OWNER' => null,
            'CAT_SELECT' => '*',
            'DAYS' => '',
            'IMAGE_SELECT' => '*',
            'VIDEO_SELECT' => '*',
            'FILTER' => '',
            'GALLERY_SORT' => 'title ASC',
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__gallery_slideshow_screen() : object
    {
        $carousel_entries = new Tempcode();

        $carousel_entries->attach(do_template('GALLERY_SLIDESHOW_CAROUSEL_ENTRY', [
            '_GUID' => '4aebcb6841a8f31809762ec990b8931f',
            'INDEX' => strval(0),
            'DESCRIPTION' => lorem_sentence(),
            '_TITLE' => lorem_phrase(),
            'TYPE' => 'image',
            'ID' => placeholder_numeric_id(),
            'VIEWS' => '123',
            'ADD_DATE_RAW' => placeholder_date_raw(),
            'EDIT_DATE_RAW' => '',
            'SUBMITTER' => placeholder_numeric_id(),
            'IS_CURRENT' => true,
            'IMAGE_URL' => placeholder_image_url(),
            'FULL_URL' => placeholder_url(),
            'VIEW_URL' => placeholder_url(),
            'VIEW_URL_2' => placeholder_url(),
            '_EDIT_URL' => placeholder_url(),
            'CAT' => 'root',
            'COMMENTS_OPTIONS_HASH' => '',
        ]));

        return lorem_globalise(do_lorem_template('GALLERY_SLIDESHOW_SCREEN', [
            'SLIDESHOW_ID' => placeholder_codename(),
            '_TOTAL_ITEMS' => placeholder_number(),
            'TOTAL_ITEMS' => placeholder_number(),
            'CURRENT_INDEX' => placeholder_number(),
            'CURRENT_URL' => placeholder_image_url(),
            'CURRENT_VIDEO' => null,
            'CURRENT_TYPE' => 'image',
            'CURRENT_COMMENT_DETAILS' => '',
            'CAROUSEL_ENTRIES' => $carousel_entries,
        ]), null, '', true);
    }

    /**
     * Uninstall default content.
     */
    public function uninstall_test_content()
    {
        require_code('galleries2');

        $to_delete = $GLOBALS['SITE_DB']->query_select('images', ['id'], [$GLOBALS['SITE_DB']->translate_field_ref('title') => lorem_phrase()]);
        foreach ($to_delete as $record) {
            delete_image($record['id']);
        }

        $to_delete = $GLOBALS['SITE_DB']->query_select('videos', ['id'], [$GLOBALS['SITE_DB']->translate_field_ref('title') => lorem_phrase()]);
        foreach ($to_delete as $record) {
            delete_video($record['id']);
        }

        $to_delete = $GLOBALS['SITE_DB']->query_select('galleries', ['name'], [$GLOBALS['SITE_DB']->translate_field_ref('fullname') => lorem_phrase()]);
        foreach ($to_delete as $record) {
            delete_gallery($record['name']);
        }
    }

    /**
     * Install default content.
     */
    public function install_test_content()
    {
        require_code('galleries2');

        if ($GLOBALS['SITE_DB']->query_select_value_if_there('galleries', 'name', ['name' => 'lorem_1']) === null) {
            add_gallery('lorem_1', lorem_phrase(), lorem_chunk(), '', 'root', 1, 1, 0);
            require_code('permissions2');
            set_global_category_access('galleries', 'lorem_1');
        }
        if ($GLOBALS['SITE_DB']->query_select_value_if_there('galleries', 'name', ['name' => 'lorem_2']) === null) {
            add_gallery('lorem_2', lorem_phrase(), lorem_chunk(), '', 'root', 1, 1, 1, GALLERY_LAYOUT_MODE_CAROUSEL);
            require_code('permissions2');
            set_global_category_access('galleries', 'lorem_2');

            add_gallery('member_' . strval(get_member()) . '_lorem_2', lorem_phrase(), lorem_chunk(), '', 'lorem_2', 1, 1, 1, GALLERY_LAYOUT_MODE_CAROUSEL);
            require_code('permissions2');
            set_global_category_access('galleries', 'lorem_2');
        }

        add_image(lorem_phrase(), 'lorem_1', lorem_chunk(), placeholder_image_url(), 1, 1, 1, 1, '');

        add_video(lorem_phrase(), 'lorem_1', lorem_chunk(), placeholder_image_url(), placeholder_image_url(), 1, 1, 1, 1, '', 5, 100, 100);

        add_image(lorem_phrase(), 'lorem_2', lorem_chunk(), placeholder_image_url(), 1, 1, 1, 1, '');

        add_image(lorem_phrase(), 'member_' . strval(get_member()) . '_lorem_2', lorem_chunk(), placeholder_image_url(), 1, 1, 1, 1, '');
    }

    /**
     * Find available predefined content, and what is installed.
     *
     * @return array A map of available predefined content codenames, and details (if installed, and title)
     */
    public function enumerate_predefined_content() : array
    {
        require_lang('homepage_hero_slider');

        $default_homepage_hero_slider_category = 'homepage_hero_slider';

        $default_homepage_hero_slides_urls = [
            'data/images/homepage_hero_slider/bastei_bridge.jpg',
            'data/images/homepage_hero_slider/rustic.jpg',
            'data/images/homepage_hero_slider/waterfall.jpg',
        ];

        $where = [];
        foreach ($default_homepage_hero_slides_urls as $url) {
            $where[] = '(' . db_string_equal_to('cat', $default_homepage_hero_slider_category) . ' AND ' . db_string_equal_to('url', $url) . ')';
        }
        $or_list = implode(' OR ', $where);

        $installed = ($GLOBALS['SITE_DB']->query_select_value('images', 'COUNT(*)', [], ' AND (' . $or_list . ')') > 0);

        return [
            'have_default_homepage_hero_slides' => [
                'title' => do_lang_tempcode('HAVE_DEFAULT_HOMEPAGE_HERO_SLIDES'),
                'description' => do_lang_tempcode('DESCRIPTION_HAVE_DEFAULT_HOMEPAGE_HERO_SLIDES'),
                'installed' => $installed,
            ],
        ];
    }

    /**
     * Install predefined content.
     *
     * @param  ?array $content A list of predefined content labels to install (null: all)
     */
    public function install_predefined_content(?array $content = null)
    {
        if ((($content === null) || (in_array('have_default_homepage_hero_slides', $content))) && (!has_predefined_content('galleries', 'have_default_homepage_hero_slides'))) {
            require_lang('homepage_hero_slider');
            require_code('galleries');
            require_code('galleries2');
            require_code('lang3');

            add_gallery('homepage_hero_slider', 'Homepage Hero Slider', 'Slides for the homepage hero slider', '', 'root', 1, 1, 0, GALLERY_LAYOUT_MODE_DEFAULT, '', '', '', '', '', '', '', 0, 0);
            set_global_category_access('galleries', 'homepage_hero_slider');

            $image_owner_id = get_first_admin_user();
            foreach (['bastei_bridge', 'rustic', 'waterfall'] as $i => $img) {
                $slider_insert_map = null;
                $langs = find_all_langs();
                $langs = [user_lang() => $langs[user_lang()]] + $langs;
                foreach ($langs as $lang => $lang_type) {
                    if ((is_file(get_file_base() . '/lang/' . $lang . '/critical_error.ini')) || (is_file(get_file_base() . '/lang_custom/' . $lang . '/critical_error.ini'))) { // Make sure it's a reasonable looking pack, not just a stub
                        $slide_contents = trim('
{+START,INCLUDE,GALLERY_HOMEPAGE_HERO_SLIDE}
HEADLINE=' . do_lang('DEFAULT_SLIDE' . strval($i + 1) . '_HEADLINE', null, null, null, $lang) . '
SUBLINE=' . do_lang('DEFAULT_SLIDE' . strval($i + 1) . '_SUBLINE', null, null, null, $lang) . '
TEXT=' . do_lang('DEFAULT_SLIDE' . strval($i + 1) . '_TEXT', null, null, null, $lang) . '
LINK1_URL=' . do_lang('DEFAULT_LINK1_URL', escape_html(get_brand_base_url()), null, null, $lang) . '
LINK1_TEXT=' . do_lang('DEFAULT_LINK1_TEXT', null, null, null, $lang) . '
LINK2_URL=' . do_lang('DEFAULT_LINK2_URL', escape_html(get_brand_base_url()), null, null, $lang) . '
LINK2_TEXT=' . do_lang('DEFAULT_LINK2_TEXT', null, null, null, $lang) . '
{+END}
{$,page hint: no_wysiwyg}
                        ');

                        if ($lang == user_lang()) {
                            $slider_insert_map = insert_lang('the_description', $slide_contents, 3, null, true, null, null, true);
                        } elseif (multi_lang_content()) {
                            insert_lang('the_description', $slide_contents, 3, null, true, $slider_insert_map['the_description'], $lang, true);
                        }
                    }
                }

                add_image(
                    lang_code_to_default_content('title', 'DEFAULT_SLIDE_X', false, 2, null, integer_format($i + 1)),
                    'homepage_hero_slider',
                    $slider_insert_map,
                    'data/images/homepage_hero_slider/' . $img . '.jpg',
                    1,
                    0,
                    0,
                    0,
                    '',
                    $image_owner_id,
                    null,
                    null,
                    0
                );
            }
        }
    }

    /**
     * Uninstall predefined content.
     *
     * @param  ?array $content A list of predefined content labels to uninstall (null: all)
     */
    public function uninstall_predefined_content(?array $content = null)
    {
        if ((($content === null) || (in_array('have_default_homepage_hero_slides', $content))) && (has_predefined_content('galleries', 'have_default_homepage_hero_slides'))) {
            require_code('galleries2');

            $default_homepage_hero_slider_category = 'homepage_hero_slider';

            // Delete default slide images
            $default_homepage_hero_slides_urls = [
                'data/images/homepage_hero_slider/bastei_bridge.jpg',
                'data/images/homepage_hero_slider/rustic.jpg',
                'data/images/homepage_hero_slider/waterfall.jpg',
            ];
            $where = [];
            foreach ($default_homepage_hero_slides_urls as $url) {
                $where[] = '(' . db_string_equal_to('cat', $default_homepage_hero_slider_category) . ' AND ' . db_string_equal_to('url', $url) . ')';
            }
            $rows = $GLOBALS['SITE_DB']->query('SELECT id FROM ' . get_table_prefix() . 'images WHERE ' . implode(' OR ', $where));
            foreach ($rows as $row) {
                delete_image($row['id']);
            }

            // Delete the category as well if now empty
            $image_count = $GLOBALS['SITE_DB']->query_select_value('images', 'COUNT(*)', ['cat' => $default_homepage_hero_slider_category]);
            if ($image_count == 0) {
                $GLOBALS['SITE_DB']->query_delete('galleries', ['name' => $default_homepage_hero_slider_category]);
            }
        }
    }
}
