<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    news_shared
 */

/**
 * Hook class.
 */
class Hook_addon_registry_news_shared
{
    /**
     * Get a list of file permissions to set.
     *
     * @param  boolean $runtime Whether to include wildcards represented runtime-created chmoddable files
     * @return array File permissions to set
     */
    public function get_chmod_array(bool $runtime = false) : array
    {
        return [];
    }

    /**
     * Get the version of the software this addon is for (used in generating the TAR filename).
     *
     * @return float Version number
     */
    public function get_version() : float
    {
        return cms_version_number();
    }

    /**
     * Get the minimum required version of the website software needed to use this addon.
     *
     * @return float Minimum required website software version
     */
    public function get_min_cms_version() : float
    {
        return 11.0;
    }

    /**
     * Get the maximum compatible version of the website software to use this addon.
     *
     * @return ?float Maximum compatible website software version (null: no maximum version currently)
     */
    public function get_max_cms_version() : ?float
    {
        return 11.9;
    }

    /**
     * Get the addon category.
     *
     * @return string The category
     */
    public function get_category() : string
    {
        return 'Information Display';
    }

    /**
     * Get the description of the addon.
     *
     * @return string Description of the addon
     */
    public function get_description() : string
    {
        return '(Common files needed for RSS and News addons)';
    }

    /**
     * Get a list of tutorials that apply to this addon.
     *
     * @return array List of tutorials
     */
    public function get_applicable_tutorials() : array
    {
        return [
            'tut_news',
            'tut_adv_news',
            'tut_site_messaging',
        ];
    }

    /**
     * Get a mapping of dependency types.
     *
     * @return array A structure specifying dependency information
     */
    public function get_dependencies() : array
    {
        return [
            'requires' => [],
            'recommends' => [],
            'conflicts_with' => [],
        ];
    }

    /**
     * Explicitly say which icon should be used.
     *
     * @return URLPATH Icon
     */
    public function get_default_icon() : string
    {
        return 'themes/default/images/icons/admin/component.svg';
    }

    /**
     * Get a list of files that belong to this addon.
     *
     * @return array List of files
     */
    public function get_file_list() : array
    {
        return [
            'lang/EN/news.ini',
            'sources/hooks/systems/addon_registry/news_shared.php',
            'themes/default/css/news.css',
            'themes/default/templates/NEWS_BOX.tpl',
            'themes/default/templates/NEWS_BRIEF.tpl',
        ];
    }

    /**
     * Get mapping between template names and the method of this class that can render a preview of them.
     *
     * @return array The mapping
     */
    public function tpl_previews() : array
    {
        return [
            'templates/NEWS_BOX.tpl' => 'news_piece_summary',
        ];
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__news_piece_summary() : object
    {
        return lorem_globalise(do_lorem_template('NEWS_BOX', [
            'BLOG' => lorem_phrase(),
            'AUTHOR_URL' => placeholder_url(),
            'TAGS' => placeholder_tags(),
            'CATEGORY' => lorem_phrase(),
            '_CATEGORY' => placeholder_numeric_id(),
            'CATEGORY_URL' => placeholder_url(),
            'REP_IMAGE_URL' => placeholder_image_url(),
            'AUTHOR' => lorem_phrase(),
            '_AUTHOR' => lorem_phrase(),
            'SUBMITTER' => placeholder_numeric_id(),
            'AVATAR' => lorem_phrase(),
            'NEWS_TITLE' => lorem_phrase(),
            'NEWS_TITLE_PLAIN' => lorem_phrase(),
            'DATE' => placeholder_date(),
            'NEWS' => lorem_phrase(),
            'COMMENTS' => lorem_phrase(),
            'VIEW' => lorem_phrase(),
            'ID' => placeholder_numeric_id(),
            'FULL_URL' => placeholder_url(),
            'COMMENT_COUNT' => lorem_phrase(),
            'READ_MORE' => lorem_sentence(),
            'TRUNCATE' => false,
            'FIRSTTIME' => placeholder_date(),
            'LASTTIME' => placeholder_date(),
            'CLOSED' => false,
            'FIRSTUSERNAME' => lorem_word(),
            'LASTUSERNAME' => lorem_word(),
            'FIRSTMEMBERID' => placeholder_numeric_id(),
            'LASTMEMBERID' => placeholder_numeric_id(),
            'DATE_RAW' => placeholder_date_raw(),
            'GIVE_CONTEXT' => true,
            'FORUM_ID' => placeholder_numeric_id(),
            'BRIEF_CONTENT' => lorem_sentence_html(),
        ]), null, '', true);
    }
}
