<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    quizzes
 */

/**
 * Hook class.
 */
class Hook_addon_registry_quizzes
{
    /**
     * Get a list of file permissions to set.
     *
     * @param  boolean $runtime Whether to include wildcards represented runtime-created chmoddable files
     * @return array File permissions to set
     */
    public function get_chmod_array(bool $runtime = false) : array
    {
        return [];
    }

    /**
     * Get the version of the software this addon is for (used in generating the TAR filename).
     *
     * @return float Version number
     */
    public function get_version() : float
    {
        return cms_version_number();
    }

    /**
     * Get the minimum required version of the website software needed to use this addon.
     *
     * @return float Minimum required website software version
     */
    public function get_min_cms_version() : float
    {
        return 11.0;
    }

    /**
     * Get the maximum compatible version of the website software to use this addon.
     *
     * @return ?float Maximum compatible website software version (null: no maximum version currently)
     */
    public function get_max_cms_version() : ?float
    {
        return 11.9;
    }

    /**
     * Get the addon category.
     *
     * @return string The category
     */
    public function get_category() : string
    {
        return 'New Features';
    }

    /**
     * Get the description of the addon.
     *
     * @return string Description of the addon
     */
    public function get_description() : string
    {
        return 'Construct competitions, surveys, and tests, for members to perform. Highly configurable, and comes with administrative tools to handle the results.';
    }

    /**
     * Get a list of tutorials that apply to this addon.
     *
     * @return array List of tutorials
     */
    public function get_applicable_tutorials() : array
    {
        return [
            'tut_quizzes',
        ];
    }

    /**
     * Get a mapping of dependency types.
     *
     * @return array A structure specifying dependency information
     */
    public function get_dependencies() : array
    {
        return [
            'requires' => [],
            'recommends' => [],
            'conflicts_with' => [],
        ];
    }

    /**
     * Explicitly say which icon should be used.
     *
     * @return URLPATH Icon
     */
    public function get_default_icon() : string
    {
        return 'themes/default/images/icons/menu/rich_content/quiz.svg';
    }

    /**
     * Get a list of files that belong to this addon.
     *
     * @return array List of files
     */
    public function get_file_list() : array
    {
        return [
            'adminzone/pages/modules/admin_quiz.php',
            'cms/pages/modules/cms_quiz.php',
            'lang/EN/quiz.ini',
            'site/pages/modules/quiz.php',
            'sources/hooks/blocks/side_stats/quiz.php',
            'sources/hooks/modules/admin_import_types/quizzes.php',
            'sources/hooks/modules/admin_newsletter/quiz.php',
            'sources/hooks/modules/members/quiz.php',
            'sources/hooks/modules/search/quiz.php',
            'sources/hooks/systems/actionlog/quizzes.php',
            'sources/hooks/systems/addon_registry/quizzes.php',
            'sources/hooks/systems/commandr_fs/quizzes.php',
            'sources/hooks/systems/config/points_ADD_QUIZ.php',
            'sources/hooks/systems/config/quiz_show_stats_count_total_open.php',
            'sources/hooks/systems/config/search_quiz.php',
            'sources/hooks/systems/content_meta_aware/quiz.php',
            'sources/hooks/systems/notifications/quiz_results.php',
            'sources/hooks/systems/page_groupings/quiz.php',
            'sources/hooks/systems/points/quiz__add.php',
            'sources/hooks/systems/points/quiz__pass.php',
            'sources/hooks/systems/preview/quiz.php',
            'sources/hooks/systems/privacy/quizzes.php',
            'sources/hooks/systems/sitemap/quiz.php',
            'sources/hooks/systems/tasks/export_quiz.php',
            'sources/quiz.php',
            'sources/quiz2.php',
            'themes/default/css/quizzes.css',
            'themes/default/images/icons/menu/cms/quiz/find_winners.svg',
            'themes/default/images/icons/menu/cms/quiz/index.html',
            'themes/default/images/icons/menu/cms/quiz/quiz_results.svg',
            'themes/default/images/icons/menu/rich_content/quiz.svg',
            'themes/default/images/icons_monochrome/menu/cms/quiz/find_winners.svg',
            'themes/default/images/icons_monochrome/menu/cms/quiz/index.html',
            'themes/default/images/icons_monochrome/menu/cms/quiz/quiz_results.svg',
            'themes/default/images/icons_monochrome/menu/rich_content/quiz.svg',
            'themes/default/javascript/quizzes.js',
            'themes/default/templates/MEMBER_QUIZ_ENTRIES.tpl',
            'themes/default/templates/QUIZ_ARCHIVE_SCREEN.tpl',
            'themes/default/templates/QUIZ_BOX.tpl',
            'themes/default/templates/QUIZ_DONE_SCREEN.tpl',
            'themes/default/templates/QUIZ_RESULTS.tpl',
            'themes/default/templates/QUIZ_RESULTS_SCREEN.tpl',
            'themes/default/templates/QUIZ_RESULT_SCREEN.tpl',
            'themes/default/templates/QUIZ_SCREEN.tpl',
            'themes/default/text/QUIZ_SURVEY_ANSWERS_MAIL.txt',
            'themes/default/text/QUIZ_TEST_ANSWERS_MAIL.txt',
        ];
    }

    /**
     * Get mapping between template names and the method of this class that can render a preview of them.
     *
     * @return array The mapping
     */
    public function tpl_previews() : array
    {
        return [
            'templates/QUIZ_RESULTS_SCREEN.tpl' => 'administrative__quiz_results_screen',
            'templates/QUIZ_RESULT_SCREEN.tpl' => 'administrative__quiz_result_screen',
            'templates/QUIZ_RESULTS.tpl' => 'quiz_results',
            'templates/QUIZ_BOX.tpl' => 'quiz_archive_screen',
            'templates/QUIZ_ARCHIVE_SCREEN.tpl' => 'quiz_archive_screen',
            'templates/QUIZ_SCREEN.tpl' => 'quiz_screen',
            'text/QUIZ_SURVEY_ANSWERS_MAIL.txt' => 'quiz_survey_answers_mail',
            'text/QUIZ_TEST_ANSWERS_MAIL.txt' => 'quiz_test_answers_mail',
            'templates/QUIZ_DONE_SCREEN.tpl' => 'quiz_done_screen',
            'templates/MEMBER_QUIZ_ENTRIES.tpl' => 'member_quiz_entries',
        ];
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__administrative__quiz_results_screen() : object
    {
        $fields = new Tempcode();
        foreach (placeholder_array() as $k => $v) {
            $fields->attach(do_lorem_template('MAP_TABLE_FIELD', [
                'ABBR' => '',
                'NAME' => lorem_phrase(),
                'VALUE' => lorem_phrase(),
            ]));
        }
        $summary = do_lorem_template('MAP_TABLE', [
            'FIELDS' => $fields,
            'RESPONSIVE' => true,
        ]);

        return lorem_globalise(do_lorem_template('QUIZ_RESULTS_SCREEN', [
            'TITLE' => lorem_screen_title(),
            'SUMMARY' => $summary,
            'RESULTS' => placeholder_table(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__administrative__quiz_result_screen() : object
    {
        $given_answers_arr = [];
        foreach ([true, false, null] as $was_correct) {
            $given_answers_arr[] = [
                'QUESTION' => lorem_phrase(),
                'GIVEN_ANSWER' => lorem_phrase(),
                'WAS_CORRECT' => $was_correct,
                'CORRECT_ANSWER' => lorem_phrase(),
                'CORRECT_EXPLANATION' => lorem_paragraph_html(),
            ];
        }

        return lorem_globalise(do_lorem_template('QUIZ_RESULT_SCREEN', [
            'TITLE' => lorem_screen_title(),
            'USERNAME' => lorem_phrase(),
            'MEMBER_URL' => placeholder_url(),
            'DATE' => placeholder_date(),
            '_DATE' => placeholder_date_raw(),
            'ENTRY_ID' => placeholder_numeric_id(),
            'QUIZ_NAME' => lorem_phrase(),
            'GIVEN_ANSWERS_ARR' => $given_answers_arr,
            'PASSED' => true,
            'TYPE' => lorem_phrase(),
            '_TYPE' => 'TEST',
            'MARKS' => placeholder_number(),
            'POTENTIAL_EXTRA_MARKS' => placeholder_number(),
            'OUT_OF' => placeholder_number(),
            'MARKS_RANGE' => placeholder_number(),
            'PERCENTAGE_RANGE' => placeholder_number(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__quiz_results() : object
    {
        $given_answers_arr = [];
        foreach ([true, false, null] as $was_correct) {
            $given_answers_arr[] = [
                'QUESTION' => lorem_phrase(),
                'GIVEN_ANSWER' => lorem_phrase(),
                'WAS_CORRECT' => $was_correct,
                'CORRECT_ANSWER' => lorem_phrase(),
                'CORRECT_EXPLANATION' => lorem_paragraph_html(),
            ];
        }

        return lorem_globalise(do_lorem_template('QUIZ_RESULTS', [
            'GIVEN_ANSWERS_ARR' => $given_answers_arr,
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__quiz_archive_screen() : object
    {
        $content_tests = new Tempcode();
        $content_competitions = new Tempcode();
        $content_surveys = new Tempcode();
        foreach (placeholder_array() as $k => $v) {
            $link = do_lorem_template('QUIZ_BOX', [
                'GIVE_CONTEXT' => true,
                'TYPE' => lorem_word(),
                '_TYPE' => lorem_word(),
                'DATE' => placeholder_date(),
                '_DATE' => placeholder_date_raw(),
                'URL' => placeholder_url(),
                'NAME' => lorem_phrase(),
                'START_TEXT' => lorem_phrase(),
                'TIMEOUT' => placeholder_number(),
                '_TIMEOUT' => placeholder_number(),
                'REDO_TIME' => placeholder_number(),
                'POINTS' => placeholder_number(),
                '_POINTS' => placeholder_number(),
                'ID' => placeholder_numeric_id(),
                'SUBMITTER' => placeholder_numeric_id(),
                'PERCENTAGE' => placeholder_number(),
                'OPEN_TIME' => placeholder_date_raw(),
                'CLOSE_TIME' => placeholder_date_raw(),
            ]);
        }
        $content_surveys->attach($link);
        $content_tests->attach($link);
        $content_competitions->attach($link);

        return lorem_globalise(do_lorem_template('QUIZ_ARCHIVE_SCREEN', [
            'TITLE' => lorem_screen_title(),
            'CONTENT_SURVEYS' => $content_surveys,
            'CONTENT_COMPETITIONS' => $content_competitions,
            'CONTENT_TESTS' => $content_tests,
            'PAGINATION' => placeholder_pagination(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__quiz_screen() : object
    {
        // This is for getting the $cms.doAjaxRequest() javascript function.
        $warning_details = do_lorem_template('WARNING_BOX', [
            'WARNING' => lorem_phrase(),
        ]);

        return lorem_globalise(do_lorem_template('QUIZ_SCREEN', [
            'TAGS' => placeholder_tags(),
            'ID' => placeholder_numeric_id(),
            'WARNING_DETAILS' => $warning_details,
            'URL' => placeholder_url(),
            'TITLE' => lorem_screen_title(),
            'START_TEXT' => lorem_sentence_html(),
            'FIELDS' => placeholder_fields(),
            'EDIT_URL' => placeholder_url(),
            'ALL_REQUIRED' => false,
            'TYPE' => lorem_word(),
            '_TYPE' => lorem_word(),
            'DATE' => placeholder_date(),
            '_DATE' => placeholder_date_raw(),
            'NAME' => lorem_phrase(),
            'TIMEOUT' => placeholder_number(),
            '_TIMEOUT' => placeholder_number(),
            'REDO_TIME' => placeholder_number(),
            'POINTS' => placeholder_number(),
            '_POINTS' => placeholder_number(),
            'GIVE_CONTEXT' => true,
            'SUBMITTER' => placeholder_numeric_id(),
            'PERCENTAGE' => placeholder_number(),
            'OPEN_TIME' => placeholder_date_raw(),
            'CLOSE_TIME' => placeholder_date_raw(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__quiz_survey_answers_mail() : object
    {
        $given_answers = new Tempcode();
        foreach (placeholder_array() as $k => $v) {
            $given_answers->attach(lorem_phrase());
        }

        $given_answers_arr = [];
        foreach ([true, false, null] as $was_correct) {
            $given_answers_arr[] = [
                'QUESTION' => lorem_phrase(),
                'GIVEN_ANSWER' => lorem_phrase(),
                'WAS_CORRECT' => $was_correct,
                'CORRECT_ANSWER' => lorem_phrase(),
                'CORRECT_EXPLANATION' => lorem_paragraph_html(),
            ];
        }

        return lorem_globalise(do_lorem_template('QUIZ_SURVEY_ANSWERS_MAIL', [
            'ENTRY_ID' => placeholder_numeric_id(),
            'QUIZ_NAME' => lorem_phrase(),
            'GIVEN_ANSWERS_ARR' => $given_answers_arr,
            'GIVEN_ANSWERS' => $given_answers,
            'MEMBER_PROFILE_URL' => placeholder_url(),
            'USERNAME' => lorem_phrase(),
            'FORUM_DRIVER' => null,
        ], null, false, null, '.txt', 'text'), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__quiz_test_answers_mail() : object
    {
        $unknowns = new Tempcode();
        foreach (placeholder_array() as $k => $v) {
            $unknowns->attach(lorem_phrase());
        }

        $corrections = new Tempcode();
        foreach (placeholder_array() as $k => $v) {
            $corrections->attach(lorem_phrase());
        }

        $given_answers = new Tempcode();
        foreach (placeholder_array() as $k => $v) {
            $given_answers->attach(lorem_phrase());
        }

        $given_answers_arr = [];
        foreach ([true, false, null] as $was_correct) {
            $given_answers_arr[] = [
                'QUESTION' => lorem_phrase(),
                'GIVEN_ANSWER' => lorem_phrase(),
                'WAS_CORRECT' => $was_correct,
                'CORRECT_ANSWER' => lorem_phrase(),
                'CORRECT_EXPLANATION' => lorem_paragraph_html(),
            ];
        }

        return lorem_globalise(do_lorem_template('QUIZ_TEST_ANSWERS_MAIL', [
            'ENTRY_ID' => placeholder_numeric_id(),
            'QUIZ_NAME' => lorem_phrase(),
            'GIVEN_ANSWERS_ARR' => $given_answers_arr,
            'GIVEN_ANSWERS' => $given_answers,
            'UNKNOWNS' => $unknowns,
            'CORRECTIONS' => $corrections,
            'RESULT' => lorem_phrase(),
            'USERNAME' => lorem_phrase(),
        ], null, false, null, '.txt', 'text'), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__quiz_done_screen() : object
    {
        $given_answers_arr = [];
        foreach ([true, false, null] as $was_correct) {
            $given_answers_arr[] = [
                'QUESTION' => lorem_phrase(),
                'GIVEN_ANSWER' => lorem_phrase(),
                'WAS_CORRECT' => $was_correct,
                'CORRECT_ANSWER' => lorem_phrase(),
                'CORRECT_EXPLANATION' => lorem_paragraph_html(),
            ];
        }

        return lorem_globalise(do_lorem_template('QUIZ_DONE_SCREEN', [
            'TITLE' => lorem_screen_title(),
            'ENTRY_ID' => placeholder_numeric_id(),
            'QUIZ_NAME' => lorem_phrase(),
            'GIVEN_ANSWERS_ARR' => $given_answers_arr,
            'CORRECTIONS' => lorem_phrase(),
            'PASSED' => true,
            'POINTS_DIFFERENCE' => placeholder_number(),
            'RESULT' => lorem_phrase(),
            'TYPE' => do_lang('SURVEY'),
            '_TYPE' => 'SURVEY',
            'MESSAGE' => lorem_phrase(),
            'REVEAL_ANSWERS' => true,
            'MARKS' => placeholder_number(),
            'POTENTIAL_EXTRA_MARKS' => placeholder_number(),
            'OUT_OF' => placeholder_number(),
            'MARKS_RANGE' => placeholder_number(),
            'PERCENTAGE_RANGE' => placeholder_number(),
            'AFFIRMATIONS' => lorem_phrase(),
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__member_quiz_entries() : object
    {
        $categories = [];
        $categories[do_lang('OTHER')] = [
            'QUIZZES' => [],
            'RUNNING_MARKS' => placeholder_number(),
            'RUNNING_OUT_OF' => placeholder_number(),
            'RUNNING_PERCENTAGE' => placeholder_number(),
            'RUNNING_MARKS__CREDIT' => placeholder_number(),
            'RUNNING_OUT_OF__CREDIT' => placeholder_number(),
            'RUNNING_PERCENTAGE__CREDIT' => placeholder_number(),
        ];
        $categories[do_lang('OTHER')]['QUIZZES'][] = [
            'QUIZ_NAME' => lorem_phrase(),
            'QUIZ_START_TEXT' => lorem_paragraph_html(),
            'QUIZ_ID' => placeholder_numeric_id(),
            'QUIZ_URL' => placeholder_url(),
            'ENTRY_ID' => placeholder_numeric_id(),
            'ENTRY_DATE' => placeholder_date(),
            '_ENTRY_DATE' => placeholder_date_raw(),
            'OUT_OF' => placeholder_number(),
            'MARKS_RANGE' => placeholder_number(),
            'PERCENTAGE_RANGE' => placeholder_number(),
            'PASSED' => true,
            'POINTS' => placeholder_number(),
        ];

        return lorem_globalise(do_lorem_template('MEMBER_QUIZ_ENTRIES', [
            'CATEGORIES' => $categories,
            'MEMBER_ID' => placeholder_numeric_id(),
            'SORTING' => placeholder_table(),
            'DELETE_URL' => placeholder_url(),
        ]), null, '', true);
    }

    /**
     * Uninstall default content.
     */
    public function uninstall_test_content()
    {
        require_code('quiz2');

        $to_delete = $GLOBALS['SITE_DB']->query_select('quizzes', ['id'], [$GLOBALS['SITE_DB']->translate_field_ref('q_name') => lorem_phrase()]);
        foreach ($to_delete as $record) {
            delete_quiz($record['id']);
        }
    }

    /**
     * Install default content.
     */
    public function install_test_content()
    {
        require_code('quiz2');
        require_lang('quiz');

        add_quiz(lorem_phrase(), null, lorem_paragraph(), lorem_paragraph(), lorem_paragraph(), '', 50, null, null, 0, 0, 'SURVEY', 1, do_lang('EXAMPLE_QUESTIONS'));
    }
}
