<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    syndication
 */

/**
 * Hook class.
 */
class Hook_addon_registry_syndication
{
    /**
     * Get a list of file permissions to set.
     *
     * @param  boolean $runtime Whether to include wildcards represented runtime-created chmoddable files
     * @return array File permissions to set
     */
    public function get_chmod_array(bool $runtime = false) : array
    {
        return [];
    }

    /**
     * Get the version of the software this addon is for (used in generating the TAR filename).
     *
     * @return float Version number
     */
    public function get_version() : float
    {
        return cms_version_number();
    }

    /**
     * Get the minimum required version of the website software needed to use this addon.
     *
     * @return float Minimum required website software version
     */
    public function get_min_cms_version() : float
    {
        return 11.0;
    }

    /**
     * Get the maximum compatible version of the website software to use this addon.
     *
     * @return ?float Maximum compatible website software version (null: no maximum version currently)
     */
    public function get_max_cms_version() : ?float
    {
        return 11.9;
    }

    /**
     * Get the addon category.
     *
     * @return string The category
     */
    public function get_category() : string
    {
        return 'Information Display';
    }

    /**
     * Get the description of the addon.
     *
     * @return string Description of the addon
     */
    public function get_description() : string
    {
        return 'Syndicate RSS/Atom feeds of your content.';
    }

    /**
     * Get a list of tutorials that apply to this addon.
     *
     * @return array List of tutorials
     */
    public function get_applicable_tutorials() : array
    {
        return [
            'tut_integration',
            'tut_news',
            'tut_adv_news',
            'tut_site_messaging',
        ];
    }

    /**
     * Get a mapping of dependency types.
     *
     * @return array A structure specifying dependency information
     */
    public function get_dependencies() : array
    {
        return [
            'requires' => [
                'syndication_blocks',
            ],
            'recommends' => [],
            'conflicts_with' => [],
            'previously_in_addon' => ['core_syndication'],
        ];
    }

    /**
     * Explicitly say which icon should be used.
     *
     * @return URLPATH Icon
     */
    public function get_default_icon() : string
    {
        return 'themes/default/images/icons/links/rss.svg';
    }

    /**
     * Get a list of files that belong to this addon.
     *
     * @return array List of files
     */
    public function get_file_list() : array
    {
        return [
            'backend.php',
            'data/backend_cloud.php',
            'sources/hooks/systems/addon_registry/syndication.php',
            'sources/hooks/systems/non_active_urls/news_rss_cloud.php',
            'sources/hooks/systems/rss/.htaccess',
            'sources/hooks/systems/rss/index.html',
            'sources/rss2.php',
            'sources_custom/hooks/systems/rss/.htaccess',
            'sources_custom/hooks/systems/rss/index.html',
            'themes/default/xml/ATOM_ENTRY.xml',
            'themes/default/xml/ATOM_WRAPPER.xml',
            'themes/default/xml/ATOM_XSLT.xml',
            'themes/default/xml/OPML_WRAPPER.xml',
            'themes/default/xml/OPML_XSLT.xml',
            'themes/default/xml/RSS_ABBR.xml',
            'themes/default/xml/RSS_CLOUD.xml',
            'themes/default/xml/RSS_ENTRY.xml',
            'themes/default/xml/RSS_ENTRY_COMMENTS.xml',
            'themes/default/xml/RSS_WRAPPER.xml',
            'themes/default/xml/RSS_XSLT.xml',
        ];
    }

    /**
     * Get mapping between template names and the method of this class that can render a preview of them.
     *
     * @return array The mapping
     */
    public function tpl_previews() : array
    {
        return [
            'xml/RSS_ENTRY_COMMENTS.xml' => 'rss_wrapper',
            'xml/RSS_XSLT.xml' => 'rss_xslt',
            'xml/ATOM_XSLT.xml' => 'atom_xslt',
            'xml/OPML_XSLT.xml' => 'opml_xslt',
            'xml/OPML_WRAPPER.xml' => 'opml_wrapper',
            'xml/RSS_CLOUD.xml' => 'rss_wrapper',
            'xml/ATOM_ENTRY.xml' => 'atom_wrapper',
            'xml/ATOM_WRAPPER.xml' => 'atom_wrapper',
            'xml/RSS_WRAPPER.xml' => 'rss_wrapper',
            'xml/RSS_ENTRY.xml' => 'rss_wrapper',
            'xml/RSS_ABBR.xml' => 'rss_wrapper',
        ];
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__rss_wrapper() : object
    {
        $comments = do_lorem_template('RSS_ENTRY_COMMENTS', ['COMMENT_URL' => placeholder_url(), 'ID' => placeholder_codename()], null, false, null, '.xml', 'xml');

        $content = do_lorem_template('RSS_ABBR', [], null, false, null, '.xml', 'xml');
        $content->attach(do_lorem_template('RSS_ENTRY', [
            'TITLE' => lorem_phrase(),
            'SUMMARY' => lorem_paragraph_html(),
            'VIEW_URL' => placeholder_url(),
            'DATE' => placeholder_date(),
            'EDIT_DATE' => placeholder_date(),
            'CATEGORY_RAW' => lorem_word(),
            'CATEGORY' => lorem_word(),
            'AUTHOR' => lorem_word(),
            'IF_COMMENTS' => $comments,
            'ID' => placeholder_codename(),
        ], null, false, null, '.xml', 'xml'));

        $cloud = do_lorem_template('RSS_CLOUD', [
            'TYPE' => 'news',
            'DOMAIN' => 'example.com',
            'PORT' => '80',
            'PATH' => '/cms',
        ], null, false, null, '.xml', 'xml');

        return do_lorem_template('RSS_WRAPPER', [
            'MODE' => lorem_word(),
            'MODE_NICE' => lorem_word(),
            'SELECT' => lorem_word_2(),
            'DATE' => placeholder_date(),
            'LOGO_URL' => placeholder_image_url(),
            'CONTENT' => $content,
            'CUTOFF' => placeholder_number(),
            'ABOUT' => lorem_paragraph(),
            'VERSION' => lorem_word(),
            'RSS_CLOUD' => '',
        ], null, false, null, '.xml', 'xml');
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__atom_wrapper() : object
    {
        $content = do_lorem_template('ATOM_ENTRY', [
            'TITLE' => lorem_phrase(),
            'SUMMARY' => lorem_paragraph_html(),
            'VIEW_URL' => placeholder_url(),
            'DATE' => placeholder_date(),
            'EDIT_DATE' => placeholder_date(),
            'CATEGORY_RAW' => lorem_word(),
            'CATEGORY' => lorem_word(),
            'AUTHOR' => lorem_word(),
            'IF_COMMENTS' => '',
            'ID' => placeholder_codename(),
        ], null, false, null, '.xml', 'xml');

        return do_lorem_template('ATOM_WRAPPER', [
            'MODE' => lorem_word(),
            'MODE_NICE' => lorem_word(),
            'SELECT' => lorem_word_2(),
            'DATE' => placeholder_date(),
            'LOGO_URL' => placeholder_image_url(),
            'CONTENT' => $content,
            'CUTOFF' => placeholder_number(),
            'ABOUT' => lorem_paragraph(),
            'VERSION' => lorem_word(),
            'RSS_CLOUD' => '',
        ], null, false, null, '.xml', 'xml');
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__rss_xslt() : object
    {
        require_lang('rss');

        return do_lorem_template('RSS_XSLT', [
            'JAVASCRIPT_XSL_MOPUP' => '',
        ], null, false, null, '.xml', 'xml');
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__atom_xslt() : object
    {
        require_lang('rss');

        return do_lorem_template('ATOM_XSLT', [
            'JAVASCRIPT_XSL_MOPUP' => '',
        ], null, false, null, '.xml', 'xml');
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__opml_xslt() : object
    {
        require_lang('rss');

        return do_lorem_template('OPML_XSLT', [
            'JAVASCRIPT_XSL_MOPUP' => '',
        ], null, false, null, '.xml', 'xml');
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__opml_wrapper() : object
    {
        require_lang('rss');

        return do_lorem_template('OPML_WRAPPER', [
            'FEEDS' => placeholder_array(),
            'ABOUT' => lorem_phrase(),
            'DATE' => placeholder_date(),
        ], null, false, null, '.xml', 'xml');
    }
}
