<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    syndication_blocks
 */

/**
 * Hook class.
 */
class Hook_addon_registry_syndication_blocks
{
    /**
     * Get a list of file permissions to set.
     *
     * @param  boolean $runtime Whether to include wildcards represented runtime-created chmoddable files
     * @return array File permissions to set
     */
    public function get_chmod_array(bool $runtime = false) : array
    {
        return [];
    }

    /**
     * Get the version of the software this addon is for (used in generating the TAR filename).
     *
     * @return float Version number
     */
    public function get_version() : float
    {
        return cms_version_number();
    }

    /**
     * Get the minimum required version of the website software needed to use this addon.
     *
     * @return float Minimum required website software version
     */
    public function get_min_cms_version() : float
    {
        return 11.0;
    }

    /**
     * Get the maximum compatible version of the website software to use this addon.
     *
     * @return ?float Maximum compatible website software version (null: no maximum version currently)
     */
    public function get_max_cms_version() : ?float
    {
        return 11.9;
    }

    /**
     * Get the addon category.
     *
     * @return string The category
     */
    public function get_category() : string
    {
        return 'Information Display';
    }

    /**
     * Get the description of the addon.
     *
     * @return string Description of the addon
     */
    public function get_description() : string
    {
        return 'Show RSS and Atom feeds from other websites.';
    }

    /**
     * Get a list of tutorials that apply to this addon.
     *
     * @return array List of tutorials
     */
    public function get_applicable_tutorials() : array
    {
        return [
            'tut_news',
            'tut_adv_news',
            'tut_site_messaging',
        ];
    }

    /**
     * Get a mapping of dependency types.
     *
     * @return array A structure specifying dependency information
     */
    public function get_dependencies() : array
    {
        return [
            'requires' => [
                'news',
            ],
            'recommends' => [],
            'conflicts_with' => [],
        ];
    }

    /**
     * Explicitly say which icon should be used.
     *
     * @return URLPATH Icon
     */
    public function get_default_icon() : string
    {
        return 'themes/default/images/icons/links/rss.svg';
    }

    /**
     * Get a list of files that belong to this addon.
     *
     * @return array List of files
     */
    public function get_file_list() : array
    {
        return [
            'sources/blocks/bottom_rss.php',
            'sources/blocks/main_rss.php',
            'sources/blocks/side_rss.php',
            'sources/hooks/modules/admin_setupwizard/syndication_blocks.php',
            'sources/hooks/systems/addon_registry/syndication_blocks.php',
            'sources/hooks/systems/commandr_commands/feed_display.php',
            'sources/hooks/systems/config/is_on_rss.php',
            'sources/hooks/systems/config/is_rss_advertised.php',
            'sources/hooks/systems/config/rss_update_time.php',
            'themes/default/css/rss.css',
            'themes/default/templates/BLOCK_MAIN_RSS.tpl',
            'themes/default/templates/BLOCK_MAIN_RSS_SUMMARY.tpl',
            'themes/default/templates/BLOCK_SIDE_RSS.tpl',
            'themes/default/templates/BLOCK_SIDE_RSS_SUMMARY.tpl',
        ];
    }

    /**
     * Get mapping between template names and the method of this class that can render a preview of them.
     *
     * @return array The mapping
     */
    public function tpl_previews() : array
    {
        return [
            'templates/BLOCK_SIDE_RSS_SUMMARY.tpl' => 'block_side_rss',
            'templates/BLOCK_SIDE_RSS.tpl' => 'block_side_rss',
            'templates/BLOCK_MAIN_RSS_SUMMARY.tpl' => 'block_main_rss',
            'templates/BLOCK_MAIN_RSS.tpl' => 'block_main_rss',
        ];
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__block_side_rss() : object
    {
        require_lang('news');

        $content = new Tempcode();
        foreach (placeholder_array() as $k => $v) {
            $content->attach(do_lorem_template('BLOCK_SIDE_RSS_SUMMARY', [
                'FEED_URL' => placeholder_url(),
                'FULL_URL' => placeholder_url(),
                'NEWS_TITLE' => lorem_phrase(),
                'DATE' => placeholder_date(),
                'DATE_RAW' => placeholder_date_raw(),
                'SUMMARY' => lorem_paragraph(),
                'TICKER' => lorem_phrase(),
            ]));
        }

        return lorem_globalise(do_lorem_template('BLOCK_SIDE_RSS', [
            'BLOCK_ID' => lorem_word(),
            'FEED_URL' => placeholder_url(),
            'TITLE' => lorem_phrase(),
            'CONTENT' => $content,
            'TICKER' => true,
        ]), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__block_main_rss() : object
    {
        require_lang('news');
        require_css('news');

        $content = new Tempcode();
        foreach (placeholder_array() as $k => $v) {
            $content->attach(do_lorem_template('BLOCK_MAIN_RSS_SUMMARY', [
                'NEWS_TITLE' => lorem_phrase(),
                'FEED_URL' => placeholder_url(),
                'DATE' => placeholder_date(),
                'AUTHOR' => lorem_phrase(),
                'CATEGORY_IMG' => placeholder_image_url(),
                'CATEGORY' => lorem_phrase(),
                'FULL_URL' => placeholder_link(),
                'FULL_URL_RAW' => placeholder_url(),
                'NEWS' => lorem_paragraph_html(),
                'NEWS_FULL' => lorem_paragraph_html(),
                'BLOCK_ID' => placeholder_codename(),
                'DATE_RAW' => placeholder_date_raw(),
            ]));
        }

        return lorem_globalise(do_lorem_template('BLOCK_MAIN_RSS', [
            'BLOCK_ID' => lorem_word(),
            'FEED_URL' => placeholder_url(),
            'TITLE' => lorem_phrase(),
            'COPYRIGHT' => lorem_phrase(),
            'AUTHOR' => lorem_phrase(),
            'CONTENT' => $content,
        ]), null, '', true);
    }
}
