<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    validation
 */

/**
 * Hook class.
 */
class Hook_addon_registry_validation
{
    /**
     * Get a list of file permissions to set.
     *
     * @param  boolean $runtime Whether to include wildcards represented runtime-created chmoddable files
     * @return array File permissions to set
     */
    public function get_chmod_array(bool $runtime = false) : array
    {
        return [];
    }

    /**
     * Get the version of the software this addon is for (used in generating the TAR filename).
     *
     * @return float Version number
     */
    public function get_version() : float
    {
        return cms_version_number();
    }

    /**
     * Get the minimum required version of the website software needed to use this addon.
     *
     * @return float Minimum required website software version
     */
    public function get_min_cms_version() : float
    {
        return 11.0;
    }

    /**
     * Get the maximum compatible version of the website software to use this addon.
     *
     * @return ?float Maximum compatible website software version (null: no maximum version currently)
     */
    public function get_max_cms_version() : ?float
    {
        return 11.9;
    }

    /**
     * Get the addon category.
     *
     * @return string The category
     */
    public function get_category() : string
    {
        return 'Community';
    }

    /**
     * Get the description of the addon.
     *
     * @return string Description of the addon
     */
    public function get_description() : string
    {
        return 'Subject member\'s to validation (approval) of their content submissions, and enable/disable content.';
    }

    /**
     * Get a list of tutorials that apply to this addon.
     *
     * @return array List of tutorials
     */
    public function get_applicable_tutorials() : array
    {
        return [
            'tut_censor',
        ];
    }

    /**
     * Get a mapping of dependency types.
     *
     * @return array A structure specifying dependency information
     */
    public function get_dependencies() : array
    {
        return [
            'requires' => [],
            'recommends' => [],
            'conflicts_with' => [],
            'previously_in_addon' => ['core_unvalidated', 'unvalidated'],
        ];
    }

    /**
     * Explicitly say which icon should be used.
     *
     * @return URLPATH Icon
     */
    public function get_default_icon() : string
    {
        return 'themes/default/images/icons/menu/adminzone/audit/not_validated.svg';
    }

    /**
     * Get a list of files that belong to this addon.
     *
     * @return array List of files
     */
    public function get_file_list() : array
    {
        return [
            'adminzone/pages/modules/admin_validation.php',
            'lang/EN/validation.ini',
            'sources/hooks/blocks/main_staff_checklist/validation.php',
            'sources/hooks/systems/addon_registry/validation.php',
            'sources/hooks/systems/notifications/content_validated.php',
            'sources/hooks/systems/notifications/needs_validation.php',
            'sources/hooks/systems/page_groupings/validation.php',
            'sources/hooks/systems/validation/.htaccess',
            'sources/hooks/systems/validation/index.html',
            'sources/validation.php',
            'sources_custom/hooks/systems/validation/.htaccess',
            'sources_custom/hooks/systems/validation/index.html',
            'themes/default/images/icons/menu/adminzone/audit/not_validated.svg',
            'themes/default/images/icons_monochrome/menu/adminzone/audit/not_validated.svg',
            'themes/default/templates/VALIDATION_SCREEN.tpl',
            'themes/default/templates/VALIDATION_SECTION.tpl',
            'themes/default/text/VALIDATION_REQUEST_MAIL.txt',
        ];
    }

    /**
     * Get mapping between template names and the method of this class that can render a preview of them.
     *
     * @return array The mapping
     */
    public function tpl_previews() : array
    {
        return [
            'templates/VALIDATION_SECTION.tpl' => 'administrative__validation_screen',
            'templates/VALIDATION_SCREEN.tpl' => 'administrative__validation_screen',
            'text/VALIDATION_REQUEST_MAIL.txt' => 'administrative__validation_request_mail',
        ];
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__administrative__validation_request_mail() : object
    {
        return lorem_globalise(do_lorem_template('VALIDATION_REQUEST_MAIL', [
            'USERNAME' => lorem_word(),
            'CONTENT_TYPE_LABEL' => lorem_phrase(),
            'TITLE' => lorem_phrase(),
            'TYPE' => lorem_phrase(),
            'ID' => placeholder_codename(),
            'URL' => placeholder_url(),
        ], null, false, null, '.txt', 'text'), null, '', true);
    }

    /**
     * Get a preview(s) of a (group of) template(s), as a full standalone piece of HTML in Tempcode format.
     * Uses sources/lorem.php functions to place appropriate stock-text. Should not hard-code things, as the code is intended to be declarative.
     * Assumptions: You can assume all Lang/CSS/JavaScript files in this addon have been pre-required.
     *
     * @return Tempcode Preview
     */
    public function tpl_preview__administrative__validation_screen() : object
    {
        $section = do_lorem_template('VALIDATION_SECTION', [
            'TITLE' => lorem_phrase(),
            'CONTENT' => lorem_phrase(),
        ]);
        return lorem_globalise(do_lorem_template('VALIDATION_SCREEN', [
            'TITLE' => lorem_screen_title(),
            'SECTIONS' => $section,
            'TEXT' => lorem_paragraph_html(),
        ]), null, '', true);
    }
}
