<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    wiki
 */

/**
 * Hook class.
 */
class Hook_ajax_tree_choose_wiki_page
{
    /**
     * Run function for ajax-tree hooks. Generates XML for a tree list, which is interpreted by JavaScript and expanded on-demand (via new calls).
     *
     * @param  ?ID_TEXT $id The ID to do under (null: root)
     * @param  array $options Options being passed through
     * @param  ?ID_TEXT $default The ID to select by default (null: none)
     * @return string XML in the special category,entry format
     */
    public function run(?string $id, array $options, ?string $default = null) : string
    {
        if (!addon_installed('wiki')) {
            return '<result></result>';
        }

        require_code('wiki');
        require_lang('wiki');

        $wiki_seen = [];
        $tree = get_wiki_page_tree($wiki_seen, ($id === null) ? null : intval($id), null, null, false, false, ($id === null) ? 0 : 1);

        $levels_to_expand = array_key_exists('levels_to_expand', $options) ? ($options['levels_to_expand']) : intval(get_value('levels_to_expand__' . substr(get_class($this), 5), null, true));
        $options['levels_to_expand'] = max(0, $levels_to_expand - 1);

        $stripped_id = $id;

        $out = '';

        $out .= '<options>' . xmlentities(json_encode($options)) . '</options>' . "\n";

        if (!has_actual_page_access(null, 'wiki')) {
            $tree = [];
        }

        foreach ($tree as $t) {
            $_id = strval($t['id']);

            if ($stripped_id === $_id) {
                continue; // Possible when we look under as a root
            }

            $title = $t['title'];
            $has_children = ($t['child_count'] != 0);
            $selectable = true;

            $tag = 'category'; // category
            $out .= '<' . $tag . ' id="' . xmlentities($_id) . '" title="' . xmlentities($title) . '" has_children="' . ($has_children ? 'true' : 'false') . '" selectable="' . ($selectable ? 'true' : 'false') . '"></' . $tag . '>' . "\n";

            if ($levels_to_expand > 0) {
                $out .= '<expand>' . xmlentities($_id) . '</expand>' . "\n";
            }
        }

        if ($id === null) {
            $orphans = $GLOBALS['SITE_DB']->query('SELECT p.id,p.title,p.add_date FROM ' . get_table_prefix() . 'wiki_pages p WHERE NOT EXISTS(SELECT * FROM ' . get_table_prefix() . 'wiki_children WHERE child_id=p.id) ORDER BY add_date DESC', 50/*reasonable limit*/, 0, false, false, ['title' => 'SHORT_TRANS']);
            foreach ($orphans as $i => $orphan) {
                $orphans[$i]['_title'] = get_translated_text($orphan['title']);
            }
            if (count($orphans) < 50) {
                sort_maps_by($orphans, '_title', false, true);
            }

            foreach ($orphans as $orphan) {
                if (!has_category_access(get_member(), 'wiki_page', strval($orphan['id']))) {
                    continue;
                }

                if ($orphan['id'] == db_get_first_id()) {
                    continue;
                }

                $title = $orphan['_title'];

                $_id = strval($orphan['id']);
                $title = $orphan['_title'];
                $has_children = ($GLOBALS['SITE_DB']->query_select_value('wiki_children', 'COUNT(*)', ['parent_id' => $orphan['id']]) != 0);
                $selectable = true;

                $tag = 'category'; // category
                $out .= '<' . $tag . ' id="' . $_id . '" title="' . xmlentities($title) . '" has_children="' . ($has_children ? 'true' : 'false') . '" selectable="' . ($selectable ? 'true' : 'false') . '"></' . $tag . '>' . "\n";
            }
        }

        $tag = 'result'; // result
        return '<' . $tag . '>' . $out . '</' . $tag . '>';
    }

    /**
     * Generate a simple selection list for the ajax-tree hook. Returns a normal <select> style <option>-list, for fallback purposes.
     *
     * @param  ?ID_TEXT $id The ID to do under (null: root) - not always supported
     * @param  array $options Options being passed through
     * @param  ?ID_TEXT $it The ID to select by default (null: none)
     * @return Tempcode The nice list
     */
    public function simple(?string $id, array $options, ?string $it = null) : object
    {
        if (!addon_installed('wiki')) {
            return new Tempcode();
        }

        require_lang('wiki');
        require_code('wiki');

        return create_selection_list_wiki_page_tree(($it === null) ? null : intval($it));
    }
}
