<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    cns_warnings
 */

/**
 * Hook class.
 */
class Hook_cns_warnings_probation
{
    /**
     * Get details for this hook.
     *
     * @return ?array The details (null: hook disabled)
     */
    public function get_details() : ?array
    {
        if (!addon_installed('cns_warnings')) {
            return null;
        }

        return [
            'order' => 4,
        ];
    }

    /**
     * Generate punitive action text from a punitive action database row.
     *
     * @param  array $row The database row
     * @return string The punitive action text
     */
    public function generate_text(array $row) : string
    {
        if (!addon_installed('cns_warnings')) {
            return '';
        }

        require_lang('cns_warnings');

        switch ($row['p_action']) {
            case '_PUNITIVE_PROBATION':
                return do_lang('_PUNITIVE_PROBATION', $row['p_param_a']);

            default:
                return '';
        }
    }

    /**
     * Render form fields for the warnings screen.
     *
     * @param  Tempcode $add_text Tempcode to be included on the intro paragraph of the warnings screen (passed by reference)
     * @param  Tempcode $fields The fields to be rendered (passed by reference)
     * @param  Tempcode $hidden The hidden fields to be included (passed by reference)
     * @param  boolean $new Whether it is a new warning/punishment record
     * @param  LONG_TEXT $explanation The explanation for the warning/punishment record
     * @param  BINARY $is_warning Whether to make this a formal warning
     * @param  MEMBER $member_id The member the warning is for
     * @param  BINARY $spam_mode Whether this is a spam warning
     * @param  ?AUTO_LINK $post_id The ID of the forum post of which we clicked warn (null: we are not warning on a forum post)
     * @param  ?SHORT_TEXT $ip_address The IP address of the poster (null: we are not warning on a forum post)
     */
    public function get_form_fields(object &$add_text, object &$fields, object &$hidden, bool $new, string $explanation, int $is_warning, int $member_id, int $spam_mode, ?int $post_id, ?string $ip_address)
    {
        if (!addon_installed('cns_warnings')) {
            return;
        }

        if (!$new) {
            return;
        }

        if (has_privilege(get_member(), 'probate_members')) {
            $fields->attach(form_input_integer(do_lang_tempcode('EXTEND_PROBATION'), do_lang_tempcode('DESCRIPTION_EXTEND_PROBATION'), 'probation', 0, true));
        }
    }

    /**
     * Actualise punitive actions.
     * Note that this assumes action was applied through the warnings form, and that post parameters still exist.
     *
     * @param  array $punitive_messages Punitive action text to potentially be included in the PT automatically (passed by reference)
     * @param  AUTO_LINK $warning_id The ID of the warning that was created for this punitive action
     * @param  MEMBER $member_id The member this warning is being applied to
     * @param  SHORT_TEXT $username The username of the member this warning is being applied to
     * @param  SHORT_TEXT $explanation The defined explanation for this warning
     * @param  LONG_TEXT $message The message to be sent as a PT (passed by reference; you should generally use $punitive_text instead if you want to add PT text)
     */
    public function actualise_punitive_action(array &$punitive_messages, int $warning_id, int $member_id, string $username, string $explanation, string &$message)
    {
        if (!addon_installed('cns_warnings')) {
            return;
        }

        $probation = post_param_integer('probation', 0);

        // Add probation time to the member if we requested and have the privilege
        if (has_privilege(get_member(), 'probate_members')) {
            if ($probation != 0) {
                $probation_expiration_time = $GLOBALS['FORUM_DRIVER']->get_member_row_field($member_id, 'm_probation_expiration_time');
                if (($probation_expiration_time === null) || ($probation_expiration_time < time())) {
                    $probation_expiration_time = time();
                }
                $probation_expiration_time += $probation * 60 * 60 * 24;
                $GLOBALS['FORUM_DB']->query_update('f_members', ['m_probation_expiration_time' => $probation_expiration_time], ['id' => $member_id], '', 1);

                require_code('cns_general_action2');
                cns_mod_log_it('START_PROBATION', strval($member_id), $username, $explanation);

                $GLOBALS['FORUM_DB']->query_insert('f_warnings_punitive', [
                    'p_warning_id' => $warning_id,
                    'p_member_id' => $member_id,
                    'p_ip_address' => '',
                    'p_email_address' => '',
                    'p_hook' => 'probation',
                    'p_action' => '_PUNITIVE_PROBATION',
                    'p_param_a' => strval($probation),
                    'p_param_b' => '',
                    'p_reversed' => 0,
                ]);

                $punitive_messages[] = do_lang('PUNITIVE_PROBATION', strval($probation), strval(get_timezoned_date_time($probation_expiration_time, false, false, $member_id)), null, null, false);
            }
        }
    }

    /**
     * Actualiser to undo a certain type of punitive action.
     *
     * @param  array $punitive_action The database row for the punitive action being undone
     * @param  array $warning The database row for the warning associated with the punitive action being undone
     */
    public function undo_punitive_action(array $punitive_action, array $warning)
    {
        $error = new Tempcode();
        if (!addon_installed__messaged('cns_warnings', $error)) {
            warn_exit($error);
        }

        if (get_forum_type() != 'cns') {
            warn_exit(do_lang_tempcode('INTERNAL_ERROR', escape_html('08c35fb23fb75a87bf728eefcc760c47')));
        }

        $member_id = intval($punitive_action['p_member_id']);
        $probation = intval($punitive_action['p_param_a']);

        $probation_expiration_time = $GLOBALS['FORUM_DRIVER']->get_member_row_field($member_id, 'm_probation_expiration_time');
        if ($probation_expiration_time !== null) {
            $GLOBALS['FORUM_DB']->query_update('f_members', ['m_probation_expiration_time' => $probation_expiration_time - $probation * 60 * 60 * 24], ['id' => $member_id], '', 1);
        }

        require_code('cns_general_action2');
        cns_mod_log_it('STOP_PROBATION', strval($warning['id']), $GLOBALS['FORUM_DRIVER']->get_username($member_id));
    }

    /**
     * Return information for the standing profile tab.
     *
     * @param  MEMBER $member_id_of The member whose profile we are viewing
     * @param  MEMBER $member_id_viewing The member who is viewing the profile
     * @param  array $warning_ids Array of formal warning IDs against this member for checking against queried punitive actions
     * @return array Array of maps with information about this punitive action
     */
    public function get_stepper(int $member_id_of, int $member_id_viewing, array $warning_ids) : array
    {
        if (!addon_installed('cns_warnings') || get_forum_type() != 'cns') {
            return [];
        }

        $info = [];
        $probation = $GLOBALS['FORUM_DRIVER']->get_member_row_field($member_id_of, 'm_probation_expiration_time');
        if (($probation !== null) && ($probation > time())) {
            require_code('temporal');
            $info[] = [
                'icon' => 'menu/adminzone/security',
                'text' => do_lang_tempcode('STANDING_PROBATION_TEXT', escape_html(get_timezoned_date_time($probation, false, false, $member_id_viewing))),
            ];
        }

        return [
            [
                'order' => 80,
                'label' => do_lang('STANDING_PROBATION'),
                'explanation' => do_lang('DESCRIPTION_STANDING_PROBATION'),
                'icon' => 'menu/adminzone/security',
                'active' => !empty($info),
                'active_color' => 'warning',
                'info' => $info,
            ],
        ];
    }
}
