<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    securitylogging
 */

/**
 * Hook class.
 */
class Hook_cns_warnings_syndicate
{
    /**
     * Get details for this hook.
     *
     * @return ?array The details (null: hook disabled)
     */
    public function get_details() : ?array
    {
        if (!addon_installed('cns_warnings') || !addon_installed('securitylogging')) {
            return null;
        }

        return [
            'order' => 3,
        ];
    }

    /**
     * Generate punitive action text from a punitive action database row.
     *
     * @param  array $row The database row
     * @return string The punitive action text
     */
    public function generate_text(array $row) : string
    {
        if (!addon_installed('cns_warnings') || !addon_installed('securitylogging')) {
            return '';
        }

        require_lang('cns_warnings');

        switch ($row['p_action']) {
            case '_PUNITIVE_STOP_FORUM_SPAM':
                return do_lang('_PUNITIVE_STOP_FORUM_SPAM', $row['p_ip_address'], $row['p_email_address']);

            default:
                return '';
        }
    }

    /**
     * Render form fields for the warnings screen.
     *
     * @param  Tempcode $add_text Tempcode to be included on the intro paragraph of the warnings screen (passed by reference)
     * @param  Tempcode $fields The fields to be rendered (passed by reference)
     * @param  Tempcode $hidden The hidden fields to be included (passed by reference)
     * @param  boolean $new Whether it is a new warning/punishment record
     * @param  LONG_TEXT $explanation The explanation for the warning/punishment record
     * @param  BINARY $is_warning Whether to make this a formal warning
     * @param  MEMBER $member_id The member the warning is for
     * @param  BINARY $spam_mode Whether this is a spam warning
     * @param  ?AUTO_LINK $post_id The ID of the forum post of which we clicked warn (null: we are not warning on a forum post)
     * @param  ?SHORT_TEXT $ip_address The IP address of the poster (null: we are not warning on a forum post)
     */
    public function get_form_fields(object &$add_text, object &$fields, object &$hidden, bool $new, string $explanation, int $is_warning, int $member_id, int $spam_mode, ?int $post_id, ?string $ip_address)
    {
        if (!addon_installed('cns_warnings') || !addon_installed('securitylogging')) {
            return;
        }

        if (!$new) {
            return;
        }

        if ((get_option('stopforumspam_api_key') != '') || (get_option('spam_use_tornevall') == '1')) {
            require_lang('submitban');
            $fields->attach(form_input_tick(do_lang_tempcode('SYNDICATE_TO_STOPFORUMSPAM'), do_lang_tempcode('DESCRIPTION_SYNDICATE_TO_STOPFORUMSPAM'), 'stopforumspam', ($spam_mode == 1)));
        }
    }

    /**
     * Actualise punitive actions.
     * Note that this assumes action was applied through the warnings form, and that post parameters still exist.
     *
     * @param  array $punitive_messages Punitive action text to potentially be included in the PT automatically (passed by reference)
     * @param  AUTO_LINK $warning_id The ID of the warning that was created for this punitive action
     * @param  MEMBER $member_id The member this warning is being applied to
     * @param  SHORT_TEXT $username The username of the member this warning is being applied to
     * @param  SHORT_TEXT $explanation The defined explanation for this warning
     * @param  LONG_TEXT $message The message to be sent as a PT (passed by reference; you should generally use $punitive_text instead if you want to add PT text)
     */
    public function actualise_punitive_action(array &$punitive_messages, int $warning_id, int $member_id, string $username, string $explanation, string &$message)
    {
        if (!addon_installed('cns_warnings') || !addon_installed('securitylogging')) {
            return;
        }

        $stopforumspam = post_param_integer('stopforumspam', 0);

        // Syndicate spam report if requested
        if ($stopforumspam == 1) {
            $banned_ip = $GLOBALS['FORUM_DRIVER']->get_member_row_field($member_id, 'm_ip_address');
            $email_address = $GLOBALS['FORUM_DRIVER']->get_member_email_address($member_id);
            require_code('failure');
            require_code('failure_spammers');
            syndicate_spammer_report($banned_ip, $username, $email_address, $explanation, false);

            require_code('cns_general_action2');
            cns_mod_log_it('MARK_AS_SPAMMER', strval($member_id), $username, $explanation);

            $GLOBALS['FORUM_DB']->query_insert('f_warnings_punitive', [
                'p_warning_id' => $warning_id,
                'p_member_id' => $member_id,
                'p_ip_address' => $banned_ip,
                'p_email_address' => $email_address,
                'p_hook' => 'syndicate',
                'p_action' => '_PUNITIVE_STOP_FORUM_SPAM',
                'p_param_a' => '',
                'p_param_b' => '',
                'p_reversed' => 0,
            ]);

            $punitive_messages[] = do_lang('PUNITIVE_STOP_FORUM_SPAM', null, null, null, null, false);
        }
    }

    /**
     * Return information for the standing profile tab.
     *
     * @param  MEMBER $member_id_of The member whose profile we are viewing
     * @param  MEMBER $member_id_viewing The member who is viewing the profile
     * @param  array $warning_ids Array of formal warning IDs against this member for checking against queried punitive actions
     * @return array Array of maps with information about this punitive action
     */
    public function get_stepper(int $member_id_of, int $member_id_viewing, array $warning_ids) : array
    {
        if (!addon_installed('cns_warnings') || !addon_installed('securitylogging')) {
            return [];
        }

        $info = [];
        foreach ($warning_ids as $warning_id) {
            $row = $GLOBALS['SITE_DB']->query_select_value_if_there('f_warnings_punitive', 'id', ['p_warning_id' => $warning_id, 'p_action' => '_PUNITIVE_STOP_FORUM_SPAM']);
            if ($row !== null) {
                $info[] = [
                    'icon' => 'menu/adminzone/security',
                    'text' => do_lang_tempcode('STANDING_DANGER_SYNDICATE_TEXT'),
                ];
                break;
            }
        }

        return [
            [
                'order' => 100,
                'label' => do_lang('STANDING_DANGER'),
                'explanation' => do_lang('DESCRIPTION_STANDING_DANGER'),
                'icon' => 'buttons/no',
                'active' => !empty($info),
                'active_color' => 'danger',
                'info' => $info,
            ],
        ];
    }
}
