<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    leader_board
 */

require_code('resource_fs');

/**
 * Hook class.
 */
class Hook_commandr_fs_leader_board extends Resource_fs_base
{
    public $file_resource_type = 'leader_board';

    /**
     * Standard Commandr-fs function for seeing how many resources are. Useful for determining whether to do a full rebuild.
     *
     * @param  ID_TEXT $resource_type The resource type
     * @return integer How many resources there are
     */
    public function get_resources_count(string $resource_type) : int
    {
        return $GLOBALS['SITE_DB']->query_select_value('leader_boards', 'COUNT(*)');
    }

    /**
     * Standard Commandr-fs function for searching for a resource by label.
     *
     * @param  ID_TEXT $resource_type The resource type
     * @param  LONG_TEXT $label The resource label
     * @return array A list of resource IDs
     */
    public function find_resource_by_label(string $resource_type, string $label) : array
    {
        $_ret = $GLOBALS['SITE_DB']->query_select('leader_boards', ['id'], ['lb_title' => $label], 'ORDER BY id');
        $ret = [];
        foreach ($_ret as $r) {
            $ret[] = strval($r['id']);
        }
        return $ret;
    }

    /**
     * Whether the filesystem hook is active.
     *
     * @return boolean Whether it is
     */
    public function is_active() : bool
    {
        return (addon_installed('points')) && (addon_installed('leader_board'));
    }

    /**
     * Standard Commandr-fs date fetch function for resource-fs hooks. Defined when getting an edit date is not easy.
     *
     * @param  array $row Resource row (not full, but does contain the ID)
     * @param  ID_TEXT $category Parent category (blank: root / not applicable)
     * @return ?TIME The edit date or add date, whichever is higher (null: could not find one)
     */
    protected function _get_file_edit_date(array $row, string $category = '') : ?int
    {
        $query = 'SELECT MAX(date_and_time) FROM ' . get_table_prefix() . 'actionlogs WHERE ' . db_string_equal_to('param_a', strval($row['id'])) . ' AND  (' . db_string_equal_to('the_type', 'ADD_LEADER_BOARD') . ' OR ' . db_string_equal_to('the_type', 'EDIT_LEADER_BOARD') . ')';
        return $GLOBALS['SITE_DB']->query_value_if_there($query);
    }

    /**
     * Standard Commandr-fs add function for resource-fs hooks. Adds some resource with the given label and properties.
     *
     * @param  LONG_TEXT $filename Filename OR Resource label
     * @param  string $path The path (blank: root / not applicable)
     * @param  array $properties Properties (may be empty, properties given are open to interpretation by the hook but generally correspond to database fields)
     * @param  ?ID_TEXT $force_type Resource type to try to force (null: do not force)
     * @return ~ID_TEXT The resource ID (false: error, could not create via these properties / here)
     */
    public function file_add(string $filename, string $path, array $properties, ?string $force_type = null)
    {
        list($properties, $label) = $this->_file_magic_filter($filename, $path, $properties, $this->file_resource_type);

        require_code('leader_board2');

        $board_type = $this->_default_property_str($properties, 'type');
        if ($board_type == '') {
            $board_type = 'holders';
        }
        $member_count = $this->_default_property_int($properties, 'member_count');
        if ($member_count == 0) {
            $member_count = 10;
        }
        $timeframe = $this->_default_property_str($properties, 'timeframe');
        if ($timeframe == '') {
            $timeframe = 'month';
        }
        $rolling = $this->_default_property_int($properties, 'rolling');
        $include_staff = $this->_default_property_int($properties, 'include_staff');
        $calculate_voting_power = $this->_default_property_int($properties, 'calculate_voting_power');

        $id = add_leader_board($label, $board_type, $member_count, $timeframe, $rolling, $include_staff, null, $calculate_voting_power);

        if (isset($properties['groups'])) {
            table_from_portable_rows('leader_boards_groups', $properties['groups'], ['lb_leader_board_id' => $id], TABLE_REPLACE_MODE_BY_EXTRA_FIELD_DATA);
        }

        $this->_resource_save_extend($this->file_resource_type, strval($id), $filename, $label, $properties);

        return strval($id);
    }

    /**
     * Standard Commandr-fs load function for resource-fs hooks. Finds the properties for some resource.
     *
     * @param  SHORT_TEXT $filename Filename
     * @param  string $path The path (blank: root / not applicable). It may be a wildcarded path, as the path is used for content-type identification only. Filenames are globally unique across a hook; you can calculate the path using ->search.
     * @return ~array Details of the resource (false: error)
     */
    public function file_load(string $filename, string $path)
    {
        list($resource_type, $resource_id) = $this->file_convert_filename_to_id($filename);

        $rows = $GLOBALS['SITE_DB']->query_select('leader_boards', ['*'], ['id' => intval($resource_id)], '', 1);
        if (!array_key_exists(0, $rows)) {
            return false;
        }
        $row = $rows[0];

        $groups = table_to_portable_rows('leader_boards_groups', [], ['lb_leader_board_id' => intval($resource_id)]);

        $properties = [
            'title' => $row['lb_title'],
            'type' => $row['lb_type'],
            'member_count' => $row['lb_member_count'],
            'timeframe' => $row['lb_timeframe'],
            'rolling' => $row['lb_rolling'],
            'include_staff' => $row['lb_include_staff'],
            'calculate_voting_power' => $row['lb_calculate_voting_power'],
            'groups' => $groups,
        ];
        $this->_resource_load_extend($resource_type, $resource_id, $properties, $filename, $path);
        return $properties;
    }

    /**
     * Standard Commandr-fs edit function for resource-fs hooks. Edits the resource to the given properties.
     *
     * @param  ID_TEXT $filename The filename
     * @param  string $path The path (blank: root / not applicable)
     * @param  array $properties Properties (may be empty, properties given are open to interpretation by the hook but generally correspond to database fields)
     * @param  boolean $explicit_move Whether we are definitely moving (as opposed to possible having it in multiple positions)
     * @return ~ID_TEXT The resource ID (false: error, could not create via these properties / here)
     */
    public function file_edit(string $filename, string $path, array $properties, bool $explicit_move = false)
    {
        list($resource_type, $resource_id) = $this->file_convert_filename_to_id($filename);
        list($properties, $label) = $this->_file_magic_filter($filename, $path, $properties, $this->file_resource_type);

        if ($resource_id === null) {
            return false;
        }

        require_code('leader_board2');

        $board_type = $this->_default_property_str($properties, 'type');
        if ($board_type == '') {
            $board_type = 'holders';
        }
        $member_count = $this->_default_property_int($properties, 'member_count');
        if ($member_count == 0) {
            $member_count = 10;
        }
        $timeframe = $this->_default_property_str($properties, 'timeframe');
        if ($timeframe == '') {
            $timeframe = 'month';
        }
        $rolling = $this->_default_property_int($properties, 'rolling');
        $include_staff = $this->_default_property_int($properties, 'include_staff');
        $calculate_voting_power = $this->_default_property_int($properties, 'calculate_voting_power');

        edit_leader_board(intval($resource_id), $label, $board_type, $member_count, $timeframe, $rolling, $include_staff, null, $calculate_voting_power);

        if (isset($properties['groups'])) {
            table_from_portable_rows('leader_boards_groups', $properties['groups'], ['lb_leader_board_id' => intval($resource_id)], TABLE_REPLACE_MODE_BY_EXTRA_FIELD_DATA);
        }

        $this->_resource_save_extend($this->file_resource_type, $resource_id, $filename, $label, $properties);

        return $resource_id;
    }

    /**
     * Standard Commandr-fs delete function for resource-fs hooks. Deletes the resource.
     *
     * @param  ID_TEXT $filename The filename
     * @param  string $path The path (blank: root / not applicable)
     * @return boolean Success status
     */
    public function file_delete(string $filename, string $path) : bool
    {
        list($resource_type, $resource_id) = $this->file_convert_filename_to_id($filename);

        if ($resource_id === null) {
            return false;
        }

        require_code('leader_board2');
        delete_leader_board(intval($resource_id));

        return true;
    }
}
