<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    calendar
 */

/**
 * Hook class.
 */
class Hook_content_meta_aware_event extends Hook_CMA
{
    /**
     * Get content type details.
     *
     * @param  ?ID_TEXT $zone The zone to link through to (null: autodetect)
     * @param  boolean $get_extended_data Populate additional data that is somewhat costly to compute (add_url, archive_url)
     * @return ?array Map of content-type info (null: disabled)
     */
    public function info(?string $zone = null, bool $get_extended_data = false) : ?array
    {
        if (!addon_installed('calendar')) {
            return null;
        }

        return [
            'support_custom_fields' => true,

            'content_type_label' => 'calendar:EVENT',
            'content_type_universal_label' => 'Calendar event',

            'db' => $GLOBALS['SITE_DB'],
            'extra_where_sql' => 'r.e_member_calendar IS NULL',
            'table' => 'calendar_events',
            'id_field' => 'id',
            'id_field_numeric' => true,
            'parent_category_field' => 'e_type',
            'parent_category_meta_aware_type' => 'calendar_type',
            'is_category' => false,
            'is_entry' => true,
            'category_field' => 'e_type', // For category permissions
            'permission_module' => 'calendar', // For category permissions
            'parent_spec__table_name' => 'calendar_types',
            'parent_spec__parent_name' => null,
            'parent_spec__field_name' => 'id',
            'category_is_string' => false,

            'title_field' => 'e_title',
            'title_field_dereference' => true,
            'description_field' => 'e_content',
            'description_field_dereference' => true,
            'description_field_supports_comcode' => true,
            'image_field' => null,
            'image_field_is_theme_image' => false,
            'alternate_icon_theme_image' => 'icons/menu/rich_content/calendar',

            'view_page_link_pattern' => '_SEARCH:calendar:view:_WILD',
            'edit_page_link_pattern' => '_SEARCH:cms_calendar:_edit:_WILD',
            'view_category_page_link_pattern' => '_SEARCH:calendar:browse:_WILD',
            'add_url' => ($get_extended_data && has_submit_permission('mid', get_member(), get_ip_address(), 'cms_calendar')) ? (get_module_zone('cms_calendar') . ':cms_calendar:add') : null,
            'archive_url' => $get_extended_data ? ((($zone !== null) ? $zone : get_module_zone('calendar')) . ':calendar') : null,

            'support_url_monikers' => true,

            'views_field' => 'e_views',
            'order_field' => null,
            'submitter_field' => 'e_submitter',
            'author_field' => null,
            'add_time_field' => 'e_add_date',
            'edit_time_field' => 'e_edit_date',
            'date_field' => 'e_next_recurrence_time',
            'validated_field' => 'validated',
            'additional_sort_fields' => [
                'start_time' => [
                    'ASC' => 'e_start_year ASC,e_start_month ASC,e_start_day ASC,e_start_hour ASC,e_start_minute ASC',
                    'DESC' => 'e_start_year DESC,e_start_month DESC,e_start_day DESC,e_start_hour DESC,e_start_minute DESC',
                ],
            ],

            'seo_type_code' => 'event',

            'feedback_type_code' => 'events',

            'search_hook' => 'calendar',
            'rss_hook' => 'calendar',
            'attachment_hook' => 'calendar',
            'notification_hook' => 'calendar_event',
            'sitemap_hook' => 'event',

            'addon_name' => 'calendar',

            'cms_page' => 'cms_calendar',
            'module' => 'calendar',

            'commandr_filesystem_hook' => 'calendar',
            'commandr_filesystem__is_folder' => false,

            'support_revisions' => false,

            'support_privacy' => true,

            'support_content_reviews' => true,

            'support_spam_heuristics' => 'post',

            'actionlog_regexp' => '\w+_EVENT',

            'default_prominence_weight' => PROMINENCE_WEIGHT_HIGHEST | PROMINENCE_FLAG_ACTIVE_ONLY,
            'default_prominence_flags' => 0,

            // No support for recurrences due to complexity of that
            'active_only_extra_where_sql' => '(r.e_next_recurrence_time IS NOT NULL OR r.e_previous_recurrence_time IS NOT NULL AND r.e_previous_recurrence_time>' . strval(time() - 60 * 60 * 24 * 2) . ')', // Only show those in future or recent past
            'prominence_custom_sort' => '(' . strval(time()) . '-' . db_function('LEAST', [db_function('ABS', [strval(time()) . '-' . db_function('COALESCE', ['e_previous_recurrence_time', strval(PHP_INT_MAX)])]), db_function('ABS', [strval(time()) . '-' . db_function('COALESCE', ['e_next_recurrence_time', strval(PHP_INT_MAX)])])]) . ')',
            'prominence_custom_sort_dir' => 'ASC',
        ];
    }

    /**
     * Render a content box for a content row.
     *
     * @param  array $row The database row for the content
     * @param  ID_TEXT $zone The zone to display in
     * @param  boolean $give_context Whether to include context (i.e. say WHAT this is, not just show the actual content)
     * @param  boolean $include_breadcrumbs Whether to include breadcrumbs (if there are any)
     * @param  ?ID_TEXT $root Virtual root to use (null: none)
     * @param  boolean $attach_to_url_filter Whether to copy through any filter parameters in the URL, under the basis that they are associated with what this box is browsing
     * @param  ID_TEXT $guid Overridden GUID to send to templates (blank: none)
     * @return Tempcode Results
     */
    public function render_box(array $row, string $zone, bool $give_context = true, bool $include_breadcrumbs = true, ?string $root = null, bool $attach_to_url_filter = false, string $guid = '') : object
    {
        require_code('calendar');

        return render_event_box($row, $zone, $give_context, $guid);
    }

    /**
     * Create a selection list.
     *
     * @param  ?string $id The pre-selected ID (null: none selected)
     * @return Tempcode List
     */
    public function create_selection_list(?string $id = null) : object
    {
        require_code('calendar');

        return create_selection_list_events(null, ($id === null) ? null : intval($id));
    }
}
