<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    news
 */

/**
 * Hook class.
 */
class Hook_content_meta_aware_news_category extends Hook_CMA
{
    /**
     * Get content type details.
     *
     * @param  ?ID_TEXT $zone The zone to link through to (null: autodetect)
     * @param  boolean $get_extended_data Populate additional data that is somewhat costly to compute (add_url, archive_url)
     * @return ?array Map of content-type info (null: disabled)
     */
    public function info(?string $zone = null, bool $get_extended_data = false) : ?array
    {
        if (!addon_installed('news')) {
            return null;
        }

        return [
            'support_custom_fields' => false,

            'content_type_label' => 'news:NEWS_CATEGORY',
            'content_type_universal_label' => 'News category',

            'db' => $GLOBALS['SITE_DB'],
            'table' => 'news_categories',
            'id_field' => 'id',
            'id_field_numeric' => true,
            'parent_category_field' => null,
            'parent_category_meta_aware_type' => 'news_category',
            'is_category' => true,
            'is_entry' => false,
            'permission_module' => 'news', // For category permissions
            'parent_spec__table_name' => null,
            'parent_spec__parent_name' => null,
            'parent_spec__field_name' => null,
            'category_field' => 'id', // For category permissions
            'category_is_string' => false,

            'title_field' => 'nc_title',
            'title_field_dereference' => true,
            'description_field' => null,
            'description_field_dereference' => true,
            'description_field_supports_comcode' => true,
            'image_field' => ['nc_img', 'CALL: generate_news_category_image_url'],
            'image_field_is_theme_image' => false,
            'alternate_icon_theme_image' => null,

            'view_page_link_pattern' => '_SEARCH:news:browse:_WILD',
            'edit_page_link_pattern' => '_SEARCH:cms_news:_edit_category:_WILD',
            'view_category_page_link_pattern' => '_SEARCH:news:browse:_WILD',
            'add_url' => ($get_extended_data && has_submit_permission('mid', get_member(), get_ip_address(), 'cms_news')) ? (get_module_zone('cms_news') . ':cms_news:add') : null,
            'archive_url' => $get_extended_data ? ((($zone !== null) ? $zone : get_module_zone('news')) . ':news') : null,

            'support_url_monikers' => true,

            'views_field' => null,
            'order_field' => null,
            'submitter_field' => null,
            'author_field' => null,
            'add_time_field' => null,
            'edit_time_field' => null,
            'date_field' => null,
            'validated_field' => null,
            'additional_sort_fields' => [
                'entries_count' => '((SELECT COUNT(*) FROM ' . get_table_prefix() . 'news n WHERE n.news_category=r.id) + (SELECT COUNT(*) FROM ' . get_table_prefix() . 'news_category_entries nc WHERE nc.news_entry_category=r.id))',
            ],

            'seo_type_code' => 'news_category',

            'feedback_type_code' => null,

            'search_hook' => null,
            'rss_hook' => null,
            'attachment_hook' => null,
            'notification_hook' => null,
            'sitemap_hook' => 'news_category',

            'addon_name' => 'news',

            'cms_page' => 'cms_news',
            'module' => 'news',

            'commandr_filesystem_hook' => 'news',
            'commandr_filesystem__is_folder' => true,

            'support_revisions' => false,

            'support_privacy' => false,

            'support_content_reviews' => true,

            'support_spam_heuristics' => null,

            'actionlog_regexp' => '\w+_NEWS_CATEGORY',

            'default_prominence_weight' => PROMINENCE_WEIGHT_NONE,
            'default_prominence_flags' => 0,
            'prominence_custom_sort' => '(SELECT MAX(date_and_time) FROM ' . get_table_prefix() . 'news WHERE news_category=r.id AND validated=1)',
            'prominence_custom_sort_dir' => 'DESC',
        ];
    }

    /**
     * Get headings of special relevant data this content type supports.
     *
     * @return array A map of heading codenames to Tempcode labels
     */
    public function get_special_keymap_headings() : array
    {
        $headings = [];

        $headings['entry_count'] = do_lang_tempcode('ENTRIES');

        return $headings;
    }

    /**
     * Get special relevant data this content type supports.
     *
     * @param  array $row Database row
     * @return array A map of heading codenames to Tempcode values
     */
    public function get_special_keymap(array $row) : array
    {
        $keymap = [];

        $num_entries = $GLOBALS['SITE_DB']->query_select_value('news', 'COUNT(*)', ['validated' => 1, 'news_category' => $row['id']]);
        $num_entries += $GLOBALS['SITE_DB']->query_select_value('news n JOIN ' . get_table_prefix() . 'news_category_entries c ON c.news_entry=n.id', 'COUNT(*)', ['validated' => 1, 'news_entry_category' => $row['id']]);
        $keymap['entry_count'] = escape_html(integer_format($num_entries, 0));

        return $keymap;
    }

    /**
     * Render a content box for a content row.
     *
     * @param  array $row The database row for the content
     * @param  ID_TEXT $zone The zone to display in
     * @param  boolean $give_context Whether to include context (i.e. say WHAT this is, not just show the actual content)
     * @param  boolean $include_breadcrumbs Whether to include breadcrumbs (if there are any)
     * @param  ?ID_TEXT $root Virtual root to use (null: none)
     * @param  boolean $attach_to_url_filter Whether to copy through any filter parameters in the URL, under the basis that they are associated with what this box is browsing
     * @param  ID_TEXT $guid Overridden GUID to send to templates (blank: none)
     * @return Tempcode Results
     */
    public function render_box(array $row, string $zone, bool $give_context = true, bool $include_breadcrumbs = true, ?string $root = null, bool $attach_to_url_filter = false, string $guid = '') : object
    {
        require_code('news');

        return render_news_category_box($row, $zone, $give_context, $attach_to_url_filter, null, $guid);
    }

    /**
     * Create a selection list.
     *
     * @param  ?string $id The pre-selected ID (null: none selected)
     * @return Tempcode List
     */
    public function create_selection_list(?string $id = null) : object
    {
        require_code('news');

        return create_selection_list_news_categories(($id === null) ? null : intval($id));
    }
}

/**
 * Find an entry image.
 *
 * @param  array $row Database row of entry
 * @return string The image URL
 */
function generate_news_category_image_url(array $row) : string
{
    if (!addon_installed('news')) {
        return '';
    }

    require_code('news');
    return get_news_category_image_url($row['nc_img']);
}
