<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    wiki
 */

/**
 * Hook class.
 */
class Hook_content_meta_aware_wiki_post extends Hook_CMA
{
    /**
     * Get content type details.
     *
     * @param  ?ID_TEXT $zone The zone to link through to (null: autodetect)
     * @param  boolean $get_extended_data Populate additional data that is somewhat costly to compute (add_url, archive_url)
     * @return ?array Map of content-type info (null: disabled)
     */
    public function info(?string $zone = null, bool $get_extended_data = false) : ?array
    {
        if (!addon_installed('wiki')) {
            return null;
        }

        return [
            'support_custom_fields' => true,

            'content_type_label' => 'wiki:WIKI_POST',
            'content_type_universal_label' => 'Wiki+ Post',

            'db' => $GLOBALS['SITE_DB'],
            'table' => 'wiki_posts',
            'id_field' => 'id',
            'id_field_numeric' => true,
            'parent_category_field' => 'page_id',
            'parent_category_meta_aware_type' => 'wiki_page',
            'is_category' => false,
            'is_entry' => true,
            'category_field' => 'page_id', // For category permissions
            'permission_module' => 'wiki_page', // For category permissions
            'parent_spec__table_name' => 'wiki_children',
            'parent_spec__parent_name' => 'parent_id',
            'parent_spec__field_name' => 'child_id',
            'category_is_string' => false,

            'title_field' => null,
            'title_field_dereference' => false,
            'description_field' => 'the_message',
            'description_field_dereference' => true,
            'description_field_supports_comcode' => true,
            'image_field' => null,
            'image_field_is_theme_image' => false,
            'alternate_icon_theme_image' => 'icons/menu/rich_content/wiki',

            'view_page_link_pattern' => '_SEARCH:wiki:find_post:_WILD',
            'edit_page_link_pattern' => '_SEARCH:wiki:post:post_id=_WILD',
            'view_category_page_link_pattern' => '_SEARCH:wiki:browse:_WILD',
            'add_url' => ($get_extended_data && has_submit_permission('low', get_member(), get_ip_address(), 'wiki')) ? (get_module_zone('wiki') . ':wiki:add_post') : null,
            'archive_url' => $get_extended_data ? ((($zone !== null) ? $zone : get_module_zone('wiki')) . ':wiki') : null,

            'support_url_monikers' => false,

            'views_field' => 'wiki_views',
            'order_field' => null,
            'submitter_field' => 'member_id',
            'author_field' => null,
            'add_time_field' => 'date_and_time',
            'edit_time_field' => 'edit_date',
            'date_field' => 'date_and_time',
            'validated_field' => 'validated',
            'validation_is_minor' => true,
            'additional_sort_fields' => null,

            'seo_type_code' => null,

            'feedback_type_code' => null,

            'search_hook' => 'wiki_posts',
            'rss_hook' => null,
            'attachment_hook' => 'wiki_post',
            'notification_hook' => 'wiki',
            'sitemap_hook' => null,

            'addon_name' => 'wiki',

            'cms_page' => 'wiki',
            'module' => 'wiki',

            'commandr_filesystem_hook' => 'wiki',
            'commandr_filesystem__is_folder' => false,

            'support_revisions' => true,

            'support_privacy' => false,

            'support_content_reviews' => false,

            'support_spam_heuristics' => 'post',

            'actionlog_regexp' => '\w+_WIKI_POST',

            'default_prominence_weight' => PROMINENCE_WEIGHT_NONE,
            'default_prominence_flags' => 0,
        ];
    }

    /**
     * Get content title of a content row.
     *
     * @param  array $row The database row for the content
     * @param  integer $render_type A FIELD_RENDER_* constant
     * @param  boolean $falled_back_to_id Whether this has had to fall back to an ID due to missing title (returned by reference)
     * @param  boolean $resource_fs_style Whether to use the content API as resource-fs requires (may be slightly different)
     * @return mixed Content title (string or Tempcode, depending on $render_type)
     */
    public function get_title(array $row, int $render_type = 1, bool &$falled_back_to_id = false, bool $resource_fs_style = false)
    {
        $falled_back_to_id = false;
        $ret = parent::get_title($row, $render_type, $falled_back_to_id, $resource_fs_style);
        if ($falled_back_to_id) {
            if (array_key_exists('page_id', $row)) {
                $wiki_page_title = $GLOBALS['SITE_DB']->query_select_value_if_there('wiki_pages', 'title', ['id' => $row['page_id']]);
                if ($wiki_page_title !== null) {
                    $ret = do_lang('wiki:WIKI_POST_ON', $wiki_page_title);
                } else {
                    $ret = do_lang('wiki:WIKI_POST_ON', ltrim($ret, '#'));
                }
            } else {
                $ret = do_lang('wiki:WIKI_POST_ON', ltrim($ret, '#'));
            }
        }
        return $ret;
    }

    /**
     * Render a content box for a content row.
     *
     * @param  array $row The database row for the content
     * @param  ID_TEXT $zone The zone to display in
     * @param  boolean $give_context Whether to include context (i.e. say WHAT this is, not just show the actual content)
     * @param  boolean $include_breadcrumbs Whether to include breadcrumbs (if there are any)
     * @param  ?ID_TEXT $root Virtual root to use (null: none)
     * @param  boolean $attach_to_url_filter Whether to copy through any filter parameters in the URL, under the basis that they are associated with what this box is browsing
     * @param  ID_TEXT $guid Overridden GUID to send to templates (blank: none)
     * @return Tempcode Results
     */
    public function render_box(array $row, string $zone, bool $give_context = true, bool $include_breadcrumbs = true, ?string $root = null, bool $attach_to_url_filter = false, string $guid = '') : object
    {
        require_code('wiki');

        return render_wiki_post_box($row, $zone, $give_context, $include_breadcrumbs, ($root === null) ? null : intval($root), $guid);
    }
}
