<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    core_privacy
 */

/**
 * Hook class.
 */
class Hook_cron_privacy_purging
{
    /**
     * Get info from this hook.
     *
     * @param  ?TIME $last_run Last time run (null: never)
     * @param  ?boolean $calculate_num_queued Calculate the number of items queued, if possible (null: the hook may decide / low priority)
     * @return ?array Return a map of info about the hook (null: disabled)
     */
    public function info(?int $last_run, ?bool $calculate_num_queued) : ?array
    {
        return [
            'label' => 'Privacy purging (database)',
            'num_queued' => null,
            'minutes_between_runs' => 60 * 24,
            'enabled_by_default' => true,
        ];
    }

    /**
     * Run function for system scheduler hooks. Searches for things to do. ->info(..., true) must be called before this method.
     *
     * @param  ?TIME $last_run Last time run (null: never)
     */
    public function run(?int $last_run)
    {
        disable_php_memory_limit();

        require_code('privacy');

        $hook_obs = find_all_hook_obs('systems', 'privacy', 'Hook_privacy_');
        foreach ($hook_obs as $hook_ob) {
            $details = $hook_ob->info();
            if ($details !== null) {
                foreach ($details['database_records'] as $table_name => $table_details) {
                    if (($table_details['retention_handle_method'] != PRIVACY_METHOD__LEAVE) && ($table_details['timestamp_field'] !== null)) {
                        $table_action = $table_details['retention_handle_method'];
                        $this->handle_for_table($hook_ob, $table_name, $table_details, $table_action);
                    }
                }
            }
        }
    }

    /**
     * Run the task hook.
     *
     * @param  object $hook_ob Privacy object
     * @param  string $table_name Table name
     * @param  array $table_details Table details
     * @param  integer $table_action A PRIVACY_METHOD_* constant
     */
    protected function handle_for_table(object $hook_ob, string $table_name, array $table_details, int $table_action)
    {
        $db = get_db_for($table_name);

        $selection_sql = $hook_ob->get_selection_sql($table_name, $table_details, $table_action, true);
        if ($selection_sql == '') { // Blank result means we should not run for this table (no filters / retention period)
            return;
        }

        $sql = 'SELECT * FROM ' . $db->get_table_prefix() . $table_name;
        $sql .= $selection_sql;
        $rows = $db->query($sql);

        foreach ($rows as $row) {
            switch ($table_action) {
                case PRIVACY_METHOD__ANONYMISE:
                    $hook_ob->anonymise($table_name, $table_details, $row);
                    break;

                case PRIVACY_METHOD__DELETE:
                    $hook_ob->delete($table_name, $table_details, $row);
                    break;
            }
        }
    }
}
