<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    health_check
 */

/**
 * Hook class.
 */
class Hook_health_check_marketing_seo_robotstxt extends Hook_Health_Check
{
    /*
    Hook covers both SEO and Security issues in robots.txt
    */

    protected $category_label = 'robots.txt';

    /**
     * Standard hook run function to run this category of health checks.
     *
     * @param  ?array $sections_to_run Which check sections to run (null: all)
     * @param  integer $check_context The current state of the website (a CHECK_CONTEXT__* constant)
     * @param  boolean $manual_checks Mention manual checks
     * @param  boolean $automatic_repair Do automatic repairs where possible
     * @param  ?boolean $use_test_data_for_pass Should test data be for a pass [if test data supported] (null: no test data)
     * @param  ?array $urls_or_page_links List of URLs and/or page-links to operate on, if applicable (null: those configured)
     * @param  ?array $comcode_segments Map of field names to Comcode segments to operate on, if applicable (null: N/A)
     * @param  boolean $show_unusable_categories Whether to include categories that might not be accessible for some reason
     * @return array A pair: category label, list of results
     */
    public function run(?array $sections_to_run, int $check_context, bool $manual_checks = false, bool $automatic_repair = false, ?bool $use_test_data_for_pass = null, ?array $urls_or_page_links = null, ?array $comcode_segments = null, bool $show_unusable_categories = false) : array
    {
        $this->process_checks_section('testRobotsTxtValidity', 'robots.txt validity', $sections_to_run, $check_context, $manual_checks, $automatic_repair, $use_test_data_for_pass, $urls_or_page_links, $comcode_segments);
        $this->process_checks_section('testRobotsTxtCorrectness', 'robots.txt correctness', $sections_to_run, $check_context, $manual_checks, $automatic_repair, $use_test_data_for_pass, $urls_or_page_links, $comcode_segments);
        $this->process_checks_section('testRobotsTxtCompleteness', 'robots.txt completeness', $sections_to_run, $check_context, $manual_checks, $automatic_repair, $use_test_data_for_pass, $urls_or_page_links, $comcode_segments);
        $this->process_checks_section('testRobotsTxtSitemapLinkage', 'Sitemap linkage', $sections_to_run, $check_context, $manual_checks, $automatic_repair, $use_test_data_for_pass, $urls_or_page_links, $comcode_segments);

        return [$this->category_label, $this->results];
    }

    /**
     * Run a section of health checks.
     *
     * @param  integer $check_context The current state of the website (a CHECK_CONTEXT__* constant)
     * @param  boolean $manual_checks Mention manual checks
     * @param  boolean $automatic_repair Do automatic repairs where possible
     * @param  ?boolean $use_test_data_for_pass Should test data be for a pass [if test data supported] (null: no test data)
     * @param  ?array $urls_or_page_links List of URLs and/or page-links to operate on, if applicable (null: those configured)
     * @param  ?array $comcode_segments Map of field names to Comcode segments to operate on, if applicable (null: N/A)
     */
    public function testRobotsTxtValidity(int $check_context, bool $manual_checks = false, bool $automatic_repair = false, ?bool $use_test_data_for_pass = null, ?array $urls_or_page_links = null, ?array $comcode_segments = null)
    {
        if ($check_context == CHECK_CONTEXT__INSTALL) {
            $this->log('Skipped; we are running from installer.');
            return;
        }
        if ($check_context == CHECK_CONTEXT__SPECIFIC_PAGE_LINKS) {
            $this->log('Skipped; running on specific page links.');
            return;
        }

        $rules = $this->robots_parse(null, true);

        $this->assertTrue($rules !== null, '[tt]robots.txt[/tt] not found on domain root, or not web accessible. This file is necessary to tell web crawlers what to do and not to do.');
    }

    /**
     * Run a section of health checks.
     *
     * @param  integer $check_context The current state of the website (a CHECK_CONTEXT__* constant)
     * @param  boolean $manual_checks Mention manual checks
     * @param  boolean $automatic_repair Do automatic repairs where possible
     * @param  ?boolean $use_test_data_for_pass Should test data be for a pass [if test data supported] (null: no test data)
     * @param  ?array $urls_or_page_links List of URLs and/or page-links to operate on, if applicable (null: those configured)
     * @param  ?array $comcode_segments Map of field names to Comcode segments to operate on, if applicable (null: N/A)
     */
    public function testRobotsTxtCorrectness(int $check_context, bool $manual_checks = false, bool $automatic_repair = false, ?bool $use_test_data_for_pass = null, ?array $urls_or_page_links = null, ?array $comcode_segments = null)
    {
        if ($check_context == CHECK_CONTEXT__INSTALL) {
            $this->log('Skipped; we are running from installer.');
            return;
        }
        if ($check_context == CHECK_CONTEXT__SPECIFIC_PAGE_LINKS) {
            $this->log('Skipped; running on specific page links.');
            return;
        }

        $rules = $this->robots_parse(null, true);
        if ($rules === null) {
            $this->stateCheckSkipped('No robots.txt file');
            return;
        }

        $url = $this->get_page_url();

        $google_allowed = $this->robots_allowed($url, 'Googlebot', true);
        $other_allowed = $this->robots_allowed($url, 'Googlebot', false); // We'll still check for Google, just with the other way of doing precedence

        if ($check_context == CHECK_CONTEXT__TEST_SITE) {
            if (is_local_machine(get_base_url_hostname())) {
                return; // Google cannot access anyway
            }

            if ($google_allowed == $other_allowed) {
                $this->assertTrue(!$google_allowed, 'Site not blocked by [tt]robots.txt[/tt]; test sites should be blocked so they do not get indexed.');
            } else {
                $this->assertTrue(!$google_allowed, 'Site not blocked on Google by [tt]robots.txt[/tt] as per Google\'s way of implementing robots standard; test sites should be blocked so they do not get indexed.');
                $this->assertTrue(!$other_allowed, 'Site not blocked on Google by [tt]robots.txt[/tt] as per standard (non-Google) way of implementing robots standard; test sites should be blocked so they do not get indexed.');
            }
        } else {
            if ($google_allowed == $other_allowed) {
                $this->assertTrue($google_allowed, 'Live site blocked by [tt]robots.txt[/tt]');
            } else {
                $this->assertTrue($google_allowed, 'Live site blocked on Google by [tt]robots.txt[/tt] as per Google\'s way of implementing robots standard');
                $this->assertTrue($other_allowed, 'Live site blocked on Google by [tt]robots.txt[/tt] as per standard (non-Google) way of implementing robots standard');
            }
        }

        /*
        This shows how the inconsistency works...

        Standard block:
        User-Agent: *
        Disallow: /
        Allow: /cms
        (Disallow takes precedence due to order of rules)

        Google block:
        User-Agent: *
        Allow: /
        Disallow: /cms
        (Disallow takes precedence due to specificity)

        Consistent block:
        User-Agent: *
        Disallow: /cms
        Allow: /
        (Disallow takes precedence both due to order of rules and specificity)
        */
    }

    /**
     * Run a section of health checks.
     *
     * @param  integer $check_context The current state of the website (a CHECK_CONTEXT__* constant)
     * @param  boolean $manual_checks Mention manual checks
     * @param  boolean $automatic_repair Do automatic repairs where possible
     * @param  ?boolean $use_test_data_for_pass Should test data be for a pass [if test data supported] (null: no test data)
     * @param  ?array $urls_or_page_links List of URLs and/or page-links to operate on, if applicable (null: those configured)
     * @param  ?array $comcode_segments Map of field names to Comcode segments to operate on, if applicable (null: N/A)
     */
    public function testRobotsTxtCompleteness(int $check_context, bool $manual_checks = false, bool $automatic_repair = false, ?bool $use_test_data_for_pass = null, ?array $urls_or_page_links = null, ?array $comcode_segments = null)
    {
        if ($check_context == CHECK_CONTEXT__INSTALL) {
            $this->log('Skipped; we are running from installer.');
            return;
        }
        if ($check_context == CHECK_CONTEXT__SPECIFIC_PAGE_LINKS) {
            $this->log('Skipped; running on specific page links.');
            return;
        }

        $rules = $this->robots_parse(null, true);
        if ($rules === null) {
            $this->stateCheckSkipped('No robots.txt file');
            return;
        }

        $scripts = [ // Really bad if these get indexed on Google
            'adminzone/',
        ];
        foreach ($scripts as $script) {
            $url = get_base_url() . '/' . $script;
            $allowed = $this->robots_allowed($url, 'Googlebot', true);
            $this->assertTrue(!$allowed, 'robots.txt should be blocking [tt]' . $script . '[/tt] because this is a high-security area.');
        }
    }

    /**
     * Run a section of health checks.
     *
     * @param  integer $check_context The current state of the website (a CHECK_CONTEXT__* constant)
     * @param  boolean $manual_checks Mention manual checks
     * @param  boolean $automatic_repair Do automatic repairs where possible
     * @param  ?boolean $use_test_data_for_pass Should test data be for a pass [if test data supported] (null: no test data)
     * @param  ?array $urls_or_page_links List of URLs and/or page-links to operate on, if applicable (null: those configured)
     * @param  ?array $comcode_segments Map of field names to Comcode segments to operate on, if applicable (null: N/A)
     */
    public function testRobotsTxtSitemapLinkage(int $check_context, bool $manual_checks = false, bool $automatic_repair = false, ?bool $use_test_data_for_pass = null, ?array $urls_or_page_links = null, ?array $comcode_segments = null)
    {
        if ($check_context == CHECK_CONTEXT__INSTALL) {
            $this->log('Skipped; we are running from installer.');
            return;
        }
        if ($check_context == CHECK_CONTEXT__SPECIFIC_PAGE_LINKS) {
            $this->log('Skipped; running on specific page links.');
            return;
        }

        $rules = $this->robots_parse(null);

        if ($rules === null) {
            $this->stateCheckSkipped('No robots.txt file');
            return;
        }

        $found = [];
        foreach ($rules as $_rule) {
            list($key, $rule) = $_rule;

            if ($key == 'sitemap') {
                $found[] = $rule;
            }
        }

        $expected_sitemap_url = get_base_url() . '/data_custom/sitemaps/index.xml';

        $ok = false;
        $sitemap_rule = null;
        foreach ($found as $i => $rule) {
            $this->assertTrue(strpos($rule, '://') !== false, 'Sitemap URL is relative, should be absolute');
            $ok_here = ($rule == $expected_sitemap_url);

            if ($ok_here) {
                $sitemap_rule = $rule;
            }

            $ok = $ok || $ok_here;
        }
        if ($sitemap_rule !== null) {
            $this->assertTrue($ok, 'Sitemap URL is ' . $sitemap_rule . ' but we expected ' . $expected_sitemap_url);
        }

        if ($check_context == CHECK_CONTEXT__TEST_SITE) {
            $this->assertTrue(empty($found), 'Sitemap directive found in robots.txt but this is a test site and we should not have one');
        }

        if ($check_context == CHECK_CONTEXT__LIVE_SITE) {
            $this->assertTrue(!empty($found), 'No Sitemap directive found in robots.txt; you should set one to ' . $expected_sitemap_url);
        }
    }

    /**
     * Find whether the robots.txt allows a URL to be indexed.
     *
     * @param  URLPATH $url The URL
     * @param  string $user_agent The user-agent
     * @param  boolean $google_style Whether to evaluate robots.txt like Google would (Google is slightly non-standard)
     * @return boolean Whether robots access is allowed
     */
    protected function robots_allowed(string $url, string $user_agent, bool $google_style) : bool
    {
        $rules = $this->robots_parse($user_agent);

        if ($rules === null) {
            return true;
        }

        $url_path = cms_parse_url_safe($url, PHP_URL_PATH);

        $best_precedence = 0;
        $allowed = true;
        foreach ($rules as $_rule) {
            list($key, $rule) = $_rule;

            switch ($key) {
                case 'allow':
                case 'disallow':
                    if ($rule == '') {
                        break; // Ignored rule
                    }

                    if (preg_match('#^' . $rule . '#', $url_path) != 0) {
                        if ($google_style) {
                            if (strlen($rule) > $best_precedence) {
                                $allowed = ($key == 'allow');
                                $best_precedence = strlen($rule);
                            }
                        } else {
                            return ($key == 'allow');
                        }
                    }

                    break;
            }
        }
        return $allowed;
    }

    /**
     * Parse our domain's robots.txt.
     *
     * @param  ?string $user_agent The user-agent (null: N/A)
     * @param  boolean $error_messages Show error messages for any parsing issues
     * @return ?array List of rules (null: could not parse)
     */
    protected function robots_parse(?string $user_agent, bool $error_messages = false) : ?array
    {
        // The best specification is by Google now:
        //  https://developers.google.com/search/reference/robots_txt

        $base_url = get_base_url();
        $base_url_path = cms_parse_url_safe($base_url, PHP_URL_PATH);
        $robots_url = preg_replace('#' . preg_quote($base_url_path, '#') . '$#', '', $base_url) . '/robots.txt';

        $agents_regexp = preg_quote('*');
        if ($user_agent !== null) {
            $agents_regexp .= '|' . preg_quote($user_agent, '#');
        }

        $contents = http_get_contents($robots_url, ['trigger_error' => false]);
        if ($contents === null) {
            return null;
        }
        $robots_lines = explode("\n", $contents);

        // Go through lines
        $rules = [];
        $following_rules_apply = false;
        $best_following_rules_apply = 0;
        $just_did_ua_line = false;
        $did_some_ua_line = false;
        foreach ($robots_lines as $line) {
            $line = trim($line);

            // Skip blank lines
            if ($line == '') {
                continue;
            }

            // Skip comment lines
            if ($line[0] == '#') {
                continue;
            }

            // The following rules only apply if the User-Agent matches
            $matches = [];
            if (preg_match('#^User-Agent:(.*)#i', $line, $matches) != 0) {
                $agent_spec = $matches[1];
                $_following_rules_apply = (preg_match('#(' . $agents_regexp . ')#i', $agent_spec) != 0); // It's a bit weird how "googlebot-xxx" would match but "google" would not, but that's the standard (and there's justification when you think about it)
                if ($_following_rules_apply) {
                    if (strlen($agent_spec) >= $best_following_rules_apply) {
                        $following_rules_apply = true;
                        $best_following_rules_apply = strlen($agent_spec);
                        $rules = []; // Reset rules, as now this is the best scoring rules section (we don't merge sections)
                    }
                } elseif (!$just_did_ua_line) {
                    $following_rules_apply = false;
                }

                $just_did_ua_line = true;
                $did_some_ua_line = true;

                continue;
            }

            // Record rules
            if (preg_match('#^([\w-]+):\s*(.*)\s*$#i', $line, $matches) != 0) {
                $key = cms_strtolower_ascii($matches[1]);
                $value = trim($matches[2]);

                $core_rule = ($key == 'allow') || ($key == 'disallow');

                if ($error_messages) {
                    $this->assertTrue(in_array($key, ['allow', 'disallow', 'sitemap', 'crawl-delay']), 'Unrecognised [tt]robots.txt[/tt] rule:' . $key);

                    if ($core_rule) {
                        $this->assertTrue($did_some_ua_line, 'Floating [tt]' . cms_ucwords_ascii($key) . '[/tt] outside of any User-Agent section of [tt]robots.txt[/tt]');
                    }
                }

                if ($following_rules_apply) {
                    // Add rules that apply to array for testing
                    if ($core_rule) {
                        $rule = addcslashes($value, '#\+?^[](){}|-'); // Escape things that are in regexps but should be literal here
                        $rule = str_replace('*', '.*', $rule); // * wild cards are ".*" in a regexp
                        // "$" remains unchanged

                        $rules[] = [$key, $rule];
                    } else {
                        $rules[] = [$key, $value];
                    }
                }

                $just_did_ua_line = false;

                continue;
            }

            // Unrecognised line
            if ($error_messages) {
                $this->assertTrue(false, 'Unrecognised line in [tt]robots.txt[/tt]:' . $line);
            }
        }

        return $rules;
    }
}
