<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    health_check
 */

/**
 * Hook class.
 */
class Hook_health_check_performance_bloat extends Hook_Health_Check
{
    protected $category_label = 'Bloated data';

    /**
     * Standard hook run function to run this category of health checks.
     *
     * @param  ?array $sections_to_run Which check sections to run (null: all)
     * @param  integer $check_context The current state of the website (a CHECK_CONTEXT__* constant)
     * @param  boolean $manual_checks Mention manual checks
     * @param  boolean $automatic_repair Do automatic repairs where possible
     * @param  ?boolean $use_test_data_for_pass Should test data be for a pass [if test data supported] (null: no test data)
     * @param  ?array $urls_or_page_links List of URLs and/or page-links to operate on, if applicable (null: those configured)
     * @param  ?array $comcode_segments Map of field names to Comcode segments to operate on, if applicable (null: N/A)
     * @param  boolean $show_unusable_categories Whether to include categories that might not be accessible for some reason
     * @return array A pair: category label, list of results
     */
    public function run(?array $sections_to_run, int $check_context, bool $manual_checks = false, bool $automatic_repair = false, ?bool $use_test_data_for_pass = null, ?array $urls_or_page_links = null, ?array $comcode_segments = null, bool $show_unusable_categories = false) : array
    {
        $this->process_checks_section('testTableSize', 'Table size (row count)', $sections_to_run, $check_context, $manual_checks, $automatic_repair, $use_test_data_for_pass, $urls_or_page_links, $comcode_segments);
        $this->process_checks_section('testDatabaseSize', 'Database size', $sections_to_run, $check_context, $manual_checks, $automatic_repair, $use_test_data_for_pass, $urls_or_page_links, $comcode_segments);
        $this->process_checks_section('testDirectorySize', 'Directory size', $sections_to_run, $check_context, $manual_checks, $automatic_repair, $use_test_data_for_pass, $urls_or_page_links, $comcode_segments);
        $this->process_checks_section('testLogSize', 'Log size', $sections_to_run, $check_context, $manual_checks, $automatic_repair, $use_test_data_for_pass, $urls_or_page_links, $comcode_segments);

        return [$this->category_label, $this->results];
    }

    /**
     * Run a section of health checks.
     *
     * @param  integer $check_context The current state of the website (a CHECK_CONTEXT__* constant)
     * @param  boolean $manual_checks Mention manual checks
     * @param  boolean $automatic_repair Do automatic repairs where possible
     * @param  ?boolean $use_test_data_for_pass Should test data be for a pass [if test data supported] (null: no test data)
     * @param  ?array $urls_or_page_links List of URLs and/or page-links to operate on, if applicable (null: those configured)
     * @param  ?array $comcode_segments Map of field names to Comcode segments to operate on, if applicable (null: N/A)
     */
    public function testTableSize(int $check_context, bool $manual_checks = false, bool $automatic_repair = false, ?bool $use_test_data_for_pass = null, ?array $urls_or_page_links = null, ?array $comcode_segments = null)
    {
        if ($check_context == CHECK_CONTEXT__INSTALL) {
            $this->log('Skipped; we are running from installer.');
            return;
        }
        if ($check_context == CHECK_CONTEXT__SPECIFIC_PAGE_LINKS) {
            $this->log('Skipped; running on specific page links.');
            return;
        }

        $tables = [ // Should generally be consistent with database_relations.php TABLE_PURPOSE__NO_BACKUPS
            'autosave' => 100000,
            'cache' => 1000000,
            'cache_on' => 10000,
            'cached_comcode_pages' => 10000,
            'captchas' => 10000,
            'catalogue_cat_treecache' => 1000000,
            'catalogue_childcountcache' => 100000,
            'chat_active' => 100000,
            'chat_events' => 10000000,
            'cron_caching_requests' => 10000,
            'post_tokens' => 1000000,
            'digestives_tin' => 1000000,
            'incoming_uploads' => 10000,
            'stats' => 10000000,
            'stats_events' => 10000000,
            'stats_link_tracker' => 10000000,
            'edit_pings' => 10000,
            'hackattack' => 1000000,
            'logged_mail_messages' => 100000,
            'messages_to_render' => 100000,
            'sessions' => 1000000,
            'sitemap_cache' => 100000,
            'temp_block_permissions' => 10000000,
            'url_title_cache' => 100000,
            'urls_checked' => 100000,
        ];

        foreach ($tables as $table => $max_threshold) {
            if ($GLOBALS['SITE_DB']->table_exists($table)) {
                $cnt = $GLOBALS['SITE_DB']->query_select_value($table, 'COUNT(*)');
                $this->assertTrue($cnt < $max_threshold, 'Volatile-defined table [tt]' . $table . '[/tt] is very large @ ' . integer_format($cnt) . ' records');
            }
        }
    }

    /**
     * Run a section of health checks.
     *
     * @param  integer $check_context The current state of the website (a CHECK_CONTEXT__* constant)
     * @param  boolean $manual_checks Mention manual checks
     * @param  boolean $automatic_repair Do automatic repairs where possible
     * @param  ?boolean $use_test_data_for_pass Should test data be for a pass [if test data supported] (null: no test data)
     * @param  ?array $urls_or_page_links List of URLs and/or page-links to operate on, if applicable (null: those configured)
     * @param  ?array $comcode_segments Map of field names to Comcode segments to operate on, if applicable (null: N/A)
     */
    public function testDatabaseSize(int $check_context, bool $manual_checks = false, bool $automatic_repair = false, ?bool $use_test_data_for_pass = null, ?array $urls_or_page_links = null, ?array $comcode_segments = null)
    {
        if ($check_context == CHECK_CONTEXT__INSTALL) {
            $this->log('Skipped; we are running from installer.');
            return;
        }
        if ($check_context == CHECK_CONTEXT__SPECIFIC_PAGE_LINKS) {
            $this->log('Skipped; running on specific page links.');
            return;
        }

        if (strpos(get_db_type(), 'mysql') === false) {
            $this->stateCheckSkipped('Can only check when running MySQL (or MariaDB)');
            return;
        }

        $query = 'SELECT SUM(data_length) AS size_data_length,SUM(index_length) AS size_index_length FROM information_schema.TABLES WHERE ' . db_string_equal_to('table_schema', get_db_site()) . ' AND table_name LIKE \'' . db_encode_like(get_table_prefix() . '%') . '\'';
        $db_size = $GLOBALS['SITE_DB']->query($query, null, 0, true);
        if (empty($db_size)) {
            $this->stateCheckSkipped('Failed to check database size, possible lack of permissions');
        } else {
            $db_size = @intval($db_size['size_data_length']) + @intval($db_size['size_index_length']);
            require_code('files');
            $max_threshold = intval(get_option('hc_database_threshold'));
            $this->assertTrue($db_size < $max_threshold * 1024 * 1024, 'Database is very large @ ' . clean_file_size($db_size));
        }
    }

    /**
     * Run a section of health checks.
     *
     * @param  integer $check_context The current state of the website (a CHECK_CONTEXT__* constant)
     * @param  boolean $manual_checks Mention manual checks
     * @param  boolean $automatic_repair Do automatic repairs where possible
     * @param  ?boolean $use_test_data_for_pass Should test data be for a pass [if test data supported] (null: no test data)
     * @param  ?array $urls_or_page_links List of URLs and/or page-links to operate on, if applicable (null: those configured)
     * @param  ?array $comcode_segments Map of field names to Comcode segments to operate on, if applicable (null: N/A)
     */
    public function testDirectorySize(int $check_context, bool $manual_checks = false, bool $automatic_repair = false, ?bool $use_test_data_for_pass = null, ?array $urls_or_page_links = null, ?array $comcode_segments = null)
    {
        if ($check_context == CHECK_CONTEXT__INSTALL) {
            $this->log('Skipped; we are running from installer.');
            return;
        }
        if ($check_context == CHECK_CONTEXT__SPECIFIC_PAGE_LINKS) {
            $this->log('Skipped; running on specific page links.');
            return;
        }

        require_code('files2');

        $mb = 1024 * 1024;
        $directories = [
            'caches/static' => 2048,
            'caches/lang' => 128,
            'caches/persistent' => 256,
            'caches/self_learning' => 256,
            'uploads/incoming' => 1024,
            'uploads/captcha' => 1024,
            'temp' => 512,
            'data_custom/errors' => 256,
        ];
        foreach (array_keys(find_all_langs()) as $lang) {
            $directories['themes/' . $GLOBALS['FORUM_DRIVER']->get_theme('') . '/templates_cached/' . $lang] = 48;
        }
        foreach ($directories as $dir => $max_threshold_size_in_mb) {
            if (file_exists(get_file_base() . '/' . $dir)) {
                $size = get_directory_size(get_file_base() . '/' . $dir);
                $this->assertTrue($size < $mb * $max_threshold_size_in_mb, 'Directory [tt]' . $dir . '[/tt] is very large @ ' . clean_file_size($size));
            }
        }

        $directories = [
            'uploads/incoming' => 100,
            'temp' => 100,
            'data_custom' => 100,
            'data_custom/errors' => 100,
        ];
        foreach ($directories as $dir => $max_contents_threshold) {
            if (file_exists(get_file_base() . '/' . $dir)) {
                $count = count(get_directory_contents(get_file_base() . '/' . $dir, '', IGNORE_ACCESS_CONTROLLERS, false));
                $this->assertTrue($count < $max_contents_threshold, 'Directory [tt]' . $dir . '[/tt] now contains ' . integer_format($count) . ' files, should be kept under ' . integer_format($max_contents_threshold));
            }
        }
    }

    /**
     * Run a section of health checks.
     *
     * @param  integer $check_context The current state of the website (a CHECK_CONTEXT__* constant)
     * @param  boolean $manual_checks Mention manual checks
     * @param  boolean $automatic_repair Do automatic repairs where possible
     * @param  ?boolean $use_test_data_for_pass Should test data be for a pass [if test data supported] (null: no test data)
     * @param  ?array $urls_or_page_links List of URLs and/or page-links to operate on, if applicable (null: those configured)
     * @param  ?array $comcode_segments Map of field names to Comcode segments to operate on, if applicable (null: N/A)
     */
    public function testLogSize(int $check_context, bool $manual_checks = false, bool $automatic_repair = false, ?bool $use_test_data_for_pass = null, ?array $urls_or_page_links = null, ?array $comcode_segments = null)
    {
        if ($check_context == CHECK_CONTEXT__INSTALL) {
            $this->log('Skipped; we are running from installer.');
            return;
        }
        if ($check_context == CHECK_CONTEXT__SPECIFIC_PAGE_LINKS) {
            $this->log('Skipped; running on specific page links.');
            return;
        }

        require_code('files');

        $log_threshold = 1024 * 256; // 256 kb

        $log_files = [];

        $log_files[] = 'error_log'; // PHP may auto-create these at startup

        $path = get_file_base();
        foreach (array_merge(find_all_zones(), ['data_custom', 'data']) as $dir) {
            $dh = @opendir($path . '/' . $dir);
            if ($dh !== false) {
                while (($f = readdir($dh)) !== false) {
                    if ((substr($f, -4) == '.log') || (substr($f, -7) == 'log.php') || (substr($f, -7) == 'log.txt')) {
                        $log_files[] = $dir . '/' . $f;
                    }
                }
                closedir($dh);
            }
        }

        foreach ($log_files as $f) {
            $_path = $path . '/' . $f;

            if (!is_file($_path)) {
                continue;
            }

            $size = filesize($_path);
            $this->assertTrue($size < $log_threshold, 'Size of [tt]' . $f . '[/tt] log is very large @ ' . clean_file_size($size));
        }
    }
}
