<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    health_check
 */

/**
 * Hook class.
 */
class Hook_health_check_security_hackattack extends Hook_Health_Check
{
    protected $category_label = 'Hack-attacks';

    /**
     * Standard hook run function to run this category of health checks.
     *
     * @param  ?array $sections_to_run Which check sections to run (null: all)
     * @param  integer $check_context The current state of the website (a CHECK_CONTEXT__* constant)
     * @param  boolean $manual_checks Mention manual checks
     * @param  boolean $automatic_repair Do automatic repairs where possible
     * @param  ?boolean $use_test_data_for_pass Should test data be for a pass [if test data supported] (null: no test data)
     * @param  ?array $urls_or_page_links List of URLs and/or page-links to operate on, if applicable (null: those configured)
     * @param  ?array $comcode_segments Map of field names to Comcode segments to operate on, if applicable (null: N/A)
     * @param  boolean $show_unusable_categories Whether to include categories that might not be accessible for some reason
     * @return array A pair: category label, list of results
     */
    public function run(?array $sections_to_run, int $check_context, bool $manual_checks = false, bool $automatic_repair = false, ?bool $use_test_data_for_pass = null, ?array $urls_or_page_links = null, ?array $comcode_segments = null, bool $show_unusable_categories = false) : array
    {
        $this->process_checks_section('testOverseasAccess', 'Overseas access', $sections_to_run, $check_context, $manual_checks, $automatic_repair, $use_test_data_for_pass, $urls_or_page_links, $comcode_segments);
        $this->process_checks_section('testHackAttacks', 'Attack frequency', $sections_to_run, $check_context, $manual_checks, $automatic_repair, $use_test_data_for_pass, $urls_or_page_links, $comcode_segments);
        $this->process_checks_section('testFailedLogins', 'Failed logins', $sections_to_run, $check_context, $manual_checks, $automatic_repair, $use_test_data_for_pass, $urls_or_page_links, $comcode_segments);
        $this->process_checks_section('testRateLimitSpike', 'Rate-limit spiking', $sections_to_run, $check_context, $manual_checks, $automatic_repair, $use_test_data_for_pass, $urls_or_page_links, $comcode_segments);

        return [$this->category_label, $this->results];
    }

    /**
     * Run a section of health checks.
     *
     * @param  integer $check_context The current state of the website (a CHECK_CONTEXT__* constant)
     * @param  boolean $manual_checks Mention manual checks
     * @param  boolean $automatic_repair Do automatic repairs where possible
     * @param  ?boolean $use_test_data_for_pass Should test data be for a pass [if test data supported] (null: no test data)
     * @param  ?array $urls_or_page_links List of URLs and/or page-links to operate on, if applicable (null: those configured)
     * @param  ?array $comcode_segments Map of field names to Comcode segments to operate on, if applicable (null: N/A)
     */
    public function testOverseasAccess(int $check_context, bool $manual_checks = false, bool $automatic_repair = false, ?bool $use_test_data_for_pass = null, ?array $urls_or_page_links = null, ?array $comcode_segments = null)
    {
        if ($check_context == CHECK_CONTEXT__INSTALL) {
            $this->log('Skipped; we are running from installer.');
            return;
        }
        if ($check_context == CHECK_CONTEXT__SPECIFIC_PAGE_LINKS) {
            $this->log('Skipped; running on specific page links.');
            return;
        }

        if (addon_installed('stats')) {
            if (get_db_type() == 'xml') { // Too much data
                $ok = false;
            } else {
                $test = $GLOBALS['SITE_DB']->query_select_value_if_there('ip_country', 'id');
                $ok = ($test !== null);
            }

            if ($ok) {
                require_code('locations');

                $admin_groups = $GLOBALS['FORUM_DRIVER']->get_super_admin_groups();
                $members = $GLOBALS['FORUM_DRIVER']->member_group_query($admin_groups);
                foreach ($members as $member) {
                    $id = $GLOBALS['FORUM_DRIVER']->mrow_member_id($member);
                    $username = $GLOBALS['FORUM_DRIVER']->mrow_username($member);

                    $countries = [];
                    $rows = $GLOBALS['SITE_DB']->query_select('stats', ['DISTINCT ip'], ['member_id' => $id], 'AND date_and_time>' . strval(time() - 60 * 60 * 24 * 7));
                    foreach ($rows as $row) {
                        $country = geolocate_ip($row['ip']);
                        if ($country !== null) {
                            $countries[] = '[tt]' . $country . '[/tt]';
                        }
                    }

                    $this->assertTrue(count(array_unique($countries)) <= 1, 'Admin account "' . $username . '" appears to have logged in from multiple countries (' . implode(', ', $countries) . '). The account may have been compromised.');
                }
            } else {
                $this->stateCheckSkipped('Geolocation data not installed so cannot do admin country checks');
            }
        } else {
            $this->stateCheckSkipped('Could not find geolocation history without the stats addon being installed');
        }
    }

    /**
     * Run a section of health checks.
     *
     * @param  integer $check_context The current state of the website (a CHECK_CONTEXT__* constant)
     * @param  boolean $manual_checks Mention manual checks
     * @param  boolean $automatic_repair Do automatic repairs where possible
     * @param  ?boolean $use_test_data_for_pass Should test data be for a pass [if test data supported] (null: no test data)
     * @param  ?array $urls_or_page_links List of URLs and/or page-links to operate on, if applicable (null: those configured)
     * @param  ?array $comcode_segments Map of field names to Comcode segments to operate on, if applicable (null: N/A)
     */
    public function testHackAttacks(int $check_context, bool $manual_checks = false, bool $automatic_repair = false, ?bool $use_test_data_for_pass = null, ?array $urls_or_page_links = null, ?array $comcode_segments = null)
    {
        if ($check_context == CHECK_CONTEXT__INSTALL) {
            $this->log('Skipped; we are running from installer.');
            return;
        }
        if ($check_context == CHECK_CONTEXT__SPECIFIC_PAGE_LINKS) {
            $this->log('Skipped; running on specific page links.');
            return;
        }

        // TODO: configurable threshold
        $threshold = 50;

        $sql = 'SELECT COUNT(*) FROM ' . get_table_prefix() . 'hackattack WHERE date_and_time>' . strval(time() - 60 * 60 * 24);
        $num_failed = $GLOBALS['SITE_DB']->query_value_if_there($sql);
        $this->assertTrue($num_failed < $threshold, 'High number of hack-attacks triggered in the last 24 hours, at ' . integer_format($num_failed, 0));
    }

    /**
     * Run a section of health checks.
     *
     * @param  integer $check_context The current state of the website (a CHECK_CONTEXT__* constant)
     * @param  boolean $manual_checks Mention manual checks
     * @param  boolean $automatic_repair Do automatic repairs where possible
     * @param  ?boolean $use_test_data_for_pass Should test data be for a pass [if test data supported] (null: no test data)
     * @param  ?array $urls_or_page_links List of URLs and/or page-links to operate on, if applicable (null: those configured)
     * @param  ?array $comcode_segments Map of field names to Comcode segments to operate on, if applicable (null: N/A)
     */
    public function testFailedLogins(int $check_context, bool $manual_checks = false, bool $automatic_repair = false, ?bool $use_test_data_for_pass = null, ?array $urls_or_page_links = null, ?array $comcode_segments = null)
    {
        if ($check_context == CHECK_CONTEXT__INSTALL) {
            $this->log('Skipped; we are running from installer.');
            return;
        }
        if ($check_context == CHECK_CONTEXT__SPECIFIC_PAGE_LINKS) {
            $this->log('Skipped; running on specific page links.');
            return;
        }

        // TODO: configurable threshold
        $threshold = 50 + intval($GLOBALS['FORUM_DRIVER']->get_num_members() / 25); // There will be a slight increase in frequency with more members present

        $sql = 'SELECT COUNT(*) FROM ' . get_table_prefix() . 'failedlogins WHERE date_and_time>' . strval(time() - 60 * 60 * 24);
        $num_failed = $GLOBALS['SITE_DB']->query_value_if_there($sql);
        $this->assertTrue($num_failed < $threshold, 'High number of failed login attempts in the last 24 hours, at ' . integer_format($num_failed, 0));
    }

    /**
     * Run a section of health checks.
     *
     * @param  integer $check_context The current state of the website (a CHECK_CONTEXT__* constant)
     * @param  boolean $manual_checks Mention manual checks
     * @param  boolean $automatic_repair Do automatic repairs where possible
     * @param  ?boolean $use_test_data_for_pass Should test data be for a pass [if test data supported] (null: no test data)
     * @param  ?array $urls_or_page_links List of URLs and/or page-links to operate on, if applicable (null: those configured)
     * @param  ?array $comcode_segments Map of field names to Comcode segments to operate on, if applicable (null: N/A)
     */
    public function testRateLimitSpike(int $check_context, bool $manual_checks = false, bool $automatic_repair = false, ?bool $use_test_data_for_pass = null, ?array $urls_or_page_links = null, ?array $comcode_segments = null)
    {
        if ($check_context != CHECK_CONTEXT__LIVE_SITE) {
            $this->log('Skipped; we are not running from a live site.');
            return;
        }

        global $RATE_LIMITING_DATA;
        $RATE_LIMITING_DATA = [];

        $rate_limiter_path = get_custom_file_base() . '/data_custom/rate_limiter.php';
        if (is_file($rate_limiter_path)) {
            $fp = fopen($rate_limiter_path, 'rb');
            flock($fp, LOCK_SH);
            include($rate_limiter_path);
            flock($fp, LOCK_UN);
            fclose($fp);
        }

        $threshold_sample = intval(get_option('hc_requests_window_size'));
        $threshold_rps = floatval(get_option('hc_requests_per_second_threshold'));

        $threshold_sample_compound = intval(get_option('hc_compound_requests_window_size'));
        $threshold_rps_compound = floatval(get_option('hc_compound_requests_per_second_threshold'));

        /*  Test
        $RATE_LIMITING_DATA = [
            '1.2.3.4' => array_fill(0, 30, time()),
        ];
        */

        if (!empty($RATE_LIMITING_DATA)) {
            global $SITE_INFO;
            $rate_limit_time_window = empty($SITE_INFO['rate_limit_time_window']) ? 10 : intval($SITE_INFO['rate_limit_time_window']);

            $times_compound = [];

            foreach ($RATE_LIMITING_DATA as $ip => $times) {
                $requests_per_second = floatval(count($times)) / floatval($rate_limit_time_window);
                $ok = (count($times) < $threshold_sample) || ($requests_per_second < $threshold_rps);
                $this->assertTrue($ok, 'Heavy visitor load @ ' . float_format($requests_per_second, 2, true) . ' PHP requests per second (for a sample size over ' . integer_format($threshold_sample) . ') requests from IP ' . $ip);

                $times_compound = array_merge($times_compound, $times);
            }

            $requests_per_second = floatval(count($times_compound)) / floatval($rate_limit_time_window);
            $ok = (count($times_compound) < $threshold_sample) || ($requests_per_second < $threshold_rps);
            $this->assertTrue($ok, 'Heavy visitor load @ ' . float_format($requests_per_second, 2, true) . ' PHP requests per second (for a sample size over ' . integer_format($threshold_sample_compound) . ') requests from all IPs together');
        }
    }
}
