<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    cns_forum
 */

/**
 * Hook class.
 */
class Hook_notification_cns_pts extends Hook_Notification
{
    /**
     * Find the initial setting that members have for a notification code (only applies to the member_could_potentially_enable members).
     *
     * @param  ID_TEXT $notification_code Notification code
     * @param  ?SHORT_TEXT $category The category within the notification code (null: none)
     * @param  MEMBER $member_id The member the notification would be for
     * @return integer Initial setting
     */
    public function get_initial_setting(string $notification_code, ?string $category, int $member_id) : int
    {
        return A__STATISTICAL;
    }

    /**
     * Find a bitmask of settings (e-mail, SMS, etc) a notification code supports for listening on.
     *
     * @param  ID_TEXT $notification_code Notification code
     * @return integer Allowed settings
     */
    public function allowed_settings(string $notification_code) : int
    {
        $ret = A__ALL & ~A_INSTANT_PT; // Topic, so do not allow PT
        if ($notification_code == 'cns_new_pt') {
            if (get_option('pt_notifications_as_web') == '0') {
                $ret = $ret & ~A_WEB_NOTIFICATION;
            }
        }
        return $ret;
    }

    /**
     * Get a list of all the notification codes this hook can handle.
     * (Addons can define hooks that handle whole sets of codes, so hooks are written so they can take wide authority).
     *
     * @return array List of codes (mapping between code names, and a pair: section and labelling for those codes)
     */
    public function list_handled_codes() : array
    {
        if (get_forum_type() != 'cns') {
            return [];
        }

        if (!addon_installed('cns_forum')) {
            return [];
        }

        $list = [];
        $list['cns_new_pt'] = [do_lang('MESSAGES'), do_lang('cns:NOTIFICATION_TYPE_cns_new_pt')];
        $list['cns_topic_invite'] = [do_lang('MESSAGES'), do_lang('cns:NOTIFICATION_TYPE_cns_topic_invite')];
        return $list;
    }

    /**
     * Get a list of members who have enabled this notification (i.e. have permission to AND have chosen to or are defaulted to).
     *
     * @param  ID_TEXT $notification_code Notification code
     * @param  ?SHORT_TEXT $category The category within the notification code (null: none)
     * @param  ?array $to_member_ids List of member IDs we are restricting to (null: no restriction). This effectively works as a intersection set operator against those who have enabled.
     * @param  ?integer $from_member_id The member ID doing the sending. Either a MEMBER or a negative number (e.g. A_FROM_SYSTEM_UNPRIVILEGED) (null: current member)
     * @param  integer $start Start position (for pagination)
     * @param  integer $max Maximum (for pagination)
     * @return array A pair: Map of members to their notification setting, and whether there may be more
     */
    public function list_members_who_have_enabled(string $notification_code, ?string $category = null, ?array $to_member_ids = null, ?int $from_member_id = null, int $start = 0, int $max = 300) : array
    {
        $members = [];
        if ($notification_code == 'cns_new_pt') {
            $members = $this->_all_members_who_have_enabled($notification_code, $category, $to_member_ids, $start, $max);
            if (empty($to_member_ids)) { // Only if we're not actually sending a PT now (notification should happen if sending a PT to someone who can't send them, with default settings)
                $members = $this->_all_members_who_have_enabled_with_privilege($members, 'use_pt', $notification_code, $category, $to_member_ids, $start, $max);
            }
            return $members;
        }
        $members = $this->_all_members_who_have_enabled_with_zone_access($members, 'forum', $notification_code, $category, $to_member_ids, $start, $max);
        if ($notification_code == 'cns_topic_invite') {
            $members = $this->_all_members_who_have_enabled($notification_code, $category, $to_member_ids, $start, $max);
            $members = $this->_all_members_who_have_enabled_with_privilege($members, 'use_pt', $notification_code, $category, $to_member_ids, $start, $max);
            $members = $this->_all_members_who_have_enabled_with_zone_access($members, 'forum', $notification_code, $category, $to_member_ids, $start, $max);
            return $members;
        }
        return parent::list_members_who_have_enabled($notification_code, $category, $to_member_ids, $from_member_id, $start, $max);
    }
}
