<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    points
 */

/**
 * Hook class.
 */
class Hook_notification_points extends Hook_Notification
{
    /**
     * Find the initial setting that members have for a notification code (only applies to the member_could_potentially_enable members).
     *
     * @param  ID_TEXT $notification_code Notification code
     * @param  ?SHORT_TEXT $category The category within the notification code (null: none)
     * @param  MEMBER $member_id The member the notification would be for
     * @return integer Initial setting
     */
    public function get_initial_setting(string $notification_code, ?string $category, int $member_id) : int
    {
        if ($notification_code == 'points_transaction') {
            return A_DAILY_EMAIL_DIGEST | A_WEB_NOTIFICATION; // Could have a lot of them, so use an e-mail digest to avoid annoying users
        }
        if ($notification_code == 'point_escrows') {
            return A__ALL & ~A_INSTANT_PT;
        }
        if ($notification_code == 'point_escrows_staff_active') {
            return A__ALL & ~A_INSTANT_PT;
        }
        return A_NA;
    }

    /**
     * Get a list of all the notification codes this hook can handle.
     * (Addons can define hooks that handle whole sets of codes, so hooks are written so they can take wide authority).
     *
     * @return array List of codes (mapping between code names, and a pair: section and labelling for those codes)
     */
    public function list_handled_codes() : array
    {
        if (!addon_installed('points')) {
            return [];
        }

        $list = [];
        $list['points_transaction'] = [do_lang('ACTIVITY'), do_lang('points:NOTIFICATION_TYPE_points_transaction')];
        $list['point_escrows'] = [do_lang('ACTIVITY'), do_lang('points:NOTIFICATION_TYPE_point_escrows')];
        $list['point_escrows_staff_active'] = [do_lang('ACTIVITY'), do_lang('points:NOTIFICATION_TYPE_point_escrows_staff_active')];
        $list['point_escrows_staff_passive'] = [do_lang('ACTIVITY'), do_lang('points:NOTIFICATION_TYPE_point_escrows_staff_passive')];
        return $list;
    }

    /**
     * Find whether a member could enable this notification (i.e. have permission to).
     *
     * @param  ID_TEXT $notification_code Notification code
     * @param  MEMBER $member_id Member to check against
     * @param  ?SHORT_TEXT $category The category within the notification code (null: none)
     * @return boolean Whether they could
     */
    public function member_could_potentially_enable(string $notification_code, int $member_id, ?string $category = null) : bool
    {
        if ($notification_code == 'point_escrows_staff_active' || $notification_code == 'point_escrows_staff_passive') {
            if (!addon_installed('points')) {
                return false;
            }

            return ((has_privilege($member_id, 'moderate_points_escrow')) && (has_actual_page_access($member_id, 'points')));
        }
        return parent::member_could_potentially_enable($notification_code, $member_id, $category);
    }

    /**
     * Get a list of members who have enabled this notification (i.e. have permission to AND have chosen to or are defaulted to).
     *
     * @param  ID_TEXT $notification_code Notification code
     * @param  ?SHORT_TEXT $category The category within the notification code (null: none)
     * @param  ?array $to_member_ids List of member IDs we are restricting to (null: no restriction). This effectively works as a intersection set operator against those who have enabled.
     * @param  ?integer $from_member_id The member ID doing the sending. Either a MEMBER or a negative number (e.g. A_FROM_SYSTEM_UNPRIVILEGED) (null: current member)
     * @param  integer $start Start position (for pagination)
     * @param  integer $max Maximum (for pagination)
     * @return array A pair: Map of members to their notification setting, and whether there may be more
     */
    public function list_members_who_have_enabled(string $notification_code, ?string $category = null, ?array $to_member_ids = null, ?int $from_member_id = null, int $start = 0, int $max = 300) : array
    {
        if ($notification_code == 'points_transaction' || $notification_code == 'point_escrows') {
            $members = $this->_all_members_who_have_enabled($notification_code, $category, $to_member_ids, $start, $max);
            $members = $this->_all_members_who_have_enabled_with_page_access($members, 'points', $notification_code, $category, $to_member_ids, $start, $max);
        }
        if ($notification_code == 'point_escrows_staff_active' || $notification_code == 'point_escrows_staff_passive') {
            $members = $this->_all_members_who_have_enabled($notification_code, $category, $to_member_ids, $start, $max);
            $members = $this->_all_members_who_have_enabled_with_privilege($members, 'moderate_points_escrow', $notification_code, $category, $to_member_ids, $start, $max);
            $members = $this->_all_members_who_have_enabled_with_page_access($members, 'points', $notification_code, $category, $to_member_ids, $start, $max);
        }

        return $members;
    }
}
