<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    points
 */

/**
 * Hook class.
 */
class Hook_rss_points
{
    /**
     * Check if the given member has access to view this feed.
     *
     * @param  MEMBER $member_id The member trying to access this feed
     * @return boolean Whether the member has access
     */
    public function has_access(int $member_id) : bool
    {
        if (!addon_installed('points')) {
            return false;
        }

        if (!has_actual_page_access($member_id, 'points')) {
            return false;
        }
        if (!has_privilege($member_id, 'view_points_ledger')) {
            return false;
        }

        return true;
    }

    /**
     * Run function for RSS hooks.
     *
     * @param  string $_filters A list of categories we accept from
     * @param  TIME $cutoff Cutoff time, before which we do not show results from
     * @param  string $prefix Prefix that represents the template set we use
     * @set RSS_ ATOM_
     * @param  string $date_string The standard format of date to use for the syndication type represented in the prefix
     * @param  integer $max The maximum number of entries to return, ordering by date
     * @return ?array A pair: The main syndication section, and a title (null: error)
     */
    public function run(string $_filters, int $cutoff, string $prefix, string $date_string, int $max) : ?array
    {
        if (!$this->has_access(get_member())) {
            return null;
        }

        $filters = selectcode_to_sqlfragment($_filters, 'receiving_member', 'f_members', null, 'receiving_member', 'id', true, true); // Note that the parameters are fiddled here so that category-set and record-set are the same, yet SQL is returned to deal in an entirely different record-set (entries' record-set)

        require_lang('points');
        require_code('points');

        $content = new Tempcode();
        $rows = $GLOBALS['SITE_DB']->query('SELECT * FROM ' . $GLOBALS['SITE_DB']->get_table_prefix() . 'points_ledger WHERE ' . $filters . ' AND date_and_time>' . strval($cutoff) . ' ORDER BY date_and_time DESC', $max);
        foreach ($rows as $row) {
            $id = strval($row['id']);

            $author = '';
            if (($row['anonymous'] == 0) || (has_privilege(get_member(), 'trace_anonymous_points_transactions'))) {
                $author = $GLOBALS['FORUM_DRIVER']->get_username($row['sending_member'], false, USERNAME_DEFAULT_BLANK);
            } else {
                $author = do_lang('ANON');
            }

            $news_date = date($date_string, $row['date_and_time']);
            $edit_date = '';

            $to = $GLOBALS['FORUM_DRIVER']->get_username($row['receiving_member']);
            $news_title = xmlentities(do_lang('POINTS_RSS_LINE', $to, integer_format($row['amount_gift_points'] + $row['amount_points'])));
            $summary = xmlentities(get_translated_text($row['reason']));
            $news = '';

            $category = '';
            $category_raw = '';

            $view_url = points_url($row['receiving_member'], true);

            $if_comments = new Tempcode();

            $content->attach(do_template($prefix . 'ENTRY', ['VIEW_URL' => $view_url, 'SUMMARY' => $summary, 'EDIT_DATE' => $edit_date, 'IF_COMMENTS' => $if_comments, 'TITLE' => $news_title, 'CATEGORY_RAW' => $category_raw, 'CATEGORY' => $category, 'AUTHOR' => $author, 'ID' => $id, 'NEWS' => $news, 'DATE' => $news_date], null, false, null, '.xml', 'xml'));
        }

        return [$content, do_lang('POINTS')];
    }
}
