<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    catalogues
 */

/**
 * Hook class.
 */
class Hook_sitemap_catalogue_category extends Hook_sitemap_content
{
    protected $content_type = 'catalogue_category';
    protected $screen_type = 'category';

    // If we have a different content type of entries, under this content type
    protected $entry_content_type = ['catalogue_entry'];
    protected $entry_sitetree_hook = ['catalogue_entry'];

    /**
     * Get the permission page that nodes matching $page_link in this hook are tied to.
     * The permission page is where privileges may be overridden against.
     *
     * @param  string $page_link The page-link
     * @return ?ID_TEXT The permission page (null: none)
     */
    public function get_privilege_page(string $page_link) : ?string
    {
        return 'cms_catalogues';
    }

    /**
     * Find what fields we should select for the Sitemap to be buildable. We don't want to select too much for perf reasons.
     * Also find out what language fields we should load up for the table (returned by reference).
     *
     * @param  ?array $cma_info CMA info (null: standard for this hook)
     * @param  ?string $table_alias Table alias (null: none)
     * @param  ?array $lang_fields_filtered List of language fields to load (null: not passed)
     * @return array Map between field name and field type
     */
    protected function select_fields(?array $cma_info = null, ?string $table_alias = null, ?array &$lang_fields_filtered = null) : array
    {
        if ($cma_info === null) {
            $cma_info = $this->_get_cma_info();
        }

        $ret = parent::select_fields($cma_info, $table_alias, $lang_fields_filtered);
        if ($cma_info['table'] == 'catalogue_categories') {
            $ret[] = 'cc_parent_id';
        }
        return $ret;
    }

    /**
     * Find details of a position in the Sitemap.
     *
     * @param  ID_TEXT $page_link The page-link we are finding
     * @param  ?mixed $callback Callback function to send discovered page-links to (null: return)
     * @param  ?array $valid_node_types List of node types we will return/recurse-through (null: no limit)
     * @param  ?integer $child_cutoff Maximum number of children before we cut off all children (null: no limit)
     * @param  ?integer $max_recurse_depth How deep to go from the Sitemap root (null: no limit)
     * @param  integer $recurse_level Our recursion depth (used to limit recursion, or to calculate importance of page-link, used for instance by XML Sitemap [deeper is typically less important])
     * @param  integer $options A bitmask of SITEMAP_GEN_* options
     * @param  ID_TEXT $zone The zone we will consider ourselves to be operating in (needed due to transparent redirects feature)
     * @param  integer $meta_gather A bitmask of SITEMAP_GATHER_* constants, of extra data to include
     * @param  ?array $row Database row (null: lookup)
     * @param  boolean $return_anyway Whether to return the structure even if there was a callback. Do not pass this setting through via recursion due to memory concerns, it is used only to gather information to detect and prevent parent/child duplication of default entry points.
     * @return ?array Node structure (null: working via callback / error)
     */
    public function get_node(string $page_link, $callback = null, ?array $valid_node_types = null, ?int $child_cutoff = null, ?int $max_recurse_depth = null, int $recurse_level = 0, int $options = 0, string $zone = '_SEARCH', int $meta_gather = 0, ?array $row = null, bool $return_anyway = false) : ?array
    {
        if (!addon_installed('catalogues')) {
            return null;
        }

        if (!$this->check_for_looping($page_link)) {
            return null;
        }

        $_ = $this->_create_partial_node_structure($page_link, $callback, $valid_node_types, $child_cutoff, $max_recurse_depth, $recurse_level, $options, $zone, $meta_gather, $row);
        if ($_ === null) {
            return null;
        }
        list($content_id, $row, $partial_struct) = $_;

        // level 0 = root
        // level 1 = zone
        if ($recurse_level == 2) {
            $sitemap_priority = SITEMAP_IMPORTANCE_MEDIUM;
        } else {
            $sitemap_priority = SITEMAP_IMPORTANCE_LOW;
        }

        $struct = [
            'sitemap_priority' => $sitemap_priority,
            'sitemap_refreshfreq' => 'weekly',

            'privilege_page' => $this->get_privilege_page($page_link),

            'edit_url' => build_url(['page' => 'cms_catalogues', 'type' => '_edit_category', 'id' => $content_id], get_module_zone('cms_catalogues')),
        ] + $partial_struct;

        if (($row['cc_parent_id'] === null) && ($GLOBALS['SITE_DB']->query_select_value_if_there('catalogues', 'c_is_tree', ['c_name' => $row['c_name']]) == 1)) {
            $struct['extra_meta']['is_a_category_tree_root'] = true;
        }

        if (!$this->_check_node_permissions($struct, $options)) {
            return null;
        }

        // Sometimes page groupings link direct to catalogue categories, so search for an icon
        $row_x = $this->_load_row_from_page_groupings(null, $meta_gather, $zone, 'catalogues', 'category', $content_id);
        if (!empty($row_x)) {
            if (($options & SITEMAP_GEN_LABEL_CONTENT_TYPES) == 0) {
                $struct['title'] = null;
            }
            $struct['extra_meta']['image'] = null;
            $struct['extra_meta']['icon'] = null;
            $this->_ameliorate_with_row($options, $struct, $row_x, $meta_gather);
        }

        if ($callback !== null) {
            call_user_func($callback, $struct);
        }

        // Categories done after node callback, to ensure sensible ordering
        $children = $this->_get_children_nodes($content_id, $page_link, $callback, $valid_node_types, $child_cutoff, $max_recurse_depth, $recurse_level, $options, $zone, $meta_gather, $row, '', null, $GLOBALS['SITE_DB']->translate_field_ref('cc_title'));
        $struct['children'] = $children;

        return ($callback === null || $return_anyway) ? $struct : null;
    }
}
