<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    catalogues
 */

/**
 * Hook class.
 */
class Hook_task_export_catalogue
{
    /**
     * Run the task hook.
     *
     * @param  ID_TEXT $catalogue_name The catalogue to export
     * @param  ?string $file_type The file type to export with (null: default)
     * @return ?array A tuple of at least 2: Return mime-type, content (either Tempcode, or a string, or a filename and file-path pair to a temporary file), map of HTTP headers if transferring immediately, map of ini_set commands if transferring immediately (null: show standard success message)
     */
    public function run(string $catalogue_name, ?string $file_type = null) : ?array
    {
        if (!addon_installed('catalogues')) {
            return null;
        }

        require_code('catalogues');

        $catalogue_row = $GLOBALS['SITE_DB']->query_select('catalogues', ['*'], ['c_name' => $catalogue_name]);
        if ($catalogue_row === null) {
            $catalogue_row = [];
        }
        if (isset($catalogue_row[0])) {
            $catalogue_row = $catalogue_row[0];
        }

        $category_names = [];

        require_code('files_spreadsheets_write');
        if ($file_type === null) {
            $file_type = spreadsheet_write_default();
        }
        $filename = $catalogue_name . '-' . date('Y-m-d') . '.' . $file_type;
        $outfile_path = null;
        $sheet_writer = spreadsheet_open_write($outfile_path, $filename, CMS_Spreadsheet_Writer::ALGORITHM_RAW);

        $fields = $GLOBALS['SITE_DB']->query_select('catalogue_fields', ['*'], ['c_name' => $catalogue_name], 'ORDER BY cf_order,' . $GLOBALS['SITE_DB']->translate_field_ref('cf_name'));
        global $CAT_FIELDS_CACHE;
        $CAT_FIELDS_CACHE[$catalogue_name] = $fields;
        $header_row = ['ID', 'CATEGORY'];
        foreach ($fields as $k) {
            $header_row[] = get_translated_text($k['cf_name']);
        }
        $sheet_writer->write_row($header_row);

        $num_rows = $GLOBALS['SITE_DB']->query_select_value('catalogue_entries', 'COUNT(*)', ['c_name' => $catalogue_name]);

        $start = 0;
        do {
            $entry_rows = $GLOBALS['SITE_DB']->query_select('catalogue_entries', ['*'], ['c_name' => $catalogue_name], 'ORDER BY ce_add_date ASC', 4000, $start);

            foreach ($entry_rows as $i => $entry_row) {
                task_log($this, 'Exporting catalogue row', $i, $num_rows);

                if ($entry_row === null) {
                    $entry_row = [];
                }
                if (isset($entry_row[0])) {
                    $entry_row = $entry_row[0];
                }

                $details = get_catalogue_entry_field_values($catalogue_name, $entry_row);

                if (!isset($category_names[$entry_row['cc_id']])) {
                    if (!array_key_exists($entry_row['cc_id'], $category_names)) {
                        $category_names[$entry_row['cc_id']] = get_translated_text($GLOBALS['SITE_DB']->query_select_value('catalogue_categories', 'cc_title', ['id' => $entry_row['cc_id']]));
                    }
                }

                $row = [$entry_row['id'], $category_names[$entry_row['cc_id']]];
                foreach ($details as $val) {
                    $row[] = $val['effective_value_pure'];
                }

                $sheet_writer->write_row($row);
            }

            $start += 4000;
        } while (!empty($entry_rows));
        $sheet_writer->close();

        $headers = [];
        $headers['Content-Type'] = $sheet_writer->get_mime_type();
        $headers['Content-Disposition'] = 'attachment; filename="' . escape_header($filename) . '"';

        $ini_set = [];
        $ini_set['ocproducts.xss_detect'] = '0';

        return [$sheet_writer->get_mime_type(), [$filename, $outfile_path], $headers, $ini_set];
    }
}
