<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    shopping
 */

/**
 * Hook class.
 */
class Hook_task_export_shopping_orders
{
    /**
     * Run the task hook.
     *
     * @param  ?AUTO_LINK $filter_id Filter by order ID (null: do not filter)
     * @param  ID_TEXT $filter_username Filter by customer username (blank: do not filter)
     * @param  ID_TEXT $filter_txn_id Filter by transaction ID (blank: do not filter)
     * @param  SHORT_TEXT $_filter_order_status Filter by comma-delimited order statuses (blank: do not filter)
     * @param  ?TIME $filter_start Only include orders on or after this date/time (null: do not filter)
     * @param  ?TIME $filter_end Only include orders on or before this date/time (null: do not filter)
     * @param  ?string $file_type The file type to export with (null: default)
     * @return ?array A tuple of at least 2: Return mime-type, content (either Tempcode, or a string, or a filename and file-path pair to a temporary file), map of HTTP headers if transferring immediately, map of ini_set commands if transferring immediately (null: show standard success message)
     */
    public function run(?int $filter_id = null, string $filter_username = '', string $filter_txn_id = '', string $_filter_order_status = '', ?int $filter_start = null, ?int $filter_end = null, ?string $file_type = null) : ?array
    {
        if (!addon_installed('shopping')) {
            return null;
        }

        require_code('ecommerce');

        $filter_order_status = explode(',', $_filter_order_status);

        $where = '1=1';
        $filename = 'orders_';
        if ($filter_id !== null) {
            $where .= ' AND id=' . strval($filter_id);
            $filename = 'order_' . strval($filter_id);
        }
        if ($filter_username != '') {
            $member_id = $GLOBALS['FORUM_DRIVER']->get_member_from_username($filter_username);
            if ($member_id !== null) {
                $where .= ' AND member_id=' . strval($member_id);
                $filename .= '_' . $filter_username;
            } else {
                warn_exit(do_lang_tempcode('_MEMBER_NO_EXIST', escape_html($filter_username)));
            }
        }
        if ($filter_txn_id != '') {
            $where .= ' AND ' . db_string_equal_to('txn_id', $filter_txn_id);
            $filename .= '_' . $filter_txn_id;
        }
        if ($_filter_order_status != '') {
            $where .= ' AND (';
            foreach ($filter_order_status as $key => $status) {
                if ($key > 0) {
                    $where .= ' OR ';
                }
                $where .= db_string_equal_to('order_status', $status);
            }
            $where .= ')';
            $filename .= '_' . $_filter_order_status;
        }
        if ($filter_start !== null) {
            $where .= ' AND add_date>=' . strval($filter_start);
            $filename .= '_' . strval($filter_start);
        }
        if ($filter_end !== null) {
            $where .= ' AND add_date<=' . strval($filter_end);
            $filename .= '_' . strval($filter_end);
        }

        require_code('files_spreadsheets_write');
        if ($file_type === null) {
            $file_type = spreadsheet_write_default();
        }
        $filename .= '.' . $file_type;

        $outfile_path = null;
        $sheet_writer = spreadsheet_open_write($outfile_path, $filename);

        $max = 500;
        $start = 0;

        $query = 'FROM ' . get_table_prefix() . 'shopping_orders o' . $GLOBALS['SITE_DB']->singular_join('ecom_trans_addresses', 'a', 'o.txn_id=a.a_txn_id', 'id', 'MIN', 'LEFT JOIN') . ' WHERE ' . $where;

        $max_rows = $GLOBALS['SITE_DB']->query_value_if_there('SELECT COUNT(*) ' . $query);

        do {
            $rows = $GLOBALS['SITE_DB']->query('SELECT o.*,o.id AS o_id,a.* ' . $query . ' ORDER BY o.add_date', $max, $start);

            foreach ($rows as $i => $_order) {
                task_log($this, 'Processing shopping order row', $i, $max_rows);

                $order = [];

                $order[do_lang('ORDER_NUMBER')] = strval($_order['o_id']);

                $order[do_lang('ORDERED_DATE')] = get_timezoned_date_time($_order['add_date']);

                $order[do_lang('ORDER_STATUS')] = do_lang($_order['order_status']);

                $order[do_lang('PRICE')] = float_format($_order['total_price']);

                $order[do_lang(get_option('tax_system'))] = float_format($_order['total_tax']);

                $order[do_lang('SHIPPING_COST')] = float_format($_order['total_shipping_cost']);

                $order[do_lang('ORDERED_PRODUCTS')] = get_ordered_product_list_string($_order['o_id']);

                $order[do_lang('ORDERED_BY')] = $GLOBALS['FORUM_DRIVER']->get_username($_order['member_id']);

                // Put address together
                $address = [];
                if ($_order['a_firstname'] !== null) {
                    if ($_order['a_firstname'] . $_order['a_lastname'] != '') {
                        $address[] = trim($_order['a_firstname'] . ' ' . $_order['a_lastname']);
                    }
                    if ($_order['a_street_address'] != '') {
                        $address[] = $_order['a_street_address'];
                    }
                    if ($_order['a_city'] != '') {
                        $address[] = $_order['a_city'];
                    }
                    if ($_order['a_county'] != '') {
                        $address[] = $_order['a_county'];
                    }
                    if ($_order['a_state'] != '') {
                        $address[] = $_order['a_state'];
                    }
                    if ($_order['a_post_code'] != '') {
                        $address[] = $_order['a_post_code'];
                    }
                    if ($_order['a_country'] != '') {
                        $address[] = $_order['a_country'];
                    }
                    if ($_order['a_email'] != '') {
                        $address[] = do_lang('EMAIL_ADDRESS') . ': ' . $_order['a_email'];
                    }
                    if ($_order['a_phone'] != '') {
                        $address[] = do_lang('PHONE_NUMBER') . ': ' . $_order['a_phone'];
                    }
                }
                $full_address = implode(', ', $address);
                $order[do_lang('SHIPPING_ADDRESS')] = $full_address;

                $sheet_writer->write_row($order);
            }

            $start += $max;
        } while (!empty($rows));

        $sheet_writer->close();

        $headers = [];
        $headers['Content-Type'] = $sheet_writer->get_mime_type();
        $headers['Content-Disposition'] = 'attachment; filename="' . escape_header($filename) . '"';

        $ini_set = [];
        $ini_set['ocproducts.xss_detect'] = '0';

        return [$sheet_writer->get_mime_type(), [$filename, $outfile_path], $headers, $ini_set];
    }
}
