<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    newsletter
 */

/**
 * Hook class.
 */
class Hook_task_send_newsletter
{
    /**
     * Run the task hook.
     *
     * @param  integer $message_id The newsletter message in the newsletter archive
     * @param  LONG_TEXT $message_raw The newsletter message
     * @param  SHORT_TEXT $subject The newsletter subject
     * @param  LANGUAGE_NAME $lang The language
     * @param  array $send_details A map describing what newsletters the newsletter is being sent to
     * @param  BINARY $html_only Whether to only send in HTML format
     * @param  EMAIL $from_email Override the e-mail address the mail is sent from (blank: staff address)
     * @param  string $from_name Override the name the mail is sent from (blank: site name)
     * @param  integer $priority The message priority (1=urgent, 3=normal, 5=low)
     * @range  1 5
     * @param  array $spreadsheet_data Spreadsheet data of extra subscribers (blank: none). This is in the same newsletter spreadsheet format that we export elsewhere.
     * @param  ID_TEXT $mail_template The template used to show the e-mail
     * @return ?array A tuple of at least 2: Return mime-type, content (either Tempcode, or a string, or a filename and file-path pair to a temporary file), map of HTTP headers if transferring immediately, map of ini_set commands if transferring immediately (null: show standard success message)
     */
    public function run(int $message_id, string $message_raw, string $subject, string $lang, array $send_details, int $html_only, string $from_email, string $from_name, int $priority, array $spreadsheet_data, string $mail_template) : ?array
    {
        if (!addon_installed('newsletter')) {
            return null;
        }

        $auto_pause = (get_option('newsletter_auto_pause') == '1');

        if ($auto_pause) {
            require_code('config2');
            set_option('newsletter_paused', '1');
        }

        require_code('newsletter');
        require_lang('newsletter');
        require_code('mail');

        $last_cron = get_value('last_cron');

        $blocked = newsletter_block_list();

        disable_php_memory_limit();

        $using_drip_queue = ($last_cron !== null) || (get_option('newsletter_paused') == '1');

        if ($using_drip_queue) {
            $already_queued = collapse_2d_complexity('d_to_email', 'tmp', $GLOBALS['SITE_DB']->query_select('newsletter_drip_send', ['d_to_email', '1 AS tmp'], ['d_message_id' => $message_id]));
        } else {
            $already_queued = [];
        }

        $max = 300;
        $max_rows = null;
        $start = 0;
        $count = 0;
        do {
            $_subscribers = newsletter_who_send_to($send_details, $lang, $start, $max, $spreadsheet_data);
            if ($start == 0) {
                list($subscribers, $totals) = $_subscribers;
                $max_rows = array_sum($totals);
            } else {
                list($subscribers) = $_subscribers;
            }

            $insert_maps = [ // We will do very efficient mass-inserts (making index maintenance and disk access much more efficient)
                'd_inject_time' => [],
                'd_message_id' => [],
                'd_message_binding' => [],
                'd_to_email' => [],
                'd_to_name' => [],
            ];

            // Send to all
            foreach ($subscribers as $email_address => $subscriber_map) {
                task_log($this, 'Processing recipient newsletter into queue', $count, $max_rows);

                if (isset($blocked[$email_address])) {
                    continue;
                }

                if (isset($already_queued[$email_address])) {
                    continue;
                }
                $already_queued[$email_address] = true;

                if ($using_drip_queue) {
                    $insert_map = [
                        'd_inject_time' => time(),
                        'd_message_id' => $message_id,
                        'd_message_binding' => json_encode($subscriber_map),
                        'd_to_email' => $email_address,
                        'd_to_name' => $subscriber_map['name'],
                    ];
                    foreach ($insert_map as $key => $val) {
                        $insert_maps[$key][] = $val;
                    }

                    if (($using_drip_queue) && (count($insert_maps['d_to_email']) >= 100)) {
                        $GLOBALS['SITE_DB']->query_insert('newsletter_drip_send', $insert_maps);
                        $insert_maps = [];
                    }
                } else { // Unlikely to use this code path, but we should support operation without the system scheduler in those rare cases
                    $message_wrapped = newsletter_prepare($message_raw, $subject, $lang, $subscriber_map['forename'], $subscriber_map['surname'], $subscriber_map['name'], $email_address, $subscriber_map['send_id'], $subscriber_map['hash']);
                    $is_html = newsletter_is_html($message_wrapped);

                    dispatch_mail(
                        $subject,
                        $message_wrapped,
                        '',
                        [$email_address],
                        [$subscriber_map['name']],
                        $from_email,
                        $from_name,
                        [
                            'priority' => $priority,
                            'no_cc' => true,
                            'as_admin' => true,
                            'in_html' => $is_html,
                            'mail_template' => $mail_template,
                            'bypass_queue' => true,
                            'smtp_sockets_use' => (get_option('newsletter_smtp_sockets_use') == '1'),
                            'smtp_sockets_host' => get_option('newsletter_smtp_sockets_host'),
                            'smtp_sockets_port' => intval(get_option('newsletter_smtp_sockets_port')),
                            'smtp_sockets_username' => get_option('newsletter_smtp_sockets_username'),
                            'smtp_sockets_password' => get_option('newsletter_smtp_sockets_password'),
                            'smtp_from_address' => get_option('newsletter_smtp_from_address'),
                            'enveloper_override' => (get_option('newsletter_enveloper_override') == '1'),
                            'allow_ext_images' => (get_option('newsletter_allow_ext_images') == '1'),
                            'website_email' => get_option('newsletter_website_email'),
                            'is_bulk' => true,
                        ]
                    );
                }

                $count++;

                if (function_exists('gc_collect_cycles')) {
                    gc_collect_cycles(); // Stop problem with PHP leaking memory
                }
            }
            $start += $max;

            if (($using_drip_queue) && (!empty($insert_maps['d_to_email']))) {
                $GLOBALS['SITE_DB']->query_insert('newsletter_drip_send', $insert_maps);
            }
        } while (array_key_exists(0, $subscribers));

        if ($count == 0) {
            return ['text/html', do_lang_tempcode('NEWSLETTER_NO_TARGET')];
        }

        if ($auto_pause) {
            require_code('notifications');
            $subject = do_lang('NEWSLETTER_PAUSED_SUBJECT');
            $newsletter_manage_url = build_url(['page' => 'admin_newsletter'], get_module_zone('admin_newsletter'), [], false, false, true);
            $message = do_lang('NEWSLETTER_PAUSED_BODY', escape_html($newsletter_manage_url->evaluate()));
            dispatch_notification('newsletter_paused', 'newsletter_' . strval(time()), $subject, $message, null, null, ['priority' => 4, 'create_ticket' => true]);
        }

        $newsletter_manage_url = build_url(['page' => 'admin_newsletter'], get_module_zone('admin_newsletter'));
        if (!$using_drip_queue) {
            return ['text/html', do_lang_tempcode('SENT_NEWSLETTER', escape_html($newsletter_manage_url->evaluate()), escape_html(integer_format($count)))];
        }
        return ['text/html', do_lang_tempcode($auto_pause ? 'SENDING_NEWSLETTER_TO_PAUSED_QUEUE' : 'SENDING_NEWSLETTER_TO_QUEUE', escape_html($newsletter_manage_url->evaluate()), escape_html(integer_format($count)))];
    }
}
