<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    import
 */

/**
 * Standard code module initialisation function.
 *
 * @ignore
 */
function init__import()
{
    require_code('urls2');
}

/**
 * Load lots that the importer needs to run.
 */
function load_import_deps()
{
    require_all_lang();
    require_code('cns_groups');
    require_code('cns_members');
    require_code('cns_moderation_action');
    require_code('cns_posts_action');
    require_code('cns_polls_action');
    require_code('cns_members_action');
    require_code('cns_groups_action');
    require_code('cns_general_action');
    require_code('cns_forums_action');
    require_code('cns_topics_action');
    require_code('cns_moderation_action2');
    require_code('cns_posts_action2');
    require_code('cns_polls_action2');
    require_code('cns_members_action2');
    require_code('cns_groups_action2');
    require_code('cns_general_action2');
    require_code('cns_forums_action2');
    require_code('cns_topics_action2');
    require_css('importing');
    require_all_core_cms_code();
}

/**
 * Switch Conversr to run over the local site-DB connector. Useful when importing and our forum driver is actually connected to a forum other than Conversr.
 */
function cns_over_local()
{
    $GLOBALS['MSN_DB'] = $GLOBALS['FORUM_DB'];
    $GLOBALS['FORUM_DB'] = $GLOBALS['SITE_DB'];
}

/**
 * Undo cns_over_local.
 */
function cns_over_msn()
{
    $GLOBALS['FORUM_DB'] = $GLOBALS['MSN_DB'];
    $GLOBALS['MSN_DB'] = null;
}

/**
 * Returns the NEW ID of an imported old ID, for the specified importation type. Whether it returns null or gives an error message depends on $fail_ok.
 *
 * @param  ID_TEXT $type An importation type code, from those the software has defined (E.g. 'download', 'news', ...)
 * @param  string $id_old The source (old, original) ID of the mapping
 * @param  boolean $fail_ok If it is okay to fail to find a mapping
 * @return ?AUTO_LINK The new ID (null: not found)
 */
function import_id_remap_get(string $type, string $id_old, bool $fail_ok = false) : ?int
{
    static $remap_cache = [];
    if ((array_key_exists($type, $remap_cache)) && (array_key_exists($id_old, $remap_cache[$type]))) {
        return $remap_cache[$type][$id_old];
    }

    $value = $GLOBALS['SITE_DB']->query_select_value_if_there('import_id_remap', 'id_new', ['id_session' => get_session_id(), 'id_type' => $type, 'id_old' => $id_old]);
    if ($value === null) {
        if ($fail_ok) {
            i_timed_refresh(); // We could end up in a long loop of null values, so we need to check refresh time.
            return null;
        }
        warn_exit(do_lang_tempcode('IMPORT_NOT_IMPORTED', escape_html($type), escape_html($id_old)));
    }
    $remap_cache[$type][$id_old] = $value;

    return $value;
}

/**
 * Check to see if the given ID of the given type has been imported (if it has a mapping).
 *
 * @param  ID_TEXT $type An importation type code, from those the software has defined
 * @param  string $id_old The source (old, original) ID of the mapping
 * @return boolean Whether it has been imported
 */
function import_check_if_imported(string $type, string $id_old) : bool
{
    send_http_output_ping();

    $test = $GLOBALS['SITE_DB']->query_select_value_if_there('import_id_remap', 'id_new', ['id_session' => get_session_id(), 'id_type' => $type, 'id_old' => $id_old]);
    $imported = $test !== null;

    // Could end up in a long loop of already-imported content; so we need to check refresh time.
    if ($imported) {
        i_timed_refresh();
    }

    return $imported;
}

/**
 * Set the NEW ID for an imported old ID, which also tacitly indicates completion of importing an item of some type of content. This mapping (old ID to new ID) may be used later for importing related content that requires the new identifier. import_id_remap_get is the inverse of this function.
 *
 * @param  ID_TEXT $type An importation type code, from those the software has defined
 * @param  string $id_old The source (old, original) ID of the mapping
 * @param  AUTO_LINK $id_new The destination (new) ID of the mapping
 */
function import_id_remap_put(string $type, string $id_old, int $id_new)
{
    $GLOBALS['SITE_DB']->query_insert('import_id_remap', ['id_session' => get_session_id(), 'id_type' => $type, 'id_old' => $id_old, 'id_new' => $id_new]);

    // Mark that we did something, and also check execution time / if we need to refresh
    global $I_REFRESH_DID_SOMETHING;
    $I_REFRESH_DID_SOMETHING = true;
    i_timed_refresh();
}

/**
 * Add a word to the word-filter.
 *
 * @param  SHORT_TEXT $word Word to add to the word-filter
 * @param  SHORT_TEXT $replacement Replacement (blank: block entirely)
 * @param  ?ID_TEXT $match_type The WORDFILTER_MATCH_TYPE_* to use (null: default to WORDFILTER_MATCH_TYPE_FULL)
 */
function add_wordfilter_word(string $word, string $replacement = '', ?string $match_type = null)
{
    require_code('wordfilter');

    if ($match_type === null) {
        $match_type = WORDFILTER_MATCH_TYPE_FULL;
    }
    $test = $GLOBALS['SITE_DB']->query_select_value_if_there('wordfilter', 'word', ['word' => $word]);
    if ($test === null) {
        $GLOBALS['SITE_DB']->query_insert('wordfilter', ['word' => $word, 'w_replacement' => $replacement, 'w_match_type' => $match_type]);
    }
}

/**
 * Detect if we have reached our timerout, more accurately than set_time_limit could.
 */
function i_timed_refresh()
{
    static $time = null;
    if ($time === null) {
        $time = time();
    }
    global $I_REFRESH_TIME;
    if (($I_REFRESH_TIME == 0) || !array_key_exists('I_REFRESH_URL', $GLOBALS)) {
        return;
    }
    if (time() >= $time + $I_REFRESH_TIME) {
        i_force_refresh();
    }
}

/**
 * Force a page refresh due to maximum execution timeout / tick-function determined timeout.
 */
function i_force_refresh()
{
    if ((strpos($GLOBALS['I_REFRESH_URL'], "\n") !== false) || (strpos($GLOBALS['I_REFRESH_URL'], "\r") !== false)) {
        log_hack_attack_and_exit('HEADER_SPLIT_HACK');
    }

    global $I_REFRESH_URL, $I_REFRESH_TIME, $I_REFRESH_DID_SOMETHING;

    $post = build_keep_post_fields([], true);
    $refresh = do_template('JS_REFRESH', ['_GUID' => '748a4b4d970990453271f229f29a0a09', 'FORM_NAME' => 'redir-form']);

    $title = get_screen_title('IMPORT');
    $url = $I_REFRESH_URL;
    foreach ($GLOBALS as $key => $val) {
        if (preg_match('#^JUMPSTART_#', $key) != 0) {
            $url = preg_replace('#&' . preg_quote($key, '#') . '=\d+#', '', $url);
            $url .= '&' . $key . '=' . strval($val);
        }
    }

    if ($I_REFRESH_DID_SOMETHING) {
        $middle = do_template('REDIRECT_POST_METHOD_SCREEN', [
            '_GUID' => '63ad793efebb1abe05d2aa27a2b7de38',
            'REFRESH' => $refresh,
            'TITLE' => $title,
            'TEXT' => do_lang_tempcode('REFRESH_TIMEOUT_REACHED', strval($I_REFRESH_TIME)),
            'URL' => $url,
            'POST' => $post,
        ]);
    } else {
        warn_exit(do_lang_tempcode('REFRESH_TIMEOUT_REACHED_NOTHING_DONE', strval($I_REFRESH_TIME)));
    }

    $echo = globalise($middle, null, '', true);
    $echo->evaluate_echo();
    exit();
}

/**
 * Cleanup after an import has finished.
 */
function post_import_cleanup()
{
    // Quick and simple decaching. No need to be smart about this.
    delete_value('cns_member_count');
    delete_value('cns_topic_count');
    delete_value('cns_post_count');
}

/**
 * Turn index maintenance off to help speed import, or back on.
 *
 * @param  boolean $on Whether index maintenance should be on
 */
function set_database_index_maintenance(bool $on)
{
    if (strpos(get_db_type(), 'mysql') !== false) {
        push_db_scope_check(false);

        $tables = $GLOBALS['SITE_DB']->query_select('db_meta', ['DISTINCT m_table']);
        foreach ($tables as $table) {
            $tbl = $table['m_table'];
            $GLOBALS['SITE_DB']->query('ALTER TABLE ' . $GLOBALS['SITE_DB']->get_table_prefix() . $tbl . ' ' . ($on ? 'ENABLE' : 'DISABLE') . ' KEYS');
        }

        pop_db_scope_check();
    }
}

/**
 * Sort a list of imports according to dependencies to help ensure imports are listed in the order they should be run.
 * Note: this is not very performant, but it is assumed the import tool will rarely ever be used by multiple people at the same time.
 *
 * @param  array $imports List of imports
 * @param  ?array $dependencies Map of imports to their dependencies (null: no dependencies)
 * @param  boolean $silent_fail Whether to return null if an infinite loop / cyclic dependency occurred opposed to fatally exiting
 * @return ?array The newly-sorted imports (null: Error occurred and $silent_fail was true)
 */
function sort_imports_by_dependencies(array $imports, ?array $dependencies = null, bool $silent_fail = false) : ?array
{
    // No processing necessary if there are no dependencies
    if (($dependencies === null) || empty($dependencies)) {
        return $imports;
    }

    $_imports = [];

    // Imports with no dependencies should go first
    foreach ($imports as $import) {
        if (!isset($dependencies[$import]) || empty($dependencies[$import])) {
            $_imports[] = $import;
            unset($dependencies[$import]); // We want to ignore these imports in future foreach loops as we already determined these go first
        }
    }

    // Return immediately if there are no more dependencies to process; otherwise the loop below will exit with $fatal_exit = true.
    if (empty($dependencies)) {
        return $_imports;
    }

    // Now loop through the imports with dependencies until we've processed all of them
    $fatal_exit = false;
    do {
        $fatal_exit = true; // This should be set to false if we processed at least 1 import this iteration. Otherwise, bail from the loop.
        foreach ($dependencies as $import => $dependency_array) {
            // Remove all the dependencies we already processed as going before this import
            foreach ($dependency_array as $i => $dependency) {
                if (in_array($dependency, $_imports)) {
                    unset($dependencies[$import][$i]);
                    $fatal_exit = false; // We made a dependency check modification, so we can safely assume no infinite loop happened
                }
            }

            // If all dependencies are set to go before this import, we can safely set this import to go next.
            if (empty($dependency_array)) {
                $_imports[] = $import;
                unset($dependencies[$import]);
                $fatal_exit = false; // We added an import to our array; we know we are not in an infinite loop.
            }
        }
    } while ((count($dependencies) > 0) && !$fatal_exit);

    // Did we exit the loop prematurely (probably because of a cyclic dependency)?
    if ($fatal_exit) {
        if ($silent_fail) {
            return null;
        }
        fatal_exit(do_lang_tempcode('INTERNAL_ERROR', escape_html('551ae9687c04595d9c044d00e1367663')));
    }

    return $_imports;
}
