<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    core
 */

/**
 * Get the contents of an HTML page.
 * HTML isn't great... no dynamicness/reconfigurability at all.
 * We prefer Comcode with [html]HTML goes here[/html] usage.
 *
 * @param  PATH $string The relative (to the software's base directory) path to the HTML page
 * @param  ?PATH $file_base The file base to load from (null: standard)
 * @return string The page
 */
function load_html_page(string $string, ?string $file_base = null) : string
{
    if ($file_base === null) {
        $file_base = get_file_base();
    }

    global $PAGE_STRING;
    if ($PAGE_STRING === null) {
        $PAGE_STRING = $string;
    }

    $html = cms_file_get_contents_safe($file_base . '/' . $string, FILE_READ_LOCK);

    // Post-processing
    if (stripos($html, '<html') !== false) {
        $matches = [];

        // Fix links to anything in same dir, by assuming either a page in same zone -- or uploads/website_specific, or next to html files, or in root
        $link_attributes = ['src', 'href', 'action', 'data', 'codebase', 'background'];
        foreach ($link_attributes as $attribute) {
            $num_matches = preg_match_all('#<[^<>]* ' . $attribute . '="([^&"]+\.[^&"\.]+)"[^<>]*>#mis', $html, $matches);
            for ($i = 0; $i < $num_matches; $i++) {
                $old_url = html_entity_decode($matches[1][$i], ENT_QUOTES);

                $zone = '_SELF';
                if ($old_url[0] == '/') {
                    $old_url = substr($old_url, 1);
                    $zone = '';
                }
                $possible_zone = str_replace('/', '_', dirname($old_url));
                if ($possible_zone == '.') {
                    $possible_zone = '';
                }
                if (($possible_zone != '') && ($possible_zone != get_zone_name()) && (file_exists(get_file_base() . '/' . $possible_zone . '/index.php'))) {
                    $zone = $possible_zone;
                }

                if (substr($old_url, -4) == '.htm') {
                    $_new_url = build_url(['page' => basename(substr($old_url, 0, strlen($old_url) - 4))], $zone);
                    $new_url = $_new_url->evaluate();
                } elseif (substr($old_url, -5) == '.html') {
                    $_new_url = build_url(['page' => basename(substr($old_url, 0, strlen($old_url) - 5))], $zone);
                    $new_url = $_new_url->evaluate();
                } else {
                    $new_url = $old_url;
                    if (url_is_local($old_url)) {
                        // Strip out query strings and fragments as this will cause is_file to fail
                        $old_url_parts_a = explode('?', $old_url);
                        $old_url_sanitised_a = $old_url_parts_a[0];
                        $old_url_parts_b = explode('#', $old_url_sanitised_a);
                        $old_url_sanitised_b = $old_url_parts_b[0];

                        if (is_file(get_custom_file_base() . '/' . dirname($string) . '/' . urldecode($old_url))) { // HTML pages dir
                            $dirname = dirname($string);
                            if ($dirname == '.') {
                                $dirname = '';
                            }
                            $new_url = get_base_url() . '/' . (($dirname == '') ? '' : ($dirname . '/')) . $old_url_sanitised_b;
                        } elseif (is_file(get_custom_file_base() . '/' . get_zone_name() . '/' . urldecode($old_url_sanitised_b))) { // Zone dir
                            $new_url = get_base_url() . '/' . ((get_zone_name() == '') ? '' : (get_zone_name() . '/')) . $old_url_sanitised_b;
                        } elseif (is_file(get_custom_file_base() . '/' . urldecode($old_url_sanitised_b))) { // Root dir
                            $new_url = get_base_url() . '/' . $old_url_sanitised_b;
                        } else { // uploads/website_specific
                            $new_url = get_base_url() . '/uploads/website_specific/' . $old_url_sanitised_b;
                        }
                    }
                }

                $html = str_replace(' ' . $attribute . '="' . $old_url . '"', ' ' . $attribute . '="' . $new_url . '"', $html);
            }
        }

        // Extract script, style, and link elements from head
        if (preg_match('#<\s*head[^<>]*>(.*)<\s*/\s*head\s*>#mis', $html, $matches) != 0) {
            $head = $matches[1];

            $head_patterns = ['#<\s*script.*<\s*/\s*script\s*>#misU', '#<\s*link[^<>]*>#misU', '#<\s*style.*<\s*/\s*style\s*>#misU'];
            foreach ($head_patterns as $pattern) {
                $num_matches = preg_match_all($pattern, $head, $matches);
                for ($i = 0; $i < $num_matches; $i++) {
                    attach_to_screen_header($matches[0][$i]);
                }
            }
        }

        // Extra meta keywords and description, and title
        global $SEO_KEYWORDS, $SEO_DESCRIPTION;
        if (preg_match('#<\s*meta\s+name\s*=\s*"keywords"\s+content="([^"]*)"#mi', $html, $matches) != 0) {
            $SEO_KEYWORDS = explode(',', @html_entity_decode(trim($matches[1]), ENT_QUOTES));
        }
        if (preg_match('#<\s*meta\s+name\s*=\s*"description"\s+content="([^"]*)"#mi', $html, $matches) != 0) {
            $SEO_DESCRIPTION = @html_entity_decode(trim($matches[1]), ENT_QUOTES);
        }
        if (preg_match('#<\s*title\s*>([^<>]*)<\s*/\s*title\s*>#mis', $html, $matches) != 0) {
            set_short_title(@html_entity_decode(trim($matches[1]), ENT_QUOTES));
        }

        // Extract body
        if (preg_match('#<\s*body[^>]*>(.*)<\s*/\s*body\s*>#mis', $html, $matches) != 0) {
            $html = $matches[1];
        } else {
            $html = '';
        }
    }

    // Run hooks for modifying the HTML page
    $hook_obs = find_all_hook_obs('systems', 'site_html_pages', 'Hook_site_html_pages_');
    foreach ($hook_obs as $hook => $ob) {
        $ob->run($html, $string, $file_base); // HTML is passed by reference
    }

    $GLOBALS['SCREEN_TEMPLATE_CALLED'] = '';

    return $html;
}
