<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    core
 */

/**
 * Standard code module initialisation function.
 *
 * @ignore
 */
function init__web_resources2()
{
    global $CSS_COMPILE_ACTIVE_THEME;
    $CSS_COMPILE_ACTIVE_THEME = 'default';
}

/**
 * Compile a JavaScript file.
 *
 * @param  ID_TEXT $j Name of the JS file
 * @param  PATH $js_cache_path Full path to the JS file
 * @param  boolean $minify Whether to also do minification
 * @param  ?ID_TEXT $theme Theme to use (null: current theme)
 */
function js_compile(string $j, string $js_cache_path, bool $minify = true, ?string $theme = null)
{
    $old_limit = cms_extend_time_limit(TIME_LIMIT_EXTEND__SLOW);

    cms_profile_start_for('js_compile');

    global $KEEP_MARKERS, $SHOW_EDIT_LINKS;
    $temp_keep_markers = $KEEP_MARKERS;
    $temp_show_edit_links = $SHOW_EDIT_LINKS;
    $KEEP_MARKERS = false;
    $SHOW_EDIT_LINKS = false;
    $tpl_params = [];
    require_code('tempcode');
    $js = do_template($j, $tpl_params, null, false, null, '.js', 'javascript', $theme);
    $KEEP_MARKERS = $temp_keep_markers;
    $SHOW_EDIT_LINKS = $temp_show_edit_links;
    global $ATTACHED_MESSAGES_RAW;
    $num_msgs_before = count($ATTACHED_MESSAGES_RAW);
    $out = $js->evaluate();
    $num_msgs_after = count($ATTACHED_MESSAGES_RAW);
    $success_status = ($num_msgs_before == $num_msgs_after);
    if ($minify) {
        $out = js_minify($out);
    }

    if (($out == '') || ($minify)) {
        $contents = $out;
    } else {
        $contents = '/* DO NOT EDIT. THIS IS A CACHE FILE AND WILL GET OVERWRITTEN RANDOMLY.' . "\n" . 'INSTEAD EDIT THE TEMPLATE FROM WITHIN THE ADMIN ZONE, OR BY MANUALLY EDITING A JAVASCRIPT_CUSTOM OVERRIDE. */' . "\n\n" . $out;
    }
    require_code('files');
    $success_status = cms_file_put_contents_safe($js_cache_path, $contents, FILE_WRITE_FAILURE_SILENT | FILE_WRITE_FIX_PERMISSIONS | FILE_WRITE_SYNC_FILE | FILE_WRITE_BOM);
    if (!$success_status) {
        @touch($js_cache_path, time() - 60 * 60 * 24); // Fudge it so it's going to auto expire. We do have to write the file as it's referenced, but we want it to expire instantly so that any errors will reshow.
    } elseif (get_value('disable_web_resources_static_compression') !== '1') {
        compress_cms_stub_file($js_cache_path);
    }

    cms_profile_end_for('js_compile', $j);

    cms_set_time_limit($old_limit);
}

/**
 * Compile a CSS file.
 *
 * @param  ID_TEXT $active_theme The theme the file is being loaded for
 * @param  ID_TEXT $theme The theme the file is in
 * @param  ID_TEXT $c Name of the CSS file
 * @param  PATH $full_path Full path to the CSS file
 * @param  PATH $css_cache_path Full path to where the cached CSS file will go
 * @param  boolean $minify Whether to also do minification
 */
function css_compile(string $active_theme, string $theme, string $c, string $full_path, string $css_cache_path, bool $minify = true)
{
    cms_profile_start_for('css_compile');

    initialise_css_tempcode_context($active_theme);

    list($success_status, $out) = _css_compile($active_theme, $theme, $c, $full_path, $minify);
    require_code('files');
    $success_status = cms_file_put_contents_safe($css_cache_path, $out, FILE_WRITE_FAILURE_SILENT | FILE_WRITE_FIX_PERMISSIONS | FILE_WRITE_SYNC_FILE | FILE_WRITE_BOM);
    if (!$success_status) {
        @touch($css_cache_path, time() - 60 * 60 * 24); // Fudge it so it's going to auto expire. We do have to write the file as it's referenced, but we want it to expire instantly so that any errors will reshow.
    } elseif (get_value('disable_web_resources_static_compression') !== '1') {
        compress_cms_stub_file($css_cache_path);
    }

    cms_profile_end_for('css_compile', $c);
}

/**
 * Return a specific compiled CSS file.
 *
 * @param  ID_TEXT $active_theme The theme the file is being loaded for
 * @param  string $theme Theme name
 * @param  string $c The CSS file required
 * @param  PATH $full_path Full path to CSS file (file is in non-compiled Tempcode format)
 * @param  boolean $minify Whether to also do minification
 * @return array A pair: success status, The text of the compiled file
 *
 * @ignore
 */
function _css_compile(string $active_theme, string $theme, string $c, string $full_path, bool $minify = true) : array
{
    // Book-keeping, then loading up the CSS text
    global $KEEP_MARKERS, $SHOW_EDIT_LINKS;
    $keep_markers = $KEEP_MARKERS;
    $show_edit_links = $SHOW_EDIT_LINKS;
    $KEEP_MARKERS = false;
    $SHOW_EDIT_LINKS = false;
    if (($theme != 'default') && (!is_file($full_path))) {
        $theme = 'default';
    }
    if ($GLOBALS['RECORD_TEMPLATES_USED']) {
        record_template_used('css/' . $c . '.css');
    }
    require_code('tempcode_compiler');
    global $ATTACHED_MESSAGES_RAW;
    $num_msgs_before = count($ATTACHED_MESSAGES_RAW);
    $suffix = '.' . get_file_extension($full_path);
    $css = _do_template($theme, (strpos($full_path, '/css_custom/') !== false) ? '/css_custom/' : '/css/', $c, $c, user_lang(), $suffix, $active_theme);
    $out = $css->evaluate();
    $num_msgs_after = count($ATTACHED_MESSAGES_RAW);
    global $CSS_COMPILE_ACTIVE_THEME;
    $CSS_COMPILE_ACTIVE_THEME = $active_theme;

    // Strip empty comments (would have encapsulated Tempcode comments)
    $out = cms_preg_replace_safe('#/\*\s*\*/#', '', $out);

    // The language may need to filter it
    global $LANG_FILTER_OB;
    if (isset($LANG_FILTER_OB)) {
        $out = $LANG_FILTER_OB->filter_css($c, $out);
    }

    // Minification
    if ($minify) {
        $out = css_minify($out);
    }

    // "Do not edit the cache file" warning
    if ($c != 'no_cache') {
        if (($out != '') && (!$minify)) {
            $out = '/* DO NOT EDIT. THIS IS A CACHE FILE AND WILL GET OVERWRITTEN RANDOMLY.' . "\n" . 'INSTEAD EDIT THE CSS FROM WITHIN THE ADMIN ZONE, OR BY MANUALLY EDITING A CSS_CUSTOM OVERRIDE. */' . "\n\n" . $out;
        }
    }

    // Book-keeping then finish
    $KEEP_MARKERS = $keep_markers;
    $SHOW_EDIT_LINKS = $show_edit_links;
    if ($num_msgs_after > $num_msgs_before) { // Was an error (e.g. missing theme image), so don't cache so that the error will be visible on refresh and hence debugged
        return [false, $out];
    }
    return [true, $out];
}

/**
 * Compress a file to Gzip, and save with a stem of .gz.
 *
 * @param  PATH $stub_file Full path to the file to compress
 * @return boolean Success status
 */
function compress_cms_stub_file(string $stub_file) : bool
{
    $a = compress_cms_stub_file_br($stub_file);
    $b = compress_cms_stub_file_gz($stub_file);
    return $a || $b;
}

/**
 * Compress a file to Brotli, and save with a stem of .br.
 *
 * @param  PATH $stub_file Full path to the file to compress
 * @return boolean Success status
 */
function compress_cms_stub_file_br(string $stub_file) : bool
{
    if (!supports_brotli(false)) {
        return false;
    }

    $data = @cms_file_get_contents_safe($stub_file, FILE_READ_LOCK | FILE_READ_BOM);
    if ($data === false) {
        return false;
    }

    require_code('files');
    $compressed = cms_brotli_compress($data, 11 /*max*/);
    if ($compressed === false) {
        return false;
    }
    cms_file_put_contents_safe($stub_file . '.br', $compressed, FILE_WRITE_FAILURE_SILENT | FILE_WRITE_FIX_PERMISSIONS | FILE_WRITE_SYNC_FILE);
    return true;
}

/**
 * Compress a file, and save with a stem of .gz.
 *
 * @param  PATH $stub_file Full path to the file to compress
 * @return boolean Success status
 */
function compress_cms_stub_file_gz(string $stub_file) : bool
{
    if (!supports_gzip(false)) {
        return false;
    }

    $data = @cms_file_get_contents_safe($stub_file, FILE_READ_LOCK | FILE_READ_BOM);
    if ($data === false) {
        return false;
    }

    require_code('files');
    $compressed = cms_gzencode($data, 9 /*max*/);
    if ($compressed === false) {
        return false;
    }
    cms_file_put_contents_safe($stub_file . '.gz', $compressed, FILE_WRITE_FAILURE_SILENT | FILE_WRITE_FIX_PERMISSIONS | FILE_WRITE_SYNC_FILE);
    return true;
}

/**
 * Minimise the given JavaScript.
 *
 * @param  string $js JavaScript to minimise
 * @return string Minimised JavaScript
 */
function js_minify(string $js) : string
{
    if (strpos(substr($js, 0, 1000), 'no minify') !== false) {
        return str_replace('/*no minify*/', '', $js);
    }

    $before_ts = ini_get('ocproducts.type_strictness');
    cms_ini_set('ocproducts.type_strictness', '0');
    $before_xd = ini_get('ocproducts.xss_detect');
    cms_ini_set('ocproducts.xss_detect', '0');

    require_code('jsmin');

    if (class_exists('\JShrink\Minifier')) {
        $js = \JShrink\Minifier::minify($js);
    }

    cms_ini_set('ocproducts.type_strictness', $before_ts);
    cms_ini_set('ocproducts.xss_detect', $before_xd);

    return $js;
}

/**
 * cssmin.php - A simple CSS minifier.
 * --
 *
 * <code>
 * include("cssmin.php");
 * file_put_contents("path/to/target.css", cssmin::minify(file_get_contents("path/to/source.css")));
 * </code>
 * --
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
 * BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * --
 *
 * @author     Joe Scylla <joe.scylla@gmail.com>
 * @copyright  2008 Joe Scylla <joe.scylla@gmail.com>
 * @license    http://opensource.org/licenses/mit-license.php MIT License
 * @version    1.0 (2008-01-31)
 * @package    core
 */

/**
 * Minifies stylesheet definitions.
 *
 * @param  string $v Stylesheet definitions as string
 * @return string Minified stylesheet definitions
 */
function css_minify(string $v) : string
{
    $search = ['/\/\*[\d\D]*?\*\/|\t+/', '/\s+/'];
    $replace = ['', ' '];
    $v = preg_replace($search, $replace, $v);
    $search = ['/\\;\s/', '/\s+\{\\s+/', '/\\:\s+\\#/', '/,\s+/i', '/\\:\s+\\\'/i', '/\\:\s+([0-9]+|[A-F]+)/i'];
    $replace = [';', '{', ':#', ',', ':\'', ':$1'];
    $v = preg_replace($search, $replace, $v);
    $v = str_replace("\n", '', $v);
    return trim($v);
}
